(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * encoding_example.thy - 
 * Copyright (C) 2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This File is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)


header {* Encoding Recursive Data Structures *}

theory digraph_manuell
imports 
  OCL
begin


ML {* rep_encoder_common.log_level := 3 *}

text_raw{* \label{thy:tree} *} 

load_xmi "digraph_ocl.xmi"
 

thm level1.tree.Node.model.set

thm level1.tree.Node.model.set_fun

thm level1.tree.Node.model.set_fun.mono

thm  Node.model.hol_invariant

thm Node.model.hol_invariant.typeconstraint_left_def

thm Node.model.hol_invariant.inv_3_def

section {* Playing with Equalities *}

thm tree.Node.strict_sve
thm tree.Cnode.strict_sve

lemma "((x::(('a,'b,'c) U_tree St \<Rightarrow>  'a Cnode)) \<MathOclStrictValueEq> (y::(('a,'b,'c) U_tree St \<Rightarrow>  'a Cnode)))  
       =  (((x ->oclAsType(level1.tree.Node)) \<MathOclStrictValueEq> (y ->oclAsType(level1.tree.Node)))
           \<MathOclSand>  
           (level1.tree.Cnode.color x)  \<MathOclStrictEq>  (level1.tree.Cnode.color y))" 
oops

thm tree.Node.strong_sve
thm tree.Cnode.strong_sve


section {* Characterstic Sets for Recursive Object Structures *}

thm Node_def   (* the char set *)
thm Cnode_def

thm tree.get_Node_def

thm tree.is_Node_def

thm tree.invariant1  (*is 2nd level invariant.*)

thm Node.invariant1.inv_3_def

thm Cnode.invariant1.inv_1_def

(*
lemma shouldhold:
"\<tau> \<MathOclValid> tree.is_Node x \<Longrightarrow> \<tau> \<MathOclValid> tree.invariant1 x "
*)


text {* 
\begin{enumerate}
\item map \OCL formulae onto two-valued logic, i.e., 
\item construct the following invariant functions:
  \begin{enumerate}
  \item[user-defined] as defined by the user 
  \item[relax-refs]   relax definedness of references
  \item[relax-attrib]  relax definedness of all attribute types
  \end{enumerate}

  \begin{itemize}
  \item Klassen mit allen gegenseitig recursive Abhaengigkeiten
  \item umsetzen der Invariante auf zweiwertige Logik
  \item liberale invarianten erzeugen
  \end{itemize}
  
\end{enumerate}
*}

constdefs 
  Node_set_fun_L :: "('a,'b,'c) U_tree St \<Rightarrow>
                       (('a,'b,'c) U_tree St \<Rightarrow> ('a,'b) Node)set \<Rightarrow> 
                       (('a,'b,'c) U_tree St \<Rightarrow> ('a,'b) Node)set"
 "Node_set_fun_L \<equiv>  \<lambda> \<tau> . (\<lambda> X. {self. 
                (
                 (\<tau> \<MathOclValid> \<MathOclIsDefined> (level1.tree.Node.i self) \<longrightarrow> \<tau> \<MathOclValid> invariant1.inv_3 self)
               \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> (level1.tree.Node.left self) \<longrightarrow> (level1.tree.Node.left self) \<in> X) 
               \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> (level1.tree.Node.right self) \<longrightarrow> (level1.tree.Node.right self) \<in> X) )})"

constdefs 
  Node_set_fun :: "('a,'b,'c) U_tree St \<Rightarrow>
                    (('a,'b,'c) U_tree St \<Rightarrow> ('a,'b) Node)set \<Rightarrow> 
                    (('a,'b,'c) U_tree St \<Rightarrow> ('a,'b) Node)set"
 "Node_set_fun \<equiv>  \<lambda> \<tau> . (\<lambda> X. {self. 
                ((\<tau> \<MathOclValid> \<MathOclIsDefined> (level1.tree.Node.i self) \<and> \<tau> \<MathOclValid> invariant1.inv_3 self)
               \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> (level1.tree.Node.left self) \<and>  
                             (level1.tree.Node.left self) \<in> X) 
               \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> (level1.tree.Node.right self) \<and> 
                             (level1.tree.Node.right self) \<in> X) )})"


constdefs Node_set_L :: 
    " ('a,'b,'c)U_tree St \<Rightarrow> (('a,'b,'c)U_tree St \<Rightarrow> ('a,'b) Node) set"
    " Node_set_L \<equiv> \<lambda> \<tau> . (gfp (Node_set_fun_L \<tau>))"

constdefs Node_set :: 
    " ('a,'b,'c)U_tree St \<Rightarrow> (('a,'b,'c)U_tree St \<Rightarrow> ('a,'b) Node) set"
    " Node_set \<equiv> \<lambda> \<tau> . (gfp (Node_set_fun \<tau>))"
constdefs Cnode_set_fun_L :: "('a,'b,'c) U_tree St \<Rightarrow>
                       (('a,'b,'c) U_tree St \<Rightarrow> ('a) Cnode)set \<Rightarrow> 
                       (('a,'b,'c) U_tree St \<Rightarrow> ('a) Cnode)set"
        "Cnode_set_fun_L \<equiv>  \<lambda> \<tau> . (\<lambda> X. {self. 
                ( 
            ((self ->oclAsType(level1.tree.Node))\<in> (Node_set_fun_L \<tau> ((\<lambda> x. x ->oclAsType(level1.tree.Node)) ` X)))
           \<and>           
                       ((\<tau> \<MathOclValid> \<MathOclIsDefined> ((level1.tree.Node.left (self ->oclAsType( level1.tree.Node )))->oclAsType(level1.tree.Cnode))) \<longrightarrow> 
                               (((level1.tree.Node.left (self ->oclAsType( level1.tree.Node ))) ->oclAsType(level1.tree.Cnode)) \<in> X)) \<and> 
                  ((\<tau> \<MathOclValid> \<MathOclIsDefined> ((level1.tree.Node.right (self ->oclAsType( level1.tree.Node )))->oclAsType(level1.tree.Cnode))) \<longrightarrow> 
                              (( (level1.tree.Node.right (self ->oclAsType( level1.tree.Node ))) ->oclAsType(level1.tree.Cnode)) \<in> X)) \<and>
                  ((\<tau> \<MathOclValid> \<MathOclIsDefined> (level1.tree.Cnode.color self)) \<longrightarrow> 
                            (\<tau> \<MathOclValid> \<MathOclIsDefined> (level1.tree.Node.right (self ->oclAsType( level1.tree.Node )))) \<longrightarrow> 
                                \<not> (\<tau> \<MathOclValid> level1.tree.Cnode.color (level1.tree.Node.right (self ->oclAsType( level1.tree.Node )) 
                                                                                             ->oclAsType( level1.tree.Cnode ))
                                   \<MathOclStrongEq> level1.tree.Cnode.color self))
                )})"


constdefs Cnode_set_fun :: "('a,'b,'c) U_tree St \<Rightarrow>
                       (('a,'b,'c) U_tree St \<Rightarrow> ('a) Cnode)set \<Rightarrow> 
                       (('a,'b,'c) U_tree St \<Rightarrow> ('a) Cnode)set"
    "Cnode_set_fun \<equiv>  \<lambda> \<tau> . (\<lambda> X. {self. 
          (
            ((self ->oclAsType(level1.tree.Node))\<in> (Node_set_fun \<tau> ((\<lambda> x. x ->oclAsType(level1.tree.Node)) ` X)))
           \<and> 
           (\<tau> \<MathOclValid> \<MathOclIsDefined> (level1.tree.Cnode.color self)) \<and> 
            ((\<tau> \<MathOclValid> \<MathOclIsDefined> ((level1.tree.Node.left (self ->oclAsType( level1.tree.Node )))->oclAsType(level1.tree.Cnode)))  \<and> 
                         (((level1.tree.Node.left (self ->oclAsType( level1.tree.Node ))) ->oclAsType(level1.tree.Cnode)) \<in> X)) \<and> 
            ((\<tau> \<MathOclValid> \<MathOclIsDefined> ((level1.tree.Node.right (self ->oclAsType( level1.tree.Node )))->oclAsType(level1.tree.Cnode)))  \<and>  
                        (( (level1.tree.Node.right (self ->oclAsType( level1.tree.Node ))) ->oclAsType(level1.tree.Cnode)) \<in> X)) \<and>
            ((\<tau> \<MathOclValid> \<MathOclIsDefined> (level1.tree.Cnode.color self))  \<and>  
                      (\<tau> \<MathOclValid> \<MathOclIsDefined> ((level1.tree.Node.right (self ->oclAsType( level1.tree.Node )))->oclAsType(level1.tree.Cnode)))  \<and>  
                          \<not> (\<tau> \<MathOclValid> level1.tree.Cnode.color (level1.tree.Node.right (self ->oclAsType( level1.tree.Node )) 
                                                                                       ->oclAsType( level1.tree.Cnode ))
                             \<MathOclStrongEq> level1.tree.Cnode.color self))
          )})"



constdefs Cnode_set :: 
    " ('a,'b,'c)U_tree St \<Rightarrow> (('a,'b,'c)U_tree St \<Rightarrow> ('a) Cnode) set"
    " Cnode_set \<equiv> \<lambda> \<tau> . (gfp (Cnode_set_fun \<tau>))"

constdefs Cnode_set_L :: 
    " ('a,'b,'c)U_tree St \<Rightarrow> (('a,'b,'c)U_tree St \<Rightarrow> ('a) Cnode) set"
    " Cnode_set_L \<equiv> \<lambda> \<tau> . (gfp (Cnode_set_fun_L \<tau>))"


lemma Node_set_fun_mono: "mono (Node_set_fun \<tau>)" 
  by(auto simp add: mono_def Node_set_fun_def)

lemma Cnode_set_fun_mono: "mono (Cnode_set_fun \<tau>)"
  by(auto simp add: Cnode_set_fun_def mono_def  Node_set_fun_def) 


lemma Node_set_fun_L_mono: "mono (Node_set_fun_L \<tau>)" 
  by(auto simp add: mono_def Node_set_fun_L_def)

lemma Cnode_set_fun_L_mono: "mono (Cnode_set_fun_L \<tau>)" 
  by(auto simp add: mono_def Cnode_set_fun_L_def Node_set_fun_L_def)

(* level_1 relations between liberal and non-liberal invariants *)

lemma Node_set_L_includes_Node_set : 
"Node_set \<tau>  \<subseteq> Node_set_L \<tau>"
apply(simp add: Node_set_def Node_set_L_def)
apply(rule gfp_mono)
apply(simp add: Node_set_fun_def Node_set_fun_L_def)
apply auto
done

lemma Cnode_set_L_includes_Cnode_set : 
"Cnode_set \<tau>  \<subseteq> Cnode_set_L \<tau>"
apply(simp add: Cnode_set_def Cnode_set_L_def)
apply(rule gfp_mono)
apply(simp add: Cnode_set_fun_def Cnode_set_fun_L_def)
apply(simp add: Node_set_fun_def Node_set_fun_L_def)
apply auto
done




lemma Cnode_cast_id[simp]:
"self \<in> gfp (Cnode_set_fun \<tau>) \<Longrightarrow>
(self ->oclAsType( level1.tree.Node )) ->oclAsType( level1.tree.Cnode ) = self"
sorry

(* Relevant theorems from the universe theory are: *)
thm level1.tree.Node_2_Cnode_def
thm level1.tree.Cnode_2_Node_def
thm cast_Cnode_id
thm cast_Node_id

lemma Cnode_cast_inj[simp]:
"inj (\<lambda>x. x ->oclAsType( level1.tree.Cnode ))"
sorry


lemma image_cong:
"\<lbrakk>!!x . x \<in> S \<Longrightarrow> f x = g x\<rbrakk> \<Longrightarrow> f ` S = g ` S"
by (auto simp: image_def)


lemma huxx_left:
"selfa \<in> gfp (Cnode_set_fun \<tau>) \<Longrightarrow>
 (level1.tree.Node.left (selfa ->oclAsType( level1.tree.Node ))
               \<in> (\<lambda>self. self ->oclAsType( level1.tree.Node ))` gfp (Cnode_set_fun \<tau>))  =
 ((level1.tree.Node.left (selfa ->oclAsType( level1.tree.Node )) ->oclAsType( level1.tree.Cnode))
               \<in> gfp (Cnode_set_fun \<tau>))"
apply auto
apply (rotate_tac 1)
apply (thin_tac "?X")
apply(subst (asm) gfp_unfold [OF Cnode_set_fun_mono])
apply(subst (asm) Cnode_set_fun_def)
apply(simp)
apply (auto)
apply (subst Fun.inj_image_mem_iff[symmetric, 
                                   of "%x. x ->oclAsType( level1.tree.Cnode )"]) 
apply(simp_all add: image_compose[symmetric] comp_def
           cong: image_cong)
done


lemma huxx_right:
"selfa \<in> gfp (Cnode_set_fun \<tau>) \<Longrightarrow>
 (level1.tree.Node.right (selfa ->oclAsType( level1.tree.Node ))
               \<in> (\<lambda>self. self ->oclAsType( level1.tree.Node ))` gfp (Cnode_set_fun \<tau>))  =
 ((level1.tree.Node.right (selfa ->oclAsType( level1.tree.Node )) ->oclAsType( level1.tree.Cnode))
               \<in> gfp (Cnode_set_fun \<tau>))"
apply auto
apply (rotate_tac 1)
apply (thin_tac "?X")
apply(subst (asm) gfp_unfold [OF Cnode_set_fun_mono])
apply(subst (asm) Cnode_set_fun_def)
apply(simp)
apply (auto)
apply (subst Fun.inj_image_mem_iff[symmetric, 
                                   of "%x. x ->oclAsType( level1.tree.Cnode )"]) 
apply(simp_all add: image_compose[symmetric] comp_def
           cong: image_cong)
done



lemma Node_set_includes_Cnode_set : 
"((\<lambda> self. self ->oclAsType( level1.tree.Node)) ` (Cnode_set \<tau>))  \<subseteq> Node_set \<tau>"
apply(simp add: Cnode_set_def Node_set_def)
apply auto
apply(rule_tac X="(\<lambda> self. self ->oclAsType( level1.tree.Node)) ` (gfp (Cnode_set_fun \<tau>))" 
               in coinduct [OF Node_set_fun_mono]) 
apply(auto) 
apply(subst Node_set_fun_def)
apply simp
apply(thin_tac "self \<in> gfp (Cnode_set_fun \<tau>)")
apply(frule huxx_left)
apply(frule huxx_right)
apply(subst (asm) gfp_unfold [OF Cnode_set_fun_mono])
apply(subst (asm) Cnode_set_fun_def)
apply(subst (asm) Node_set_fun_def)
apply(simp)
done



lemma Node_inv_0 :  "(self \<in> (Node_set \<tau>)) = 
    ((\<tau> \<MathOclValid> \<MathOclIsDefined> level1.tree.Node.i self) \<and>
     (\<tau> \<MathOclValid> invariant1.inv_3 self) \<and>
     (\<tau> \<MathOclValid> \<MathOclIsDefined> level1.tree.Node.left self) \<and>
     (level1.tree.Node.left self \<in> (Node_set \<tau>)) \<and>
     (\<tau> \<MathOclValid> \<MathOclIsDefined> level1.tree.Node.right self) \<and>
     level1.tree.Node.right self \<in> (Node_set \<tau>))"

apply(subst Node_set_def)
apply(subst gfp_unfold[OF Node_set_fun_mono])
apply(subst Node_set_fun_def)
apply(simp add:  Node_set_def)
done

lemma Cnode_inv_0 :  
      "(self \<in> (Cnode_set \<tau>)) = 
        ( \<tau> \<MathOclValid> \<MathOclIsDefined> (level1.tree.Node.i (self ->oclAsType(level1.tree.Node))) 
                 \<and>  (\<tau> \<MathOclValid> invariant1.inv_3 (self ->oclAsType(level1.tree.Node))) \<and> 
         (\<tau> \<MathOclValid> \<MathOclIsDefined> level1.tree.Cnode.color self) \<and>
         (\<tau> \<MathOclValid> \<MathOclIsDefined> level1.tree.Node.i (self ->oclAsType( level1.tree.Node ))) \<and>
         (\<tau> \<MathOclValid> invariant1.inv_3 (self ->oclAsType( level1.tree.Node ))) \<and>
         (\<tau> \<MathOclValid> \<MathOclIsDefined> level1.tree.Node.left
            (self ->oclAsType( level1.tree.Node ))
                ->oclAsType( level1.tree.Cnode )) \<and>
         (\<tau> \<MathOclValid> \<MathOclIsDefined> level1.tree.Node.left
            (self ->oclAsType( level1.tree.Node ))) \<and>
         (level1.tree.Node.left
             (self ->oclAsType( level1.tree.Node ))
                  ->oclAsType( level1.tree.Cnode )
                 \<in> (Cnode_set \<tau>)) \<and>
   (\<tau> \<MathOclValid> \<MathOclIsDefined> level1.tree.Node.right
            (self ->oclAsType( level1.tree.Node ))) \<and>
         (\<tau> \<MathOclValid> \<MathOclIsDefined> level1.tree.Node.right
            (self ->oclAsType( level1.tree.Node )) 
                   ->oclAsType( level1.tree.Cnode )) \<and>
         (level1.tree.Node.right
                  (self ->oclAsType( level1.tree.Node )) 
                      ->oclAsType( level1.tree.Cnode )
                \<in> (Cnode_set \<tau>)) \<and>
         (\<tau> \<MathOclValid> \<MathOclIsDefined> level1.tree.Cnode.color self) \<and>
         (\<tau> \<MathOclValid> \<MathOclIsDefined> level1.tree.Node.right
            (self ->oclAsType( level1.tree.Node ))
                  ->oclAsType( level1.tree.Cnode )) \<and>
         (\<not> \<tau> \<MathOclValid> level1.tree.Cnode.color
            (level1.tree.Node.right
              (self ->oclAsType( level1.tree.Node )) 
                  ->oclAsType( level1.tree.Cnode ))
            \<MathOclStrongEq> level1.tree.Cnode.color self))"
apply(subst Cnode_set_def)
apply(subst gfp_unfold[OF Cnode_set_fun_mono])
apply(subst Cnode_set_fun_def)

apply(simp add:  Cnode_set_def Node_set_fun_def)
apply(auto)
apply (subst Fun.inj_image_mem_iff[symmetric, 
                                   of "%x. x ->oclAsType( level1.tree.Cnode )"])
apply(simp_all add: image_compose[symmetric] comp_def
           cong: image_cong)

apply (subst Fun.inj_image_mem_iff[symmetric, 
                                   of "%x. x ->oclAsType( level1.tree.Cnode )"],simp)
 

apply(simp_all add: image_compose[symmetric] comp_def
           cong: image_cong)
done

(* Generelle Funktion von "invariant1":
   Akkumulation der constraints auf den Attributen.
   Eigentlich mte auch die Definiertheit *aller*
   Attribute bei uns so erfat werden !!!
   Sonst gelingt nicht der Nachweis, da Node_set \<tau> kein undef
   enthlt !!! *) 
  

(* Bemerkung: Paulson schlgt fr gegenseitige
   Rekursion die Summenkonstruktion vor.
   Das geht bei uns genauso - nur nehmen wir
   gleich OclAny. Siehe Paulsons Papier.
 *)

(* Bemerkung: Wenn man berall die Defi niertheit hat,
   distribuiert OclAnd ber HOLAnd und man kann sogar 
   die Aussage als UC darstellen. *) 

(* Nun wrappen wir das Ganze auf level2: *)

constdefs Node' :: "('a,'b,'c) U_tree St \<Rightarrow> (('a,'b) Node) Set_0"
         "Node' \<equiv> \<lambda>\<tau>. Abs_Set_0 \<HolOclLiftLeft>(\<lambda> f. f \<tau>) ` (Node_set \<tau>)\<HolOclLiftRight>"

constdefs Cnode' :: "('a,'b,'c) U_tree St \<Rightarrow> (('a) Cnode) Set_0"
         "Cnode' \<equiv> \<lambda>\<tau>. Abs_Set_0 \<HolOclLiftLeft>(\<lambda> f. f \<tau>) ` (Cnode_set \<tau>)\<HolOclLiftRight>"


constdefs Node'' ::  "('a,'b,'c) U_tree St \<Rightarrow> (('a,'b) Node) Set_0"
         "Node'' \<equiv> lift0(Abs_Set_0 \<HolOclLiftLeft>\<Union> (\<tau>::('a,'b,'c) U_tree St). 
                                          (\<lambda> f. f \<tau>) ` (Node_set \<tau>) -
                                   {\<bottom>} \<HolOclLiftRight>)"

constdefs Cnode'' ::  "('a,'b,'c) U_tree St \<Rightarrow> (('a) Cnode) Set_0"
         "Cnode'' \<equiv> lift0(Abs_Set_0 \<HolOclLiftLeft>\<Union> (\<tau>::('a,'b,'c) U_tree St). 
                                          (\<lambda> f. f \<tau>) ` (Cnode_set \<tau>)  -
                                   {\<bottom>}\<HolOclLiftRight>)"

lemma aux : "(\<tau> \<MathOclValid> \<MathOclIsUndefined> x) = (down = x \<tau>)"
by(auto simp: localValidUnDefined2sem)

lemma Node''_defined[simp]: " \<MathOclIsDefined> Node'' = \<MathOclTrue>"
apply(rule ext)
apply(simp add: Node''_def OclIsDefined_def OclTrue_def
                lift0_def lift1_def DEF_def)
done

lemma Node''_defined[simp]: " \<MathOclIsDefined> Cnode'' = \<MathOclTrue>"
apply(rule ext)
apply(simp add: Cnode''_def OclIsDefined_def OclTrue_def
                lift0_def lift1_def DEF_def)
done


(* does not work, but I guess no longer relevant. 

lemma Node_defined[simp]: " \<MathOclIsDefined> Node' = \<MathOclTrue>"
apply (rule ext)
apply(simp add: Node'_def OclIsDefined_def OclTrue_def
                lift0_def lift1_def DEF_def)
apply(simp add:DEF_def[symmetric])
apply(rule DEF_Abs_Set)
apply(simp add: Node_set_def)
apply(subst gfp_unfold[OF Node_set_fun_mono])
apply(subst Node_set_fun_def)
apply(simp add: image_def aux)
sorry


lemma Cnode_defined[simp]: " \<MathOclIsDefined> Cnode' = \<MathOclTrue>"
apply (rule ext)
apply(simp add: Cnode'_def OclIsDefined_def OclTrue_def
                lift0_def lift1_def DEF_def)
apply(simp add:DEF_def[symmetric])
apply(rule DEF_Abs_Set)
apply(simp add: Cnode_set_def)
apply(subst gfp_unfold[OF Cnode_set_fun_mono])
apply(subst Cnode_set_fun_def)
apply(simp add: image_def aux)
sorry
 does not work, but I guess no longer relevant. *)



(* mit der Konsequenz: *)
lemma Node_inv_2 :
assumes self_def : " \<tau> \<MathOclValid> \<MathOclIsDefined> self "
and     consist  : "\<tau> \<MathOclValid> \<MathOclIsDefined> self \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> (invariant1.inv_3 self)"
shows   "(self \<MathOclIncludes> Node') \<tau> =  
             (  (invariant1.inv_3 self)
              \<MathOclAnd> (\<MathOclIsUndefined>(level1.tree.Node.left  self) \<MathOclOr> (level1.tree.Node.left  self)\<MathOclIncludes>Node') 
              \<MathOclAnd> (\<MathOclIsUndefined>(level1.tree.Node.right self) \<MathOclOr> (level1.tree.Node.right self)\<MathOclIncludes>Node')) \<tau> "
apply(insert self_def, insert consist[OF self_def])
apply(simp only: equiv_D_local, auto)
apply(drule includes_charn1)
apply(simp add: Node'_def)
apply(subgoal_tac "\<bottom> \<notin> (\<lambda>f. f \<tau>) ` Node_set \<tau>")
apply(subgoal_tac "(self \<tau> \<in> (\<lambda>f. f \<tau>) ` Node_set \<tau>) = (self \<in> Node_set \<tau>)")
apply simp
apply (subst (asm) Node_inv_0)
(*
apply (subst localValidAnd2conj, simp)
apply simp_all
prefer 3
apply (subst localValidAnd2conj, simp_all add: Node'_def[symmetric])
apply (simp_all add: strong_definedness)
apply safe
apply (subst localValidOr2disj, simp)
(* und hier wirds sinnlos weil das pattern dieser rule nicht ganz passt:
   weak definedness invariant \<dots> *)
apply(simp)
*)
sorry

(* Und nun definieren ganz einfach: *)

constdefs Node_inv :: "(('a,'b,'c) U_tree St \<Rightarrow> (('b,'c) Node)) \<Rightarrow> 
                       (('a,'b,'c) U_tree St) Boolean"
                      "Node_inv self \<equiv> self \<MathOclIncludes> Node'"

(* woraus sich dann  das theorem *)

lemma Node_inv_rec:
assumes self_def : " \<tau> \<MathOclValid> \<MathOclIsDefined> self "
and     consist  : "\<tau> \<MathOclValid> \<MathOclIsDefined> self \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> (invariant1.inv_3 self)"
shows"  (Node_inv self)                              \<tau> = 
        (  invariant1.inv_3 self
         \<MathOclAnd> (\<MathOclIsUndefined> level1.tree.Node.left  self \<MathOclOr> Node_inv (level1.tree.Node.left self)) 
         \<MathOclAnd> (\<MathOclIsUndefined> level1.tree.Node.right self \<MathOclOr> Node_inv (level1.tree.Node.right self))) \<tau> "
apply(insert self_def, insert consist[OF self_def])
apply(simp add: Node_inv_def)
apply(rule trans)
apply(erule Node_inv_2, auto)
done

(* ergibt. *)



(* Verbleibende Grosse Frage: Zu welchem Zweck und Ende braucht man 
   co-induction ? *)

lemma def_coinductZ:
     "[| A==gfp(f);  mono(f);  X \<subseteq> f(X Un A) |] ==> X \<subseteq> A"
apply(auto intro!: coinduct)
done



lemma " \<tau> \<MathOclValid> ((self \<MathOclIncludes> Cnode' ) \<MathOclImplies> ( (self ->oclAsType(level1.tree.Node)) \<MathOclIncludes> Node' ))"
sorry 


end
