(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * stack.thy - 
 * Copyright (C) 2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This File is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)


header {* Encoding a Stack in \UML/\OCL *}

theory stack 
imports
  OCL
begin

text {*
  We use the class diagram presented in Fig.~\ref{fig:stack_classdiag} together with the 
  following \OCL formulae shown in \autoref{lst:stacl}.

\lstinputlisting[style=displayocl,float,caption={Modelling a Stack: \OCL specification},
                 label={lst:stacl}]{stack.ocl}

\begin{figure}
\begin{center}
\includegraphics[scale=.6]{figures/stack}
\end{center}
\caption{Modelling a Stack: Data Model\label{fig:stack_classdiag}}
\end{figure}

*}


load_xmi "stack_ocl.xmi"

text {* An intersting system property would be *}
lemma "\<lbrakk>  \<MathOclValid>  \<MathOclIsDefined> s0 \<rbrakk> \<Longrightarrow> \<MathOclValid>  (push s0 e0 s1) \<MathOclImplies>  ((top s1 e1) \<MathOclAnd> (e1 \<MathOclStrongEq> e0 )) "
oops
text {* or *}
lemma "\<lbrakk>  \<MathOclValid>  \<MathOclIsDefined> s0 \<rbrakk> \<Longrightarrow>  size (s0) \<ge>  size (pop s0)"
oops

text {*
  Note that such system properties specified in \holocl allow 
  the use of operations that are not side-effect free (e.g. push) 
*}
end
