(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Let.thy - Definition of "let .. in .. end"
 * Copyright (C) 2005,2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* The \OCL let expression *}

theory OCL_Let 
imports 
  "$HOLOCL_HOME/src/Lifting" 
begin
 text_raw{* \label{thy:OCL_Let} *}
text {* 
  The \OCL let expression is a purely syntactical abbrevation 
  (see~\cite[p. A-26]{omg:ocl:2003}) therefore it can be directly 
  mapped to the \HOL let.
*}
constdefs
  OclLet   :: "[('a,'b)VAL, ('a,'b)VAL \<Rightarrow> ('a,'b\<Colon>bot)VAL] \<Rightarrow> ('a,'b)VAL"
 "OclLet    \<equiv>  Let"

syntax  (* standard syntax *)
    "_OclLet_std"        :: "[('a,'b)VAL,('a,'b)VAL \<Rightarrow> ('a,'b\<Colon>bot)VAL] \<Rightarrow> ('a,'b)VAL"
                           ("(\<OclLet> (_)/ \<OclIn> (_)/\<OclEnd>)" 15) 

syntax  (* ascii syntax *)
    "_OclLet_ascii"        :: "[('a,'b)VAL,('a,'b)VAL \<Rightarrow> ('a,'b\<Colon>bot)VAL] \<Rightarrow> ('a,'b)VAL"
                           ("(let (_)/ in (_)/end)" 15) 


syntax  (xsymbols)
    "_OclLet_math"       :: "[('a,'b\<Colon>bot)VAL] \<Rightarrow> ('a,'b)VAL"
                           ("(\<MathOclLet> (_)/ \<MathOclIn> (_)/\<MathOclEnd>)" 15) 

(*<*)
ML {* fun flat (ls: 'c list list) : 'c list = foldr (op @) [] ls; *}
parse_translation {* flat(map (emb trans_const) ["OclLet"]) *}
print_translation {* map emb_print ["OclLet"] *}
(*>*)
end
