(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL into Isabelle/HOL              
 *                                                                            
 * ROOT.ML - main "ROOT.ML" file for HOL-OCL
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)


val ord = Char.ord;
val chr = Char.chr;
val print = TextIO.print ;

 
 structure TextIO =
 struct
   open TextIO;
    
      fun inputLine is =
          (case TextIO.inputLine is of
        "" => NONE
	    | s => SOME s)
        handle IO.Io _ => raise Interrupt;
end;
(*
 TextIO.inputLine
*)

nonfix prefix;

val build_home = pwd();

use "config.sml";


val _ = if  config.xmi_support andalso (String.isSubstring "polyml" ml_system) 
        then (use "../lib/su4sml/src/contrib/HashTable.sml")
        else ();

val _ = if config.xmi_support 
	then (cd "../lib/su4sml/src"; use "ROOT.ML"; cd "../../../src")
        else ();

fun flat (ls: 'c list list) : 'c list = foldr (op @) [] ls;
infix prefix;

val holocl_version = Int.toString(holocl_ver_major)^"."^Int.toString(holocl_ver_minor)^"."
                        ^Int.toString(holocl_ver_micro)^holocl_ver_tag

val isabelle_version = version
val version = "HOL-OCL "^holocl_version^" , based on "^version;


fun holocl_home () = if (getenv "HOLOCL_HOME") <> "" 
		  then getenv "HOLOCL_HOME"
		  else ".";

fun infostr _ =
let
   val collectiontype = case config.collectiontype of
                     config.smashed   => "smashed collection types"
                   | config.unsmashed => "unsmashed collection types"
   val universe = case config.universe of
                       config.referential    => "referential universe"
                     | config.nonreferential => "nonreferential universe"
   val xmi = "XMI import "^(if config.xmi_support then "" else " not ")^"supported";
in
   ("\n")^
   ("             HOL-OCL "^holocl_version^"     \n")^
   ("       a shallow embedding of OCL into Isabelle/HOL           \n")^
   ("        http://www.brucker.ch/research/holocl/                \n")^
   (" Copyright (c) 2000-2006 Achim D. Brucker and Burkhart Wolff  \n")^
   ("                                                              \n")^
   (" Configuration:                                               \n")^
   (" - "^collectiontype^"\n")^
   (" - "^universe^"\n")^
   (" - "^xmi^"\n")^
   (" - HOLOCL_HOME="^(holocl_home ())^"\n" )^
   (" - Isabelle: "^(isabelle_version)^"\n" )

end;


fun info _ = writeln (infostr())

fun writeInfoFile fname  = File.write (Path.unpack fname) (infostr());


print_depth 0;
val goals_limit=0;

no_document use_thy "Accessible_Part";
no_document use_thy "Multiset";

(*********************************************************************)
(* loading fundamental OCL theory                                    *)
(*********************************************************************)

use_thy "OCL_Logic";

(*********************************************************************)
(* setup thy_morpher                                                 *)
(*********************************************************************)
(*
cd "prover";
use "thy_morpher.ML";
cd "..";
*)
(*
(* start thy-morpher *)
Thy_morpher.thy_morpher (theory "OCL");
*)

(*********************************************************************)
(* loading OCL theory                                                *)
(*********************************************************************)

use_thy "OCL";

(*********************************************************************)
(* loading class loader                                              *)
(*********************************************************************)
cd build_home;
val _ = if config.xmi_support 
	then let 
	val _ = writeln (" loading rep encoder")
	in (cd "rep_encoder/"; use "root.sml"; cd "../") end
        else ();
(*********************************************************************)
(* misc                                                              *)
(*********************************************************************)
ProofGeneral.write_keywords "HOL-OCL";
print_depth 10;

