(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_StateAccess.thy - State accessing operations as defined in the standard.
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* OclAny *}

theory OCL_OclAny 
imports 
    "$HOLOCL_HOME/src/library/collection/$COLLECTION/OCL_Set"
begin
text_raw{* \label{thy:OCL_OclAny} *}

subsection {* Level 0 *}
setup {* [Theory.add_path "UML_OCL.level0"] *}

constdefs
  mk_OclAny       :: "'\<alpha> OclAny_0 \<Rightarrow> '\<alpha> U"
  "mk_OclAny      \<equiv> Inl" 
  get_OclAny      :: "'\<alpha> U \<Rightarrow> '\<alpha> OclAny_0"
  "get_OclAny     \<equiv>  sum_case (\<lambda>x. x) (\<lambda>x. \<some> x. True)"
  is_OclAny       ::  "'\<alpha> OclAny_0 \<Rightarrow> bool"
  "is_OclAny obj  \<equiv> DEF(obj)"
  is_OclAny_univ  :: "'\<alpha> U \<Rightarrow> bool"
  "is_OclAny_univ \<equiv>  sum_case (\<lambda>x. True) (\<lambda>x. False)"

lemma get_mk_OclAny_id: "get_OclAny(mk_OclAny x) = x"
  by (simp add: get_OclAny_def mk_OclAny_def)

lemma mk_get_OclAny_id: "is_OclAny_univ x \<Longrightarrow> mk_OclAny(get_OclAny x) = x"
   apply (simp add: get_OclAny_def mk_OclAny_def is_OclAny_univ_def)
  apply (case_tac x, simp, simp)   
  done
 
lemma is_OclAny_univ_implies_is_get: "\<lbrakk>DEF(get_OclAny x); is_OclAny_univ x \<rbrakk> 
  \<Longrightarrow> is_OclAny(get_OclAny x)"
  apply (simp add: get_OclAny_def mk_OclAny_def is_OclAny_univ_def is_OclAny_def)
  done


lemma is_mk_OclAny:
  "\<lbrakk>  is_OclAny obj \<rbrakk> \<Longrightarrow> is_OclAny_univ (mk_OclAny obj)" 
  apply (simp add: is_OclAny_univ_def  get_OclAny_def is_OclAny_def 
                   mk_OclAny_def)
  done

setup {* [Theory.add_path "/"] *}


subsection {* Level 1 *}
setup {* [Theory.add_path "UML_OCL.level1"] *}

constdefs
  mk_OclAny       :: "('\<tau> , '\<alpha> OclAny_0)VAL \<Rightarrow> ('\<tau> , '\<alpha> U) VAL"
  "mk_OclAny      \<equiv> lift1 level0.mk_OclAny"
  get_OclAny      :: "('\<tau> , '\<alpha> U)VAL \<Rightarrow> ('\<tau> , '\<alpha> OclAny_0) VAL"
  "get_OclAny     \<equiv> lift\<^sub>1 level0.get_OclAny"
  is_OclAny       ::   "('\<tau> ,'\<alpha> OclAny_0) VAL   \<Rightarrow>  ('\<tau>)Boolean"
  "is_OclAny  \<equiv> lift\<^sub>1(strictify (lift \<circ> level0.is_OclAny))"
  is_OclAny_univ  :: "('\<tau> ,'\<alpha> U ) VAL   \<Rightarrow>  ('\<tau>)Boolean"
  "is_OclAny_univ \<equiv>  lift\<^sub>1(lift \<circ> level0.is_OclAny_univ)"
    OclAny  :: "('\<tau>, '\<alpha> OclAny_0 Set_0) VAL"
  "OclAny  \<equiv> lift0(Abs_Set_0 \<HolOclLiftLeft>lift ` UNIV\<HolOclLiftRight>)"

lemma get_mk_OclAny_id :
  "\<tau>\<MathOclValid> level1.is_OclAny obj \<Longrightarrow>  level1.get_OclAny (level1.mk_OclAny obj) = obj"
  apply (unfold level1.is_OclAny_univ_def level1.is_OclAny_def level1.get_OclAny_def 
    level1.mk_OclAny_def OclLocalValid_def)
  apply(rule ext)
  apply(simp add: lift1_def level0.get_mk_OclAny_id)
done

lemma mk_get_OclAny_id :
  "\<tau>\<MathOclValid> level1.is_OclAny_univ univ  
   \<Longrightarrow> level1.mk_OclAny (level1.get_OclAny univ) \<tau>= univ \<tau>" 
  apply (frule isDefined_if_valid)
  apply (frule  localValidDefined2sem [THEN iffD1])
  apply (unfold level1.is_OclAny_univ_def level1.is_OclAny_def level1.get_OclAny_def 
                level1.mk_OclAny_def OclLocalValid_def)
  apply(simp add:  OclLocalValid_def  lift1_def lift0_def level0.mk_get_OclAny_id OclTrue_def)
done

lemma  is_OclAny_univ_implies_is_get: 
  "\<lbrakk> DEF(level1.get_OclAny univ \<tau>); (\<tau>\<MathOclValid> level1.is_OclAny_univ univ) \<rbrakk> 
   \<Longrightarrow> (\<tau>\<MathOclValid>   level1.is_OclAny (level1.get_OclAny univ))"
  apply (frule isDefined_if_valid)
  apply (frule  localValidDefined2sem [THEN iffD1])
  apply (unfold level1.is_OclAny_univ_def level1.is_OclAny_def level1.get_OclAny_def 
    level1.mk_OclAny_def OclLocalValid_def)
  apply(simp add:  OclLocalValid_def   lift1_def lift0_def level0.is_OclAny_univ_implies_is_get 
                   OclTrue_def strictify_def DEF_def )
done

lemma is_mk_OclAny:
  "\<tau>\<MathOclValid> level1.is_OclAny obj 
   \<Longrightarrow> (\<tau>\<MathOclValid>  level1.is_OclAny_univ (level1.mk_OclAny obj))"
  apply (frule isDefined_if_valid)
  apply (frule  localValidDefined2sem [THEN iffD1])
  apply (unfold level1.is_OclAny_univ_def level1.is_OclAny_def level1.get_OclAny_def 
    level1.mk_OclAny_def OclLocalValid_def)
  apply(simp add:  OclLocalValid_def   lift1_def lift0_def OclTrue_def strictify_def DEF_def  )
  apply(case_tac "obj \<tau>= down")
  apply(auto simp: level0.is_mk_OclAny)
done


setup {* [Theory.add_path "/"] *}

subsection {* Level 2 *}
setup {* [Theory.add_path "UML_OCL"] *}

constdefs
  mk_OclAny       :: "('\<tau> , '\<alpha> OclAny_0)VAL \<Rightarrow> ('\<tau> , '\<alpha> U) VAL"
  "mk_OclAny      \<equiv> level1.mk_OclAny"
  get_OclAny      :: "('\<tau> , '\<alpha> U)VAL \<Rightarrow> ('\<tau> , '\<alpha> OclAny_0) VAL"
  "get_OclAny     \<equiv> level1.get_OclAny"
  is_OclAny       ::   "('\<tau> ,'\<alpha> OclAny_0) VAL   \<Rightarrow>  ('\<tau>)Boolean"
  "is_OclAny  \<equiv>  level1.is_OclAny"
  is_OclAny_univ  :: "('\<tau> ,'\<beta> U ) VAL   \<Rightarrow>  ('\<tau>) Boolean"
  "is_OclAny_univ \<equiv>  level1.is_OclAny_univ"
  OclAny  :: "('\<tau>, '\<alpha> OclAny_0 Set_0) VAL"
  "OclAny  \<equiv> level1.OclAny"

lemma get_mk_OclAny_id :
  "\<tau>\<MathOclValid> UML_OCL.is_OclAny obj 
   \<Longrightarrow>  UML_OCL.get_OclAny (UML_OCL.mk_OclAny obj) = obj"
  apply (unfold UML_OCL.is_OclAny_def UML_OCL.get_OclAny_def UML_OCL.mk_OclAny_def)
  apply(simp add: level1.get_mk_OclAny_id)
done

lemma mk_get_OclAny_id :
  "\<tau>\<MathOclValid> UML_OCL.is_OclAny_univ univ  
   \<Longrightarrow> UML_OCL.mk_OclAny (UML_OCL.get_OclAny univ) \<tau>= univ \<tau>" 
  apply (unfold UML_OCL.is_OclAny_def UML_OCL.get_OclAny_def UML_OCL.mk_OclAny_def 
         UML_OCL.is_OclAny_univ_def)
  apply(simp add: level1.mk_get_OclAny_id)
done
 
lemma  is_OclAny_univ_implies_is_get: 
  "\<lbrakk> DEF(UML_OCL.get_OclAny univ \<tau>); (\<tau>\<MathOclValid> UML_OCL.is_OclAny_univ univ) \<rbrakk> 
   \<Longrightarrow> (\<tau>\<MathOclValid>   UML_OCL.is_OclAny (UML_OCL.get_OclAny univ))"
  apply (unfold UML_OCL.is_OclAny_def UML_OCL.get_OclAny_def UML_OCL.mk_OclAny_def 
                UML_OCL.is_OclAny_univ_def)
  apply(simp add: level1.is_OclAny_univ_implies_is_get)
done
 
lemma is_mk_OclAny:
  "\<tau>\<MathOclValid> UML_OCL.is_OclAny obj 
   \<Longrightarrow> (\<tau>\<MathOclValid>  UML_OCL.is_OclAny_univ (UML_OCL.mk_OclAny obj))"
  apply (unfold UML_OCL.is_OclAny_def UML_OCL.get_OclAny_def UML_OCL.mk_OclAny_def 
                UML_OCL.is_OclAny_univ_def)
  apply(simp add: level1.is_mk_OclAny)
done
 
setup {* [Theory.add_path "/"] *}

section{* Re-typing or Casting.*}
text{* 
   The standard describes an operation
   that ``casts'' an object to one of its subtypes.
   Note that our OCL-encoder provides for each new class
   an own definition for this polymorphic scheme. 
*}
consts
   OclAsType ::"[('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<beta> Set_0)VAL] \<Rightarrow> ('\<tau>,'\<beta>::bot)VAL" 
                                    ("_ ->oclAsType'( _ ')" [66,65]65) 



section{* Checking Dynamic Types of ``Kinds'' *}
text{* As a consequence of the difference between dynamic and
  static typing in object-oriented languages, 
  a test for the ``kind'' of an object is provided.
  It checks for the dynamic type 

  Note that this operator is in fact defined by our \OCL encoder
  and uses the internally defined test functions. *}
constdefs
  OclIsKindOf :: "[('\<tau>,'\<alpha>::bot) VAL, ('\<tau>,'\<beta>::collection)VAL] \<Rightarrow> '\<tau> Boolean"
                                           ("_ ->oclIsKindOf'( _ ')" [66,65]65) 
 "OclIsKindOf self CharSet \<equiv> self \<MathOclIncludes>  CharSet"
  
text{* This simple fundamental definition is the fruit of our 
  generalization of the \OCL set theory. *}
constdefs
  OclIsTypeOf :: "[('\<tau>, '\<alpha>::bot) VAL, ('\<tau>, '\<beta>::collection) VAL] => '\<tau> Boolean"
	("_ ->oclIsTypeOf'( _ ')" [66,65]65)
  "OclIsTypeOf self NarrowCharSet \<equiv> self \<MathOclIncludes> NarrowCharSet"

end

