(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Boolean.thy - Semantics of Boolean expressions
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* \OCL Boolean *}

theory OCL_Boolean 
  imports 
    "$HOLOCL_HOME/src/OCL_Universe" 
    "$HOLOCL_HOME/src/OCL_Let"
begin 
text_raw{* \label{thy:OCL_Boolean} *}
section {* The Characteristic Set of the Type Boolean *}

text {* Case distinction over a boolean *}
lemma boolean_cases_sem:
  assumes undefC: "x = \<bottom> \<Longrightarrow> R"
  and     trueC:  "x = \<HolOclLiftLeft>True\<HolOclLiftRight>  \<Longrightarrow> R"
  and     falseC: "x = \<HolOclLiftLeft>False\<HolOclLiftRight>  \<Longrightarrow> R"
  shows           "R"
  apply(rule disjE)
  prefer 2 apply(erule undefC)
  prefer 2 apply(rule disjE)
  prefer 2 apply(rotate_tac -1, erule trueC)
  prefer 2 apply(rotate_tac -1, erule falseC)
  apply(auto simp: not_down_exists_lift)
  done


constdefs
  is_Boolean_0    :: "'\<alpha> U \<Rightarrow> bool"
  "is_Boolean_0   \<equiv> sum_case (\<lambda>x. False) 
                              (sum_case (\<lambda>x. False) 
                              (sum_case (\<lambda>x. False)
                              (sum_case (\<lambda>x. True)(\<lambda>x. False))))"
  get_Boolean_0   :: "'\<alpha> U \<Rightarrow> Boolean_0"
  "get_Boolean_0  \<equiv> sum_case (\<lambda>x. \<some> x. True)
                      (sum_case (\<lambda>x. \<some> x. True)
                      (sum_case (\<lambda>x. \<some> x. True)
                      (sum_case (\<lambda>x. x)(\<lambda>x. \<some> x. True))))"
  mk_Boolean_0    :: "Boolean_0 \<Rightarrow> '\<alpha> U"
  "mk_Boolean_0   \<equiv> Inr \<circ> Inr \<circ> Inr \<circ> Inl"

lemma get_mk_Boolean_id_0: "get_Boolean_0(mk_Boolean_0 x) = x"
  apply (simp add: get_Boolean_0_def mk_Boolean_0_def)
  done
lemma is_mk_Boolean_0: "is_Boolean_0(mk_Boolean_0 x) = True"
  apply (simp add: is_Boolean_0_def mk_Boolean_0_def)
  done
lemma mk_get_Boolean_id_0: "is_Boolean_0 x \<Longrightarrow> mk_Boolean_0(get_Boolean_0 x) = x"
  apply (simp add: get_Boolean_0_def mk_Boolean_0_def is_Boolean_0_def )
  apply (case_tac x, simp, simp)
  apply (case_tac b, simp, simp)
  apply (case_tac ba, simp, simp)
  apply (case_tac bb, simp, simp)
  done

(*<*)
ML{* OCL_strict_syntax_checking:=true; 
     OCL_syntax_mode:=math; *}
(*>*)

constdefs
  Boolean_0 :: "('\<tau>, Boolean_0 Set_0) VAL"
  "Boolean_0 \<equiv> lift\<^sub>0(Abs_Set_0 \<HolOclLiftLeft>lift ` UNIV\<HolOclLiftRight>)"


text {*
  Semantics of Boolean expressions, i.e., the logical operators of the
  language. They build a Kleene-Logic\index{Kleene-Logic},i.e., the usual 
  algebraic laws hold. Moreover, all operators are context passing\index{context
  passing}. This results in proof support based on extensive ternary
  case distinctions ($P(X)$ holds iff $P(\isasymMathOclUndefined)$ 
  and $P(\isasymMathOclTrue)$ and $P(\isasymMathOclFalse)$.
  *}


section {* \OCL Boolean expressions *}

subsection {* The Core of the Language *} 
text {* 
  These operators are defined via lifting principles from \HOL
  (see~\cite{brucker.ea:embedding:2003}). A more textbook-like
  ``definition'' of \OCL (following the standard) is derived in the
  theorem section.
*}

constdefs (* the definedness predicate is a non-strict operation *)          
  OclIsDefined   :: "('\<tau>,'b\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"
 "OclIsDefined   \<equiv> lift\<^sub>1(lift \<circ> DEF)" 

  OclTrue        :: "'\<tau> Boolean" 
 "OclTrue        \<equiv> lift\<^sub>0(\<HolOclLiftLeft>True\<HolOclLiftRight>)"

  OclFalse       :: "'\<tau> Boolean" 
 "OclFalse       \<equiv> lift\<^sub>0(\<HolOclLiftLeft>False\<HolOclLiftRight>)"

  OclNot         :: "'\<tau> Boolean \<Rightarrow> '\<tau> Boolean" 
 "OclNot         \<equiv> lift\<^sub>1(strictify(lift \<circ> Not \<circ> drop))"


constdefs

  (* and is a non-strict operation *)
  OclAnd   :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
 "OclAnd   \<equiv> lift\<^sub>2(\<lambda>x y. if DEF x 
                  then if DEF y then \<HolOclLiftLeft>(\<HolOclDropLeft>x\<HolOclDropRight>) \<and> (\<HolOclDropLeft>y\<HolOclDropRight>)\<HolOclLiftRight>
                       else (if \<HolOclDropLeft>x\<HolOclDropRight> then \<bottom> else \<HolOclLiftLeft>False\<HolOclLiftRight>)
                  else if DEF y 
                       then (if \<HolOclDropLeft>y\<HolOclDropRight> then \<bottom> else \<HolOclLiftLeft>False\<HolOclLiftRight>)
                       else \<bottom>)"
           (* exception for if : only one "usual" syntax. *)
  OclIf    :: "['\<tau> Boolean, ('\<tau>,'\<alpha>\<Colon>bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> ('\<tau>,'\<alpha>)VAL"
 "OclIf    \<equiv> lift\<^sub>3(\<lambda>x y z. strictify(\<lambda>a. if \<HolOclDropLeft>a\<HolOclDropRight> then y else z) x)"


 (* strict version *)
  OclSand   :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
 "OclSand   \<equiv>  lift\<^sub>2(strictify(\<lambda> x. strictify(\<lambda> y. 
                          \<HolOclLiftLeft>( \<HolOclDropLeft>x\<HolOclDropRight> ) \<and> ( \<HolOclDropLeft>y\<HolOclDropRight> )\<HolOclLiftRight> )))"


text{* The following statements represent the technical setup of
   Isabelle's syntax engine for this core-language. All these
   technical setups can be ignored in first reasing. *}

syntax  (* standard syntax *)
    "_OclIsDefined_std" :: "('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"
                           ("\<OclIsDefined> (_)" [60]60)
    "_OclIsUndefined_std"::"('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"
                           ("\<OclIsUndefined> (_)" [60]60)
    "_OclTrue_std"      :: "'\<tau> Boolean"  
                           ("\<OclTrue>")
    "_OclFalse_std"     :: "'\<tau> Boolean"
                           ("\<OclFalse>")
    "_OclNot_std"       :: "'\<tau> Boolean \<Rightarrow> '\<tau> Boolean"
                           ("\<OclNot> (_)" [59]59)
    "_OclAnd_std"       :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_  \<OclAnd>/ _)"[57,58]57)  
    "_OclIf_std"        :: "['\<tau> Boolean,('\<tau>,'b)VAL,('\<tau>,'b\<Colon>bot)VAL] \<Rightarrow> ('\<tau>,'b)VAL"
                           ("(\<OclIf> (_)/ \<OclThen> (_)/\<OclElse> (_)/\<OclEndif>)" 15) 
    "_OclSand_std"       :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_  \<OclSand>/ _)"[57,58]57)  
 

syntax  (* ascii syntax *)
    "_OclIsDefined_ascii" :: "('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"
                           ("isdef (_)" [60]60)
    "_OclIsUndefined_ascii":: "('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"
                           ("isundef (_)" [60]60)
    "_OclTrue_ascii"      :: "'\<tau> Boolean"  
                           ("TRUE")
    "_OclFalse_ascii"     :: "'\<tau> Boolean"
                           ("FALSE")
    "_OclNot_ascii"       :: "'\<tau> Boolean \<Rightarrow> '\<tau> Boolean"
                           ("not (_)" [59]59)

    "_OclAnd_ascii"       :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ and _)"[57,58]57)
    "_OclIf_ascii"        :: "['\<tau> Boolean,('\<tau>,'b)VAL,('\<tau>,'b\<Colon>bot)VAL] \<Rightarrow> ('\<tau>,'b)VAL"
                           ("(if (_)/ then (_)/else (_)/endif)" 15) 
    "_OclSand_ascii"       :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ sand _)"[57,58]57)

syntax  (xsymbols)
    "_OclIsDefined_math":: "('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"
                           ("\<MathOclIsDefined> (_)" [60]60)
    "_OclIsUndefined_math":: "('\<tau>,'\<alpha>\<Colon>bot) VAL \<Rightarrow> '\<tau> Boolean"
                           ("\<MathOclIsUndefined> (_)" [60]60)
    "_OclTrue_math"     :: "'\<tau> Boolean"  
                           ("\<MathOclTrue>")
    "_OclFalse_math"    :: "'\<tau> Boolean"
                           ("\<MathOclFalse>")
    "_OclNot_math"      :: "'\<tau> Boolean \<Rightarrow> '\<tau> Boolean"
                           ("\<MathOclNot> (_)" [59]59)
    "_OclAnd_math"      :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_  \<MathOclAnd>/ _)"[57,58]57)
    "_OclIf_math"       :: "['\<tau> Boolean,('\<tau>,'\<alpha>\<Colon>bot)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> ('\<tau>,'\<alpha>)VAL"
                           ("(\<MathOclIf> (_)/ \<MathOclThen> (_)/\<MathOclElse> (_)/\<MathOclEndif>)" 15) 
    "_OclSand_math"      :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_  \<MathOclSand>/ _)"[57,58]57)

(*<*)
ML {*
 fun flat (ls: 'c list list) : 'c list = foldr (op @) [] ls
*}
(*>*)

(*<*)
parse_translation {*
  let fun not_handler mode name (S::R)  =
          if !OCL_strict_syntax_checking andalso !OCL_syntax_mode <> mode 
          then error(String.concat["strict checking: No ",
                                   OCL_syntax2string mode,
                                   " syntax allowed here"])
          else list_comb(Syntax.const "OclNot" $ 
                          (Syntax.const "OclIsDefined" $ S),R) 
      
  in (emb not_handler "OclIsUndefined") @
     flat(map (emb trans_const) ["OclIsDefined","OclTrue","OclFalse",
                                 "OclNot","OclAnd","OclIf", "OclSand"])
  end;
*}

print_translation {*
  let fun Not_handler ((Const("OclIsDefined",_)$R)::S) = 
         (let val name = "OclIsUndefined"
              val name = ("_"^name^"_"^(OCL_syntax2string(!OCL_syntax_mode)))
          in list_comb(Syntax.const name $ R,S) end)
         |Not_handler S =
          let val name = "OclNot"
              val name = ("_"^name^"_"^(OCL_syntax2string(!OCL_syntax_mode)))
          in list_comb (Syntax.const name,S) end

  in  ("OclNot",Not_handler):: 
      map emb_print ["OclIsDefined","OclTrue","OclFalse",
                     "OclAnd","OclIf","OclSand"]
  end;
*}
(*>*)

subsection {* Is \OCL Boolean Faithful to the OCL Standard? *} 
text {*\label{Boolean:CombinatorStyleVsTextbookStyle}*}
text {* 
   In this section, we present background material
   and the proof work for the \autoref{sec:textb-vs.-comb}. 
*}

constdefs
  semfun  :: "'\<alpha> \<Rightarrow> '\<alpha>"  ("Sem [[(_)]]")
  "Sem[[x]] \<equiv> x"

text {* 
   We show for our first strict operation in \OCL, the
   \inlineocl{not} operator, that it is in fact an instance of the
   standards definition scheme: 
*}

lemma not_faithfully_represented:
  "Sem[[\<MathOclNot> X]] \<gamma>   = (if Sem[[X]] \<gamma>  \<noteq>  \<bottom>  
                     then \<HolOclLiftLeft>\<not> \<HolOclDropLeft> Sem[[X]]\<gamma> \<HolOclDropRight>\<HolOclLiftRight>  
                     else \<bottom>)"
  by(simp add:OclNot_def 
              DEF_def o_def 
              lift0_def lift1_def lift2_def
              semfun_def )

text {* 
  The proof is trivial and canonic: it consists
  of the unfolding of all combinator definitions
  (they are just abbreviations of re-occuring patterns
  in the textbook style definitions used to make
  these patterns explicit!) and the semantic function
  which is a mere syntactic marker in our context. 
*}


text {* 
  The definition of basic operations for the special
  case of boolean operations is done by a truth 
  table~\cite[table A 2, page A-12]{omg:ocl:2003}.
  We represent this table for the elementary "and" case
  as following: 
*}

lemma and_faithfully_represented:
  "Sem[[X  \<MathOclAnd>  Y]] \<gamma> = 
      (if      Sem[[X]] \<gamma> = \<HolOclLiftLeft>False\<HolOclLiftRight> \<and> Sem[[Y]] \<gamma> = \<HolOclLiftLeft>False\<HolOclLiftRight> then \<HolOclLiftLeft>False\<HolOclLiftRight>
       else if Sem[[X]] \<gamma> = \<HolOclLiftLeft>False\<HolOclLiftRight> \<and> Sem[[Y]] \<gamma> = \<HolOclLiftLeft>True\<HolOclLiftRight>  then \<HolOclLiftLeft>False\<HolOclLiftRight>
       else if Sem[[X]] \<gamma> = \<HolOclLiftLeft>True\<HolOclLiftRight>  \<and> Sem[[Y]] \<gamma> = \<HolOclLiftLeft>False\<HolOclLiftRight> then \<HolOclLiftLeft>False\<HolOclLiftRight>
       else if Sem[[X]] \<gamma> = \<HolOclLiftLeft>True\<HolOclLiftRight>  \<and> Sem[[Y]] \<gamma> = \<HolOclLiftLeft>True\<HolOclLiftRight>  then \<HolOclLiftLeft>True\<HolOclLiftRight>
       else if Sem[[X]] \<gamma> = \<HolOclLiftLeft>False\<HolOclLiftRight> \<and> Sem[[Y]] \<gamma> = \<bottom>      then \<HolOclLiftLeft>False\<HolOclLiftRight>
       else if Sem[[X]] \<gamma> = \<HolOclLiftLeft>True\<HolOclLiftRight>  \<and> Sem[[Y]] \<gamma> = \<bottom>      then \<bottom>
       else if Sem[[X]] \<gamma> = \<bottom>      \<and> Sem[[Y]] \<gamma> = \<HolOclLiftLeft>False\<HolOclLiftRight> then \<HolOclLiftLeft>False\<HolOclLiftRight>
       else if Sem[[X]] \<gamma> = \<bottom>      \<and> Sem[[Y]] \<gamma> = \<HolOclLiftLeft>True\<HolOclLiftRight>  then \<bottom>
       else if Sem[[X]] \<gamma> = \<bottom>      \<and> Sem[[Y]] \<gamma> = \<bottom>      then \<bottom>
       else \<bottom>)"
  apply(simp add:OclAnd_def 
              DEF_def o_def strictify_def
              lift0_def lift1_def lift2_def
              semfun_def )

  apply (cases "X \<gamma> = down")
  apply (cases "Y \<gamma> = down")
  apply (simp_all add: not_down_exists_lift )
  by  auto


text {*   
  For the operators \inlineocl{or}, \inlineocl{xor}, \inlineocl{and}
  \inlineocl{implies}, we refrain from an explicit proof of
  conformance with ~\cite[table A 2, page A-12]{omg:ocl:2003}.
  Instead, we preferred to define
  these operators as usual in terms of \inlineocl+and+ and
  \inlineocl+not+.  However, the subsequent lemmas
  @{text "OclUndefined__or__True"} represent another way to directly
  check that our definitions meet the standard.  
*}
  

subsection {* The Extended Core of the Language *}

constdefs
  OclOr           :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean"   
  "OclOr S T      \<equiv> \<MathOclNot> (\<MathOclNot> S \<MathOclAnd> \<MathOclNot> T)"

  OclSor           :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean"   
  "OclSor S T      \<equiv> \<MathOclNot> (\<MathOclNot> S \<MathOclSand> \<MathOclNot> T)"

text{* Alternative definitions for the use with \inlineholocl{ocl_setup_op}: *}
lemma OclOr_alt_def: "OclOr \<equiv> lift\<^sub>2 (\<lambda> x y.
      (if (DEF x)
       then (if (DEF y) then \<HolOclLiftLeft>(\<HolOclDropLeft>x\<HolOclDropRight> \<or> \<HolOclDropLeft>y\<HolOclDropRight>)\<HolOclLiftRight>
             else (if \<HolOclDropLeft>x\<HolOclDropRight> then \<HolOclLiftLeft>True\<HolOclLiftRight> else \<bottom>))
       else (if (DEF y) then (if \<HolOclDropLeft>y\<HolOclDropRight> then \<HolOclLiftLeft>True\<HolOclLiftRight> else \<bottom>) else \<bottom>)))"
  apply(rule eq_reflection, (rule ext)+)
  by(simp add: OclOr_def OclAnd_def OclNot_def ss_lifting)

lemma OclSor_alt_def: 
  "OclSor \<equiv> lift\<^sub>2 (strictify (\<lambda> x. strictify (\<lambda> y.\<HolOclLiftLeft>(\<HolOclDropLeft>x\<HolOclDropRight> \<or> \<HolOclDropLeft>y\<HolOclDropRight>)\<HolOclLiftRight>)))"
  apply(rule eq_reflection, (rule ext)+)
  by(simp add: OclSor_def OclNot_def OclSand_def ss_lifting')


syntax  (* standard syntax *)
    "_OclOr_std"       :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<OclOr>/ _)" [53,54]53)
    "_OclSor_std"       :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<OclSor>/ _)" [53,54]53)

syntax  (* standard syntax *)
    "_OclOr_ascii"     :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ or/ _)" [53,54]53)
    "_OclSor_ascii"     :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ sor/ _)" [53,54]53)

syntax  (* standard syntax *)
    "_OclOr_math"      :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<MathOclOr>/ _)" [53,54]53)
    "_OclSor_math"      :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 

                           ("(_ \<MathOclSor>/ _)" [53,54]53)
(*<*)
parse_translation {*
  let
  in flat(map (emb trans_const) ["OclOr","OclSor"])
  end;
*}

print_translation {*
  let
  in map emb_print ["OclOr","OclSor"]
  end;
*}
(*>*)

constdefs
  OclXor          :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean"   
  "OclXor S T     \<equiv> (S \<MathOclOr> T) \<MathOclAnd> \<MathOclNot> (S \<MathOclAnd> T)"

  OclImplies      :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean"   
  "OclImplies S T \<equiv>  (\<MathOclNot> S) \<MathOclOr> T" 

text{* alternative definitions for the use with \inlineholocl{ocl_setup_op} *}
lemma OclXor_alt_def: "OclXor \<equiv> lift\<^sub>2 (\<lambda> x y.
      (if (DEF x)
       then (if (DEF y) then \<HolOclLiftLeft>(\<HolOclDropLeft>x\<HolOclDropRight> \<noteq> \<HolOclDropLeft>y\<HolOclDropRight>)\<HolOclLiftRight> else \<bottom>)
       else \<bottom>))"
  apply(rule eq_reflection, (rule ext)+)
  by(simp add: OclXor_def OclOr_alt_def OclAnd_def OclNot_def 
     ss_lifting)

lemma OclImplies_alt_def: "OclImplies \<equiv> lift\<^sub>2 (\<lambda> x y.
      (if (DEF x)
       then (if (DEF y) then \<HolOclLiftLeft>(\<HolOclDropLeft>x\<HolOclDropRight> \<longrightarrow> \<HolOclDropLeft>y\<HolOclDropRight>)\<HolOclLiftRight>
             else (if \<HolOclDropLeft>x\<HolOclDropRight> then \<bottom> else \<HolOclLiftLeft>True\<HolOclLiftRight>))
       else (if (DEF y) then (if \<HolOclDropLeft>y\<HolOclDropRight> then \<HolOclLiftLeft>True\<HolOclLiftRight> else \<bottom>) else \<bottom>)))"
  apply(rule eq_reflection, (rule ext)+)
  by(simp add: OclImplies_def OclAnd_def OclNot_def
               OclOr_alt_def  OclIsDefined_def ss_lifting)

text {*		 
  There is an inconsistency between the normative
  part~\cite[Chapter 11]{omg:ocl:2003} and the
  informative~\cite[Appendix A]{omg:ocl:2003} the standard. For
  reasons discussed below, we chose the \inlineocl{implies} version
  defined by~{\cite[Table A-2, page A-12]{omg:ocl:2003}} and define
  the other version later.
  \OCLglitch{The truth for \inlineocl{implies} does not fulfill the 
  normative requirement.
  }
*}
constdefs
  OclSxor          :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean"   
  "OclSxor S T     \<equiv> (S \<MathOclSor> T) \<MathOclSand> \<MathOclNot> (S \<MathOclSand> T)"

  OclSimplies      :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean"   
  "OclSimplies S T \<equiv> (\<MathOclNot> S) \<MathOclSor> T"

text{* Alternative definitions for the use with \inlineholocl{ocl_setup_op} *}
lemma OclSxor_alt_def: 
  "OclSxor \<equiv> lift\<^sub>2 (strictify (\<lambda> x. strictify (\<lambda> y.\<HolOclLiftLeft>(\<HolOclDropLeft>x\<HolOclDropRight> \<noteq> \<HolOclDropLeft>y\<HolOclDropRight>)\<HolOclLiftRight>)))"
  apply(rule eq_reflection, (rule ext)+)
  by(auto simp: OclSxor_def OclSor_def OclNot_def OclSand_def 
                ss_lifting')

lemma OclSimplies_alt_def: 
  "OclSimplies \<equiv> lift\<^sub>2 (strictify (\<lambda> x. strictify (\<lambda> y.\<HolOclLiftLeft>(\<HolOclDropLeft>x\<HolOclDropRight> \<longrightarrow> \<HolOclDropLeft>y\<HolOclDropRight>)\<HolOclLiftRight>)))"
  apply(rule eq_reflection, (rule ext)+)
  by(simp add: OclSimplies_def OclSxor_def OclNot_def OclSand_def 
               OclSor_def ss_lifting')


syntax  (* standard syntax *)
    "_OclXor_std"      :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<OclXor>/ _)" [55,56]55)
    "_OclImplies_std"  :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<OclImplies>/ _)"[53,52]52)
    "_OclSxor_std"      :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<OclSxor>/ _)" [55,56]55)
    "_OclSimplies_std"  :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<OclSimplies>/ _)"[53,52]52)

syntax  (* standard syntax *)
    "_OclXor_ascii"    :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ xor/ _)" [55,56]55)
    "_OclImplies_ascii":: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ implies/ _)"[53,52]52)
    "_OclSxor_ascii"    :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ sxor/ _)" [55,56]55)
    "_OclSimplies_ascii":: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ simplies/ _)"[53,52]52)


syntax  (* standard syntax *)
    "_OclXor_math"     :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<MathOclXor>/ _)" [55,56]55)
    "_OclImplies_math" :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<MathOclImplies>/ _)"[53,52]52)
    "_OclSxor_math"     :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<MathOclSxor>/ _)" [55,56]55)
    "_OclSimplies_math" :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<MathOclSimplies>/ _)"[53,52]52)

(*<*)
parse_translation {*
  let
  in flat(map (emb trans_const) ["OclXor","OclImplies","OclSxor","OclSimplies"])
  end;
*}

print_translation {*
  let
  in map emb_print ["OclXor","OclImplies","OclSxor","OclSimplies"]
  end;
*}
(*>*)


text {*
  An \inlineocl{implies} variant for three-valued logics
  introduced by H\"ahnle~\cite{hahnle:towards:1991}:
  *}
constdefs
   OclImplies1     :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean"  
  "OclImplies1 S T \<equiv> \<MathOclIsUndefined> S \<MathOclOr> \<MathOclNot> S \<MathOclOr> T"


text {*
  An \inlineocl{implies} variant for three-valued logics
  introduced by Hennicker et al~\cite{hennicker.ea:precise:2002}.
  This variant is also the version actually used in the normative
  part of the standard.
*}
constdefs
   OclImplies2     :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean"  
  "OclImplies2 S T \<equiv> (\<MathOclNot> S) \<MathOclOr>  (S \<MathOclAnd> T)"


text{* Alternative definitions for the use with \inlineholocl{ocl_setup_op} *}
lemma OclImplies1_alt_def: "OclImplies1 \<equiv> lift\<^sub>2 (\<lambda> x y.
      (if (DEF x)
       then (if \<HolOclDropLeft>x\<HolOclDropRight> then y else \<HolOclLiftLeft>True\<HolOclLiftRight>)
       else \<HolOclLiftLeft>True\<HolOclLiftRight>))"
  apply(rule eq_reflection, (rule ext)+)
  by(simp add: OclImplies1_def OclAnd_def OclNot_def
               OclOr_alt_def  OclIsDefined_def ss_lifting)

lemma OclImplies2_alt_def: "OclImplies2 \<equiv> lift\<^sub>2 (\<lambda> x y. 
      (if (DEF x)
       then (if \<HolOclDropLeft>x\<HolOclDropRight> then y else \<HolOclLiftLeft>True\<HolOclLiftRight>)
       else \<bottom>))"
  apply(rule eq_reflection, (rule ext)+)
  by(simp add: OclImplies2_def OclAnd_def OclNot_def
               OclOr_alt_def  OclIsDefined_def ss_lifting)


syntax  (* standard syntax *)
    "_OclImplies1_std"  :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<OclVImplies>/ _)"[53,52]52)
    "_OclImplies2_std"  :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<OclVVImplies>/ _)"[53,52]52)

syntax  (* standard ascii syntax *)
    "_OclImplies1_ascii":: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ implies1/ _)"[53,52]52)
    "_OclImplies2_ascii":: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ implies2/ _)"[53,52]52)

syntax  (* standard syntax *)
    "_OclImplies1_math" :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<MathOclVImplies>/ _)"[53,52]52)
    "_OclImplies2_math" :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<MathOclVVImplies>/ _)"[53,52]52)

(*<*)
parse_translation {*
  let
  in flat(map (emb trans_const) ["OclImplies1","OclImplies2"])
  end;
*}

print_translation {*
  let
  in map emb_print ["OclImplies1","OclImplies2"]
  end;
*}

(*>*)


syntax  (* standard syntax *)
    "_OclImplies1_std" :: "'\<tau> Boolean \<Rightarrow> '\<tau> Boolean"
                           ("\<OclVImplies> (_)"   [59]59)
    "_OclImplies2_std" :: "['\<tau> Boolean, '\<tau> Boolean] \<Rightarrow> '\<tau> Boolean" 
                           ("(_ \<OclVVImplies>/ _)"[57,58]57)
    "_OclIf_std"       :: "['\<tau> Boolean,('\<tau>,'\<alpha>\<Colon>bot)VAL,('\<tau>,'\<alpha>\<Colon>bot)VAL] \<Rightarrow> ('\<tau>,'\<alpha>)VAL"
                           ("(\<OclIf> (_)/ \<OclThen> (_)/\<OclElse> (_)/\<OclEndif>)" 15)


section {* Setup of the Basics *} 


(* better move the following two proofs to OCL_logic? *)
lemma OclIsDefined_vs_DEF: " \<MathOclIsDefined> x =  \<MathOclTrue> \<Longrightarrow> DEF (x St)"
  apply (simp add: OclIsDefined_def lift1_def o_def OclTrue_def lift0_def)
  apply (drule fun_cong [of _ _ "St"] )
  apply (auto)
  done

lemma DEF_vs_OclIsDefined: 
  assumes 1:  "(\<And> St. DEF (x St))" 
  shows       "\<MathOclTrue> = \<MathOclIsDefined> x"
  apply (rule ext)
  apply (simp add: OclIsDefined_def lift1_def o_def OclTrue_def lift0_def)
  apply (rule 1)
  done

ML_setup {*
    val OclUndefined_def = get_def (theory "OCL_Undefined") "OclUndefined"
*}
ML {*
    val lift3_def = get_def (theory "Lifting") "lift3"
    val lift2_def = get_def (theory "Lifting") "lift2"
    val lift1_def = get_def (theory "Lifting") "lift1"
    val lift0_def = get_def (theory "Lifting") "lift0"
    val cp_def = get_def (theory "Lifting") "cp"
	
*}

ML {*
	       
    val strictify_def = get_def (theory "Lifting") "strictify"
    val isStrict_def  = get_def (theory "Lifting")  "isStrict"
			 
    val DEF_def = get_def  (theory "Lifting")  "DEF"
    val not_down_exists_lift = thm "not_down_exists_lift"
    val up_split = thm "up.split"
    val cp_lift0_fw = thm "cp_lift0_fw"
    val cp_lift1_fw = thm "cp_lift1_fw"
    val cp_lift2_fw = thm "cp_lift2_fw"
    val cp_lift3_fw = thm "cp_lift3_fw"
    val DEF_strictify_DEF_args  = thm "DEF_strictify_DEF_args"
    val DEF_strictify_DEF_args2 = thm "DEF_strictify_DEF_args2"
    val lift1b_undef = thm "lift1b_undef"
    val lift2_undef1a= thm "lift2_undef1a"
    val lift2_undef2a= thm "lift2_undef2a"
    val cp_compose2  = thm "cp_compose2"
    val UU_fun_def   = get_def  (theory "Lifting")  "UU_fun"
*}

ML {*		     
    val OclIsDefined_def= thm "OclIsDefined_def"
    val OclTrue_def     = thm "OclTrue_def"
    val OclFalse_def    = thm "OclFalse_def"
    val OclNot_def      = thm "OclNot_def"
    val OclAnd_def      = thm "OclAnd_def"
    val OclIf_def       = thm "OclIf_def"
    val OclOr_def       = thm "OCL_Boolean.OclOr_def"
    val OclXor_def      = thm "OCL_Boolean.OclXor_def"
    val OclImplies_def  = thm "OCL_Boolean.OclImplies_def"
    val OclImplies1_def = thm "OCL_Boolean.OclImplies1_def"
    val OclImplies2_def = thm "OCL_Boolean.OclImplies2_def"
*}

section {* The Basic Laws for the Boolean Operators *}

ML_setup {*
fun prover1 (goal,prems,name) =
    let val thm = prove_goalw (the_context()) (DEF_def::strictify_def::o_def::
                                   lift0_def::lift1_def::lift2_def::lift3_def
                                   ::prems) goal
                  (fn _ => [rtac ext 1,
                            ALLGOALS(asm_full_simp_tac 
                                        (simpset() 
                                         addsplits [split_if,up_split])),
                            ALLGOALS(asm_full_simp_tac 
                                        (simpset() 
                                         addsimps  [not_down_exists_lift])),
                            auto_tac(claset(),simpset())
                           ])
    in bind_thm(name,thm);
       Addsimps[thm]
    end;
*}

ML_setup {*
map prover1
[("\<MathOclIsDefined> \<MathOclUndefined> = \<MathOclFalse>", [OclIsDefined_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
             "OclIsDefined__undef"),
 ("\<MathOclIsDefined> \<MathOclTrue> = \<MathOclTrue>", [OclIsDefined_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
             "OclIsDefined__True"),
 ("\<MathOclIsDefined> \<MathOclFalse> = \<MathOclTrue>", [OclIsDefined_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
             "OclIsDefined__False"),
 ("\<MathOclNot> \<MathOclUndefined> = \<MathOclUndefined>", [OclNot_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
             "not__undef"),
 ("\<MathOclNot> \<MathOclTrue> = \<MathOclFalse>", [OclNot_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
             "not__True"),
 ("\<MathOclNot> \<MathOclFalse> = \<MathOclTrue>", [OclNot_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
             "not__False")];

*}


text{* This shows the following facts:  @{thm "OclIsDefined__undef"} *}
text{* @{thm "OclIsDefined__True"} *}
text{* @{thm "OclIsDefined__False"} *}
text{* @{thm "not__undef"} *}
text{* @{thm "not__True"} *}
text{* @{thm "not__False"} *}
text{* \ldots  which also prove the complience with Table A.2 in the
       standard 2.0. *}


ML {*
map prover1 
[("\<MathOclUndefined> \<MathOclAnd> \<MathOclUndefined> = \<MathOclUndefined>", [OclAnd_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
               "undef__and__undef"),
 ("\<MathOclUndefined> \<MathOclAnd> \<MathOclTrue> = \<MathOclUndefined>", [OclAnd_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
               "undef__and__True"),
 ("\<MathOclUndefined> \<MathOclAnd> \<MathOclFalse> = \<MathOclFalse>", [OclAnd_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
               "undef__and__False"),
 ("\<MathOclTrue> \<MathOclAnd> \<MathOclUndefined> = \<MathOclUndefined>", [OclAnd_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
               "True__and__undef"),
 ("\<MathOclTrue> \<MathOclAnd> \<MathOclTrue> = \<MathOclTrue>", [OclAnd_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
               "True__and__True"),
 ("\<MathOclTrue> \<MathOclAnd> \<MathOclFalse> = \<MathOclFalse>", [OclAnd_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
               "True__and__False"),
 ("\<MathOclFalse> \<MathOclAnd> \<MathOclUndefined> = \<MathOclFalse>", [OclAnd_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
               "False__and__undef"),
 ("\<MathOclFalse> \<MathOclAnd> \<MathOclTrue> = \<MathOclFalse>", [OclAnd_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
               "False__and__True"),
 ("\<MathOclFalse> \<MathOclAnd> \<MathOclFalse> = \<MathOclFalse>", [OclAnd_def,OclTrue_def,OclFalse_def,OclUndefined_def], 
               "False__and__False")];
*}



ML {*
map prover1 
[("\<MathOclUndefined> \<MathOclOr> \<MathOclUndefined> = \<MathOclUndefined>", [OclOr_def], "OclUndefined__or__undef"),
 ("\<MathOclUndefined> \<MathOclOr> \<MathOclTrue> = \<MathOclTrue>", [OclOr_def], "OclUndefined__or__True"),
 ("\<MathOclUndefined> \<MathOclOr> \<MathOclFalse> = \<MathOclUndefined>", [OclOr_def], "OclUndefined__or__False"),
 ("\<MathOclTrue> \<MathOclOr> \<MathOclUndefined> = \<MathOclTrue>", [OclOr_def], "True__or__undef"),
 ("\<MathOclTrue> \<MathOclOr> \<MathOclTrue> = \<MathOclTrue>", [OclOr_def], "True__or__True"),
 ("\<MathOclTrue> \<MathOclOr> \<MathOclFalse> = \<MathOclTrue>", [OclOr_def], "True__or__False"),
 ("\<MathOclFalse> \<MathOclOr> \<MathOclUndefined> = \<MathOclUndefined>", [OclOr_def], "False__or__undef"),
 ("\<MathOclFalse> \<MathOclOr> \<MathOclTrue> = \<MathOclTrue>", [OclOr_def], "False__or__True"),
 ("\<MathOclFalse> \<MathOclOr> \<MathOclFalse> = \<MathOclFalse>", [OclOr_def], "False__or__False")];
*}

ML {*
map prover1 
[("\<MathOclUndefined> \<MathOclXor> \<MathOclUndefined> = \<MathOclUndefined>", [OclXor_def], "OclUndefined__xor__undef"),
 ("\<MathOclUndefined> \<MathOclXor> \<MathOclTrue> = \<MathOclUndefined>", [OclXor_def], "OclUndefined__xor__True"),
 ("\<MathOclUndefined> \<MathOclXor> \<MathOclFalse> = \<MathOclUndefined>", [OclXor_def], "OclUndefined__xor__False"),
 ("\<MathOclTrue> \<MathOclXor> \<MathOclUndefined> = \<MathOclUndefined>", [OclXor_def], "True__xor__undef"),
 ("\<MathOclTrue> \<MathOclXor> \<MathOclTrue> = \<MathOclFalse>", [OclXor_def], "True__xor__True"),
 ("\<MathOclTrue> \<MathOclXor> \<MathOclFalse> = \<MathOclTrue>", [OclXor_def], "True__xor__False"),
 ("\<MathOclFalse> \<MathOclXor> \<MathOclUndefined> = \<MathOclUndefined>", [OclXor_def], "OclUndefined__xor__undef"),
 ("\<MathOclFalse> \<MathOclXor> \<MathOclTrue> = \<MathOclTrue>", [OclXor_def], "OclUndefined__xor__True"),
 ("\<MathOclFalse> \<MathOclXor> \<MathOclFalse> = \<MathOclFalse>", [OclXor_def], "OclUndefined__xor__False")];
*}


ML {*
map prover1 
[("\<MathOclUndefined> \<MathOclImplies> \<MathOclUndefined> = \<MathOclUndefined>",[OclImplies_def], "OclUndefined__implies__undef2"),
 ("\<MathOclUndefined> \<MathOclImplies> \<MathOclTrue> = \<MathOclTrue>",[OclImplies_def], "OclUndefined__implies__True2"),
 ("\<MathOclUndefined> \<MathOclImplies> \<MathOclFalse> = \<MathOclUndefined>",[OclImplies_def], "OclUndefined__implies__FALSE2"),
 ("\<MathOclTrue> \<MathOclImplies> \<MathOclUndefined> = \<MathOclUndefined>",[OclImplies_def], "True__implies__undef2"),
 ("\<MathOclTrue> \<MathOclImplies> \<MathOclTrue> = \<MathOclTrue>",[OclImplies_def], "True__implies__True2"),
 ("\<MathOclTrue> \<MathOclImplies> \<MathOclFalse> = \<MathOclFalse>",[OclImplies_def], "True__implies__False2"),
 ("\<MathOclFalse> \<MathOclImplies> \<MathOclUndefined> = \<MathOclTrue>",[OclImplies_def], "FALSE__implies__undef2"),
 ("\<MathOclFalse> \<MathOclImplies> \<MathOclTrue> = \<MathOclTrue>",[OclImplies_def], "False__implies__True2"),
 ("\<MathOclFalse> \<MathOclImplies> \<MathOclFalse> = \<MathOclTrue>",[OclImplies_def], "False__implies__False2")];
*}


ML {*
map prover1 
[("\<MathOclUndefined> \<MathOclVImplies> \<MathOclUndefined> = \<MathOclTrue>",[OclImplies1_def], "OclUndefined__implies1__undef1"),
 ("\<MathOclUndefined> \<MathOclVImplies> \<MathOclTrue> = \<MathOclTrue>",[OclImplies1_def], "OclUndefined__implies1__True1"),
 ("\<MathOclUndefined> \<MathOclVImplies> \<MathOclFalse> = \<MathOclTrue>",[OclImplies1_def], "OclUndefined__implies1__False1"),
 ("\<MathOclTrue> \<MathOclVImplies> \<MathOclUndefined> = \<MathOclUndefined>",[OclImplies1_def], "True__implies1__undef1"),
 ("\<MathOclTrue> \<MathOclVImplies> \<MathOclTrue> = \<MathOclTrue>",[OclImplies1_def], "True__implies1__True1"),
 ("\<MathOclTrue> \<MathOclVImplies> \<MathOclFalse> = \<MathOclFalse>",[OclImplies1_def], "True__implies1__False1"),
 ("\<MathOclFalse> \<MathOclVImplies> \<MathOclUndefined> = \<MathOclTrue>",[OclImplies1_def], "False__implies1__undef1"),
 ("\<MathOclFalse> \<MathOclVImplies> \<MathOclTrue> = \<MathOclTrue>",[OclImplies1_def], "False__implies1__undef1"),
 ("\<MathOclFalse> \<MathOclVImplies> \<MathOclFalse> = \<MathOclTrue>",[OclImplies1_def], "False__implies1__undef1")];
*}

ML {*
map prover1 
[("\<MathOclUndefined> \<MathOclVVImplies> \<MathOclUndefined> = \<MathOclUndefined>",[OclImplies2_def], "OclUndefined__implies2__undef"),
 ("\<MathOclUndefined> \<MathOclVVImplies> \<MathOclTrue> = \<MathOclUndefined>",[OclImplies2_def], "OclUndefined__implies2__True"),
 ("\<MathOclUndefined> \<MathOclVVImplies> \<MathOclFalse> = \<MathOclUndefined>",[OclImplies2_def], "OclUndefined__implies2__FALSE"),
 ("\<MathOclTrue> \<MathOclVVImplies> \<MathOclUndefined> = \<MathOclUndefined>",[OclImplies2_def], "True__implies2__undef"),
 ("\<MathOclTrue> \<MathOclVVImplies> \<MathOclTrue> = \<MathOclTrue>",[OclImplies2_def], "True__implies2__True"),
 ("\<MathOclTrue> \<MathOclVVImplies> \<MathOclFalse> = \<MathOclFalse>",[OclImplies2_def], "True__implies2__False"),
 ("\<MathOclFalse> \<MathOclVVImplies> \<MathOclUndefined> = \<MathOclTrue>",[OclImplies2_def], "FALSE__implies2__undef"),
 ("\<MathOclFalse> \<MathOclVVImplies> \<MathOclTrue> = \<MathOclTrue>",[OclImplies2_def], "False__implies2__True"),
 ("\<MathOclFalse> \<MathOclVVImplies> \<MathOclFalse> = \<MathOclTrue>",[OclImplies2_def], "False__implies2__False")];
*}



ML {*
map prover1
[("(\<MathOclIf>  \<MathOclUndefined> \<MathOclThen> Y \<MathOclElse> Z  \<MathOclEndif>) = \<MathOclUndefined>",
             [OclIf_def,OclUndefined_def], "if_undef"),
 ("(\<MathOclIf>  \<MathOclTrue> \<MathOclThen> Y \<MathOclElse> Z \<MathOclEndif>) = \
   \(Y::('\<tau>,'\<alpha>::bot)VAL)",    
             [OclIf_def,OclTrue_def], "if_TRUE"),
 ("(\<MathOclIf>  \<MathOclFalse> \<MathOclThen> Y \<MathOclElse> Z \<MathOclEndif>) = \
   \(Z::('\<tau>,'\<alpha>::bot)VAL)",    
             [OclIf_def,OclFalse_def], "if_FALSE")];
*}
 


section {* All operators are contextpassing *}



lemmas cp_undef = OclUndefined_def [THEN cp_lift0_fw, standard]
(*
bind_thm("cp_TRUE",   OCL_Boolean.OclTrue_def RS cp_lift0_fw)
bind_thm("cp_FALSE",  OclFalse_def RS cp_lift0_fw);
 *)
lemmas cp_OclIsDefined = OclIsDefined_def [THEN cp_lift1_fw, standard]
lemmas cp_not = OclNot_def [THEN cp_lift1_fw, standard]
lemmas cp_and = OclAnd_def [THEN cp_lift2_fw, standard]
lemmas cp_if  = OclIf_def  [THEN cp_lift3_fw, standard]

declare cp_undef [simp,intro!]   cp_OclIsDefined [simp, intro!] 
        cp_not   [simp, intro!]  cp_and [simp, intro!]  
        cp_if    [simp, intro!]


ML {*
fun prover2 (goal,prems,name) =
    let val thm = prove_goalw (the_context()) (prems) goal
                  (fn _ => [Fast_tac 1])
    in bind_thm(name,thm);
       Addsimps[thm];
       AddSIs [thm]
    end;
*}

ML {*
map prover2
[("\<And> P. \<lbrakk> cp P;cp P'\<rbrakk> \<Longrightarrow> cp(\<lambda>X. P X \<MathOclOr> P' X)", [OclOr_def],       "cp_or"),
 ("\<And> P.  \<lbrakk>cp P;cp P'\<rbrakk> \<Longrightarrow> cp(\<lambda>X. P X \<MathOclXor> P' X)", [OclXor_def],     "cp_xor"),
 ("\<And> P.  \<lbrakk>cp P;cp P'\<rbrakk> \<Longrightarrow> cp(\<lambda>X. P X \<MathOclImplies> P' X)",[OclImplies_def], "cp_implies"),
 ("\<And> P.  \<lbrakk>cp P;cp P'\<rbrakk> \<Longrightarrow> cp(\<lambda>X. P X \<MathOclVImplies> P' X)", [OclImplies1_def],"cp_implies1"),
 ("\<And> P.  \<lbrakk>cp P;cp P'\<rbrakk> \<Longrightarrow> cp(\<lambda>X. P X \<MathOclVVImplies> P' X)", [OclImplies2_def],"cp_implies2")]
*}
text{* This shows the following facts:  
@{thm "cp_or"} 
@{thm "cp_xor"} 
@{thm "cp_implies"} 
@{thm "cp_implies1"} 
@{thm "cp_implies2"} 
*}

section {* Proof Technique: High-level Trichometry on context passing contexts *}

lemma base_distinct: 
		    "X \<tau> = \<MathOclUndefined> \<tau> \<or> X \<tau> = \<MathOclTrue> \<tau> \<or> X \<tau> = \<MathOclFalse> \<tau>"
  apply (simp add: OclUndefined_def lift0_def OclTrue_def OclFalse_def)
  apply (auto)
  apply (simp add: not_down_exists_lift)
  apply (erule exE)
  apply (auto)
  done
  ML_setup {* val base_distinct = thm "base_distinct" *}

ML_setup {* 
val prems = goalw (the_context()) []  
		  "(X \<tau> ~= \<MathOclFalse> \<tau>) = (X \<tau> = \<MathOclUndefined> \<tau> | X \<tau> = \<MathOclTrue> \<tau>)";
    by(cut_facts_tac[base_distinct] 1);
    by(Step_tac 1);
    ba 1;
    by(ALLGOALS(asm_full_simp_tac ((simpset()) addsimps 
                                   [OclFalse_def,OclTrue_def,
                                    lift0_def,OclUndefined_def])));
    qed"distinct_1";

val prems = goalw (the_context()) []  
		  "(X \<tau> ~= \<MathOclTrue> \<tau>) = (X \<tau> = \<MathOclUndefined> \<tau> | X \<tau> = \<MathOclFalse> \<tau>)";
    by(cut_facts_tac[base_distinct] 1);
    by(Step_tac 1);
    ba 1;
    by(ALLGOALS(asm_full_simp_tac ((simpset()) addsimps 
                                   [OclFalse_def,OclTrue_def,lift0_def,OclUndefined_def])));
    qed"distinct_2";


val prems = goalw (the_context()) []  
		  "(X \<tau> ~= \<bottom>) = (X \<tau> = \<MathOclTrue> \<tau> | X \<tau> = \<MathOclFalse> \<tau>)";
    by(cut_facts_tac[base_distinct] 1);
    by(Step_tac 1);
    by(ALLGOALS(asm_full_simp_tac ((simpset()) addsimps 
                                   [OclFalse_def,OclTrue_def,lift0_def,OclUndefined_def])));
    by(Fast_tac 1);
    qed"distinct_3";

*}


lemma cp_distinct_core:
     "\<lbrakk> cp P; cp P';                          
        P \<MathOclUndefined> \<tau> = P' \<MathOclUndefined> \<tau>;                         
        P \<MathOclTrue> \<tau>  = P' \<MathOclTrue> \<tau>;                          
        P \<MathOclFalse> \<tau> = P' \<MathOclFalse> \<tau>                          
      \<rbrakk> \<Longrightarrow> P X \<tau> = P' X \<tau>"
apply (unfold cp_def)
apply (erule exE) ;
apply (erule exE)
apply (rotate_tac -2) 
apply simp
apply (cut_tac X = "X" and \<tau> = "\<tau>" in base_distinct)
apply (erule disjE) ;
apply (erule_tac [2] disjE)
apply (auto)
done
ML {* val cp_distinct_core = thm "cp_distinct_core" *}

lemma cp_distinct_core_P:
      "\<lbrakk> cp P; cp P';                          
         X \<tau> = \<MathOclUndefined> \<tau> \<Longrightarrow> P \<MathOclUndefined> \<tau> = P' \<MathOclUndefined> \<tau>;     
         X \<tau> = \<MathOclTrue> \<tau>  \<Longrightarrow> P \<MathOclTrue> \<tau>  = P' \<MathOclTrue> \<tau>;      
         X \<tau> = \<MathOclFalse> \<tau> \<Longrightarrow> P \<MathOclFalse> \<tau> = P' \<MathOclFalse> \<tau>      
       \<rbrakk> \<Longrightarrow> P X \<tau> = P' X \<tau>"
apply (unfold cp_def)
apply (erule exE) ;
apply (erule exE)
apply (rotate_tac -2) 
apply simp
apply (cut_tac X = "X" and \<tau> = "\<tau>" in base_distinct)
apply (erule disjE) 
apply (erule_tac [2] disjE)
apply (auto)
done

ML {*  val cp_distinct_core_P = thm "cp_distinct_core_P" *}


lemma local_validity_propagation1:
      "\<lbrakk>  X \<tau> = \<MathOclUndefined> \<tau>; 
          P \<MathOclUndefined> \<tau> = P' \<MathOclUndefined> \<tau>;    
          cp P; cp P'  
       \<rbrakk>  \<Longrightarrow> (P\<Colon>('\<tau>)Boolean \<Rightarrow> ('\<tau>,'\<alpha>)VAL) X \<tau> = P' X \<tau>"

apply (rule_tac P = "P" and P' = "P'" in cp_distinct_core_P)
apply auto
apply (simp_all add: OclTrue_def OclFalse_def OclUndefined_def lift0_def)
done

ML {* val local_validity_propagation1 = thm "local_validity_propagation1" *}

lemma local_validity_propagation2:
      "\<lbrakk>  X \<tau> = \<MathOclTrue> \<tau>; 
          P \<MathOclTrue> \<tau> = P' \<MathOclTrue> \<tau>;    
          cp P; cp P'  
       \<rbrakk> \<Longrightarrow> (P\<Colon>('\<tau>)Boolean \<Rightarrow> ('\<tau>,'\<alpha>)VAL) X \<tau> = P' X \<tau>"

apply (rule_tac P = "P" and P' = "P'" in cp_distinct_core_P)
apply auto
apply (simp_all add: OclTrue_def OclFalse_def OclUndefined_def lift0_def)
done
ML {* val local_validity_propagation2 = thm "local_validity_propagation2" *}

lemma local_validity_propagation3:
      "\<lbrakk>  X \<tau> = \<MathOclFalse> \<tau>; 
          P \<MathOclFalse> \<tau> = P' \<MathOclFalse> \<tau>;    
          cp P; cp P'  
       \<rbrakk> \<Longrightarrow> (P\<Colon>('\<tau>)Boolean \<Rightarrow> ('\<tau>,'\<alpha>)VAL) X \<tau> = P' X \<tau>"

apply (rule_tac P = "P" and P' = "P'" in cp_distinct_core_P)
apply auto
apply (simp_all add: OclTrue_def OclFalse_def OclUndefined_def lift0_def)
done

ML {* val  local_validity_propagation3 = thm "local_validity_propagation3" *}  

lemma cp_distinct: 
      "\<lbrakk> cp P; cp P';                          
        P \<MathOclUndefined> = P' \<MathOclUndefined>;                               
        P \<MathOclTrue>  = P' \<MathOclTrue>;                                
        P \<MathOclFalse> = P' \<MathOclFalse>                                
       \<rbrakk> \<Longrightarrow> P X = P' X"
apply (rule ext)
apply (drule_tac f = "P \<MathOclUndefined>" and x = "x" in fun_cong)
apply (drule_tac f = "P \<MathOclTrue>" and x = "x" in fun_cong)
apply (drule_tac f = "P \<MathOclFalse>" and x = "x" in fun_cong)
apply (rule cp_distinct_core)
back
apply (auto)
done

ML {* val cp_distinct = thm "cp_distinct" *}

lemma cp_distinct1:
      "\<lbrakk> \<MathOclIsDefined> X = \<MathOclTrue>; cp P; cp P';         
         P \<MathOclTrue>  = P' \<MathOclTrue>;                                
         P \<MathOclFalse> = P' \<MathOclFalse>                                
       \<rbrakk> \<Longrightarrow> P X = P' X"
apply (unfold cp_def OclIsDefined_def o_def lift1_def)
apply (rule ext)
apply (drule_tac f = "P \<MathOclTrue>" and x = "x" in fun_cong)
apply (drule_tac f = "P \<MathOclFalse>" and x = "x" in fun_cong)
apply (drule_tac f = "\<lambda>c. \<HolOclLiftLeft>DEF (X c) \<HolOclLiftRight>" and x = "x" in fun_cong)
apply (erule exE) ;
apply (erule exE)
apply (rotate_tac -2)
apply simp
apply (cut_tac X = "X" and \<tau> = "x" in base_distinct)
apply (erule disjE) ;
apply (erule_tac [2] disjE)
apply (simp)
apply (simp add: DEF_def lift0_def OclUndefined_def OclTrue_def)
apply (auto)
done

ML {* val cp_distinct1 = thm "cp_distinct1" *}


lemma cp_distinct2:     
      "\<lbrakk> cp P; cp P';                         
         \<And> X. X = \<MathOclUndefined> \<Longrightarrow> P X = P' X;                   
         \<And> X. X = \<MathOclTrue>  \<Longrightarrow> P X = P' X;                   
         \<And> X. X = \<MathOclFalse> \<Longrightarrow> P X = P' X                    
       \<rbrakk> \<Longrightarrow> P X = P' X"

apply (rule cp_distinct)
apply auto
done

ML {* val cp_distinct2 = thm "cp_distinct2" *}

text {*  The lemmas for @{text "st_transissibility"} allow for a new, high-level 
         proof style based on case-distictions \emph{avoiding} unfolding the 
         definition of \inlineocl{and}, \ldots
*}

ML {* 
fun Ocl_case_tac str n= 
    EVERY[(eres_inst_tac [("X",str)] cp_distinct1 n) ORELSE
          (res_inst_tac  [("X",str)] cp_distinct  n),
           TRY(Fast_tac n), TRY(Fast_tac n),   
            Auto_tac ]
*}


(* it would be a standard excercise in Isabelle hacking
   to compute the "dists"-parameter on the fly in this
   tactic - just extract the free variables in left-to-right
   order of right type ... *)

ML {*
fun prover3 (goal,dists,b,name) =
    let val thm = prove_goalw (the_context()) [] goal
                  (fn _ => map (fn x => Ocl_case_tac x 1) dists)
    in bind_thm(name,thm);
       if b then Addsimps[thm] else ()
    end;
*}

ML {*
map prover3
[("\<MathOclNot>(\<MathOclNot> X) = X",    ["X"],true,"not_not") , 
 ("X \<MathOclAnd> X = X",     ["X"],true, "and_idem"),
 ("\<MathOclFalse> \<MathOclAnd> X = \<MathOclFalse>",     ["X"],true,"and_false_dominant_1"),
 ("X \<MathOclAnd> \<MathOclFalse> = \<MathOclFalse>",     ["X"],true,"and_false_dominant_2"),
 ("\<MathOclTrue> \<MathOclAnd> X = X",     ["X"],true,"and_true_neutral_1"), 
 ("X \<MathOclAnd> \<MathOclTrue> = X",     ["X"],true,"and_true_neutral_2"),
 ("X \<MathOclAnd> Y = Y\<MathOclAnd> X",  ["X","Y"],false,"and_commute"),
 ("X \<MathOclAnd> (Y \<MathOclAnd> Z) = (X \<MathOclAnd> Y)\<MathOclAnd> Z",["X","Y","Z"],false,"and_assoc")]
*}

ML {*
map prover3
[("X \<MathOclOr> X = X",["X"],true, "or_idem"),
 ("X \<MathOclOr> \<MathOclTrue> = \<MathOclTrue>",["X"],true, "or_true_dominant_1"), 
 ("\<MathOclTrue> \<MathOclOr> X = \<MathOclTrue>",["X"],true, "or_true_dominant_2"),
 ("X \<MathOclOr> \<MathOclFalse> = X",["X"],true, "or_false_neutral_1"),
 ("\<MathOclFalse> \<MathOclOr> X = X",["X"],true, "or_false_neutral_2"),
 ("X \<MathOclOr> Y = Y \<MathOclOr> X",["X","Y"],false, "or_commute"),
 ("(X \<MathOclOr> Y) \<MathOclOr> Z = X \<MathOclOr> (Y \<MathOclOr> Z)",["X","Y","Z"],false,"or_assoc")];
*}

ML {*
map prover3
[("(X \<MathOclOr> Y)\<MathOclAnd> Z = (X \<MathOclAnd> Z) \<MathOclOr> (Y \<MathOclAnd> Z)",["Z","Y","X","X"],
  false,"and_or_distrib1"),
 ("Z \<MathOclAnd> (X \<MathOclOr> Y) = (Z \<MathOclAnd> X) \<MathOclOr> (Z \<MathOclAnd> Y)",["Z","Y","X","X"],
  false,"and_or_distrib2"),
 ("(X \<MathOclAnd> Y) \<MathOclOr> Z = (X \<MathOclOr> Z) \<MathOclAnd> (Y \<MathOclOr> Z)",["Z","Y","X","X"],
  false,"or_and_distrib1"),
 ("Z \<MathOclOr> (X \<MathOclAnd> Y) = (Z \<MathOclOr> X) \<MathOclAnd> (Z \<MathOclOr> Y)",["Z","Y","X","X"],
  false,"or_and_distrib2")];
*}

ML {*
map prover3
[("(X \<MathOclAnd> Y) = \<MathOclNot>(\<MathOclNot>(X) \<MathOclOr> \<MathOclNot>(Y))",["X","Y"],false,"de_morgan"),
 ("\<MathOclNot>(X \<MathOclAnd> Y) = (\<MathOclNot>(X) \<MathOclOr> \<MathOclNot>(Y))",["X","Y"],false,"de_morgan1"),
 ("\<MathOclNot>(X \<MathOclOr> Y) = (\<MathOclNot>(X) \<MathOclAnd> \<MathOclNot>(Y))",["X","Y"],false,"de_morgan2")];
*}

ML {*
map prover3
[("X \<MathOclXor> \<MathOclTrue> = \<MathOclNot> X",   ["X"], true, "xor_true_inverse_1"),
 ("\<MathOclTrue> \<MathOclXor> X = \<MathOclNot> X",   ["X"], true, "xor_true_inverse_2"),
 ("X \<MathOclXor> \<MathOclFalse> = X",     ["X"], true, "xor_false_neutral_1"),
 ("\<MathOclFalse> \<MathOclXor> X = X",     ["X"], true, "xor_false_neutral_2"),
 ("X \<MathOclXor> Y = Y \<MathOclXor> X", ["X","Y"],false, "xor_commute"),
 ("(X \<MathOclXor> Y) \<MathOclXor> Z = X \<MathOclXor> (Y \<MathOclXor> Z)",
                   ["X","Y","Z","Z","Y","Z"], false, "xor_assoc")]; 
*}

text{* We will not include the or-generating rewrite rules into
       the rewriter since they do not "simplify" the situation.
       As rules, they are important, of course. *}
ML {*
map prover3
[("(X \<MathOclImplies> \<MathOclFalse>) = \<MathOclNot> X",["X"], true, "implies_false"),
 ("(X \<MathOclImplies> \<MathOclTrue>) = \<MathOclTrue>",  ["X"], true, "implies_true1"),
 ("(X \<MathOclImplies> \<MathOclUndefined>) = \<MathOclNot> X \<MathOclOr> \<MathOclUndefined>",  ["X"], false, "implies_undef1"),
 ("(\<MathOclFalse> \<MathOclImplies> X) = \<MathOclTrue>",  ["X"], true, "implies_true2"),
 ("(\<MathOclTrue> \<MathOclImplies> X) = X",  ["X"], true, "implies_idem"),
 ("(\<MathOclUndefined> \<MathOclImplies> X) = X \<MathOclOr> \<MathOclUndefined>",  ["X"], false, "implies_undef2"),
 ("(X \<MathOclImplies> X) = \<MathOclNot> X \<MathOclOr> X",  ["X"], false, "implies_idem1"),
 ("X \<MathOclImplies> (Y \<MathOclAnd> Z) = ((X \<MathOclImplies> Y) \<MathOclAnd> (X \<MathOclImplies> Z))",
                   ["X","Y","Z","Z"],false, "implies_andI"),
 ("X \<MathOclImplies> (Y \<MathOclOr> Z) = ((X \<MathOclImplies> Y) \<MathOclOr> (X \<MathOclImplies> Z))",
                   ["X","Y","Z","Z"],false, "implies_orI"),
 ("((X \<MathOclAnd> Y) \<MathOclImplies> Z) = X \<MathOclImplies> (Y \<MathOclImplies> Z)",
                   ["Z","Y","X","Y","X"],false, "implies_andE"),
 ("((X \<MathOclOr> Y) \<MathOclImplies> Z) = (X \<MathOclImplies> Z) \<MathOclAnd> (Y \<MathOclImplies> Z)",
                   ["Z","Y","X","Y","X","X","X","X","Y"],false, 
                   "implies_orE"),
 ("X \<MathOclImplies> (Y \<MathOclImplies> Z) = Y \<MathOclImplies> (X \<MathOclImplies> Z)",
   ["Z","Y","X","Y","X"],false, "implies_commute")]
*}


ML {*
map prover3
[("\<MathOclUndefined>  \<MathOclVVImplies> X  = \<MathOclUndefined>",["X"], true, "implies2_undef"),
 ("\<MathOclTrue>  \<MathOclVVImplies> X  = X",["X"], true, "implies2_idem"),
 ("\<MathOclFalse>  \<MathOclVVImplies> X  = \<MathOclTrue>", ["X"], true, "implies2_true2"),
 ("(X \<MathOclVVImplies> X) = \<MathOclNot> X \<MathOclOr> X",  ["X"], false, "implies2_idem1"),
 ("X  \<MathOclVVImplies> \<MathOclUndefined>  = \<MathOclNot> X \<MathOclOr> \<MathOclUndefined>",["X"], false, "implies2_undef2"),
 ("X  \<MathOclVVImplies> \<MathOclFalse>  = \<MathOclNot> X",["X"], true, "implies2_false"),
 ("X  \<MathOclVVImplies> \<MathOclTrue>  = \<MathOclNot> X \<MathOclOr> X",  ["X"], false, "implies2_true1")
] *}


ML{*
map prover3
[("X \<MathOclVVImplies> (Y \<MathOclAnd> Z) = ((X \<MathOclVVImplies> Y) \<MathOclAnd> (X \<MathOclVVImplies> Z))",
                   ["X"],false, 
                   "implies2_andI"),
 ("X \<MathOclVVImplies> (Y \<MathOclOr> Z) = ((X \<MathOclVVImplies> Y) \<MathOclOr> (X \<MathOclVVImplies> Z))",
                   ["X"],false, 
                   "implies2_orI")
 (* ("((X \<MathOclAnd> Y) \<MathOclVVImplies> Z) = X \<MathOclVVImplies> (Y \<MathOclVVImplies> Z)",
                   ["X","Y"],false, "implies2_andE") *)
 (* ("((X \<MathOclOr> Y) \<MathOclVVImplies> Z) = (X \<MathOclVVImplies> Z) \<MathOclAnd> (Y \<MathOclVVImplies> Z)",
                   ["X","Y"],false,  "implies2_orE") *)
 (* ("X \<MathOclVVImplies> (Y \<MathOclVVImplies> Z) = Y \<MathOclVVImplies> (X \<MathOclVVImplies> Z)",
      ["X","Y"],false, "implies2_commute") *) ]
*} 

text{* The properties in comments, crucial for deduction, could not be proven.
   In fact, they do not hold due to the counter-examples: *}

lemma "\<lbrakk> X=\<MathOclUndefined>; Y=\<MathOclTrue>; Z=\<MathOclTrue> \<rbrakk> \<Longrightarrow>
       ((X \<MathOclOr> Y) \<MathOclVVImplies> Z) \<noteq> (X \<MathOclVVImplies> Z) \<MathOclAnd> (Y \<MathOclVVImplies> Z)"
apply(simp, auto simp: lift0_def OclUndefined_def OclTrue_def)
apply(drule fun_cong, simp)
done

lemma "\<lbrakk> X=\<MathOclUndefined>; Y=\<MathOclFalse> \<rbrakk> \<Longrightarrow>
       ((X \<MathOclAnd> Y) \<MathOclVVImplies> Z) \<noteq> (X \<MathOclVVImplies> (Y \<MathOclVVImplies> Z))"
apply(simp, auto simp: lift0_def OclUndefined_def OclTrue_def)
apply(drule fun_cong, simp)
done

lemma "\<lbrakk> X=\<MathOclUndefined>; Y=\<MathOclFalse> \<rbrakk> \<Longrightarrow>
       X \<MathOclVVImplies> (Y \<MathOclVVImplies> Z) \<noteq> Y \<MathOclVVImplies> (X \<MathOclVVImplies> Z)"
apply(simp, auto simp: lift0_def OclUndefined_def OclTrue_def)
apply(drule fun_cong, simp)
done

text{* All these higher logical arrangement rules \emph{do not 
   hold} for the standards definition @{thm "OclImplies2_def"} of 
   the implication, in contrast to the "classical definition" 
   @{thm "OclImplies_def"}. Although the standards choice
   is feasible, in the light of these dramatic algebraic deficiencies,
   we qualify it as glitch from the deduction point of view
   and suggest to apply the definition used in the appendix.
   \OCLglitch{Definition for implies inappropriate}
  *}   


text {*
  For the completeness of our presentation, we show the properties
  of \inlineocl{implies} introduced by H\"ahnle~\cite{hahnle:towards:1991}:
*}
ML {*
map prover3
[("(X \<MathOclVImplies> X) = \<MathOclTrue>",["X"], true, "implies1_idem"),
 ("(\<MathOclUndefined> \<MathOclVImplies> X) = \<MathOclTrue>",["X"], true, "implies1_undef"),
 ("\<And> X. \<MathOclIsDefined>(X)=\<MathOclTrue> \<Longrightarrow> (X \<MathOclVImplies> \<MathOclFalse>) = \<MathOclNot> X",
                ["X"], true, "implies1_false"),
 ("(X \<MathOclVImplies> \<MathOclTrue>) = \<MathOclTrue>",["X"], true, "implies1_true1"),
 ("(\<MathOclFalse> \<MathOclVImplies> X) = \<MathOclTrue>",["X"], true, "implies1_true2"),
 ("(\<MathOclTrue> \<MathOclVImplies> X) = X",["X"], true, "implies1_idem2"),
 ("((X \<MathOclAnd> Y) \<MathOclVImplies> Z) = X \<MathOclVImplies> (Y \<MathOclVImplies> Z)",
                ["Z","Y","X","Y","X"],false, 
                "implies1_andE"),
 ("((X \<MathOclOr> Y) \<MathOclVImplies> Z) = (X \<MathOclVImplies> Z)\<MathOclAnd> (Y \<MathOclVImplies> Z)",
                ["Z","Y","X","Y","X","X","X","X","Y"],false, 
                "implies1_orE"),
 ("X \<MathOclVImplies> (Y \<MathOclVImplies> Z) = Y \<MathOclVImplies> (X \<MathOclVImplies> Z)",
                ["X"],false, "implies1_commute"),
 ("X \<MathOclVImplies> (Y \<MathOclAnd> Z) = ((X \<MathOclVImplies> Y) \<MathOclAnd> (X \<MathOclVImplies> Z))",
                ["X"],false, "implies1_andI"),
 ("X \<MathOclVImplies> (Y \<MathOclOr> Z) = ((X \<MathOclVImplies> Y) \<MathOclOr> (X \<MathOclVImplies> Z))",
                ["X"],false, "implies1_orI")]
*}

ML {*
map prover3
[("\<MathOclNot>(\<MathOclIsDefined> X) \<MathOclOr> X \<MathOclOr> \<MathOclNot> X = \<MathOclTrue>",        ["X"],false,"kleenical0"),
 ("\<And>X. \<MathOclIsDefined> X = \<MathOclTrue> \<Longrightarrow> (X \<MathOclOr> \<MathOclNot> X) = \<MathOclTrue>", ["X"],false,"kleenical1"),
 ("\<And>X. \<MathOclIsDefined> X = \<MathOclTrue> \<Longrightarrow> (\<MathOclNot> X \<MathOclOr> X) = \<MathOclTrue>", ["X"],false,"kleenical2"),
 ("(\<MathOclIsDefined>X \<MathOclAnd> X \<MathOclAnd> \<MathOclNot> X) = \<MathOclFalse>",           ["X"],false,"absurd0"),
 ("\<And>X. \<MathOclIsDefined> X = \<MathOclTrue> \<Longrightarrow> (X \<MathOclAnd> \<MathOclNot> X) = \<MathOclFalse>", ["X"],false,"absurd1"),
 ("\<And>X. \<MathOclIsDefined> X = \<MathOclTrue> \<Longrightarrow> (\<MathOclNot> X\<MathOclAnd> X) = \<MathOclFalse>",  ["X"],false,"absurd2"),
 ("\<MathOclNot>(\<MathOclIsDefined> X) \<MathOclOr> (X \<MathOclImplies> X) = \<MathOclTrue>",       ["X"],false,"implies_assume0"),
 ("\<And>X. \<MathOclIsDefined> X = \<MathOclTrue>  \<Longrightarrow> (X \<MathOclImplies> X) = \<MathOclTrue>",  ["X"],false,"implies_assume")]
*}

ML {*
map prover3
[("\<And>X. \<MathOclIsDefined> X = \<MathOclTrue> \<Longrightarrow> \<MathOclIsDefined>(\<MathOclNot> X) = \<MathOclTrue>", ["X"],true, "isDefined_not"),
 ("\<And>Z. \<lbrakk> \<MathOclIsDefined> X = \<MathOclTrue>;\<MathOclIsDefined> Y = \<MathOclTrue> \<rbrakk> \<Longrightarrow>  \<MathOclIsDefined>(X \<MathOclAnd> Y) = \<MathOclTrue>",
                                ["X"],true, "isDefined_and"),
 ("\<And>Z. \<lbrakk> \<MathOclIsDefined> X = \<MathOclTrue>;\<MathOclIsDefined> Y = \<MathOclTrue> \<rbrakk> \<Longrightarrow>  \<MathOclIsDefined>(X \<MathOclOr> Y) = \<MathOclTrue>",
                                ["X"],true, "isDefined_or"),
 ("\<And>Z. \<lbrakk> \<MathOclIsDefined> X = \<MathOclTrue>;\<MathOclIsDefined> Y = \<MathOclTrue> \<rbrakk> \<Longrightarrow>  \<MathOclIsDefined>(X \<MathOclXor> Y) = \<MathOclTrue>",
                                ["X"],true, "isDefined_xor"),  
 ("\<And>Z. \<lbrakk> \<MathOclIsDefined> X = \<MathOclTrue>;\<MathOclIsDefined> Y = \<MathOclTrue> \<rbrakk> \<Longrightarrow>  \<MathOclIsDefined>(X \<MathOclImplies> Y) = \<MathOclTrue>",
                                ["X"],true, "isDefined_implies"),
 ("\<And>Z. \<lbrakk> \<MathOclIsDefined> X = \<MathOclTrue>;\<MathOclIsDefined> Y = \<MathOclTrue> \<rbrakk> \<Longrightarrow>  \<MathOclIsDefined>(X \<MathOclVVImplies> Y) = \<MathOclTrue>",
                                ["X"],true, "isDefined_implies2"),
 ("\<And>A. \<lbrakk> \<MathOclIsDefined> X = \<MathOclTrue>;\<MathOclIsDefined> Y = \<MathOclTrue>; \<MathOclIsDefined> Z = \<MathOclTrue>\<rbrakk> \<Longrightarrow> \
   \ \<MathOclIsDefined>(\<MathOclIf> X \<MathOclThen> Y \<MathOclElse> Z \<MathOclEndif>) = \<MathOclTrue>",
                                ["X"],true, "isDefined_if")]
 
*}

ML {*
map prover3
[("(\<MathOclIf> X \<MathOclThen> (Y::'\<tau>=>('b::bot)) \<MathOclElse> Z \<MathOclEndif>) = \
  \(\<MathOclIf> \<MathOclNot> X \<MathOclThen> Z \<MathOclElse> Y \<MathOclEndif>)",    ["X"],false,"if_swap"),
 ("\<And>Z. \<MathOclIsDefined> X = \<MathOclTrue> \<Longrightarrow> (\<MathOclIf> X \<MathOclThen> Y \<MathOclElse> Z \<MathOclEndif>) = ((X \<MathOclImplies> Y) \<MathOclAnd> (\<MathOclNot> X \<MathOclImplies> Z))",
                                     ["X"],false,"if_to_implies")]
*}

section {* The Laws  for the @{text "is_def"} Operator *}
subsection {* Some basic laws. *}

lemma isDefined_idem[simp] : "\<MathOclIsDefined>(\<MathOclIsDefined> X) = \<MathOclTrue>"
   by(simp add: OclIsDefined_def lift0_def 
                lift1_def OclUndefined_def o_def OclTrue_def)


ML {*
map prover3
[("\<MathOclIsDefined>(\<MathOclIsDefined> X \<MathOclAnd> X) = \<MathOclTrue>",   ["X"],true, "isDefined_cong1"),
 ("\<MathOclIsDefined>(\<MathOclIsDefined> X \<MathOclAnd> \<MathOclNot> X) = \<MathOclTrue>", ["X"],true, "isDefined_cong2"),
 ("\<MathOclIsDefined>(X \<MathOclAnd> \<MathOclIsDefined> X) = \<MathOclTrue>",   ["X"],true, "isDefined_cong3"),
 ("\<MathOclIsDefined>(\<MathOclNot> X \<MathOclAnd> \<MathOclIsDefined> X) = \<MathOclTrue>", ["X"],true, "isDefined_cong4")];

*}

subsection {* Some generic laws for the @{text "is_def"} operator *}
text{* This section is essentially a rephrasing of the 
  theory \theory{Lifting} in the newly established terminology
  of this theory. *}

lemma is_isdef_lift0: "\<MathOclIsDefined>(lift\<^sub>0(\<HolOclLiftLeft> c \<HolOclLiftRight>)) = \<MathOclTrue>";
  apply (unfold OclIsDefined_def lift0_def lift1_def OclTrue_def o_def)
  apply (rule ext)
  apply (auto)
  done


lemma lift1_strictify_is_isdef: 
  "\<MathOclIsDefined>(lift\<^sub>1 (strictify (\<lambda>x. \<HolOclLiftLeft>f x\<HolOclLiftRight>)) X) = \<MathOclIsDefined> X"
  apply (unfold OclIsDefined_def lift1_def OclUndefined_def o_def)
  apply (rule ext)
  apply auto
  apply (drule DEF_strictify_DEF_args)
  apply auto
  done
  

lemma lift2_strictify_is_isdef: 
  "\<MathOclIsDefined>(lift\<^sub>2 (strictify(\<lambda>x. strictify(\<lambda>y. \<HolOclLiftLeft>f x y\<HolOclLiftRight>))) X Y) =  (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply (unfold OclIsDefined_def lift1_def lift2_def OclUndefined_def 
                OclAnd_def o_def)
  apply (rule ext)
  apply auto
  apply (drule DEF_strictify_DEF_args2)
  apply (drule_tac [2] DEF_strictify_DEF_args2)
  apply auto
  done

declare lift1_strictify_is_isdef [simp] lift2_strictify_is_isdef [simp]

(* TODO: move to OCL_Undefined.thy *)
lemma lift1b_undef_fw:
"f \<equiv> lift\<^sub>1 (strictify(g)) \<Longrightarrow> f \<MathOclUndefined> = \<MathOclUndefined>"
apply (unfold OclUndefined_def lift0_def)
apply (tactic {* fold_tac [UU_fun_def] *})
apply (rule lift1b_undef)
done


lemma lift2_undef1a_fw:
"f \<equiv> lift\<^sub>2 (strictify(\<lambda>x. strictify(g x))) \<Longrightarrow> f \<MathOclUndefined> (X\<Colon>'\<tau> \<Rightarrow> '\<alpha>::bot) = \<MathOclUndefined>"
apply (unfold OclUndefined_def lift0_def)
apply ( tactic {* fold_tac [UU_fun_def] *} )
apply (rule lift2_undef1a)
done

lemma lift2_undef2a_fw:
"f \<equiv> lift\<^sub>2 (strictify(\<lambda>x. strictify(g x))) \<Longrightarrow>   f(X\<Colon>'\<tau>\<Rightarrow>'\<alpha>\<Colon>bot) \<MathOclUndefined> = \<MathOclUndefined>"
apply (unfold OclUndefined_def lift0_def)
apply (tactic {* fold_tac [UU_fun_def] *})
apply (rule lift2_undef2a)
done

ML {* val lift2_undef2a_fw = thm "lift2_undef2a_fw" *}

lemma lift1_strict_is_isdef_fw: 
"f \<equiv> lift\<^sub>1 (strictify (\<lambda>x. \<HolOclLiftLeft>g\<HolOclDropLeft>x\<HolOclDropRight>\<HolOclLiftRight>)) \<Longrightarrow>  \<MathOclIsDefined>(f X) = \<MathOclIsDefined> X"
by(auto)

lemma lift1_strictify_is_isdef_fw:
"f \<equiv> lift\<^sub>1 (strictify(\<lambda>x. \<HolOclLiftLeft>g x\<HolOclLiftRight>)) \<Longrightarrow>  \<MathOclIsDefined>(f X) = \<MathOclIsDefined> X"
by(auto)

lemma lift2_strict_is_isdef_fw:
"f \<equiv> lift\<^sub>2(strictify(\<lambda>x. strictify(\<lambda>y. \<HolOclLiftLeft>g\<HolOclDropLeft>x\<HolOclDropRight> \<HolOclDropLeft>y\<HolOclDropRight>\<HolOclLiftRight>)))\<Longrightarrow> \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
by(auto)

lemma lift2_strictify_is_isdef_fw:
"f \<equiv> lift\<^sub>2 (strictify (\<lambda>x. strictify(\<lambda>y. \<HolOclLiftLeft>g x y\<HolOclLiftRight>))) \<Longrightarrow> \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
by(auto)

subsection{* Definedness reasoning for the new lifting construction *}

lemma fw_OCL_is_isdef_lift1'_0:
  assumes f_def: "f \<equiv> (lift1' lift_arg0) (strictify (\<lambda> x. \<HolOclLiftLeft>g x\<HolOclLiftRight>))"
  shows          "\<MathOclIsDefined>(f X) = \<MathOclIsDefined> X"
  by(rule ext, simp add: f_def OclIsDefined_def ss_lifting')

lemma fw_OCL_is_isdef_lift2'_00:
  assumes f_def: "f \<equiv> (lift2' lift_arg0 lift_arg0) 
                        (strictify (\<lambda> x. (strictify (\<lambda> y. \<HolOclLiftLeft>g x y\<HolOclLiftRight>))))"
  shows          "\<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X) \<MathOclAnd> (\<MathOclIsDefined> Y)"
  by(rule ext, simp add: f_def OclIsDefined_def OclAnd_def ss_lifting')

text{* register rules for \inlineholocl{ocl_setup_op} *}

ML {*
val _ = let

  fun is_isdef_rules rname = 
    let val rname' = "fw_OCL_is_isdef" ^ rname;
    in  RULE
         {rule      = (rname', thm rname'),
          simp_tac  = fn _ => fn x => x,
          is_atomic = fn _ => true, 
          register  = default_register_simponly, 
          name_gen  = default_namegen "OCL_is_isdef_"
         }
    end

in
  add_setup_rules (
    (map is_isdef_rules ["_lift1'_0", "_lift2'_00"])
  )
end
*}


section {* Some basic laws for the\inlineocl{not} operator *}

lemma not_injective : "\<lbrakk> \<MathOclNot>X = \<MathOclNot>Y \<rbrakk> \<Longrightarrow> X = Y"
  by(drule_tac f =    "\<lambda>x. \<MathOclNot> x" in  arg_cong, simp)

lemma not_bijective[simp]: "(\<MathOclNot>X = \<MathOclNot>Y) = (X = Y)"
  by(auto intro!: not_injective)



section {* Characterizations of logical operator definedness *}
lemma isDefined_notD:  "\<MathOclIsDefined>(\<MathOclNot> X) = \<MathOclTrue> \<Longrightarrow> \<MathOclIsDefined>(X)=\<MathOclTrue>"
  by (simp add: OclNot_def o_def lift1_strictify_is_isdef)

ML_setup {*
map prover3
[("\<MathOclIsDefined>(\<MathOclNot> X)   = (\<MathOclIsDefined> X)",     ["X"],true, "isDefined_notD0"),
 ("\<MathOclIsDefined>(X \<MathOclAnd> Y) = ((\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y) \<MathOclOr> (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclNot> X) \<MathOclOr> (\<MathOclIsDefined> Y \<MathOclAnd> \<MathOclNot> Y))",
                          ["X","Y","Y"],false, "isDefined_andD0"),
 ("\<MathOclIsDefined>(X \<MathOclOr> Y) = ((\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y) \<MathOclOr> (\<MathOclIsDefined> X \<MathOclAnd> X)  \<MathOclOr> (\<MathOclIsDefined> Y \<MathOclAnd> Y))",
                          ["X","Y","Y"],false, "isDefined_orD0"),
 ("\<MathOclIsDefined>(X \<MathOclXor> Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)",["X","Y"],false, "isDefined_xorD0"),
 ("\<MathOclIsDefined>(X \<MathOclImplies> Y)=((\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y) \<MathOclOr> (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclNot> X)  \<MathOclOr> (\<MathOclIsDefined> Y \<MathOclAnd> Y))",
                          ["X","Y","Y"],false, "isDefined_impliesD0"),
 ("\<MathOclIsDefined>(X \<MathOclVVImplies> Y)= \<MathOclIsDefined> X \<MathOclAnd> (\<MathOclNot>(X \<MathOclAnd> \<MathOclNot> \<MathOclIsDefined> Y))",
                          ["X"],false, "isDefined_implies2D0"),
 ("\<MathOclIsDefined>(\<MathOclIf> X \<MathOclThen> Y \<MathOclElse> Z \<MathOclEndif>) = (\<MathOclIsDefined> X \<MathOclAnd> ((X \<MathOclAnd> \<MathOclIsDefined> Y) \<MathOclOr> (\<MathOclNot> X \<MathOclAnd> \<MathOclIsDefined> Z)))",
                          ["X"],false, "isDefined_ifD0")];
*}


text {*
   The set of rules above is of fundamental importance for the
   forward-reasoning of definedness expressions---it allows for
   computing for any core logic expression a \acs{dnf} composed of
   boolean literals and text @{text "is_def(X)"} terms (where @{text
   "X"} is either a Boolean \OCL atom (variable) or a non Boolean
   expression (\inlineocl{if})). If the latter are composed by strict
   \OCL operators, it can be further atomized to @{text "is_def(X)"}
   atoms where @{text "X"} occurs in the expression.  *}


section{* Some generic laws for the Ocl if  operator *}

lemma if_distrib_strict:
    "\<lbrakk> cp P; isStrict P \<rbrakk> \<Longrightarrow>
     P(\<MathOclIf> X \<MathOclThen> (Y::('\<tau>,'\<alpha>::bot)VAL) \<MathOclElse> Z \<MathOclEndif>) = 
      (\<MathOclIf> X \<MathOclThen> ((P Y)::('\<tau>,'\<beta>::bot)VAL) \<MathOclElse> P Z \<MathOclEndif>)"
   apply(rule_tac X = X in cp_distinct)
   apply(rule_tac P = P in cp_compose2)
   apply auto
   apply(simp add: UU_fun_def OclUndefined_def  
                   isStrict_def lift0_def)
   done

lemma if_distrib2 :
   "\<lbrakk> cp P; \<MathOclIsDefined> X = \<MathOclTrue> \<rbrakk> \<Longrightarrow>
    P(\<MathOclIf> X \<MathOclThen> (Y::('\<tau>,'\<alpha>::bot)VAL) \<MathOclElse> Z \<MathOclEndif>) = 
     (\<MathOclIf> X \<MathOclThen> ((P Y)::('\<tau>,'\<beta>::bot)VAL) \<MathOclElse> P Z \<MathOclEndif>)"
    apply(simp add: cp_def OclIf_def lift3_def,
          erule exE, rule ext)
    apply(simp add: OclIsDefined_def DEF_def OclTrue_def  
                    lift0_def lift1_def o_def)
    apply(drule_tac x=\<tau> in fun_cong, simp)
    apply(simp only: not_down_exists_lift, erule exE)
    apply simp
    done


section {* Final Setup *}

lemma and_left_idem: "P \<MathOclAnd> (P \<MathOclAnd> Q) = P \<MathOclAnd> Q"
  apply (subst and_assoc )
  apply (subst and_idem )
  apply (auto)
  done
  ML {* val and_left_idem = thm "and_left_idem" *}

lemma or_left_idem:  "P \<MathOclOr> (P \<MathOclOr> Q) = P \<MathOclOr> Q"
  apply (subst or_assoc [symmetric])
  apply (subst or_idem)
  apply (auto)
  done
  ML {* val or_left_idem = thm "or_left_idem"  *}

lemma and_left_assoc: "X \<MathOclAnd> (Y \<MathOclAnd> Z) = Y \<MathOclAnd> (X \<MathOclAnd> Z)"
  apply (simp add:and_assoc)
  apply (subst and_commute)
  apply simp
  done

lemma or_left_assoc: "X \<MathOclOr> (Y \<MathOclOr> Z) = Y \<MathOclOr> (X \<MathOclOr> Z)"
  apply (simp add:or_assoc[symmetric])
  apply (subst or_commute)
  apply simp
  done


text{* Now we prepare a setup for \ac{aci} rewriting
  of the boolean operators: These are:
  \begin{align}
      a \circ b       &= b \circ a\\
      (a \circ b) \circ c &= a \circ (b \circ c)\\
      a \circ (b \circ c) &= b \circ (a \circ c)\\
      a \circ a       &= a\\
      a \circ (a \circ b) &= a \circ b
  \end{align}
 *}

lemmas OCL_logic_ACI = 
        and_commute and_assoc[symmetric]
        and_left_assoc and_idem and_left_idem	  
        or_commute or_assoc 
        or_left_assoc or_idem 
	or_left_idem


section{* Boolean Relations *}

subsection{* OCL Equalities *}

text {* In the non-referential universe the following defines the 
        two default equalities. In the referntial universe, these
        definitions define the referntial equalities. 
*}
constdefs OclStrongEq   :: "[('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean"
         "OclStrongEq   \<equiv>  lift\<^sub>2 (\<lambda> x y. \<HolOclLiftLeft>x = y\<HolOclLiftRight> )"

          OclStrictEq   :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
         "OclStrictEq   \<equiv>  lift\<^sub>2 (strictify(\<lambda> x. 
                                strictify(\<lambda> y. \<HolOclLiftLeft>x = y\<HolOclLiftRight> )))"



consts
          OclStrongValueEq        :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean"
          OclStrictValueEq        :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean"
          OclStrongDeepValueEq    :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean"
          OclStrictDeepValueEq    :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean"



syntax  (* standard syntax *)
   "_OclStrongEq_std"  :: "[('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean"
                         ("(1_/ \<OclStrongEq> _)" [63,64]63) 
   "_OclStrictEq_std"    :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/\<OclStrictEq>_)" [63,64]63)
   "_OclStrongValueEq_std"  :: "[('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean"
                         ("(1_/ \<OclStrongValueEq> _)" [63,64]63) 
   "_OclStrictValueEq_std"    :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/\<OclStrictValueEq>_)" [63,64]63)
   "_OclStrongDeepValueEq_std"  :: "[('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean"
                         ("(1_/ \<OclStrongDeepValueEq> _)" [63,64]63) 
   "_OclStrictDeepValueEq_std"    :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/\<OclStrictDeepValueEq>_)" [63,64]63)

syntax   (* standard ascii syntax *)
   "_OclStrongEq_ascii"::"[('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean"
                         ("(1_/ `=` _)" [63,64]63) 
   "_OclStrictEq_ascii"  :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/`==`_)" [63,64]63)
   "_OclStrongValueEq_ascii"::"[('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean"
                         ("(1_/ `~=` _)" [63,64]63) 
   "_OclStrictValueEq_ascii"  :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/`~==`_)" [63,64]63)
   "_OclStrongDeepValueEq_ascii"::"[('\<tau>,'b::bot)VAL,('\<tau>,'b)VAL] \<Rightarrow> '\<tau> Boolean"
                         ("(1_/ `~=~` _)" [63,64]63) 
   "_OclStrictDeepValueEq_ascii"  :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/`~==~`_)" [63,64]63)


syntax (xsymbols) (* mathematical syntax *)
    "_OclStrongEq_math":: "[('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/ \<MathOclStrongEq> _)" [63,64]63) 
    "_OclStrictEq_math"  :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/ \<MathOclStrictEq> _)" [63,64]63)
    "_OclStrongValueEq_math":: "[('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/ \<MathOclStrongValueEq> _)" [63,64]63) 
    "_OclStrictValueEq_math"  :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/ \<MathOclStrictValueEq> _)" [63,64]63)
    "_OclStrongDeepValueEq_math":: "[('\<tau>,'\<alpha>::bot)VAL,('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/ \<MathOclStrongDeepValueEq> _)" [63,64]63) 
    "_OclStrictDeepValueEq_math"  :: "[('\<tau>,'\<alpha>::bot)VAL, ('\<tau>,'\<alpha>)VAL] \<Rightarrow> '\<tau> Boolean" 
                         ("(1_/ \<MathOclStrictDeepValueEq> _)" [63,64]63)
(*<*)
parse_translation {* flat(map (emb trans_const) 
                              ["OclStrongEq","OclStrictEq",
			       "OclStrongValueEq","OclStrictValueEq",
			       "OclStrongDeepValueEq","OclStrictDeepValueEq"]) *}

print_translation {* map emb_print 
                         ["OclStrongEq", "OclStrictEq", "OclRefEq",
			  "OclStrongValueEq","OclStrictValueEq", 
			  "OclStrongDeepValueEq","OclStrictDeepValueEq"] *}
(*>*)
syntax   
     "@OclNeq" :: "['\<tau> Boolean,'\<tau> Boolean] \<Rightarrow> '\<tau> Boolean"     
                 ("_ /`<>` _"  [51,52]51)

translations
     "X `<>` Y" == "\<MathOclNot>(X \<MathOclStrictEq> Y)"

subsection {* Is \holocl{} Faithful to the \OCL standard (Part II)? *} 
text{*\label{Boolean:CombinatorStyleVsTextbookStyle2}*}

lemma strictEq_faithfully_represented:
  "Sem[[X \<MathOclStrictEq> Y]] \<gamma> = (if    Sem[[X]] \<gamma> \<noteq> \<bottom> \<and> Sem[[Y]] \<gamma> \<noteq> \<bottom> 
                     then  \<HolOclLiftLeft> \<HolOclDropLeft> Sem[[X]]\<gamma> \<HolOclDropRight> = \<HolOclDropLeft> Sem[[Y]]\<gamma> \<HolOclDropRight>\<HolOclLiftRight> 
                     else  \<bottom>)"
  apply(simp add:OclStrictEq_def 
              DEF_def o_def 
              lift0_def lift1_def lift2_def
              semfun_def )
  apply (simp_all add: not_down_exists_lift, auto)
  done

text{* This operator---the first binary operator of a strict
  operation---represents directly the definition scheme as imposed
  by~\cite[Definition A. 16]{omg:ocl:2003}.  The operators for Set,
  Bag, Integer and Real will follow this scheme.  *}
  
(*<*)
ML{* OCL_strict_syntax_checking:=false; *}
(*>*)
end
