(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Numerals.thy -
 * Copyright (C) 2005-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* Numerals *}
 
theory OCL_Numerals 
imports 
  "$HOLOCL_HOME/src/OCL_Logic" 
begin
 text_raw{* \label{thy:OCL_Numerals} *}
instance up :: (number)number .. 
instance fun:: (type,number)number ..


subsection {* \OCL Ordering Interface *}
text {* 
  This is ony a syntactic interface for the ordering operators
  in \OCL. Concrete definitions are provided in the structures
  Integer and Real. 
  *}

consts
  OclLe              :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"    
  OclLess            :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  
  OclGe              :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  
  OclGreater         :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  


syntax  (* standard syntax *)
  "_OclLe_std"       :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"    
                         ("(1_/\<OclLe> _)" [63,64]63)
  "_OclLess_std"     :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  
                         ("(1_/\<OclLess> _)" [63,64]63)
  "_OclGe_std"       :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  
                         ("(1_/\<OclGe> _)" [63,64]63)
  "_OclGreater_std"  :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  
                         ("(1_/\<OclGreater> _)" [63,64]63)


syntax
  "_OclLe_ascii"     :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"    
                         ("(1_/`<=` _)" [63,64]63)
  "_OclLess_ascii"   :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  
                         ("(1_/`<` _)" [63,64]63)
  "_OclGe_ascii"     :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  
                         ("(1_/`>=` _)" [63,64]63)
  "_OclGreater_ascii":: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  
                         ("(1_/`>` _)" [63,64]63)

syntax (xsymbols)
  "_OclLe_math"     :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"    
                         ("(1_/\<MathOclLe> _)" [63,64]63)
  "_OclLess_math"   :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  
                         ("(1_/\<MathOclLess> _)" [63,64]63)
  "_OclGe_math"     :: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  
                         ("(1_/\<MathOclGe> _)" [63,64]63)
  "_OclGreater_math":: "[('a,'b)VAL, ('a,'b)VAL] => 'a Boolean"  
                         ("(1_/\<MathOclGreater> _)" [63,64]63)
(*<*)
parse_translation {* flat(map (emb trans_const) 
                              ["OclLe","OclLess","OclGe","OclGreater"]) *}

print_translation {* map emb_print 
                              ["OclLe","OclLess","OclGe","OclGreater"] *}

(*>*)
end

