(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_String.thy -
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* \OCL String *}
 
theory OCL_String 
imports 
  "$HOLOCL_HOME/src/library/basic/OCL_Integer" 
begin
text_raw{* \label{thy:OCL_String} *}

constdefs
  is_String_0     :: "'a U \<Rightarrow> bool" 
  "is_String_0    \<equiv> sum_case (\<lambda>x. False)
                              (sum_case (\<lambda>x. False)
                              (sum_case (\<lambda>x. False)
                              (sum_case (\<lambda>x. False)(\<lambda>x. True))))"

  get_String_0    :: "'a U \<Rightarrow> String_0"
  "get_String_0  \<equiv> sum_case (\<lambda>x. \<some> x. True)
                              (sum_case (\<lambda>x. \<some> x. True)
                              (sum_case (\<lambda>x. \<some> x. True)
                              (sum_case (\<lambda>x. \<some> x. True)(\<lambda>x. x))))"
  mk_String_0     :: "String_0 \<Rightarrow> 'a U"
  "mk_String_0    \<equiv> Inr \<circ> Inr \<circ> Inr o Inr"

lemma get_mk_String_id_0: "get_String_0(mk_String_0 x) = x"
  apply (simp add: get_String_0_def mk_String_0_def)
  done
lemma is_mk_String_0: "is_String_0(mk_String_0 x) = True"
  apply (simp add: is_String_0_def mk_String_0_def)
  done
lemma mk_get_String_id_0: "is_String_0 x \<Longrightarrow> mk_String_0(get_String_0 x) = x"
  apply (simp add: get_String_0_def mk_String_0_def is_String_0_def )
  apply (case_tac x, simp, simp)
  apply (case_tac b, simp, simp)
  apply (case_tac ba, simp, simp) 
  apply (case_tac bb, simp, simp)
 done

constdefs
   String_0  :: "('st, String_0 Set_0)  VAL"
  "String_0  \<equiv> lift0(Abs_Set_0 \<HolOclLiftLeft>lift ` UNIV\<HolOclLiftRight>)"

consts
  "OclConcat"    :: "['a String, 'a String] \<Rightarrow> 'a String"   ("_ ->concat'( _ ')" [66,65]65)
  "OclToLower"   :: "'a String \<Rightarrow> 'a String"                ("_ ->toLower'( _ ')" [66,65]65)
  "OclToUpper"   :: "'a String \<Rightarrow> 'a String"                ("_ ->toUpper'( _ ')" [66,65]65)
  "stringsize"   :: "'a String \<Rightarrow> 'a Integer"               ("_ ->size'( ')" [66])
  "OclSubstring" :: "['a String, 'a Integer, 'a Integer] \<Rightarrow> 'a String"     
                                                            ("_ ->subString'(_ _')"[66,65,65]) 

defs
  OclConcat_def:   "OclConcat    \<equiv> lift\<^sub>2(strictify(\<lambda> x\<Colon>String_0. 
                                                strictify(\<lambda> y. \<HolOclLiftLeft> \<HolOclDropLeft>x\<HolOclDropRight> @ \<HolOclDropLeft>y\<HolOclDropRight> \<HolOclLiftRight>)))"
  stringsize_def:  "stringsize   \<equiv> lift\<^sub>1(strictify(\<lambda> x\<Colon>String_0. \<HolOclLiftLeft> int(size ( \<HolOclDropLeft>x\<HolOclDropRight> )) \<HolOclLiftRight> ))"

  OclSubstring_def:"OclSubstring \<equiv> lift\<^sub>3(strictify(\<lambda> s\<Colon>String_0. 
                                                   strictify(\<lambda> l\<Colon>Integer_0. 
                                                   strictify(\<lambda> u\<Colon>Integer_0. 
                                        \<HolOclLiftLeft> (List.take (nat(( \<HolOclDropLeft>u\<HolOclDropRight> -  \<HolOclDropLeft>l\<HolOclDropRight>  + 1\<Colon>int  ))) 
                                          (List.drop (nat( \<HolOclDropLeft>l\<HolOclDropRight>  - 1\<Colon>int)) \<HolOclDropLeft>s\<HolOclDropRight>  )) \<HolOclLiftRight> ) )))"

(* Todo: Defs for toLower and toUpper *)
end


    
