(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OclUndefined.thy - Type Definition for OclUndefined
 * Copyright (C) 2004-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
*                          Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* Type Definition for OclUndefined *}

theory OCL_Undefined 
imports 
  "$HOLOCL_HOME/src/OCL_Kernel" 
begin
text_raw{* \label{thy:OCL_Undefined} *}
text {*
  As previously mentioned, all types, let it be basic
  types or class types, my contain the element @{const "\<bottom>"} 
  denoting undefinedness. In classifier types, this can be interpreted 
  as  equivalent in unboxed types to a null-reference. 
  Therefore, there is a universal constant in \OCL,
  called $\isasymMathOclUndefined$, that is constructed as follows:  
  *}
constdefs 
     OclUndefined   :: "('a,'b\<Colon>bot) VAL" 
    "OclUndefined   \<equiv> lift\<^sub>0 \<bottom>" 

syntax  (* standard syntax *)
    "_OclUndefined_std"      :: "('a,'b\<Colon>bot) VAL"   ("\<OclUndefined>")
			    
syntax  (* standard syntax *)
    "_OclUndefined_ascii"    :: "('a,'b\<Colon>bot) VAL"   ("undef")

syntax  (* standard syntax *)
    "_OclUndefined_math"     :: "('a,'b\<Colon>bot) VAL"   ("\<MathOclUndefined>")

(*<*)
parse_translation {*
  let
  in emb trans_const "OclUndefined"
  end;
*}

print_translation {*
  let
  in [emb_print "OclUndefined"]
  end;
*}
(*>*)

section{* Rules describing the effect of strictify on the \OCL level *}

lemma fw_OCL_undef_1_lift1'_0:
  assumes f_def: "f \<equiv> (lift1' lift_arg0) (strictify P)"
  shows          "f OclUndefined = OclUndefined"
  by(simp add: f_def OclUndefined_def ss_lifting')

text{* binary rules *}

lemma fw_OCL_undef_1_lift2'_00:
  assumes f_def: "f \<equiv> (lift2' lift_arg0 lift_arg0) (strictify P)"
  shows          "f OclUndefined Y = OclUndefined"
  by(simp add: f_def OclUndefined_def ss_lifting')

lemma fw_OCL_undef_1_lift2'_01:
  assumes f_def: "f \<equiv> (lift2' lift_arg0 lift_arg1) (strictify P)"
  shows          "f OclUndefined Y = OclUndefined"
  by(simp add: f_def OclUndefined_def ss_lifting')

lemma fw_OCL_undef_2_lift2'_00:
  assumes f_def: "f \<equiv> (lift2' lift_arg0 lift_arg0) (\<lambda> x. strictify (P x))"
  shows          "f X OclUndefined = OclUndefined"
  by(simp add: f_def OclUndefined_def ss_lifting')

text{* tertiary rules *}

lemma fw_OCL_undef_1_lift3'_000:
  assumes f_def: "f \<equiv> (lift3' lift_arg0 lift_arg0 lift_arg0) (strictify P)"
  shows          "f OclUndefined Y Z = OclUndefined"
  by(simp add: f_def OclUndefined_def ss_lifting')

lemma fw_OCL_undef_1_lift3'_020:
  assumes f_def: "f \<equiv> (lift3' lift_arg0 lift_arg2 lift_arg0) (strictify P)"
  shows          "f OclUndefined Y Z = OclUndefined"
  by(simp add: f_def OclUndefined_def ss_lifting')

lemma fw_OCL_undef_3_lift3'_000:
  assumes f_def: "f \<equiv> (lift3' lift_arg0 lift_arg0 lift_arg0) (\<lambda> x y. strictify (P x y))"
  shows          "f X Y OclUndefined = OclUndefined"
  by(simp add: f_def OclUndefined_def ss_lifting')

lemma fw_OCL_undef_3_lift3'_020:
  assumes f_def: "f \<equiv> (lift3' lift_arg0 lift_arg2 lift_arg0) (\<lambda> x y. strictify (P x y))"
  shows          "f X Y OclUndefined = OclUndefined"
  by(simp add: f_def OclUndefined_def ss_lifting')


(*<*)
section{* Setting up ocl\_setup\_op *}

text{* This is the setup of all strictness rules
  \todo{Move also the OLD rules to this place.}
*}

ML {*

val _ = let

  fun undef_rules position rname = 
    let val position' = Int.toString position;
        val rname' = "fw_OCL_undef_" ^ position' ^ rname;
    in  RULE
         {rule      = (rname', thm rname'),
          simp_tac  = fn _ => fn x => x,
          is_atomic = fn _ => true, 
          register  = default_register_simponly, 
          name_gen  = default_namegen ("OCL_undef_"^ position' ^"_")
         }
    end

in

  add_setup_rules (
    (map (undef_rules 1)
         ["_lift1'_0", "_lift2'_00", "_lift2'_01", "_lift3'_000", "_lift3'_020"])
    @ (map (undef_rules 2) ["_lift2'_00"])
    @ (map (undef_rules 3) ["_lift3'_000", "_lift3'_020"])
  )



end


*}

ML {*

 length (!setup_rulebase);

*}

(*>*)
end
