(* ****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Sequence.thy -
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNES 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ***************************************************************************** *)

header {* \OCL Sequence *}
theory OCL_Sequence 
imports 
  "$HOLOCL_HOME/src/library/collection/OCL_Collection" 
begin
 text_raw{* \label{thy:OCL_Sequence} *}


section{* Theorems needed in this theory but not belonging to it *}


subsection{* Boolean calculus on the semantic level *}

(* TODO: put this lemma at the right place *)
lemma boolean_cases_sem:
  assumes undefC: "X St = \<bottom> \<Longrightarrow> R"
  and     trueC:  "X St = \<HolOclLiftLeft>True\<HolOclLiftRight>  \<Longrightarrow> R"
  and     falseC: "X St = \<HolOclLiftLeft>False\<HolOclLiftRight>  \<Longrightarrow> R"
  shows           "R"
  apply(rule disjE)
  prefer 2 apply(erule undefC)
  prefer 2 apply(rule disjE)
  prefer 2 apply(rotate_tac -1, erule trueC)
  prefer 2 apply(rotate_tac -1, erule falseC)
  apply(auto)
  apply(simp add: not_down_exists_lift)
  apply(erule exE, auto)
  done

subsection{* Some cp theory to simplify the later proofwork *}
(* TODO: move this statements *)

text{* NOTE: The combinator cp2 isn't used anymore in this theory
and will be removed from the OCLBag theory as well. These theorems
are here because they give some insight on the properties of cp-ness.

There are places where some functions must be contextpassing in their
first and their second argument. This is now written explicitely. What
remains is to decide if the current style is the right one.

What is really needed very much is tool support with relation to the
proofing of properties depending on cp-ness.
*}

(*
lemma cp2_by_cp:
  "\<lbrakk> !!x. cp (P x); !!x. cp (\<lambda> y. P y x) \<rbrakk> \<Longrightarrow> cp2 P"
  by(simp add: cp2_def cp_by_cpify)

lemma cp2_imp_cp:
  "\<lbrakk> cp2 P; cp f; cp g \<rbrakk> \<Longrightarrow> cp (\<lambda> x. P (f x) (g x))"
  apply(simp add: cp2_def cp_by_cpify lift0_def)
  apply(auto)
  apply(rule_tac t="(P (f X) (g X) \<tau>)" and s="(P (\<lambda>s. (f X \<tau>)) (\<lambda>s. (g X \<tau>)) \<tau>)" in subst) 
  apply(simp)
  apply(rule_tac t="(P (f (\<lambda>s. (X \<tau>))) (g (\<lambda>s. (X \<tau>))) \<tau>)" 
             and s="(P (\<lambda>s.(f (\<lambda>s. (X \<tau>))) \<tau>) (\<lambda>s. (g (\<lambda>s. (X \<tau>))) \<tau>) \<tau>)" in subst)
  prefer 2
  apply(simp)
  apply(drule_tac x="f (\<lambda>s. (X \<tau>))" in allE)
  apply(rotate_tac -1)
  apply(drule_tac x="g (\<lambda>s. (X \<tau>))" in allE)
  apply(rotate_tac -1)
  apply(drule_tac x="\<tau>" in allE)
  prefer 2 prefer 3 prefer 4
  apply(rule sym)
  apply(assumption)+
  done

lemma cp2_cp_conv:
  "cp2 P = ((\<forall> x. cp (P x)) \<and> (\<forall> x. cp (\<lambda> y. P y x)))"
  apply(auto)
  apply(rule_tac f="\<lambda> x. x" and P="\<lambda>u ua. (P ua u)" in cp2_imp_cp)
  prefer 4
  apply(rule_tac f="\<lambda> x. x" and P="P" in cp2_imp_cp)
  prefer 7
  apply(simp_all add: cp_by_cpify cp2_def lift0_def)
  done
*)
(* TODO: which should be simplification rules and which form should they
         have.

         This concerns most, the relation of the following three rules
         and cp2_by_cp. Probably a better destruction rule would be nice. *)
(*
lemma cp2_imp_cp_unary:
  "cp2 P \<Longrightarrow> cp (\<lambda> x. P x x)"
  by(simp add: cp2_def cp_by_cpify)

lemma cp2_imp_cp_1:
  "cp2 P \<Longrightarrow> (cp (\<lambda> y. P y z))"
  apply(rule_tac f="\<lambda> x. x" and P="P" in cp2_imp_cp)
  apply(simp_all add: cp_by_cpify cp2_def lift0_def)
  done

lemma cp2_imp_cp_2:
  "cp2 P \<Longrightarrow> (cp (P x))"
  apply(rule_tac f="\<lambda> x. x" and P="\<lambda>u ua. (P ua u)" in cp2_imp_cp)
  apply(simp_all add: cp_by_cpify cp2_def lift0_def)
  done

text{* A sometimes quite useful substitution rule *}

lemma cp_subst:
  "\<lbrakk> cp P \<rbrakk> \<Longrightarrow> P x \<tau> = P (lift0 (x \<tau>)) \<tau>"
  by(subst (asm) cp_by_cpify, auto)

text{* and one formulation of a destruction rule for cp2 *}

lemma cp2D:
  "cp2 P \<Longrightarrow> (P x y \<tau>) = (P (lift0 (x \<tau>)) (lift0 (y \<tau>)) \<tau>)"
  by(unfold cp2_def, auto simp: lift0_def)


*)
(* ***************** *)
(* End of cp theory *)
(* ***************** *)

subsection{* Relating definedness and strictness of functions *}

(* TODO: how is this more nicely done? tranforming !! into \<forall> *)

lemma defopt_imp_undef_1:
  assumes defopt_P: "!! x y \<tau>. (\<tau> \<MathOclValid> \<MathOclIsDefined> (P x y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> x) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> y))"
  shows             "P \<MathOclUndefined> b = \<MathOclUndefined>"
  apply(subgoal_tac "\<forall> x y \<tau>. (\<tau> \<MathOclValid> \<MathOclIsDefined> (P x y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> x) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> y))")
  apply(rule ext)
  apply(erule_tac x="\<MathOclUndefined>" in allE)
  apply(erule_tac x="b" in allE)
  apply(erule_tac x="x" in allE)
  apply(simp add: OclIsDefined_def OclLocalValid_def
                  OclTrue_def OclUndefined_def
                  lift0_def lift1_def o_def DEF_def)
  apply(auto intro!: defopt_P)
  done

lemma defopt_imp_undef_2:
  assumes defopt_P: "!! x y \<tau>. (\<tau> \<MathOclValid> \<MathOclIsDefined> (P x y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> x) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> y))"
  shows             "P a \<MathOclUndefined> = \<MathOclUndefined>"
  apply(subgoal_tac "\<forall> x y \<tau>. (\<tau> \<MathOclValid> \<MathOclIsDefined> (P x y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> x) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> y))")
  apply(rule ext)
  apply(erule_tac x="a" in allE)
  apply(erule_tac x="\<MathOclUndefined>" in allE)
  apply(erule_tac x="x" in allE)
  apply(simp add: OclIsDefined_def OclLocalValid_def
                  OclTrue_def OclUndefined_def
                  lift0_def lift1_def o_def DEF_def)
  apply(auto intro!: defopt_P)
  done


section{* Properties of the Congruence inside the Datatype Adaption *}

text{* The following rules are transformed versions of the automatically
generated rules for datatype conversions. *}


lemma Abs_Sequence_inject_charn:
  "\<lbrakk> (\<bottom>::'a::bot) \<notin> set x; (\<bottom>::'a::bot) \<notin> set y \<rbrakk> \<Longrightarrow> 
   (((Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Sequence_0 \<HolOclLiftLeft>y\<HolOclLiftRight>)) = (x = y))"
  by(subst Abs_Sequence_0_inject, 
     simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)

(* TODO: which form is to be preferred *)

(* as an elimination rule:


lemma Abs_Sequence_inject_absurd1:
  "\<lbrakk> (Abs_Sequence_0 down) = (Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>); (\<bottom>::'a::bot) \<notin> set x \<rbrakk> \<Longrightarrow> P"
  by(subst (asm) Abs_Sequence_0_inject,
     simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)


  or as a simplifier rule:

  - are both of them needed?

*)

lemma Abs_Sequence_inject_absurd11 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x \<rbrakk> \<Longrightarrow> ((Abs_Sequence_0 down) = (Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(subst Abs_Sequence_0_inject,
     simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)

lemma Abs_Sequence_inject_absurd12 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x \<rbrakk> \<Longrightarrow> ((Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Sequence_0 down)) = False"
  by(subst Abs_Sequence_0_inject,
     simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)

lemma Abs_Sequence_inject_absurd21 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x \<rbrakk> \<Longrightarrow> ( \<bottom> = (Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(simp add: UU_Sequence_def)

lemma Abs_Sequence_inject_absurd22 [simp]:
  "\<lbrakk> \<bottom> \<notin> set x \<rbrakk> \<Longrightarrow> ((Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<bottom>) = False"
  by(simp add: UU_Sequence_def)

lemma Abs_Sequence_inject_charn:
  "\<lbrakk> (\<bottom>::'a::bot) \<notin> set x; (\<bottom>::'a::bot) \<notin> set y \<rbrakk> \<Longrightarrow> 
   (((Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Sequence_0 \<HolOclLiftLeft>y\<HolOclLiftRight>)) = (x = y))"
  by(subst Abs_Sequence_0_inject, 
     simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)


lemma Abs_Sequence_inverse_charn [simp]:
  "(\<bottom> \<notin> set x) \<Longrightarrow> Rep_Sequence_0 (Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<HolOclLiftLeft>x\<HolOclLiftRight>"
  by (subst Abs_Sequence_0_inverse, 
      simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)

lemma Abs_Sequence_inverse_charn2 [simp]:
  "(\<bottom> \<notin> set x) \<Longrightarrow> \<HolOclDropLeft>Rep_Sequence_0 (Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)\<HolOclDropRight> = x"
  by (subst Abs_Sequence_0_inverse, 
      simp_all add: OCL_Sequence_type.Sequence_0_def smash_def)
(*
lemma Abs_Sequence_cases_charn:
  assumes bottomCase : "\<lbrakk> x = \<bottom> \<rbrakk> \<Longrightarrow> P"
  assumes listCase   : "\<And>y. (\<lbrakk> x = Abs_Sequence_0 \<HolOclLiftLeft>y\<HolOclLiftRight>; \<bottom> \<notin> set y \<rbrakk> \<Longrightarrow> P)"
  shows                "P"
  apply(rule_tac x="x" in Abs_Sequence_0_cases)
  apply(case_tac "y = \<bottom>")
  apply(rule bottomCase, simp add: UU_Sequence_def)
  apply(frule not_down_exists_lift2[THEN iffD1])
  apply(rule_tac y="\<HolOclDropLeft>y\<HolOclDropRight>" in listCase)
  apply(auto simp: OCL_Sequence_type.Sequence_0_def smash_def)
  done

lemma Abs_Sequence_induct_charn:
  assumes bottomCase : "P \<bottom>"
  assumes stepCase   : "\<And>y. (\<lbrakk> \<bottom> \<notin> set y \<rbrakk> \<Longrightarrow> P (Abs_Sequence_0 \<HolOclLiftLeft>y\<HolOclLiftRight>))"
  shows                "P x"
  apply(rule_tac x="x" in Abs_Sequence_0_induct)
  apply(rule_tac  x="Abs_Sequence_0 y" in Abs_Sequence_cases_charn)
  apply(auto intro!: bottomCase elim: stepCase)
  done

lemma inj_on_Abs_Sequence: "inj_on Abs_Sequence_0 Sequence_0";
  by(rule inj_on_inverseI,rule Sequence_0.Abs_Sequence_0_inverse,assumption)

lemma inj_Rep_Sequence : "inj Rep_Sequence_0"
  by(rule inj_on_inverseI, rule Sequence_0.Rep_Sequence_0_inverse)

lemma smashed_sequence_charn: 
  "(\<bottom> \<notin> set X) = ((\<HolOclLiftLeft>X\<HolOclLiftRight>) \<in> Sequence_0)"
  by(unfold smash_def Sequence_0_def UU_Sequence_def,auto) (*characterization 1 *)

lemma UU_in_smashed_sequence [simp]: 
  "\<bottom> \<in> Sequence_0"
  by(unfold smash_def Sequence_0_def UU_Sequence_def,auto) (*characterization 2a *)

lemma down_in_smashed_sequence [simp]: 
  "down \<in> Sequence_0"
  by(unfold smash_def Sequence_0_def UU_Sequence_def,auto) (*characterization 2b *)

lemma mt_in_smashed_sequence[simp]: 
  "(\<HolOclLiftLeft>[]\<HolOclLiftRight>) \<in> Sequence_0"
  by (unfold smash_def Sequence_0_def,auto)

lemma DEF_Abs_Sequence: "\<And>X. (\<bottom> \<notin> set X) \<Longrightarrow> DEF (Abs_Sequence_0 (\<HolOclLiftLeft>X\<HolOclLiftRight>))"
  apply(unfold DEF_def UU_Sequence_def, simp)
  done

lemma DEF_Rep_Sequence: 
  "\<And>X. DEF X \<Longrightarrow> DEF (Rep_Sequence_0 X)"
  apply(unfold DEF_def UU_Sequence_def,auto)
  apply(drule_tac f = "Abs_Sequence_0" in arg_cong)
  apply(simp add: Rep_Sequence_0_inverse)
  done

lemma not_DEF_Rep_Sequence: 
  "\<And>X. \<not> DEF X \<Longrightarrow> \<not> DEF (Rep_Sequence_0 X)"
  apply(unfold DEF_def UU_Sequence_def,auto)
  apply(simp add: Abs_Sequence_0_inverse)
  done


lemma exists_lift_Sequence: 
  "\<And>X. DEF X \<Longrightarrow> \<exists>c. Rep_Sequence_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
  by (drule DEF_Rep_Sequence,simp add: DEF_X_up)

lemma exists_lift_Sequence2: 
  "\<And>X. DEF X \<Longrightarrow> \<exists>c. \<bottom> \<notin> set c \<and> Rep_Sequence_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
  apply(frule exists_lift_Sequence)
  apply(auto)
  apply(drule sym, rule swap)
  prefer 2; apply assumption;back;
  apply(subst smashed_sequence_charn)
  apply(simp add: Rep_Sequence_0)
  done

lemma Rep_Sequence_cases: 
  "Rep_Sequence_0 X = \<bottom> \<or> (\<exists>c. \<bottom> \<notin> set c \<and> Rep_Sequence_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply(case_tac "DEF X")
  apply(drule exists_lift_Sequence2)
  apply(drule_tac [2] not_DEF_Rep_Sequence)
  apply(simp_all add: DEF_def)
  done


lemma DEF_X_Sequence_0 : "DEF(X) = (\<exists>c. \<bottom> \<notin> set c \<and> Rep_Sequence_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply(insert  Rep_Sequence_cases [of "X"], auto)
  apply(drule exists_lift_Sequence,auto)
  apply(rule swap)
  prefer 2
  apply(rule not_DEF_Rep_Sequence, auto)
  done


text{* The following lemma is crucial for thy\_morpher: *} 
lemma DEF_X_Sequence: "DEF X = (\<exists>c.  \<bottom> \<notin> set c \<and> X = Abs_Sequence_0 (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply (auto simp: DEF_Abs_Sequence)
  apply (simp add:DEF_X_Sequence_0)
  apply (erule exE, rule exI, auto)
  apply (rule injD [OF inj_Rep_Sequence])
  apply (auto simp: smashed_sequence_charn Abs_Sequence_0_inverse)
  done 

text{* These four lemmas are used by ocl\_setup\_op to reason about definedness: *}

lemma lift2_strict_is_isdef_fw_Sequence_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Sequence_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Sequence_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; (\<bottom>::'a::bot) \<noteq> b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b))"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def lift0_def lift1_def lift2_def
                  strictify_def DEF_def)
  apply(rule ext)
  apply(case_tac "DEF (X St)")
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(auto simp: strictify_def DEF_def)
  apply(drule_tac b="(Y St)" in  inv_g)
  prefer 2
  apply(drule DEF_Abs_Sequence, simp add: DEF_def, auto)
  done

lemma lift2_strict_is_isdef_fw_Sequence_Sequence:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Sequence_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Sequence_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Sequence_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; \<bottom> \<notin> set b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b))"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def lift0_def lift1_def lift2_def
                  strictify_def DEF_def)
  apply(rule ext)
  apply(case_tac "DEF (X St)")
  apply(case_tac "DEF (Y St)")
  apply(drule DEF_X_Sequence_0[THEN iffD1])+
  apply(auto simp: strictify_def DEF_def)
  apply(drule_tac b="ca" in inv_g, assumption)
  apply(rotate_tac -1)
  apply(drule DEF_Abs_Sequence, simp add: DEF_def)
  done

lemma lift2_strictify_implies_LocalValid_defined_Sequence_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Sequence_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Sequence_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; (\<bottom>::'a::bot) \<noteq> b\<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b))"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Sequence_Val)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done

lemma lift2_strictify_implies_LocalValid_defined_Sequence_Sequence:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Sequence_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Sequence_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Sequence_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> \<bottom> \<notin> set a; \<bottom> \<notin> set b \<rbrakk> \<Longrightarrow> \<bottom> \<notin> set (g a b))"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Sequence_Sequence)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done
*)
section{* Characteristic Set of Sequences (parametric): *}

(* TODO: Shouldn't one take only the permutations of the list? *)

constdefs
   Sequence:: "('st, ('b::bot) Sequence_0) VAL \<Rightarrow> ('st, 'b Sequence_0 Set_0) VAL"
  "Sequence \<equiv> lift1(strictify (\<lambda>X.
                   Abs_Set_0 \<HolOclLiftLeft>(Abs_Sequence_0 o lift) ` (lists (set  \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>))\<HolOclLiftRight> ))"

ocl_setup_op [Sequence]

(*
lemma Sequence_defined:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined> X) = (\<tau> \<MathOclValid> \<MathOclIsDefined> (Sequence X))"
  apply(simp add: Sequence_def localValidDefined2sem
                  strictify_def lift1_def o_def image_def DEF_def)
  apply(rule impI, fold DEF_def)
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(rule DEF_Abs_Set, auto simp: UU_Sequence_def)
  done 
*)

constdefs
    OclMtSequence   :: "('a, 'b::bot Sequence_0) VAL"    
   "OclMtSequence  \<equiv> lift0(Abs_Sequence_0(\<HolOclLiftLeft>[]\<HolOclLiftRight>))"

    syntax  (* standard syntax *)
    "_OclMtSequence_std"   :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("\<OclMtSequence>")
    syntax  (* ascii syntax *)
    "_OclMtSequence_ascii" :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("mtSequence")
    syntax  (xsymbols)
    "_OclMtSequence_math"  :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("\<MathOclMtSequence>")
    parse_translation {* flat(map (emb trans_const) ["OclMtSequence"]) *}
    print_translation {* map emb_print ["OclMtSequence"] *}




section{* Foundational Properties of MtSequence *}

lemma RepAbs_Mt[simp]:
  "Rep_Sequence_0 (Abs_Sequence_0 \<HolOclLiftLeft>[]\<HolOclLiftRight>) = \<HolOclLiftLeft>[]\<HolOclLiftRight>"
  by(simp)
(*
lemma Abs_Sequence_Mt_DEF[simp]:
  "DEF(Abs_Sequence_0 \<HolOclLiftLeft>[]\<HolOclLiftRight>)"
  by(simp add: DEF_X_Sequence_0)
*)
(* TODO: there are more general properties that the empty
         sequence is defined in every context. Do we need that?
*)
(*
lemma OCL_is_isdef_OclMtSequence [simp]:
  "\<MathOclValid> \<MathOclIsDefined> \<MathOclMtSequence>"
  by(simp add: sym[OF global_vs_local_validity] Abs_Sequence_Mt_DEF 
               localValidDefined2sem OclMtSequence_def lift0_def)

lemma OCL_is_defopt_OclMtSequence [simp]:
  "\<tau> \<MathOclValid> \<MathOclIsDefined> \<MathOclMtSequence>"
  by(rule OCL_is_isdef_OclMtSequence[THEN valid_elim])

*)
section{* Foundational Properties of includes and excludes *}

defs 
   (* includes     :: "[('a,'b::bot) Sequence, ('a,'b) VAL] => 'a Boolean" *)
   OclIncludes_def : "OclIncludes \<equiv> lift2(strictify (\<lambda>X. strictify (\<lambda>x.
                                      \<HolOclLiftLeft>x \<in> set \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> \<HolOclLiftRight>)))"

   (* excludes     :: "[('a,'b::bot) Sequence, ('a,'b) VAL] => 'a Boolean" *)
   OclExcludes_def : "OclExcludes \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x. 
                                      \<HolOclLiftLeft>x \<notin> set \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> \<HolOclLiftRight>)))"

ocl_setup_op [OclIncludes, OclExcludes]


lemma includes_charn1 :
  "(\<tau> \<MathOclValid> x \<MathOclIncludes> S) \<Longrightarrow> ((x \<tau>) \<in> set \<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight>)" 
  apply(frule isDefined_if_valid)
  apply(drule OCL_is_defopt_OclIncludes[THEN iffD1])
  apply(simp add: OclLocalValid_def OclIsDefined_def strictify_def
                  OclNot_def OclIncludes_def
                  o_def lift0_def lift1_def lift2_def
                  OclTrue_def OclFalse_def  OclUndefined_def DEF_def) 
  done

(*

lemma includes_charn2 :
  "\<lbrakk> DEF(S \<tau>); (x \<tau>) \<in> set \<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight> \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> x \<MathOclIncludes> S"
  apply(frule DEF_X_Sequence [THEN iffD1], erule exE)
  apply(simp add: OclLocalValid_def OclIsDefined_def strictify_def
                  OclNot_def OclIncludes_def
                  o_def lift0_def lift1_def lift2_def
                  OclTrue_def OclFalse_def  OclUndefined_def DEF_def)
  apply(auto)
  done
*)

section{* Higher Properties of includes and excludes *}


(* We prefer a normal form where excludes is eliminated. *)
lemma exludes_not_includes[simp]: 
  "x \<MathOclExcludes> X = \<MathOclNot> (x::('a, 'b)VAL) \<MathOclIncludes> (X::('a, 'b::bot) Sequence)"
  apply (rule ext)
  apply (simp add:  OclExcludes_def OclIncludes_def)
  apply (simp only: OclNot_def)
  apply (simp only: OclIncludes_def OclExcludes_def lift1_def lift2_def 
                    strictify_def o_def not_def)
  apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def 
                               split add: split_if up.split)
  done

(*
lemma mtSequence_Includes_charn :
  " \<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow> 
    \<tau> \<MathOclValid> (x::('a, 'b)VAL) \<MathOclExcludes> (\<MathOclMtSequence>::('a, 'b::bot) Sequence)"
  apply(simp add: localValidDefined2sem)
  apply(simp only: exludes_not_includes OclMtSequence_def OclLocalValid_def
                  OclIncludes_def  lift0_def lift1_def lift2_def 
                  strictify_def o_def OclNot_def OclTrue_def DEF_def) 
  apply(simp add: Abs_Sequence_Mt_DEF [simplified DEF_def] strictify_def)
  done


lemma mtSequence_in_Sequence :
  " \<tau> \<MathOclValid> \<MathOclIsDefined> (S::('a, 'b::bot)Sequence) \<Longrightarrow> 
    \<tau> \<MathOclValid> (\<MathOclMtSequence>::('a,'b::bot)Sequence) \<MathOclIncludes> (Sequence  S)"
  apply(frule Sequence_defined[THEN iffD1])
  apply(simp add: localValidDefined2sem)
  apply(simp add: Sequence_def OCL_Set.OclIncludes_def OclMtSequence_def
                  lift0_def lift1_def lift2_def strictify_def
                  o_def image_def localValid2sem UU_Sequence_def UU_Set_def)
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(frule DEF_X_Set[THEN iffD1])  
  apply(frule DEF_X_Set_0[THEN iffD1])
  apply(auto)
  apply(subst (asm) Abs_Set_inject_charn)
  apply(auto simp: UU_Sequence_def)
  done


lemma X_in_Sequence_X :
  " \<tau> \<MathOclValid> \<MathOclIsDefined> (S::('a, 'b::bot)Sequence) \<Longrightarrow> 
    \<tau> \<MathOclValid> S \<MathOclIncludes> (Sequence S)"
  apply(frule Sequence_defined[THEN iffD1])
  apply(simp add: localValidDefined2sem)
  apply(simp add: Sequence_def OCL_Set.OclIncludes_def OclMtSequence_def
                  lift0_def lift1_def lift2_def strictify_def
                  o_def image_def localValid2sem UU_Sequence_def UU_Set_def)
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(frule DEF_X_Set[THEN iffD1])  
  apply(frule DEF_X_Set_0[THEN iffD1])
  apply(auto)
  apply(subst (asm) Abs_Set_inject_charn)
  apply(auto simp: UU_Sequence_def)
  done
*)

subsection{* Consequences of Smashedness of Sequences *}

text{* This is the characteristic property of smashed sequences:
       if \verb+X ->includes(x)+ is valid, we know that x must be defined. *}

lemma smashed_sequences_charn:
  "\<tau> \<MathOclValid> ((x::('a,'b)VAL) \<MathOclIncludes> (X::('a,'b::bot)Sequence)) \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsDefined> x"
  apply (simp add: OclIncludes_def lift2_def lift1_def lift0_def 
                   strictify_def OclTrue_def OclIsDefined_def o_def OclLocalValid_def)
  apply(case_tac "X \<tau> = \<bottom>", auto)
  apply(simp add: strictify_def DEF_def)
  apply(case_tac "x \<tau> = \<bottom>", auto)
  done
section{* The conversion to sequences: the identity *}

defs
   (* asSequnece     :: "('a,'b::bot) Sequence \<Rightarrow> ('a,'b::bot) Sequence" *)
   OclAsSequence_def : 
   "OclAsSequence    \<equiv> id"

lemma OclAsSequence_identity[simp]:
  "OclAsSequence (S::('a,'b::bot)Sequence) = S"
  by(simp add: OclAsSequence_def)


section{* The conversion to sets *}

defs
   (* asSet     :: "('a,'b::bot) Sequence \<Rightarrow> ('a,'b::bot) Set" *)
   OclAsSet_def : 
   "OclAsSet    \<equiv> lift1 (strictify (\<lambda>X. Abs_Set_0 \<HolOclLiftLeft>set \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>\<HolOclLiftRight>))"

subsection{* Helper theorems about the relations of the Set and Sequence datatypes *}

(*
lemma set_Sequence_eq_Set_set: 
  "(\<bottom> \<notin> set x) \<Longrightarrow>
   set \<HolOclDropLeft>(Rep_Sequence_0 (Abs_Sequence_0 \<HolOclLiftLeft>x\<HolOclLiftRight>))\<HolOclDropRight> = 
   \<HolOclDropLeft>(Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft> set x \<HolOclLiftRight>))\<HolOclDropRight>  "
  by(simp)
*)

(*
lemma DEF_Set_if_DEF_Sequence:
  "(DEF X) \<Longrightarrow> (DEF (Abs_Set_0 \<HolOclLiftLeft>(set \<HolOclDropLeft>(Rep_Sequence_0 X)\<HolOclDropRight>)\<HolOclLiftRight>))"
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(auto)
thm DEF_Abs_Set
  apply(erule DEF_Abs_Set)
  done

lemma in_over_RepAbs:
"\<bottom> \<notin> S \<Longrightarrow> (x \<in> \<HolOclDropLeft>(Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft>S\<HolOclLiftRight>))\<HolOclDropRight>)= (x \<in> S)"
by(simp)

*)
subsection{* Basic properties of OclAsSet *}

ocl_setup_op [OclAsSet]
 
(*
lemma OCL_is_defopt_OclAsSet: 
  "(\<tau> \<MathOclValid> \<MathOclIsDefined> (((self::('a,'b::bot)Sequence) ->asSet())::(('a, 'b::bot)Set))) = (\<tau> \<MathOclValid> \<MathOclIsDefined> self)"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule iffI) prefer 2
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp add: OclAsSet_def strictify_def lift1_def DEF_def)
  apply(fold DEF_def)
  apply(erule DEF_Set_if_DEF_Sequence)
  apply(auto simp: OclAsSet_def strictify_def lift1_def DEF_def)
done

lemma OCL_is_def_OclAsSet:
  "(\<MathOclIsDefined> (((self::('a,'b::bot)Sequence) ->asSet())::(('a, 'b::bot)Set))) = (\<MathOclIsDefined> self)"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule ext)
  apply(rule arg_cong) back
  apply(subgoal_tac "((St \<MathOclValid> (\<MathOclIsDefined> \<MathOclAsSet> self)) = (St \<MathOclValid> (\<MathOclIsDefined> self)))")
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule OCL_is_defopt_OclAsSet)
  done

lemma OclAsSet_finite:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,'b::bot)Sequence)) =
   (\<tau> \<MathOclValid> \<MathOclIsDefined>(((OclAsSet S)::('a,'b::bot)Set)->size()))"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule iffI)
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclAsSet_def OCL_Set.OclSize_def
                      strictify_def lift1_def DEF_def)
  apply(auto)
  done

lemma OclAsSet_mtSequence_mtSet_conv[simp]:
  "((OclAsSet (\<MathOclMtSequence>::('a,'b::bot)Sequence))) = (\<MathOclMtSet>::('a,'b::bot)Set)"
  apply(insert Abs_Sequence_Mt_DEF)
  apply(simp add: OclAsSet_def OclMtSequence_def OclMtSet_def
                  lift0_def lift1_def strictify_def DEF_def)
  done

 
subsection{* Relating includes and excludes on Sequences to their counterparts on Sets *}

lemma includes_seq2set:
  "(x::('a,'b)VAL) \<MathOclIncludes> (S::('a,'b::bot)Sequence) =
   x \<MathOclIncludes> ((OclAsSet S)::('a,'b::bot)Set)" 
  apply(rule ext)
  apply(case_tac "DEF (S xa)")
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclIncludes_def OCL_Set.OclIncludes_def
                  OclAsSet_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  done

lemma excludes_seq2set:
  "(x::('a,'b)VAL) \<MathOclExcludes> (S::('a,'b::bot)Sequence) =
   x \<MathOclExcludes> ((OclAsSet S)::('a,'b::bot)Set)"
  by(simp add: includes_seq2set)


section{* Semantic Definitions of the Standard Sequence Operations *}

(* TODO: better text ;) *)

text{* Most of the functions are speciefied as mappings to HOL.
      
       The only exception is the function sum. Because of the requirement
       that all types supporting ${plus,zero}$ should be summable.

*}
*)
defs 
   (* size      :: "('a,'b::bot )Sequence => 'a Integer" *)
   OclSize_def  : "OclSize \<equiv> lift1 (strictify (\<lambda>X. 
                             \<HolOclLiftLeft>int (Nat.size \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>)\<HolOclLiftRight>))"


   (* count     :: "[('a,'b::bot )Sequence, ('a,'b)VAL] => 'a Integer" *)
   OclCount_def : "OclCount \<equiv> lift2 (strictify   (\<lambda>X. strictify (\<lambda>x. 
                              \<HolOclLiftLeft>int (Nat.size(List.filter (op = x) \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>))\<HolOclLiftRight>)))"


   (* includesAll     :: "[('a,'b::bot)Sequence, ('a,'b)Sequence] => 'a Boolean" *)
   OclIncludesAll_def: "OclIncludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                                           \<HolOclLiftLeft>set \<HolOclDropLeft>Rep_Sequence_0 Y\<HolOclDropRight> \<subseteq> set \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>\<HolOclLiftRight>)))" 
(* OclIncludesAll_def : 
   "((self::('a,'b::bot)Sequence)->includesAll((obj :: ('a,'b)Sequence))) 
    \<equiv> (obj ->forAll(elem | (self ->includes((elem ::('a,'b)Sequence)))))" *)


   (* excludesAll     :: "[('a,'b::bot)Sequence, ('a,'b)Sequence] => 'a Boolean" *)
   OclExcludesAll_def: 
   "OclExcludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                       \<HolOclLiftLeft>set \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> \<inter> set \<HolOclDropLeft>Rep_Sequence_0 Y\<HolOclDropRight> = {}\<HolOclLiftRight>)))"
(*   OclExcludesAll_def :  
   "((self::('a,'b::bot )Sequence)->excludesAll((obj::('a,'b ) Sequence))) 
   \<equiv> (obj ->forAll(elem | (self ->includes((elem::('a, 'b ) Sequence)))))" *)

 
   (* isEmpty      :: "('a,'b::bot)Sequence => 'a Boolean" *)
   OclIsEmpty_def  :  
   "OclIsEmpty     \<equiv> lift1 (strictify (\<lambda>X. \<HolOclLiftLeft>([] = \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>)\<HolOclLiftRight>))"

 
   (* notEmpty     :: "('a,'b::bot)Sequence => 'a Boolean" *)
   OclNotEmpty_def :  
   "OclNotEmpty    \<equiv> lift1 (strictify (\<lambda>X. \<HolOclLiftLeft>([] \<noteq> \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>)\<HolOclLiftRight>))"

   (* sum          :: "('a,'b::{bot,plus,zero})Sequence) => ('a,'b::{bot,plus,zero})VAL" *)
   OclSum_def      :
   "OclSum S       \<equiv> OclIterate (S::('a,'b::{bot,plus,zero})Sequence) (\<lambda> x y. x + y) 0"
   
 
   (* TODO: OclComplement [not prescribed by the standard..is it?] *)

   (* union        :: "[('a,'b::bot)Sequence, ('a,'b)Sequence] => ('a,'b)Sequence" *)
   OclUnion_def    :
   "OclUnion       \<equiv> lift2(strictify(\<lambda>X. strictify(\<lambda>Y. Abs_Sequence_0
                       (\<HolOclLiftLeft>(concat [\<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>, \<HolOclDropLeft>Rep_Sequence_0 Y\<HolOclDropRight>])\<HolOclLiftRight>))))"



   (* asBag, asOrderedSet \<Rightarrow> see OCL_Collection_transform.thy *)
   (* flatten             \<Rightarrow> see OCL_Collection_transform.thy *)


ocl_setup_op [OclSize, OclCount, OclIncludesAll, OclExcludesAll,
              OclIsEmpty, OclNotEmpty, OclUnion]


(* TODO: setup_op doesn't generate the needed theorems for
         IncludesAll and ExcludesAll. Change specification or prove.

         The same hold for OclUnion: OCL_is_def_OclUnion and
                                     OCL_is_defopt_OclUnion are missing *)
(*
lemma OCL_is_def_OclUnion:
  "\<MathOclIsDefined>(OclUnion (X::('a,'b::bot)Sequence) Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(rule lift2_strict_is_isdef_fw_Sequence_Sequence[OF OclUnion_def])
  apply(simp)
  done
  
lemma OCL_is_defopt_OclUnion:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(OclUnion (X::('a,'b::bot)Sequence) Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(rule lift2_strictify_implies_LocalValid_defined_Sequence_Sequence[OF OclUnion_def])
  apply(simp)
  done
  
*)
section{*The iterators on Sequences*}

(* TODO: Correct this note *)
text{* Note: Quantifiers were not defined on the basis of iterate in \holocl. This
       is since iterate is undefined for all infinite collections, in contrast
       to quantifiers, which do not necessarily have computational content. *}

defs


   (* forall    :: "[('a,'b::bot Sequence) VAL, ('a,'b::bot)VAL => 'a Boolean] => 'a Boolean" *)
   OclForAll_def:  "OclForAll S P \<equiv> \<lambda> \<tau>.  
                    if DEF (S \<tau>)
                    then if  \<forall>x \<in> set \<HolOclDropLeft>Rep_Sequence_0(S \<tau>)\<HolOclDropRight>.  P (lift0 x) \<tau> = (\<HolOclLiftLeft>True\<HolOclLiftRight>) 
                         then \<HolOclLiftLeft>True\<HolOclLiftRight>
                         else if \<exists>x \<in> set \<HolOclDropLeft>Rep_Sequence_0(S \<tau>)\<HolOclDropRight>. P (lift0 x) \<tau> = (\<HolOclLiftLeft>False\<HolOclLiftRight>) 
                              then \<HolOclLiftLeft>False\<HolOclLiftRight>
                              else \<bottom>
                    else \<bottom>"

   (* select    :: "[('a,'c::bot Sequence) VAL, ('a,'b::bot)VAL => 'a Boolean] *)
   OclSelect_def:  "OclSelect S P \<equiv>  \<lambda> \<tau>. 
                    if DEF (S \<tau>)
                    then if    \<forall>x \<in> set\<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight>. DEF (P (lift0 x) \<tau>)
                         then  Abs_Sequence_0(\<HolOclLiftLeft>List.filter (\<lambda>x. P (lift0 x) \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>) 
                                                    \<HolOclDropLeft>Rep_Sequence_0(S \<tau>)\<HolOclDropRight>\<HolOclLiftRight>)
                         else \<bottom>
                    else \<bottom>"

   (* collect    ::"[('a,'d::bot Sequence) VAL, ('a,'b::bot)VAL => ('a,'c::bot)VAL] *)
   OclCollect_def: "OclCollect S P \<equiv>  \<lambda> \<tau>. 
                    if DEF (S \<tau>)
                    then if \<forall>x \<in> set\<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight>. DEF (P (\<lambda>\<tau>. x) \<tau>)
                         then Abs_Sequence_0 \<HolOclLiftLeft>map (\<lambda>x. P (lift0 x) \<tau>)  \<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight>\<HolOclLiftRight> 
                         else \<bottom>
                    else \<bottom>"

   (* iterate     ::"[ ('a,'d::bot)Sequence, 
                       [('a,'b::bot)VAL,('a,'b::bot)VAL] => ('a,'c::bot)VAL,
                       ('a,'c::bot)Sequence
                     ] 
                       => ('a,'c::bot) Sequence" *)
   OclIterate_def : 
   "OclIterate S P A \<equiv> (\<lambda> \<tau>. if DEF (S \<tau>)
                             then  (foldl (\<lambda> x y. (P (lift0 y) (lift0 x)) \<tau>) 
                                          (A \<tau>)
                                          (\<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight>))
                             else \<bottom>)"


   (* isUnique    ::"[('a,'c::bot)Sequence, ('a,'b::bot)VAL => 'a Boolean] => 'a Boolean" *)
   (* TODO: define OclIsUnique *)

   (*   any       ::"[('a,'c::bot Sequence) VAL, ('a,'b::bot)VAL => 'a Boolean] 
                     => ('a,'b::bot)VAL" *)
   OclAny_def   : "OclAny S P \<equiv>  \<lambda> \<tau>. 
                   if DEF (S \<tau>)
                   then if   \<exists> x \<in> set \<HolOclDropLeft>Rep_Sequence_0(S \<tau>)\<HolOclDropRight>. (P (lift0 x) \<tau>) = \<HolOclLiftLeft>True\<HolOclLiftRight>
                        then SOME x.(x  \<in> set \<HolOclDropLeft>Rep_Sequence_0(S \<tau>)\<HolOclDropRight> \<and> 
                                    P (lift0 x) \<tau> = (\<HolOclLiftLeft>True\<HolOclLiftRight>))
                        else \<bottom>
                   else \<bottom>" 

   (*   OclAny_def: "OclAny (S::('a,'b::bot)Sequence) P \<equiv> 
                     OclAny ((OclAsSet S)::('a,'b::bot)Set) P" *)


   (* one         ::"[('a,'c::bot Sequence) VAL, ('a,'b::bot)VAL => 'a Boolean] => 'a Boolean" *)
   OclOne_def     : "OclOne (S::('a,('b::bot) Sequence_0)VAL) P  \<equiv>
                     \<MathOclSizeLeft>(\<MathOclSelectLeft>x : S | P (x\<Colon>'a\<Colon>type \<Rightarrow> 'b\<Colon>bot)\<MathOclSelectRight>)\<MathOclSizeRight> \<MathOclStrongEq> 1"

(* TODO: finish definition

   (* collectionRange :: "[('a,'b::{bot,plus}) VAL,('a,'b) VAL] \<Rightarrow> ('a,'b) Sequence" *)
   OclCollectionRange_def :
   "OclCollectionRange \<equiv> lift2 (\<lambda> x\<Colon>Integer. (\<lambda> y. 
                           (Abs_Sequence_0 (\<HolOclLiftLeft>(map (\<lambda> z::nat. ((int z) - (\<HolOclDropLeft>x\<HolOclDropRight>))) [0..<((nat \<HolOclDropLeft>y\<HolOclDropRight>) - (nat \<HolOclDropLeft>x\<HolOclDropRight>))])\<HolOclLiftRight>))))"

*)

section{* Proving the basic properties of the iterators *}

ocl_setup_op [OclForAll, OclExists, OclSelect, OclCollect,
              OclIterate, OclAny, OclOne]

(* TODO: prove them *)

section{* Relating the operations for which it is possible to the corresponding set operations *}
(*
lemma isEmpty_seq2set:
  "OclIsEmpty(S::('a, 'b::bot) Sequence) = OclIsEmpty ((OclAsSet S)::('a,('b::bot))Set)"
  apply(rule ext)
  apply(case_tac "DEF (S x)")
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclIsEmpty_def OCL_Set.OclIsEmpty_def
                  OclAsSet_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  done

lemma notEmpty_eq_not_isEmpty:
  "OclNotEmpty(X::('a, 'b::bot) Sequence) =  \<MathOclNot> (OclIsEmpty(X))" 
  apply (rule ext)
  apply (simp add:  OclNotEmpty_def OclIsEmpty_def OclNot_def)
  apply (simp only: lift1_def lift2_def 
                    strictify_def o_def not_def)
  apply (simp_all (no_asm_use) add: DEF_def strictify_def 
                                    split add: split_if up.split)
  done

lemma isNotEmpty_seq2set:
  "OclNotEmpty(S::('a, 'b::bot) Sequence) = OclNotEmpty ((OclAsSet S)::('a,('b::bot))Set)"
  by(simp only: notEmpty_eq_not_isEmpty OCL_Set.not_empty isEmpty_seq2set)

lemma forall_by_all_includes:
  "(\<tau> \<MathOclValid> \<MathOclForAll> x \<in> (S :: (('a,('b::bot))Sequence)) \<bullet> P (x::('a,'b::bot)VAL)) = 
   (\<forall> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<MathOclImplies> (P::((('a,'b::bot)VAL)\<Rightarrow>('a,Boolean_0)VAL)) (x::('a,'b::bot)VAL))"
  apply(case_tac "DEF (S \<tau>)")
  apply(frule DEF_X_Sequence[THEN iffD1])  
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclLocalValid_def OclTrue_def OclImplies_def
                  OclForAll_def OclIncludes_def
                  strictify_def lift0_def lift1_def lift2_def DEF_def)
  oops

lemma forall_seq2set: 
  "(\<MathOclForAll> x \<in> (S :: (('a,('b::bot))Sequence)) \<bullet> P (x\<Colon>'a \<Rightarrow> 'b)) = 
   (\<MathOclForAll> x \<in> ((OclAsSet S)::('a,('b::bot))Set)  \<bullet> P x)"
  apply(simp add: OclForAll_def OCL_Set.OclForAll_def)
  apply(rule ext)
  apply(subgoal_tac "(DEF (S \<tau>)) = (DEF (((\<MathOclAsSet> S)::(('a,'b::bot)Set)) \<tau>))")
  apply(case_tac "(DEF (S \<tau>))")
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(erule exE)
  apply(subgoal_tac "\<HolOclDropLeft>(Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft> set c \<HolOclLiftRight>))\<HolOclDropRight> = set c")
  apply(simp_all add: OclAsSet_def lift0_def lift1_def strictify_def in_over_RepAbs)
  apply(auto simp: in_over_RepAbs)
  apply(frule DEF_Set_if_DEF_Sequence)
  apply(simp_all add: DEF_def)
  done

lemma exists_seq2set:
  "(\<MathOclExists> x \<in> (S :: (('a,('b::bot))Sequence)) \<bullet> P (x\<Colon>'a \<Rightarrow> 'b)) = 
   (\<MathOclExists> x \<in> ((OclAsSet S)::('a,('b::bot))Set)  \<bullet> P x)"
  by(simp add: OclExists_def forall_seq2set)

lemma includesAll_seq2set:
  "OclIncludesAll (S1::('a, 'b::bot)Sequence) (S2::('a, 'b::bot)Sequence) =
   OclIncludesAll ((OclAsSet S1)::('a, 'b::bot)Set) ((OclAsSet S2)::('a, 'b::bot)Set)"
  apply(rule ext)
  apply(case_tac "DEF (S1 x)")
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(case_tac "DEF (S2 x)", rotate_tac -1)
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclIncludesAll_def OCL_Set.OclIncludesAll_def
                  OclAsSet_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  done

lemma excludesAll_seq2set:
  "OclExcludesAll (S1::('a, 'b::bot)Sequence) (S2::('a, 'b::bot)Sequence) =
   OclExcludesAll ((OclAsSet S1)::('a, 'b::bot)Set) ((OclAsSet S2)::('a, 'b::bot)Set)"
  apply(rule ext)
  apply(case_tac "DEF (S1 x)")
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(case_tac "DEF (S2 x)", rotate_tac -1)
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp_all add: OclExcludesAll_def OCL_Set.OclExcludesAll_def
                  OclAsSet_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  apply(auto)
  done

(* TODO: discuss the definition of OclAny with bu.

         It is not clear why Any should be strict and 
         the case of no element with property P is undefined is missing
         in the definition in the OCL_Set theory.

lemma any_seq2set:
  "(OclAny (S :: (('a,('b::bot))Sequence)) (P::('a,'b::bot)VAL \<Rightarrow> ('a,Boolean_0)VAL)) = 
   (OclAny ((OclAsSet S)::('a,('b::bot))Set) P)" 
  apply(simp add: OclAny_def OCL_Set.OclAny_def)
  apply(rule ext)
  apply(subgoal_tac "(DEF (S \<tau>)) = (DEF (((\<MathOclAsSet> S)::(('a,'b::bot)Set)) \<tau>))")
  apply(case_tac "(DEF (S \<tau>))")
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(erule exE)
  apply(subgoal_tac "\<HolOclDropLeft>(Rep_Set_0 (Abs_Set_0 \<HolOclLiftLeft> set c \<HolOclLiftRight>))\<HolOclDropRight> = set c")
  apply(simp_all add: OclAsSet_def lift0_def lift1_def strictify_def in_over_RepAbs)
  apply(auto simp: in_over_RepAbs)
  apply(frule DEF_Set_if_DEF_Sequence)
  apply(simp_all add: DEF_def)
  done

*)
*)
constdefs
(* Sequence Operations *)

   OclSubSequence :: "[('a,'b::bot)Sequence,'a Integer,'a Integer]=>('a,'b ) Sequence"     
                      ("_ ->subSequence'( _ , _ ')" [66,65,65]65) 

  "OclSubSequence \<equiv> lift3 (strictify (\<lambda>s. strictify (\<lambda>l. strictify (\<lambda>u. Abs_Sequence_0
                     (\<HolOclLiftLeft>take (nat (\<HolOclDropLeft>u\<HolOclDropRight> - \<HolOclDropLeft>l\<HolOclDropRight> + 1))
                        (List.drop (nat (\<HolOclDropLeft>l\<HolOclDropRight> - 1))
                          (\<HolOclDropLeft>Rep_Sequence_0 s\<HolOclDropRight>))\<HolOclLiftRight>)))))"

constdefs

   OclFirst ::  "('a, 'b::bot ) Sequence => ('a,'b) VAL"     ("_ ->first'(')" [66]) 
  "OclFirst \<equiv> lift1 (strictify (\<lambda>S. if \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight> = [] then \<bottom>
                                 else hd (\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>)))"

   OclLast  ::  "[('a, 'b::bot) Sequence] => ('a,'b) VAL"     ("_ ->last'(')" [66]) 
  "OclLast  \<equiv> lift1 (strictify (\<lambda>S. if \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight> = [] then \<bottom>
                                 else List.last (\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>)))"

   OclAppend :: "[('a,('b::bot) ) Sequence, ('a,'b::bot ) VAL] => ('a,'b ) Sequence"
                                               ("_ ->append'( _ ')" [66,65]65)
  "OclAppend \<equiv> lift2 (strictify (\<lambda>S. strictify (\<lambda>e. Abs_Sequence_0
                  (\<HolOclLiftLeft>(concat [\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>, [e] ])\<HolOclLiftRight>))))"

constdefs
   OclPrepend :: "[('a,('b::bot) ) Sequence, ('a,'b::bot ) VAL] => ('a,'b ) Sequence"
                                               ("_ ->prepend'( _ ')" [66,65]65)
  "OclPrepend \<equiv> lift2 (strictify (\<lambda>S. strictify (\<lambda>e. Abs_Sequence_0
                   (\<HolOclLiftLeft>e # \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>\<HolOclLiftRight>))))"
defs
   (* including  :: "[('a,('b::bot) ) Sequence, ('a,'b::bot ) VAL] => ('a,'b ) Sequence" *)
   OclIncluding_def   :
   "OclIncluding  \<equiv> lift2(strictify(\<lambda>S. strictify (\<lambda>e. Abs_Sequence_0
                       (\<HolOclLiftLeft>concat [\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>, [e] ]\<HolOclLiftRight>))))"

   (* excluding  :: "[('a,'b::bot ) Sequence,('a, 'b) VAL] => ('a, 'b) Sequence" *)
   OclExcluding_def   :
   "OclExcluding  \<equiv> lift2(strictify(\<lambda>X. strictify (\<lambda>x. Abs_Sequence_0 
                      (\<HolOclLiftLeft> List.filter (\<lambda> e. x \<noteq> e) \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> \<HolOclLiftRight>))))"

constdefs
   OclAt     :: "[('a,'b::bot ) Sequence,'a Integer] => ('a, 'b) VAL"
                                               ("_ ->at'( _ ')" [66,65])
  "OclAt \<equiv> lift2(strictify(\<lambda>X. strictify(\<lambda>x. if \<HolOclDropLeft>x\<HolOclDropRight> <= 0 \<or>
                                               int(Nat.size\<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight>) <= \<HolOclDropLeft>x\<HolOclDropRight>
                                             then \<bottom> 
                                             else \<HolOclDropLeft>Rep_Sequence_0 X\<HolOclDropRight> ! nat (\<HolOclDropLeft>x\<HolOclDropRight> - 1))))"


   OclInsertAt ::  "[('a,'b::bot)Sequence,'a Integer,('a,'b)VAL] => ('a,'b )Sequence"     
                ("_ ->insertAt'( _ , _ ')" [66,65,65]) 

  "OclInsertAt \<equiv> lift3(strictify(\<lambda>S. strictify(\<lambda>p. strictify(\<lambda>e. Abs_Sequence_0
                     (\<HolOclLiftLeft>concat
                        [take (nat \<HolOclDropLeft>p\<HolOclDropRight>) (\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>),
                         [e],
                         List.drop (nat \<HolOclDropLeft>p\<HolOclDropRight>) (\<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>)
                        ]\<HolOclLiftRight>)))))"


syntax
   "@OclFinSequence" :: "args \<Rightarrow> ('a,'b Sequence_0)VAL"           ("mkSequence[(_)]")

translations
  "mkSequence[x, xs]"    == "OclIncluding (mkSequence[xs]) x"
  "mkSequence[x]"        == "OclIncluding OclMtSequence x"

lemma test : "mkSequence[1,1,1,0,1] = ?X" oops


section{* Setting up our reasoning tools about Sequences *}

subsection{* Induction *}

(*
lemma Sequence_induct_including_0:
  assumes undefCase : "P \<bottom>"
  assumes mtCase    : "P (mtSequence \<tau>)"
  assumes stepCase  : "\<And> (S::'a::bot Sequence_0) (x::'a::bot). 
                       \<lbrakk> P S; DEF S; DEF x \<rbrakk> \<Longrightarrow> (P ((OclIncluding (lift0 S) (lift0 x)) \<tau>))"
  shows               "P (S::'a::bot Sequence_0)"
  apply(rule Abs_Sequence_induct_charn, rule undefCase)
  apply(erule rev_mp)
  apply(rule rev_induct)
  apply(insert mtCase, simp add: OclMtSequence_def lift0_def)
  apply(auto)
thm OclIncluding_def
  apply(rotate_tac -1, drule_tac x="x" and S="Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>" in stepCase)
  apply(erule DEF_Abs_Sequence)
  apply(simp_all add: OclIncluding_def DEF_def neq_commute
                      lift0_def lift2_def strictify_def)
  done
  
lemma Sequence_induct_including:
  assumes undefCase : "P \<MathOclUndefined> \<tau>"
  assumes mtCase    : "P mtSequence \<tau>"
  assumes stepCase  : "\<And> (S::('a,'b::bot)Sequence) (x::('a,'b::bot)VAL). 
                       \<lbrakk> P S \<tau>; \<tau> \<MathOclValid> \<MathOclIsDefined> S; \<tau> \<MathOclValid> \<MathOclIsDefined> x\<rbrakk> \<Longrightarrow> P (OclIncluding S x) \<tau>"
  assumes cpP       : "cp P"
  shows               "P (S::('a,'b::bot)Sequence) \<tau>"
  apply(insert cpP)
  apply(frule cp_by_cpify[THEN iffD1])
  apply(erule_tac x="S" in allE)
  apply(erule_tac x="\<tau>" in allE)
  apply(simp)
  apply(frule cp_by_cpify[THEN iffD1])
  apply(rule_tac \<tau>="\<tau>" in Sequence_induct_including_0)
  apply(insert undefCase)
  apply(auto simp: OclUndefined_def intro!: mtCase)
  apply(drule stepCase)
  apply(simp_all add: localValidDefined2sem lift0_def)
  done

lemma Sequence_induct_prepend_0:
  assumes undefCase : "P \<bottom>"
  assumes mtCase    : "P (mtSequence \<tau>)"
  assumes stepCase  : "\<And> (S::'a::bot Sequence_0) (x::'a::bot). 
                       \<lbrakk>P S; DEF S; DEF x\<rbrakk> \<Longrightarrow> (P ((OclPrepend (lift0 S) (lift0 x)) \<tau>))"
  shows               "P (S::'a::bot Sequence_0)"
  apply(rule Abs_Sequence_induct_charn, rule undefCase)
  apply(erule rev_mp)
  apply(induct_tac y)
  apply(insert mtCase, simp add: OclMtSequence_def lift0_def)
  apply(auto)
  apply(rotate_tac -1, drule_tac x="a" and S="Abs_Sequence_0 \<HolOclLiftLeft>list\<HolOclLiftRight>" in stepCase)
  apply(erule DEF_Abs_Sequence)
  apply(simp_all add: OclPrepend_def DEF_def neq_commute
                      lift0_def lift2_def strictify_def)
  done

lemma Sequence_induct_prepend:
  assumes undefCase : "P \<MathOclUndefined> \<tau>"
  assumes mtCase    : "P mtSequence \<tau>"
  assumes stepCase  : "\<And> (S::('a,'b::bot)Sequence) (x::('a,'b::bot)VAL). 
                       \<lbrakk> P S \<tau>; \<tau> \<MathOclValid> \<MathOclIsDefined> S; \<tau> \<MathOclValid> \<MathOclIsDefined> x \<rbrakk> \<Longrightarrow> P (OclPrepend S x) \<tau>"
  assumes cpP       : "cp P"
  shows               "P (S::('a,'b::bot)Sequence) \<tau>"
  apply(insert cpP)
  apply(frule cp_by_cpify[THEN iffD1])
  apply(erule_tac x="S" in allE)
  apply(erule_tac x="\<tau>" in allE)
  apply(simp)
  apply(frule cp_by_cpify[THEN iffD1])
  apply(rule_tac \<tau>="\<tau>" in Sequence_induct_prepend_0)
  apply(insert undefCase mtCase)
  apply(simp_all add: OclUndefined_def)
  apply(drule stepCase)
  apply(simp_all add: localValidDefined2sem lift0_def)
  done


section{* Basic properties of the sequence operations *}

ocl_setup_op [OclAppend, OclPrepend, OclSubSequence, OclAt,
              OclFirst, OclLast, OclIncluding, OclExcluding]

(* TODO: OCL_cp, OCL_undef_1 and OCL_undef_2 theorems are generated for all 8 operations
         missing is OCL_is_def and OCL_is_defopt
         \<Rightarrow>
         they mostly don't hold as full equivalences change ocl_setup_op or prove them  *)

lemma OCL_is_def_OclIncluding:
  "\<MathOclIsDefined>(OclIncluding (X::('a,'b::bot)Sequence) (Y::('a,'b::bot)VAL)) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(rule lift2_strict_is_isdef_fw_Sequence_Val[OF OclIncluding_def])
  apply(simp)
  done
  
lemma OCL_is_defopt_OclIncluding:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(OclIncluding (X::('a,'b::bot)Sequence) (Y::('a,'b::bot)VAL))) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(rule lift2_strictify_implies_LocalValid_defined_Sequence_Val[OF OclIncluding_def])
  apply(simp)
  done

lemma OCL_is_def_OclPrepend:
  "\<MathOclIsDefined>(OclPrepend (X::('a,'b::bot)Sequence) (Y::('a,'b::bot)VAL)) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(rule lift2_strict_is_isdef_fw_Sequence_Val[OF OclPrepend_def])
  apply(simp)
  done
  
lemma OCL_is_defopt_OclPrepend:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(OclPrepend (X::('a,'b::bot)Sequence) (Y::('a,'b::bot)VAL))) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(rule lift2_strictify_implies_LocalValid_defined_Sequence_Val[OF OclPrepend_def])
  apply(simp)
  done


section{* Basic properties of the iteraters on sequences *}

subsection{* OclForall *}

lemma OCL_cp_1_OclForAll [simp,intro!]: 
  "\<lbrakk> cp (S :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Sequence)) \<rbrakk> 
   \<Longrightarrow> cp(\<lambda>X. \<MathOclForAll> Y \<in> S X \<bullet> P (Y\<Colon>'a \<Rightarrow> 'b))"
  by(simp add: forall_seq2set)

lemma OCL_cp_2_OclForAll [simp,intro!]: 
"\<lbrakk> cp (\<lambda> X St.(\<lambda>x. P (\<lambda>\<tau>. x) X St)); 
   cp (S :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Sequence)) \<rbrakk> 
 \<Longrightarrow> cp(\<lambda>X. \<MathOclForAll> Y \<in> S X \<bullet> P (Y\<Colon>'a \<Rightarrow> 'b) X) "
  by(simp add: forall_seq2set)

lemma OCL_undef_1_OclForAll [simp]:
      "(\<MathOclForAll> x \<in> (\<MathOclUndefined>::('a, 'b::bot)Sequence) \<bullet> P (x::('a, 'b)VAL)) = \<MathOclUndefined>"
  by(simp add: forall_seq2set)

subsection{* OclSelect *}

lemma OCL_cp_OclSelect [simp,intro!]: 
  "\<lbrakk> cp (\<lambda> X St.(\<lambda>x. P (\<lambda>\<tau>. x) X St)); 
     cp (S :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Sequence)) \<rbrakk> 
   \<Longrightarrow> cp(\<lambda>X. (\<MathOclSelectLeft>x : S X | (P (x\<Colon>'a \<Rightarrow> 'b::bot)) X\<MathOclSelectRight>))"
  apply(simp only: cp_def OclSelect_def lift0_def) 
  apply(erule exE)+
  apply(rule exI, rule allI, rule allI) 
  apply (simp only:)
  apply(rule_tac t = "(\<lambda>x. P (\<lambda>\<tau>. x) X St)" and 
                 s = "f (X St) St" in subst)
  prefer 2 
  ML{* Unify.search_bound:=1000; *}
  apply(rule refl)
  ML{* Unify.search_bound:=20; *}
  (* Miracle ! This works. Definitively a unification problem !!! *)
  apply simp
  done  (* temporary solution. *)
        (* TODO: improve !!! *)

subsection{* OclCollect *}

(* TODO: formulate mainly: divide collect and collectNested from each other *)
(*
lemma OCL_cp_OclCollect [simp,intro!]: 
  "\<lbrakk> cp (\<lambda> X St.(\<lambda>x. P (\<lambda>\<tau>. x) X St)); 
     cp (S :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Sequence)) \<rbrakk> 
   \<Longrightarrow> cp(\<lambda>X. (\<MathOclCollectLeft> x : S X | (P (x\<Colon>'a \<Rightarrow> 'b::bot)) X \<MathOclCollectRight>))"
  apply(simp only: cp_def OclCollect_def lift0_def) 
  apply(erule exE)+
  apply(rule exI, rule allI, rule allI) 
  apply (simp only:)
  apply(rule_tac t = "(\<lambda>x. P (\<lambda>\<tau>. x) X St)" and 
                 s = "f (X St) St" in subst)
  prefer 2 
  ML{* Unify.search_bound:=1000; *}
  apply(rule refl)
  ML{* Unify.search_bound:=20; *}
  (* Miracle ! This works. Definitively a unification problem !!! *)
  apply simp
  done  (* temporary solution. *)
        (* TODO: improve !!! *)
*)


subsection{* OclAny *}

(* TODO: this should actually be done in the OCL_Set theory and only
         referenced here via the any_seq2set correspondence *)
lemma OCL_cp_1_OclAny [simp,intro!]: 
  "\<lbrakk> cp (S :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Sequence)) \<rbrakk> 
   \<Longrightarrow> cp(\<lambda>X. OclAny (S X) (P::('a,'b::bot)VAL \<Rightarrow> ('a,Boolean_0)VAL))"
  apply(simp only: cp_def OclAny_def) 
  apply(erule exE, rule exI, rule allI, rule allI)
  apply(erule_tac x=X in allE)
  apply(erule_tac x=St in allE)
  apply(simp only:)
  done

(* TODO: formulate
lemma OCL_cp_2_OclAny [simp,intro!]: 
"\<lbrakk> cp (\<lambda> X St.(\<lambda>x. (P::('a,'b::bot)VAL \<Rightarrow> ('a,Boolean_0)VAL) (\<lambda>\<tau>. x) X St)); 
   cp (S :: (('a,'c)VAL \<Rightarrow> ('a,('b::bot))Sequence)) \<rbrakk> 
 \<Longrightarrow> cp (\<lambda>X. OclAny (S X) (P X))"
  by(simp add: forall_seq2set)
*)

lemma OclAny_charn_0:
  assumes undefCase: "Q \<MathOclUndefined> \<bottom> \<tau>"
  and     mtCase:    "\<lbrakk> \<forall> x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>. P (\<lambda> s. x) \<tau> \<noteq> \<HolOclLiftLeft>True\<HolOclLiftRight> \<rbrakk> \<Longrightarrow>
                      Q \<MathOclUndefined> (\<lambda> s. S) \<tau>"
  and     defCase  : "\<And> x. \<lbrakk> \<exists> x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>. P (\<lambda> s. x) \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>;
                             x \<in> set \<HolOclDropLeft>Rep_Sequence_0 S\<HolOclDropRight>;  P (\<lambda> s. x) \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>; DEF S \<rbrakk> \<Longrightarrow> 
                           Q (\<lambda> s. x) (\<lambda> s. S) \<tau>"
  shows              "Q (\<lambda> s. (OclAny (\<lambda> s::'a. (S::'b::bot Sequence_0)) P) \<tau>) (\<lambda> s::'a. S) \<tau>"
  apply(case_tac "DEF S")
  apply(insert undefCase mtCase)
  apply(simp_all add: localValidDefined2sem OclIsDefined_def OclAny_def
                      OclUndefined_def DEF_def
                      lift0_def lift1_def lift2_def o_def strictify_def)
  apply(case_tac "(\<exists> x \<in> set \<HolOclDropLeft>(Rep_Sequence_0 S)\<HolOclDropRight>. (P (\<lambda>s. x) \<tau>) = \<HolOclLiftLeft>True\<HolOclLiftRight>)")
  apply(rule someI2_ex, auto simp: defCase DEF_def)
  done


(* NOTE: this is a typical example were a cp_unfold tactic would be of great use *)
lemma OclAny_charn:
  assumes undefCase: "Q \<MathOclUndefined> \<bottom> \<tau>"
  and     mtCase:    "\<lbrakk> \<forall> x \<in> set \<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight>. P (\<lambda> s. x) \<tau> \<noteq> \<HolOclLiftLeft>True\<HolOclLiftRight> \<rbrakk> \<Longrightarrow>
                      Q \<MathOclUndefined> S \<tau>"
  and     defCase  : "\<And> x. \<lbrakk> \<exists> x \<in> set \<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight>. P (\<lambda> s. x) \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>;
                             x \<in> set \<HolOclDropLeft>Rep_Sequence_0 (S \<tau>)\<HolOclDropRight>;  P (\<lambda> s. x) \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>; DEF (S \<tau>) \<rbrakk> \<Longrightarrow> 
                           Q (\<lambda> s. x) S \<tau>"
  and     cpQ_1:     "\<And> y. cp (\<lambda> x. Q x y)"
  and     cpQ_2:     "\<And> x. cp (Q x)"
  shows              "Q (OclAny (S::('a,'b::bot)Sequence) P) S \<tau>"
  apply(insert cpQ_1 cpQ_2)
  apply(rule subst, rule sym)
  apply(rule_tac x="(OCL_Collection.OclAny S P)" in cp_subst, simp)
  apply(rule subst, rule sym)
  apply(rule_tac x="S" in cp_subst, simp)
  apply(rule_tac t="(lift0 (OCL_Collection.OclAny S P \<tau>))" in subst, rule sym)
  apply(rule_tac x="S" in cp_subst)
  apply(rule cp_compose2, simp_all add: lift0_def)
  apply(simp add: cp_by_cpify lift0_def)
(* TODO: this cp reasoning sucks!! *)
  apply(rule_tac S="S \<tau>" and P="P" in OclAny_charn_0) 
  apply(rule undefCase)
  apply(drule mtCase, simp add: cp_by_cpify lift0_def)
  apply(drule defCase, assumption, simp_all add: cp_by_cpify lift0_def)
  done


subsection{* OclIterate *}

text{* Contextpassingness of iterate. This has to be proved at the very
beginning to enable the use of stronger proof techniques. *}

lemma OCL_cp_OclIterate [simp,intro!]:
  "\<lbrakk>(cp F); (cp F')\<rbrakk> \<Longrightarrow> 
   (cp (\<lambda>X. (OclIterate ((F X)::(('a,'b::bot)Sequence)) (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) (F' X))))"
  apply(subst cp_by_cpify)
  apply(drule cp_by_cpify[THEN iffD1])+
  apply(simp add: OclIterate_def lift0_def lift1_def)
  done

lemma OCL_undef_1_OclIterate [simp]:
  "OclIterate (\<MathOclUndefined>::('a,'b::bot)Sequence) 
              (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) 
              A 
   = (\<MathOclUndefined>::('a,'c::bot)VAL)"
  by(simp add: OclIterate_def OclUndefined_def lift0_def DEF_def)


text{* Its characterisation on the empty sequence *}

lemma iterate_charn_mtSequence[simp]:
  "(OclIterate (mtSequence::('a,'b::bot)Sequence) 
               (P::(('a \<Rightarrow> 'b) \<Rightarrow> (('a \<Rightarrow> 'c::bot) \<Rightarrow> ('a \<Rightarrow> 'c)))) 
                A) = A"
  apply(rule ext)
  apply(subgoal_tac "x \<MathOclValid> (\<MathOclIsDefined> \<MathOclMtSequence>)")
  apply(simp_all add: localValidDefined2sem strictify_def
                      OclIterate_def OclIncluding_def OclMtSequence_def
                      lift0_def lift1_def lift2_def)
 done


text{* Its characterisation on a constant sequence including a constant value *}

lemma iterate_charn_including_0:
  assumes defS: "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Sequence)"
  assumes defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('a \<Rightarrow> 'b::bot))"
  shows "((OclIterate (OclIncluding S x) P A) \<tau>) =
         (P (lift0 (x \<tau>)) (lift0 ((OclIterate (S::(('a,'b::bot)Sequence)) P A) \<tau>)) \<tau>)"
  apply(insert defx defS)
  thm conjI[THEN OCL_is_defopt_OclIncluding[THEN iffD2]]
  apply(frule_tac Y1="x" in conjI[THEN OCL_is_defopt_OclIncluding[THEN iffD2]], assumption) 
  apply(simp add: OclLocalValid_def OclIsDefined_def strictify_def
                  lift0_def lift1_def lift2_def OclTrue_def)
  apply(rotate_tac 2)
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp add: OclLocalValid_def OclIsDefined_def strictify_def
                  OclIterate_def OclIncluding_def
                  o_def lift0_def lift1_def lift2_def
                  OclTrue_def)
  apply(auto)
  apply(rotate_tac 2, frule DEF_X_Sequence_0[THEN iffD1])
  apply(subst (asm) Abs_Sequence_inject_charn, auto)
  done

lemma iterate_charn_including:
  assumes defS: "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Sequence)"
  assumes defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('a \<Rightarrow> 'b::bot))"
  assumes cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  assumes cpP_2: "\<And> x. cp (P x)"
  shows "((OclIterate (OclIncluding S x) P A) \<tau>) =
         (P x (OclIterate (S::(('a,'b::bot)Sequence)) P A) \<tau>)"
  apply(insert defS defx cpP_1 cpP_2)
  apply(drule_tac x="x" and P="P" and A="A" in iterate_charn_including_0)
  apply(simp)
  apply(auto simp: cp_by_cpify)
  done

(* TODO: ocl_hypsubst should actually work in this context *)
(*  assumes defopt_P:  "!!x y \<tau>. (\<tau> \<MathOclValid> \<MathOclIsDefined> (P x y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> x) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> y))" *)


lemma iterate_opt_charn_including:
  assumes P_undef_1:  "!!y.  P \<MathOclUndefined> y = \<MathOclUndefined>"
  assumes P_undef_2:  "!!x.  P x \<MathOclUndefined> = \<MathOclUndefined>"
  assumes cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  assumes cpP_2: "\<And> x. cp ((P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) x)"
  shows              "(OclIterate (OclIncluding S x) P A) =
                      (P x (OclIterate (S::(('a,'b::bot)Sequence)) P A))"
  apply(insert P_undef_1 P_undef_2 cpP_1 cpP_2)
  apply(rule ext)
  apply(case_tac "xa \<MathOclValid> \<MathOclIsDefined> S")
  apply(case_tac "xa \<MathOclValid> \<MathOclIsDefined> x")
  apply(rule iterate_charn_including, simp_all)
  apply(frule isUndefined_charn_local[THEN iffD2])
  apply(rule trans, erule_tac A="x" in cp_charn, simp, simp)
  apply(rule sym[OF trans], erule_tac A="x" in cp_charn, simp, simp)
  apply(frule isUndefined_charn_local[THEN iffD2])
  apply(rule trans, erule_tac A="S" in cp_charn, simp, simp)
  apply(rule sym[OF trans], erule_tac A="S" in cp_charn)
  apply(rule_tac P="P x" in cp_compose2, simp_all)
  done


text{* iterate on prepend *}

lemma iterate_charn_prepend_0:
  assumes defS: "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Sequence)"
  assumes defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('a \<Rightarrow> 'b::bot))"
  shows "((OclIterate (OclPrepend S x) P A) \<tau>) =
         ((OclIterate (S::(('a,'b::bot)Sequence)) P (P (lift0 (x \<tau>)) (lift0 (A \<tau>)))) \<tau>)"
  apply(insert defx defS)
  thm conjI[THEN OCL_is_defopt_OclIncluding[THEN iffD2]]
  apply(frule_tac Y1="x" in conjI[THEN OCL_is_defopt_OclPrepend[THEN iffD2]], assumption) 
  apply(simp add: OclLocalValid_def OclIsDefined_def strictify_def
                  lift0_def lift1_def lift2_def OclTrue_def)
  apply(rotate_tac 2)
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(simp add: OclLocalValid_def OclIsDefined_def strictify_def
                  OclIterate_def OclPrepend_def
                  o_def lift0_def lift1_def lift2_def
                  OclTrue_def)
  apply(auto)
  apply(rotate_tac 2, frule DEF_X_Sequence_0[THEN iffD1])
  apply(subst (asm) Abs_Sequence_inject_charn, auto)
  done

lemma iterate_charn_prepend:
  assumes defS: "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Sequence)"
  assumes defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('a \<Rightarrow> 'b::bot))"
  assumes cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  assumes cpP_2: "\<And> x. cp (P x)"
  shows "((OclIterate (OclPrepend S x) P A) \<tau>) =
         ((OclIterate (S::(('a,'b::bot)Sequence)) P (P x A)) \<tau>)"
  apply(insert defS defx cpP_1 cpP_2)
  apply(drule_tac x="x" and P="P" and A="A" in iterate_charn_prepend_0)
  apply(simp_all add: cp_by_cpify OclIterate_def)
  done


subsection{* The universal and fusion property *}

text{*Fusion and universal theorems, as described in
Graham Hutton's paper,
A Tutorial on the Universality and Expressiveness of Fold,
JFP 9:4 (355-372), 1999.*}

lemma iterate_universal_0:
  assumes g_cp:    "cp g"
  assumes g_undef: "g (\<MathOclUndefined>::('a,'b::bot)Sequence) \<tau> = (\<MathOclUndefined>::('a,'c::bot)VAL) \<tau>"
  assumes g_mtSeq: "g (OclMtSequence::('a,'b::bot)Sequence) \<tau> = A \<tau>"
  assumes g_incl:  "\<And> S x. \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> S; \<tau> \<MathOclValid> \<MathOclIsDefined> x \<rbrakk> \<Longrightarrow> 
                           g (OclIncluding S x) \<tau> = P (lift0 (x \<tau>)) (lift0 (g S \<tau>)) \<tau>"
  shows            "g S \<tau> = 
                    OclIterate S (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A \<tau>"
  apply(rule_tac S="S" in Sequence_induct_including)
  apply(simp_all add: g_cp g_undef g_mtSeq)
  apply(subst iterate_charn_including_0, simp_all add: g_incl)
  done

lemma iterate_universal:
  assumes g_cp:    "cp g"
  assumes cpP_1:   "\<And> y. cp (\<lambda> x. P x y)"
  assumes cpP_2:   "\<And> x. cp (P x)"
  assumes g_undef: "g (\<MathOclUndefined>::('a,'b::bot)Sequence) \<tau> = (\<MathOclUndefined>::('a,'c::bot)VAL) \<tau>"
  assumes g_mtSeq: "g (OclMtSequence::('a,'b::bot)Sequence) \<tau> = A \<tau>"
  assumes g_incl:  "\<And> S x. \<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> S; \<tau> \<MathOclValid> \<MathOclIsDefined> x \<rbrakk> \<Longrightarrow> 
                           g (OclIncluding S x) \<tau> = P x (g S) \<tau>"
  shows            "g S \<tau> = 
                    OclIterate S (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A \<tau>"
  by(insert cpP_1 cpP_2 g_cp, rule iterate_universal_0, 
     simp_all add: g_undef g_mtSeq g_incl cp_by_cpify)

lemma iterate_fusion_0:
  "\<lbrakk> cp h;
    h (\<MathOclUndefined>::('a,'c::bot)VAL) \<tau> = (\<MathOclUndefined>::('a,'d::bot)VAL) \<tau>; 
    h B \<tau> = A \<tau>; 
    !!x y. h (Q (lift0 (x \<tau>)) (lift0 (y \<tau>))) \<tau> = 
           P (lift0 (x \<tau>)) (lift0 (h y \<tau>)) \<tau>
   \<rbrakk> \<Longrightarrow>
     h (OclIterate (S::('a,'b::bot)Sequence) 
                   (Q::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) B) \<tau> 
   = OclIterate S (P::('a,'b::bot)VAL \<Rightarrow> ('a,'d::bot)VAL \<Rightarrow> ('a,'d::bot)VAL) A \<tau>"
   apply(rule iterate_universal_0)
   apply(rule_tac P="h" in cp_compose2)
   apply(simp_all)
   apply(rule trans, rule_tac P="h" in cp_subst, simp)
   apply(subst iterate_charn_including_0)
   apply(simp_all add: cp_by_cpify)
   done

(* TODO: how would one create such a lemma? 
         and use it? *)

lemma cp_compose3: 
  "cp P \<Longrightarrow> (\<And> P'. (cp P' \<Longrightarrow> cp (\<lambda>x. P (P' x))))"
  by(frule cp_compose2, auto)


lemma iterate_fusion:
  "\<lbrakk> cp h;
    \<And> x. cp (P x);
    h (\<MathOclUndefined>::('a,'c::bot)VAL) \<tau> = (\<MathOclUndefined>::('a,'d::bot)VAL) \<tau>; 
    h B \<tau> = A \<tau>; 
    \<And> x y. h (Q x y) \<tau> = P x (h y) \<tau>
   \<rbrakk> \<Longrightarrow>
     h (OclIterate (S::('a,'b::bot)Sequence) 
                   (Q::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) B) \<tau> 
   = OclIterate S (P::('a,'b::bot)VAL \<Rightarrow> ('a,'d::bot)VAL \<Rightarrow> ('a,'d::bot)VAL) A \<tau>"
  apply(rule iterate_fusion_0)
  apply(simp_all add: cp_by_cpify)
  apply(rule trans)
  apply(rule_tac x1="y" in sym[OF cp_subst])
  apply(rule_tac P'="h" in  cp_compose2)
  apply(simp_all add: cp_by_cpify)
  done

(*TODO: make this cp reasoning more elegant #grrrrrrr# *)


subsection{* Further propierties *}

text{* Undefinedness the second part *}

lemma OCL_undef_2_OclIterate [simp]:
  assumes P_undef_2:  "!!x.  P x \<MathOclUndefined> = \<MathOclUndefined>"
  shows          "OclIterate (S::('a,'b::bot)Sequence) 
                             (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) 
                             \<MathOclUndefined> 
                  = (\<MathOclUndefined>::('a,'c::bot)VAL)"
  apply(insert P_undef_2, rule ext, rule sym)
  apply(rule iterate_universal_0)
  apply(simp_all add: lift0_def OclUndefined_def)
  done

lemma OCL_is_defopt_OclIterate:
  assumes defopt_P:  "!!x y.   (\<tau> \<MathOclValid> \<MathOclIsDefined> (P x y)) 
                             = ((\<tau> \<MathOclValid> \<MathOclIsDefined> x) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> y))"
  assumes cpP_1:   "\<And> y. cp (\<lambda> x. P x y)"
  assumes cpP_2:   "\<And> x. cp ((P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) x)"
  shows "\<tau> \<MathOclValid> \<MathOclIsDefined>(OclIterate (S::('a,'b::bot)Sequence) P A) = ((\<tau> \<MathOclValid> (\<MathOclIsDefined> S)) \<and> (\<tau> \<MathOclValid> (\<MathOclIsDefined> A)))"
  apply(insert defopt_P cpP_1 cpP_2)
  apply(rule_tac S="S" and \<tau>="\<tau>" in Sequence_induct_including, simp_all)
  apply(rule trans)
  apply(rule_tac x="(S \<MathOclIncluding> x  \<MathOclIterate>(u;ua=A | (P u ua)))" in cp_subst, simp)
  apply(subst iterate_charn_including)
  apply(simp_all)
  apply(subst OCL_is_defopt_OclIncluding)
  apply(rule trans, rule sym)
  apply(rule_tac x="(P x (OclIterate S P A))" in cp_subst, simp_all)

(* TODO: make cp reasoning easier\<dots> if possible *)

  apply(auto)
  apply(unfold cp_def)
  apply(simp add: OclLocalValid_def OclIsDefined_def OclTrue_def
                  OclIterate_def lift0_def lift1_def DEF_def)
  apply(rule exI)
  apply(rule allI)+
  apply(auto)
  done

subsection{* Rules to derive the computational rules from a correspondance to iterate *}

(* TODO: think about the correct names *)

lemma f_by_iterate_undef_1:
  "\<lbrakk> \<And> S. f (S::('a,'b::bot)Sequence) = 
     OclIterate S (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A \<rbrakk>
   \<Longrightarrow>
   f \<MathOclUndefined> = \<MathOclUndefined>"
  by(simp)
  
lemma f_by_iterate_mtSequence:
  "\<lbrakk> \<And> S. f (S::('a,'b::bot)Sequence) = 
     OclIterate S (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A \<rbrakk>
   \<Longrightarrow>
   f \<MathOclMtSequence> = A"
  by(simp)
  

lemma f_by_iterate_including:
  assumes f_by_it: "\<And> S. (f (S::('a,'b::bot)Sequence)) = 
                         (OclIterate S (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A)"
  assumes defS: "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Sequence)"
  assumes defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('a \<Rightarrow> 'b::bot))"
  assumes cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  assumes cpP_2: "\<And> x. cp (P x)"
  shows "(f (OclIncluding S x) \<tau>) = (P x (f S) \<tau>)"
  by(simp add: f_by_it iterate_charn_including defS defx cpP_1 cpP_2)

lemma f_by_iterate_opt_including:
  assumes f_by_it: "\<And> S. (f (S::('a,'b::bot)Sequence)) = 
                         (OclIterate S (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A)"
  assumes P_undef_1:  "!!y.  P \<MathOclUndefined> y = \<MathOclUndefined>"
  assumes P_undef_2:  "!!x.  P x \<MathOclUndefined> = \<MathOclUndefined>"
  assumes cpP_1:      "\<And> y. cp (\<lambda> x. P x y)"
  assumes cpP_2:      "\<And> x. cp (P x)"
  shows "f (OclIncluding S x) = P x (f S)"
  by(simp add: f_by_it iterate_opt_charn_including
               P_undef_1 P_undef_2 cpP_1 cpP_2)



(* *******************************************************************

   Reasoning about the induction and case exhaustion over lists

******************************************************************* *)




(*si: TODO

   COPY REQUIREMENTS FROM THE STANDARD

*)

section{* Lemmas required by the Standard concerning collection *}

text{* Testing the new proof method using the universal theorem *}


lemma REQ_11_7_1_1:
  "((self::('a,'b::bot)Sequence) ->size()) = 
    OclIterate self (\<lambda> (x::('a,'b::bot)VAL) (y::'a Integer). y + 1) 0"
  apply(rule ext)
  apply(rule iterate_universal)
  apply(simp_all)
  apply(simp_all add: OclSize_def OclIterate_def plus_def OclMtSequence_def
                   OclUndefined_def OclIncluding_def localValidDefined2sem
                   OCL_Integer.Zero_ocl_int_def OCL_Integer.One_ocl_int_def
                   lift0_def lift1_def lift2_def strictify_def DEF_def)
  apply(fold DEF_def)
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(auto simp: DEF_def neq_commute)
  done

(* TODO: (automatic?) setup of the computional rules *)

lemmas size_by_iterate = REQ_11_7_1_1

thm f_by_iterate_undef_1[OF size_by_iterate]
thm f_by_iterate_mtSequence[OF size_by_iterate]
thm f_by_iterate_including[OF size_by_iterate, simplified]
thm f_by_iterate_opt_including[OF size_by_iterate, simplified]

lemma REQ_11_7_1_2:     
  "(self::('a,'b::bot)Sequence) ->includes((obj::('a,'b)VAL)) =  (OclGreater (self ->count(obj)) 0)"
  apply(rule ext)
  apply(simp add: OclIncludes_def OclCount_def OclGreater_def
                  lift0_def lift1_def lift2_def strictify_def
                  OCL_Integer.Zero_ocl_int_def)
  apply(auto, fold DEF_def)
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(insert filter_empty_conv)
  apply(subst (asm) filter_empty_conv, auto)
  done


lemma REQ_11_7_1_3:     
  "(self::('a,'b::bot)Sequence) ->excludes((obj::('a,'b)VAL)) = (self ->count(obj) \<MathOclStrictEq> 0)"
  apply(rule ext)
  apply(simp add: OclExcludes_def OclCount_def OclStrictEq_def
                  lift0_def lift1_def lift2_def strictify_def
                  OCL_Integer.Zero_ocl_int_def)
  apply(insert filter_empty_conv, auto)
  done

(* TODO: ocl_hyp_subst doesn't work as assumed

         probably wrong formulation *)

text{* 
  \OCLglitch{Wrong definition of the empty sequence case in the
             correspondence of count and iterate}
   The wrong requirement is:
*}
lemma  "((self::('a,'b::bot)Sequence) ->count((obj::('a,'b)VAL))) = 
   (OclIterate self (\<lambda> (x::('a,'b::bot)VAL) (y::'a Integer). 
         \<MathOclIf>  (x \<MathOclStrictEq> obj)\<MathOclThen> (y+1) \<MathOclElse> (y) \<MathOclEndif> ) 0)"
oops
text {*
  Because in the case of \verb|(obj = \<MathOclUndefined>) and (self = \<MathOclMtSequence>)| the
  lefthand side evaluates to \verb|text \<MathOclUndefined>| whereas the righthand side evaluates to 
  @{text 0}. Therefore we use the following correct version: 
*}

lemma REQ_11_7_1_4:
  "((self::('a,'b::bot)Sequence) ->count((obj::('a,'b)VAL))) = 
   (OclIterate self 
               (\<lambda> (x::('a,'b::bot)VAL) (y::'a Integer).\<MathOclIf>  (x \<MathOclStrictEq> obj)\<MathOclThen> (y+1) \<MathOclElse> (y) \<MathOclEndif> )   
               (\<MathOclIf>    (\<MathOclIsDefined> obj) \<MathOclThen> 0 \<MathOclElse>\<MathOclUndefined> \<MathOclEndif> ) )"
  apply(rule ext)
  apply(rule iterate_universal)
  apply(simp_all)
  apply(simp_all add: OclMtSequence_def OclCount_def
                  OCL_Integer.Zero_ocl_int_def OCL_Integer.One_ocl_int_def
                  OclIf_def OclIsDefined_def DEF_def OclUndefined_def OclStrictEq_def
                  OclIncluding_def OclLocalValid_def plus_def OclTrue_def
                  lift0_def lift1_def lift2_def lift3_def strictify_def)
  apply(fold DEF_def)
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(auto simp: DEF_def neq_commute)  
  done

lemmas count_by_iterate = REQ_11_7_1_4

thm f_by_iterate_undef_1[OF count_by_iterate]
thm f_by_iterate_mtSequence[OF count_by_iterate]
thm f_by_iterate_including[OF count_by_iterate, simplified]
thm f_by_iterate_opt_including[OF count_by_iterate, simplified]
(* TODO: add simplification in the last rule to remove the last assumption *)



(* TODO: both of the following theorems are false as it was proven in Set *)
(* NOTE: this will be proved later by referring to the corresponding theorems on Set *)
lemma REQ_11_7_1_5: "(self ->includesAll(obj)) \<equiv> (obj ->forAll(elem | (self ->includes(elem))))"
oops


(* NOTE: this will be proved later by referring to the corresponding theorems on Set *)
lemma REQ_11_7_1_6:"(self ->excludesAll(obj)) \<equiv> (obj ->forAll(elem | (self ->excludes(elem))))" 
oops


lemma REQ_11_7_1_7: 
  "self ->isEmpty() = (self::('a,'b::bot)Sequence) ->size() \<MathOclStrictEq> 0"
  apply (rule ext)
  apply (simp add:  OclIsEmpty_def OclSize_def OclStrictEq_def 
                    OCL_Integer.Zero_ocl_int_def)
  apply (simp only: lift0_def lift1_def lift2_def 
                    strictify_def o_def not_def)
  apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def 
                               split add: split_if up.split)
  apply (simp only: lift0_def lift1_def lift2_def 
                    strictify_def o_def not_def OclIsDefined_def DEF_def OclTrue_def)
  apply auto
  done


lemma REQ_11_7_1_8: 
  "self ->notEmpty() = (\<MathOclNot> ((self::('a,'b::bot)Sequence) ->size() \<MathOclStrictEq> 0))"
  by(simp add: notEmpty_eq_not_isEmpty REQ_11_7_1_7)


lemma REQ_11_7_1_9:
  "OclSum S = OclIterate (S::('a,'b::{bot,plus,zero})Sequence) (\<lambda> x y. x + y) 0"
  by(simp add: OclSum_def)

(* TODO: define product? Or do we not need it? *)
(*
lemma REQ_11_7_1_10:
  OclProduct S1 S2 = \<dots> \<dots> \<dots>

*)


subsection{* Requirements of the sequence functions *}



subsection{* Requirements of the iterator functions on collections *}

lemma REQ_11_9_1_2:
  "(\<MathOclForAll> x \<in> (source::('a,'b::bot)Sequence) \<bullet> (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x) = 
   (source \<MathOclIterate>(x;y= \<MathOclTrue> | (P x) \<MathOclAnd> y))"
  apply(rule ext)
  apply(rule iterate_universal_0, simp_all)
  apply(simp_all add: OclMtSequence_def OclForAll_def OclTrue_def
                      OclIncluding_def OclAnd_def localValidDefined2sem
                      lift0_def lift1_def lift2_def strictify_def DEF_def)
  apply(fold DEF_def)
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(clarify, simp add: DEF_def)
  apply(subst (asm) neq_commute)
  apply(subgoal_tac "\<HolOclDropLeft>(Rep_Sequence_0 (Abs_Sequence_0 \<HolOclLiftLeft>(c @ [(xa x)])\<HolOclLiftRight>))\<HolOclDropRight> = c @ [(xa x)]")
  apply(rule_tac X="P (\<lambda>s. (xa x))" and St="x" in boolean_cases_sem)
  apply(auto)
  apply(rule bexI)
  prefer 2 apply(assumption, simp)
  done

lemmas forall_by_iterate = REQ_11_9_1_2

thm f_by_iterate_undef_1[OF forall_by_iterate]
thm f_by_iterate_mtSequence[OF forall_by_iterate]
thm f_by_iterate_including[OF forall_by_iterate, simplified]
thm f_by_iterate_opt_including[OF forall_by_iterate, simplified]
(* TODO: fix cp-ness *)


lemma REQ_11_9_1_1:
  "(\<MathOclExists> x \<in> (source::('a,'b::bot)Sequence) \<bullet> (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x) = 
   (source \<MathOclIterate>(x;y= \<MathOclFalse> | (P x) \<MathOclOr> y))" 
  apply(simp add: OclExists_def forall_by_iterate)
  apply(rule ext, rule iterate_fusion_0, simp_all)
(* TODO: would be nice to have the calculus laws on the semantic level *)
  apply(simp add: OclNot_def OclOr_def OclAnd_def DEF_def
                  lift0_def lift1_def lift2_def strictify_def)
  done

lemmas exists_by_iterate = REQ_11_9_1_1

thm f_by_iterate_undef_1[OF exists_by_iterate]
thm f_by_iterate_mtSequence[OF exists_by_iterate]
thm f_by_iterate_including[OF exists_by_iterate, simplified]
thm f_by_iterate_opt_including[OF exists_by_iterate, simplified]
(* TODO: fix cp-ness *)


text_raw{* 
NOTE: currently not possible as a theory about tuples is missing
\begin{verbatim}
lemma REQ_11_9_1_3:
  "OclIsUnique (source::('a,'b::bot)Sequence) (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) =
   source ->collect(iterateors | Tuple{iter = Tuple{iterators}, value=body})
          ->forAll(x,y | (x.iter <> y.iter) \<MathOclImplies> (x.value <> y.value))"
\end{verbatim}
*}


text{* 
  \OCLglitch{Wrong formulation of a deterministic variant of any}

  Assume \verb|source = [1,2] and P = \<lambda> x. \<MathOclTrue>| then the
  right-hand side of the equation evaluates to @{text 1} assuming that
  OclAsSequence is the identity function on sequences. But according to the
  standard \verb|any| can can return any value which is element of the sequence
  and for which @{text P} is true.  Therefore it could return @{text 2} which
  is not equal to @{text 1} and thus the property cannot hold.
*}


lemma REQ_11_9_1_4:
  "OclAny (source::('a,'b::bot)Sequence) (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) =
   OclFirst ((OclAsSequence (OclSelect source P)::('a,'b::bot)Sequence))"
  apply(simp, rule ext)
  apply(rule_tac S="source" in OclAny_charn)
  apply(simp_all)
  apply(simp_all add: OclFirst_def OclSelect_def OclUndefined_def
                      lift0_def lift1_def lift2_def strictify_def DEF_def)
  apply(fold DEF_def)
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(rule_tac X="P (\<lambda>s. xa)" and St="x" in boolean_cases_sem)
  apply(clarify, simp_all add: DEF_def)
  apply(auto simp: filter_empty_conv)
  oops


lemma REQ_11_9_1_5:
  "(OclOne (source::('a,'b::bot)Sequence) (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean)) = 
   (\<MathOclSizeLeft>(\<MathOclSelectLeft>x : source | (P x)\<MathOclSelectRight>)\<MathOclSizeRight> \<MathOclStrongEq> 1)"
  by(simp add: OclOne_def)


text{* NOTE: currently not possible as a the definition of collectNested is missing *}
lemma REQ_11_9_1_6:
  "(OclCollect (source::('a,'b::bot)Sequence) (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL)) = 
    (OclFlatten (OclCollectNested source P))"
  oops


subsection{* Requirements of the iterator functions on sequences *}

lemma REQ_11_9_4_1:
  "(\<MathOclSelectLeft>x : (source::('a,'b::bot)Sequence) | ((P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x)\<MathOclSelectRight>) = 
   (source \<MathOclIterate>(iterator; result = \<MathOclMtSequence> | 
                             \<MathOclIf>  (P iterator)\<MathOclThen> (OclIncluding result iterator) \<MathOclElse> (result) \<MathOclEndif> ))"
  apply(rule ext, rule iterate_universal_0, simp_all)
  apply(simp_all add: OclUndefined_def OclMtSequence_def OclIncluding_def
                      OclSelect_def OclIterate_def localValidDefined2sem
                      lift0_def lift1_def lift2_def strictify_def DEF_def)
  apply(fold DEF_def)
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(frule DEF_X_Sequence[THEN iffD1])
  apply(clarify)
  apply(simp add: OclIf_def lift0_def lift1_def lift3_def 
                  strictify_def DEF_def neq_commute)
  apply(rule_tac X="P (\<lambda>s. (xa x))" and St="x" in boolean_cases_sem)
  apply(auto)
(* TODO: how can this be automated\<dots> what is missing for the theorem provers? *)
  apply(rule bexI) prefer 2 apply(assumption, simp)
  apply(rule bexI) prefer 2 apply(assumption, simp)
  done

lemmas select_by_iterate = REQ_11_9_4_1

thm f_by_iterate_undef_1[OF select_by_iterate]
thm f_by_iterate_mtSequence[OF select_by_iterate]
thm f_by_iterate_including[OF select_by_iterate, simplified]
thm f_by_iterate_opt_including[OF select_by_iterate, simplified]
(* TODO: fix cp-ness and undefinedness *)

(* TODO: what about reject? isn't it defined or required? *)

(* TODO: add REQ_11_9_4_i  where i \<in> {2..5} *)


subsection{* Universal Quantifiers *}

text{* All this theorems are easily transformed from sets to sequences *}

lemma ForAllI:
  assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Sequence)"
  and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid> P x" 
  shows          "\<tau> \<MathOclValid> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
(*  apply(auto simp: forall_seq2set includes_seq2set
             elim!: OCL_is_defopt_OclAsSet[THEN iffD2]
             intro!: allP defS OCL_Set.isdefForAllI2) *)

  apply(subst forall_seq2set)
  apply(rule OCL_Set.ForAllI)
  apply(insert defS, erule OCL_is_defopt_OclAsSet[THEN iffD2])
  apply(rule allP, simp add: includes_seq2set)
  done

lemma undefForAllI1:
"\<lbrakk>   \<tau> \<MathOclValid> \<MathOclIsUndefined> (S::('a,('b::bot))Sequence)\<rbrakk> 
 \<Longrightarrow> \<tau> \<MathOclValid> \<MathOclIsUndefined> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b))"
by(erule_tac P = "\<lambda> X. (\<MathOclIsUndefined> (\<MathOclForAll> x \<in> X \<bullet> P x))" in subst_LJ_undef, simp_all)


lemma undefForAllI2:
  assumes exUndef1: "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('a,('b::bot))Sequence)"
  and     exUndef2: "\<tau> \<MathOclValid>  \<MathOclIsUndefined> (P x)"
  and     allP    : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid> (P x) \<MathOclOr> \<MathOclIsUndefined>(P x)" 
  and     cpP     : "cp P"
  shows             "\<tau> \<MathOclValid> \<MathOclIsUndefined>(\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
  apply(insert exUndef1)
  apply(auto simp: forall_seq2set includes_seq2set
             intro!: exUndef1 exUndef2 allP cpP OCL_Set.undefForAllI2)
  done

(*

lemma isdefForAllI1[simp, intro!]:
"\<tau> \<MathOclValid> \<MathOclIsDefined> (\<MathOclForAll> x \<in> (\<MathOclMtSequence>::('a,'b::bot)Sequence) \<bullet> P(x\<Colon>'a\<Rightarrow>'b))"
by simp
(* Not analytic, but works for rewriting \<dots> *)

lemma isdefForAllI2:
  assumes exNot1: "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('a,('b::bot))Sequence)"
  and     exNot2: "\<tau> \<MathOclValid> \<MathOclNot> (P x)"
  and     cpP   : "cp P"
  shows           "\<tau> \<MathOclValid> \<MathOclIsDefined>(\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
  apply(insert exNot1)
  apply(auto simp: forall_seq2set includes_seq2set
             intro!: exNot1 exNot2 cpP OCL_Set.isdefForAllI2)
  done

lemmas isdefForAllI3 = ForAllI[THEN isDefined_if_valid]

lemma   notForAllI:
  assumes isin : "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('a,('b::bot))Sequence)"
  and     not  : "\<tau> \<MathOclValid> \<MathOclNot> (P x)"
  and     cpP  : "cp P"
  shows          "\<tau> \<MathOclValid> \<MathOclNot> (\<MathOclForAll> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
  apply(insert isin)
  apply(auto simp: forall_seq2set includes_seq2set
             intro!: isin not cpP OCL_Set.notForAllI)
  done

lemma   ExistsI:
assumes isin : "\<tau> \<MathOclValid> x \<MathOclIncludes> (S::('a,('b::bot))Sequence)"
and     not  : "\<tau> \<MathOclValid> (P x)"
and     cpP  : "cp P"
shows          "\<tau> \<MathOclValid> (\<MathOclExists> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
by (auto simp:OclExists_def intro!: notForAllI isin not cpP)


lemma   NotExistsI:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Sequence)"
and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid>  \<MathOclNot> P x" 
shows          "\<tau> \<MathOclValid>  \<MathOclNot>(\<MathOclExists> x \<in> S \<bullet> P(x\<Colon>'a \<Rightarrow> 'b::bot))"
by (auto simp: OclExists_def intro!: ForAllI defS allP)

*)

(*

lemma SelectPartialUndef : 
assumes ainS : "\<tau> \<MathOclValid> a \<MathOclIncludes> (S::('a,('b::bot))Sequence)" 
and     allP : "\<tau> \<MathOclValid> \<MathOclIsUndefined> (P a)" 
and     cpP  : "cp P"
shows          "(\<MathOclSelectLeft>x : S | P (x::'a\<Rightarrow>'b::bot)\<MathOclSelectRight>) \<tau> = \<MathOclUndefined> \<tau>"
apply (insert ainS allP)
apply (frule isDefined_if_valid,
       drule includes_valid_def_distr[THEN iffD1])
apply (simp add: OclSelect_def OclIncluding_def OclUndefined_def
                 OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def DEF_def)
apply auto
apply (erule_tac x="a \<tau>" in ballE, simp_all)
apply (subgoal_tac "P (\<lambda>\<tau>'. a \<tau>) \<tau> = P a \<tau>", simp)
apply (rule_tac St = \<tau> and P = P in cp_charn, simp add: cpP,rule cpP)
done


lemma SelectPartialUndef2[simp] : 
assumes ainS : "\<tau> \<MathOclValid> (a::'a\<Rightarrow>'b::bot) \<MathOclIncludes> (S::('a,('b::bot))Sequence)" 
shows          "(\<MathOclSelectLeft>x : S | (\<lambda>x. \<MathOclUndefined>)((x::'a\<Rightarrow>'b::bot))\<MathOclSelectRight>) \<tau> = \<MathOclUndefined> \<tau>"
by(rule SelectPartialUndef, auto intro!: ainS)


lemma mem_OclSelect_eq:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Sequence)"
and     defa : "\<tau> \<MathOclValid> \<MathOclIsDefined>(a::('a\<Rightarrow>'b::bot))"
and     ainS : "\<tau> \<MathOclValid> a \<MathOclIncludes> S" 
and     allP : "\<And> x. \<tau> \<MathOclValid> x \<MathOclIncludes> S \<Longrightarrow> \<tau> \<MathOclValid>  \<MathOclIsDefined>(P x)" 
and     cpP  : "cp P"
shows          "(\<tau> \<MathOclValid> a \<MathOclIncludes> (\<MathOclSelectLeft>x : S | P (x::'a\<Rightarrow>'b::bot)\<MathOclSelectRight>)) 
               = (\<tau> \<MathOclValid> P a)"
apply (insert defS defa)
apply (simp add: OclSelect_def OclIncluding_def
                 OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def)
apply (frule DEF_X_Sequence[THEN iffD1])
apply (erule exE)
apply (subgoal_tac "Rep_Sequence_0 (Abs_Sequence_0 \<HolOclLiftLeft>c\<HolOclLiftRight>) = \<HolOclLiftLeft>c\<HolOclLiftRight>") prefer 2
apply (rule Abs_Sequence_0_inverse)
apply (simp_all add: smashed_sequence_charn)
apply (erule conjE)
apply (drule smashed_sequence_charn[THEN iffD2])
apply (simp add: DEF_def)


apply (subgoal_tac "(\<forall>x \<in> set c. P (\<lambda>\<tau>. x) \<tau> \<noteq> down)", simp)
apply (subgoal_tac "Abs_Sequence_0 \<HolOclLiftLeft>[x \<in> c. P (\<lambda>\<tau>. x) \<tau> = \<HolOclLiftLeft>True\<HolOclLiftRight>]\<HolOclLiftRight> \<noteq> \<bottom>", simp)
oops
(* TODO: finish

apply (subgoal_tac "a \<tau> \<in> c", simp) 
apply (rule_tac t="P a \<tau>" in subst, simp_all)
apply (rule_tac St = \<tau> and P = P in cp_charn, simp add: cpP,rule cpP)

apply (insert  ainS, simp add: OclSelect_def OclIncluding_def
                 OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def)
apply (rule DEF_Abs_Sequence[simplified DEF_def])
apply (simp)
apply (rule ballI)
apply (rule allP[simplified 
                 OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def DEF_def, simplified])
apply (simp add: defS[THEN localValidDefined2sem[THEN iffD1], simplified DEF_def]
                 strictify_def)
apply auto
done

*)


(* TODO *)
lemma nonmem_OclSelect:
assumes defS : "\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,('b::bot))Sequence)"
and     defa : "\<tau> \<MathOclValid> \<MathOclIsDefined>(a::('a\<Rightarrow>'b::bot))"
and     aninS: "\<not> (\<tau> \<MathOclValid> a \<MathOclIncludes> S)" 
and     cpP  : "cp P"
shows          "\<not> (\<tau> \<MathOclValid> a \<MathOclIncludes> (\<MathOclSelectLeft>x : S | P (x::'a\<Rightarrow>'b::bot)\<MathOclSelectRight>))"
apply (insert defS defa aninS)
apply (simp add: OclSelect_def OclIncluding_def
                 OclLocalValid_def OclIsDefined_def strictify_def
                 OclForAll_def OclNot_def OclIncludes_def
                 o_def lift0_def lift1_def lift2_def
                 OclTrue_def)
apply (frule DEF_X_Sequence [THEN iffD1], erule exE, simp add: DEF_def)
oops


(* TODO: copy rest from OCL_Set.thy *)

*)


(*************************************************************************************
 * OOOOOOLD STUFF

lemma prepend_charn1:
  "\<lbrakk>\<bottom> \<notin> set xs; xs \<noteq> [] \<rbrakk> \<Longrightarrow> 
   \<exists> (S::'b::bot Sequence_0) a. (Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>) = ((lift0 S) ->prepend( (lift0 a) )) \<tau>"
  apply(cases xs, simp_all)
  apply(rule_tac x="Abs_Sequence_0 \<HolOclLiftLeft>list\<HolOclLiftRight>" in exI)
  apply(rule_tac x="a" in exI)
  apply(simp add: OclPrepend_def
                  lift0_def lift2_def strictify_def)
  apply(auto)
  done


thm localValidDefined2sem


lemma prepend_charn3:
  "\<tau> \<MathOclValid> OclNotEmpty (S::('a,'b::bot)Sequence) \<Longrightarrow>  
   \<exists> (S'::('a,'b::bot)Sequence) a. \<tau> \<MathOclValid> (S \<MathOclStrongEq> (S' ->prepend(a)))"
  apply(frule isDefined_if_valid)
  apply(frule OCL_is_defopt_OclNotEmpty[THEN iffD1])
  apply(simp add: OclLocalValid_def OclIsDefined_def OclTrue_def
                  lift0_def lift1_def)
(*  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(auto) *)
  apply(subgoal_tac "\<bottom> \<notin> set \<HolOclDropLeft>(Rep_Sequence_0 (S \<tau>))\<HolOclDropRight>")
  apply(frule_tac \<tau>="\<tau>" in prepend_charn1, auto)
  prefer 2
  oops
(*
thm cp_strongEq[THEN cp_charn]
thm cp_charn[THEN cp_strongEq]
thm lift_drop
  apply(simp add: OclLocalValid_def OclStrongEq_def OclPrepend_def
                  OclNotEmpty_def OclTrue_def
                  lift0_def lift1_def lift2_def strictify_def)
  apply(case_tac "DEF (S \<tau>)")
  apply(frule DEF_X_Sequence_0[THEN iffD1])
  apply(auto simp: DEF_def)
  apply(frule_tac \<tau>="\<tau>" in prepend_charn1, auto)
  apply(rule_tac x="lift0 (Abs_Sequence_0 \<HolOclLiftLeft>list\<HolOclLiftRight>)" in exI)
  apply(auto simp: lift0_def)
  apply(rule exI)
  apply(rule_tac x="lift0 a" in exI)
  apply(rule_tac x="list" in exI)
*)

lemma prepend_charn2:
  "\<lbrakk>\<bottom> \<notin> set xs; xs \<noteq> [] \<rbrakk> \<Longrightarrow> 
   \<exists> (S::('a,'b::bot)Sequence) a. Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight> = (S ->prepend(a)) \<tau>"
  apply(cases xs, simp_all)
  apply(rule_tac x="lift0 (Abs_Sequence_0 \<HolOclLiftLeft>list\<HolOclLiftRight>)" in exI)
  apply(rule_tac x="lift0 a" in exI)
  apply(simp add: OclPrepend_def
                  lift0_def lift2_def strictify_def)
  apply(auto)
  done

*)
*)
end
