(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Bag.thy -
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

header {* \OCL Bag *}

theory OCL_Bag 
imports 
  "$HOLOCL_HOME/src/library/collection/smashed/OCL_Sequence" 
begin


section{* Properties of the the Datatype Adaption *}

text{* The following rules are transformed versions of the automatically
generated rules for datatype adaption *}


lemma Abs_Bag_0_inject_absurd11 [simp]:
  "\<lbrakk> count x \<bottom> = 0 \<rbrakk> \<Longrightarrow> ((Abs_Bag_0 down) = (Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(subst Abs_Bag_0_inject,
     simp_all add: OCL_Bag_type.Bag_0_def smash_def)

lemma Abs_Bag_0_inject_absurd12 [simp]:
  "\<lbrakk> count x \<bottom> = 0 \<rbrakk> \<Longrightarrow> ((Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Bag_0 down)) = False"
  by(subst Abs_Bag_0_inject,
     simp_all add: OCL_Bag_type.Bag_0_def smash_def)

lemma Abs_Bag_0_inject_absurd21 [simp]:
  "\<lbrakk> count x \<bottom> = 0 \<rbrakk> \<Longrightarrow> ( \<bottom> = (Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)) = False"
  by(simp add: UU_Bag_def)

lemma Abs_Bag_0_inject_absurd22 [simp]:
  "\<lbrakk> count x \<bottom> = 0 \<rbrakk> \<Longrightarrow> ((Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<bottom>) = False"
  by(simp add: UU_Bag_def)

lemma Abs_Bag_0_inject_charn [simp]:
  "\<lbrakk> count x (\<bottom>::'a::bot) = 0; count y (\<bottom>::'a::bot) = 0 \<rbrakk> \<Longrightarrow> 
   (((Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = (Abs_Bag_0 \<HolOclLiftLeft>y\<HolOclLiftRight>)) = (x = y))"
  by(subst Abs_Bag_0_inject, 
     simp_all add: OCL_Bag_type.Bag_0_def smash_def)

lemma Abs_Bag_0_inverse_charn [simp]:
  "(count x \<bottom> = 0) \<Longrightarrow> Rep_Bag_0 (Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>) = \<HolOclLiftLeft>x\<HolOclLiftRight>"
  by (subst Abs_Bag_0_inverse, 
      simp_all add: OCL_Bag_type.Bag_0_def smash_def)

lemma Abs_Bag_0_inverse_charn2 [simp]:
  "(count x \<bottom> = 0) \<Longrightarrow> \<HolOclDropLeft>Rep_Bag_0 (Abs_Bag_0 \<HolOclLiftLeft>x\<HolOclLiftRight>)\<HolOclDropRight> = x"
  by (subst Abs_Bag_0_inverse, 
      simp_all add: OCL_Bag_type.Bag_0_def smash_def)

lemma Abs_Bag_0_cases_charn:
  assumes bottomCase : "\<lbrakk> x = \<bottom> \<rbrakk> \<Longrightarrow> P"
  assumes listCase   : "\<And>y. (\<lbrakk> x = Abs_Bag_0 \<HolOclLiftLeft>y\<HolOclLiftRight>; count y \<bottom> = 0 \<rbrakk> \<Longrightarrow> P)"
  shows                "P"
  apply(rule_tac x="x" in Abs_Bag_0_cases)
  apply(case_tac "y = \<bottom>")
  apply(rule bottomCase, simp add: UU_Bag_def)
  apply(frule not_down_exists_lift2[THEN iffD1])
  apply(rule_tac y="\<HolOclDropLeft>y\<HolOclDropRight>" in listCase)
  apply(simp add: OCL_Bag_type.Bag_0_def smash_def DEF_def)
  apply(case_tac "(count \<HolOclDropLeft>y\<HolOclDropRight> \<bottom>) = (0\<Colon>nat)")
  apply(auto simp: OCL_Bag_type.Bag_0_def smash_def DEF_def)
  done

lemma Abs_Bag_0_induct_charn:
  assumes bottomCase : "P \<bottom>"
  assumes stepCase   : "\<And>y. (\<lbrakk> count y \<bottom> = 0 \<rbrakk> \<Longrightarrow> P (Abs_Bag_0 \<HolOclLiftLeft>y\<HolOclLiftRight>))"
  shows                "P x"
  apply(rule_tac x="x" in Abs_Bag_0_induct)
  apply(rule_tac  x="Abs_Bag_0 y" in Abs_Bag_0_cases_charn)
  apply(auto intro!: bottomCase stepCase)
  done

lemma inj_on_Abs_Bag_0: "inj_on Abs_Bag_0 Bag_0";
  by(rule inj_on_inverseI,rule Bag_0.Abs_Bag_0_inverse,assumption)

lemma inj_Rep_Bag_0 : "inj Rep_Bag_0"
  by(rule inj_on_inverseI, rule Bag_0.Rep_Bag_0_inverse)

lemma smashed_bag_charn: 
  "(count X \<bottom> = 0) = ((\<HolOclLiftLeft>X\<HolOclLiftRight>) \<in> Bag_0)"
  by(unfold smash_def Bag_0_def UU_Bag_def,auto) (*characterization 1 *)

lemma UU_in_smashed_bag [simp]: 
  "\<bottom> \<in> Bag_0"
  by(unfold smash_def Bag_0_def UU_Bag_def,auto) (*characterization 2a *)

lemma down_in_smashed_bag [simp]: 
  "down \<in> Bag_0"
  by(unfold smash_def Bag_0_def UU_Bag_def,auto) (*characterization 2b *)

lemma mt_in_smashed_bag[simp]: 
  "(\<HolOclLiftLeft>{#}\<HolOclLiftRight>) \<in> Bag_0"
  by (unfold smash_def Bag_0_def,auto)

lemma DEF_Abs_Bag: "\<And>X. (count X \<bottom> = 0) \<Longrightarrow> DEF (Abs_Bag_0 (\<HolOclLiftLeft>X\<HolOclLiftRight>))"
  apply(unfold DEF_def UU_Bag_def, simp)
  done

lemma DEF_Rep_Bag: 
  "\<And>X. DEF X \<Longrightarrow> DEF (Rep_Bag_0 X)"
  apply(unfold DEF_def UU_Bag_def,auto)
  apply(drule_tac f = "Abs_Bag_0" in arg_cong)
  apply(simp add: Rep_Bag_0_inverse)
  done

lemma not_DEF_Rep_Bag: 
  "\<And>X. \<not> DEF X \<Longrightarrow> \<not> DEF (Rep_Bag_0 X)"
  apply(unfold DEF_def UU_Bag_def,auto)
  apply(simp add: Abs_Bag_0_inverse)
  done


lemma exists_lift_Bag: 
  "\<And>X. DEF X \<Longrightarrow> \<exists>c. Rep_Bag_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
  by (drule DEF_Rep_Bag,simp add: DEF_X_up)

lemma exists_lift_Bag2: 
  "\<And>X. DEF X \<Longrightarrow> \<exists>c. count c \<bottom> = 0 \<and> Rep_Bag_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>)"
  apply(frule exists_lift_Bag)
  apply(auto simp: UU_Bag_def DEF_def)
  apply(drule sym, rule swap)
  apply assumption
  apply(unfold not_def)
  apply(subst (asm) smashed_bag_charn)
  apply(simp add: Rep_Bag_0)
  done

lemma Rep_Bag_cases: 
  "Rep_Bag_0 X = \<bottom> \<or> (\<exists>c. count c \<bottom> = 0 \<and> Rep_Bag_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply(case_tac "DEF X")
  apply(drule exists_lift_Bag2)
  apply(drule_tac [2] not_DEF_Rep_Bag)
  apply(simp_all add: DEF_def)
  done


lemma DEF_X_Bag_0 : "DEF(X) = (\<exists>c. count c \<bottom> = 0 \<and> Rep_Bag_0 X = (\<HolOclLiftLeft>c\<HolOclLiftRight>))"
  apply(insert  Rep_Bag_cases [of "X"], auto)
  apply(drule exists_lift_Bag,auto)
  apply(rule swap)
  prefer 2
  apply(rule not_DEF_Rep_Bag, auto)
  done


text{* The following lemma is crucial for thy\_morpher: *} 
lemma DEF_X_Bag: "DEF X = (\<exists>c.  count c \<bottom> = 0 \<and> X = Abs_Bag_0 (\<HolOclLiftLeft>c\<HolOclLiftRight>))"

  apply (auto simp: DEF_Abs_Bag)
  apply (simp add:DEF_X_Bag_0)
  apply (erule exE, rule exI, auto)
  apply (rule injD [OF inj_Rep_Bag_0])
  apply (auto simp: smashed_bag_charn Abs_Bag_0_inverse)
  done 

text{* This lemma is very convenient in simplifing unfolded definitions *}
lemma DEF_X_Bag':
  "\<lbrakk> X \<noteq> \<bottom> \<rbrakk>  \<Longrightarrow> 
   (\<exists>c. (count c \<bottom> = 0) \<and> (Rep_Bag_0 X = \<HolOclLiftLeft>c\<HolOclLiftRight>)) \<and> 
   (\<exists>c. (count c \<bottom> = 0) \<and> (X = (Abs_Bag_0 \<HolOclLiftLeft>c\<HolOclLiftRight>)))"
  apply(fold DEF_def)
  apply(frule DEF_X_Bag[THEN iffD1])
  apply(drule DEF_X_Bag_0[THEN iffD1])
  apply(simp)
  done
 
(* TODO: does this hurt to have both versions in the assumptions? *)
lemma Bag_sem_cases_0:
  assumes defC: "\<And> c d. \<lbrakk> X \<noteq> \<bottom>; \<bottom> \<noteq> X; 
                          (count c \<bottom> = 0); (count d \<bottom> = 0); 
                          Rep_Bag_0 X = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          X = Abs_Bag_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X"
  and   undefC: "\<lbrakk> X = \<bottom> \<rbrakk> \<Longrightarrow> P X"
  shows "P X"
  apply(rule Abs_Bag_0_cases_charn, erule undefC)
  apply(rule defC, auto)
  done

lemma Bag_sem_cases:
  assumes defC: "\<And> c d. \<lbrakk> (X \<tau>) \<noteq> \<bottom>; \<bottom> \<noteq> (X \<tau>); 
                          (count c \<bottom> = 0); (count d \<bottom> = 0); 
                          Rep_Bag_0 (X \<tau>) = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          (X \<tau>) = Abs_Bag_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X \<tau>"
  and   undefC: "\<lbrakk> (X \<tau>) = \<bottom> \<rbrakk> \<Longrightarrow> P X \<tau>"
  and      cpP: "cp P"
  shows "P X \<tau>"
  apply(rule_tac P2="P" in subst[OF sym[OF cp_subst]], rule cpP)
  apply(rule Bag_sem_cases_0)
  apply(rule_tac P1="P" in subst[OF cp_subst], rule cpP)
  apply(rule defC) prefer 7
  apply(rule_tac P1="P" in subst[OF cp_subst], rule cpP)
  apply(rule undefC, simp_all)
  done

lemma Bag_sem_cases_ext:
  assumes defC: "\<And> c d \<tau>. \<lbrakk> (X \<tau>) \<noteq> \<bottom>; \<bottom> \<noteq> (X \<tau>); 
                          (count c \<bottom> = 0); (count d \<bottom> = 0); 
                          Rep_Bag_0 (X \<tau>) = \<HolOclLiftLeft>c\<HolOclLiftRight>;
                          (X \<tau>) = Abs_Bag_0 \<HolOclLiftLeft>d\<HolOclLiftRight> \<rbrakk> \<Longrightarrow> P X \<tau> = Q X \<tau>"
  and   undefC: "\<And> \<tau>. \<lbrakk> (X \<tau>) = \<bottom> \<rbrakk> \<Longrightarrow> P X \<tau> = Q X \<tau>"
  and      cpP: "cp P"
  and      cpQ: "cp Q"
  shows "P X = Q X"
  apply(rule ext)
  apply(rule_tac X="X" in Bag_sem_cases)
  apply(rule defC) prefer 7
  apply(rule undefC)
  apply(simp_all add: cpP cpQ)
  done

text{* These four lemmas are used by \inlineocl{ocl_setup_op} 
       to reason about definedness: *}

lemma lift2_strict_is_isdef_fw_Bag_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Bag_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Bag_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> count a \<bottom> = 0; (\<bottom>::'a::bot) \<noteq> b \<rbrakk> \<Longrightarrow> count (g a b) \<bottom> = 0)"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def lift0_def lift1_def lift2_def
                  strictify_def DEF_def)
  apply(rule ext)
  apply(case_tac "DEF (X \<tau>)")
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(auto simp: strictify_def DEF_def)
  apply(drule_tac b="(Y \<tau>)" in  inv_g)
  prefer 2
  apply(drule DEF_Abs_Bag, simp add: DEF_def, auto)
  done

lemma lift2_strict_is_isdef_fw_Bag_Bag:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Bag_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Bag_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Bag_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> count a \<bottom> = 0; count b \<bottom> = 0 \<rbrakk> \<Longrightarrow> count (g a b) \<bottom> = 0)"
  shows          " \<MathOclIsDefined>(f X Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(simp add: f_def OclIsDefined_def OclAnd_def
                  o_def lift0_def lift1_def lift2_def
                  strictify_def DEF_def)
  apply(rule ext)
  apply(case_tac "DEF (X \<tau>)")
  apply(case_tac "DEF (Y \<tau>)")
  apply(drule DEF_X_Bag_0[THEN iffD1])+
  apply(auto simp: strictify_def DEF_def)
  apply(drule_tac b="ca" in inv_g, assumption)
  apply(rotate_tac -1)
  apply(drule DEF_Abs_Bag, simp add: DEF_def)
  done

lemma lift2_strictify_implies_LocalValid_defined_Bag_Val:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Bag_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Bag_0 x\<HolOclDropRight> y\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> count a \<bottom> = 0; (\<bottom>::'a::bot) \<noteq> b\<rbrakk> \<Longrightarrow> count (g a b) \<bottom> = 0)"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Bag_Val)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done

lemma lift2_strictify_implies_LocalValid_defined_Bag_Bag:
  assumes f_def: "f \<equiv> lift2(strictify(\<lambda>x. strictify(\<lambda>y. Abs_Bag_0 \<HolOclLiftLeft>g \<HolOclDropLeft>Rep_Bag_0 x\<HolOclDropRight> \<HolOclDropLeft>Rep_Bag_0 y\<HolOclDropRight>\<HolOclLiftRight>)))"
  and     inv_g: "!!a b.( \<lbrakk> count a \<bottom> = 0; count b \<bottom> = 0 \<rbrakk> \<Longrightarrow> count (g a b) \<bottom> = 0)"
  shows          " (\<tau> \<MathOclValid> \<MathOclIsDefined>(f X Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(insert f_def)
  apply(drule_tac X="X" and Y="Y" in lift2_strict_is_isdef_fw_Bag_Bag)
  apply(rule inv_g, assumption+)
  apply(simp add: OclAnd_def  OclTrue_def o_def
               OclIsDefined_def lift0_def lift1_def lift2_def
               OclLocalValid_def)
  done

section{* The conversion operators *}

subsection{* From bags to sequences *}

text{* Informally speaking the AsSequence operation returns just some sequence
  that has exactly the same elements like the given bag. 

  Note: With the current definition it does \emph{not} hold that
  \texttt{bag->seq->oset = bag->set->oset}. To get this behaviour one has
  to define AsSequence as some sequence that has the elements in the same order
  like the associated set converted to a sequence.
*}

defs
   (* asSequence     :: "('a,'b::bot) Bag \<Rightarrow> ('a,'b::bot) Sequence" *)
   OclAsSequence_def  : 
   "OclAsSequence     \<equiv> lift1 (strictify (\<lambda>X. 
                          Abs_Sequence_0 \<HolOclLiftLeft>list_of_multiset \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>))"

ocl_setup_op [OclAsSequence]

lemma OCL_is_defopt_OclAsSequence: 
  "(\<tau> \<MathOclValid> \<MathOclIsDefined> (((self::('a,'b::bot)Bag) ->asSequence())::(('a, 'b::bot)Sequence))) = (\<tau> \<MathOclValid> \<MathOclIsDefined> self)"
  apply(simp add: localValidDefined2sem OclAsSequence_def ss_lifting')
  apply(rule impI, frule DEF_X_Bag', auto)
  done

lemma OCL_is_def_OclAsSequence:
  "(\<MathOclIsDefined> (((self::('a,'b::bot)Bag) ->asSequence())::(('a, 'b::bot)Sequence))) = (\<MathOclIsDefined> self)"
  apply(rule Bag_sem_cases_ext, simp_all)
  apply(simp_all add: OclIsDefined_def OclAsSequence_def ss_lifting')
  done


subsubsection{* Characterisation theorems of the conversion from bags to sequences *}

lemma OclAsSequence_charn_0:
  assumes undefCase: "P \<MathOclUndefined> \<bottom> \<tau>"
  and     defCase  : "\<And> xs. \<lbrakk> \<HolOclDropLeft>Rep_Bag_0 B\<HolOclDropRight> = multiset_of xs; DEF B; \<bottom> \<notin> set xs \<rbrakk> \<Longrightarrow> 
                            P (\<lambda> s. (Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>)) (\<lambda> s. B) \<tau>"
  shows              "P ((OclAsSequence (\<lambda> s. (B::'b::bot Bag_0)))::('a,'b::bot)Sequence) (\<lambda> s. B) \<tau>"
  apply(insert undefCase)
  apply(simp add: localValidDefined2sem OclIsDefined_def OclAsSequence_def
                  OclUndefined_def ss_lifting')
  apply(rule impI)
  apply(frule DEF_X_Bag_0[THEN iffD1, simplified DEF_def], clarify)
  apply(rule list_of_multisetI)
  apply(rule defCase)
  apply(simp_all add: localValidDefined2sem DEF_def mem_set_multiset_eq)
  done

lemma OclAsSequence_charn:
  assumes undefCase: "P \<MathOclUndefined> \<MathOclUndefined> \<tau>"
  and     defCase  : "\<And> xs. \<lbrakk> \<HolOclDropLeft>Rep_Bag_0 (B \<tau>)\<HolOclDropRight> = multiset_of xs; \<bottom> \<notin> set xs; \<tau> \<MathOclValid> \<MathOclIsDefined>B \<rbrakk> \<Longrightarrow> 
                            P (\<lambda> s. (Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>)) B \<tau>"
  and     cp1_P    : "\<And> y. cp (\<lambda> x. P x y)"
  and     cp2_P    : "\<And> x. cp (P x)"
  shows              "P ((OclAsSequence (B::('a,'b::bot)Bag))::('a,'b::bot)Sequence) B \<tau>"
  apply(rule subst, rule_tac x1="B" in sym[OF cp_subst])
  apply(subst cp0_eq_cp[symmetric])
  apply(rule_tac cp0_app) back
  apply(rule_tac cp0_of_cp0) back
  apply(simp_all add: cp0_eq_cp cp2_P cp1_P lift0_def)
  apply(rule_tac P=P in OclAsSequence_charn_0)
  apply(insert undefCase, simp add: OclUndefined_def lift0_def)
  apply(drule defCase, simp_all add: localValidDefined2sem)
  apply(rule subst)
  apply(rule_tac P="\<lambda> x St. (P (\<lambda> s. (Abs_Sequence_0 \<HolOclLiftLeft>xs\<HolOclLiftRight>)) x) St" and A="B" in cp_charn) 
  apply(simp_all add: cp2_P)
  done

subsection{* Showing the distributivity over the equivalences *}

text{* Note: Actually it should be possible to develop the theory about the inverse
  of \texttt{multiset\_of} in a more general way by combinig the results of Hilbert\_Choice.thy
  and Multiset.thy. But I'm currently not using this approach because only few specifically
  tailored theorems are needed. *}

lemma count_filter_inv_multiset_of:
  "count M x = length (List.filter (\<lambda> y. y = x) (list_of_multiset M))"
  by(auto intro: list_of_multisetI simp: count_filter[symmetric])

lemma bag2seq_holEq:
  "((X::('a,'b::bot)Bag) = Y) = (((\<MathOclAsSequence> X)::('a,'b)Sequence) = (\<MathOclAsSequence> Y))"
  apply(auto simp: OclAsSequence_def ss_lifting' expand_fun_eq)
  apply(erule_tac x="x" in allE)
  apply(case_tac "X x \<noteq> \<bottom>", frule DEF_X_Bag')
  apply(case_tac "Y x \<noteq> \<bottom>", rotate_tac -1, frule DEF_X_Bag')
  apply(auto)
  apply(subst (asm) Abs_Sequence_0_inject_charn, auto)
  apply(subst multiset_eq_conv_count_eq)
  apply(subst count_filter_inv_multiset_of)+
  apply(simp)
  apply(case_tac "Y x \<noteq> \<bottom>", rotate_tac -1, frule DEF_X_Bag')
  apply(auto)
  done

lemma bag2seq_strictEq:
  "((X::('a,'b::bot)Bag) \<MathOclStrictEq> Y) = (((\<MathOclAsSequence> X)::('a,'b)Sequence) \<MathOclStrictEq> (\<MathOclAsSequence> Y))"
  apply(rule ext)
  apply(simp add: OclStrictEq_def OclAsSequence_def ss_lifting')
  apply(rule impI)+
  apply(frule DEF_X_Bag', rotate_tac 1)
  apply(frule DEF_X_Bag', clarify, simp)
  apply(subst Abs_Sequence_0_inject_charn, auto)
  apply(subst multiset_eq_conv_count_eq)
  apply(subst count_filter_inv_multiset_of)+
  apply(simp)
  done

lemma bag2seq_strongEq:
  "((X::('a,'b::bot)Bag) \<MathOclStrongEq> Y) = (((\<MathOclAsSequence> X)::('a,'b)Sequence) \<MathOclStrongEq> (\<MathOclAsSequence> Y))"
  apply(rule ext)
  apply(auto simp: OclStrongEq_def OclAsSequence_def ss_lifting')
  apply(frule DEF_X_Bag', clarify, simp)
  apply(frule DEF_X_Bag', clarify, simp)
  apply(frule DEF_X_Bag', rotate_tac 1)
  apply(frule DEF_X_Bag', clarify, simp)
  apply(subst (asm) Abs_Sequence_0_inject_charn, auto)
  apply(subst multiset_eq_conv_count_eq)
  apply(subst count_filter_inv_multiset_of)+
  apply(simp)
  done

lemmas ss_bag2seq = bag2seq_strictEq bag2seq_strongEq


subsection{* Simplifying multiple conversions *}

text{* asBag is the inverse of asSequence *}
lemma seq2bag_of_bag2seq_UC:
  "(OclAsBag ((OclAsSequence (B::('a,'b::bot)Bag))::('a,'b::bot)Sequence)) = B"
  apply(rule Bag_sem_cases_ext, simp_all)
  apply(simp_all add: OclAsBag_def OclAsSequence_def ss_lifting')
  done

text{* This blowup theorem can be quite useful for conversion proofs about
  functions that map from a collection type to the same collection type. *}
lemmas blowup_bag2seq = seq2bag_of_bag2seq_UC[symmetric]


subsection{* From bags to bags *}

defs
   OclAsBag_def  : "OclAsBag(self::('a, 'b::bot Bag_0) VAL) \<equiv> self"  

lemma bag2bag_id [simp]:
  "((OclAsBag (B::('a,'b::bot)Bag))::('a,'b)Bag) = B"
  by(simp add: OclAsBag_def)


subsection{* From bags to ordered sets *}

defs
   (* asOrderedSet     :: "('a,'b::bot) Bag \<Rightarrow> ('a,'b::bot) Set" *)
   OclAsOrderedSet_def : 
   "OclAsOrderedSet    \<equiv> lift1 (strictify (\<lambda>X. 
                           Abs_OrderedSet_0 \<HolOclLiftLeft>list_of_set (set_of \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>)\<HolOclLiftRight>))"

ocl_setup_op [OclAsOrderedSet]
(*
 TODO: 
change definition (if needed) in such a way that it holds: \verb|Bag\<rightarrow>Seq\<rightarrow>OSet = Bag\<rightarrow>Set\<rightarrow>OSet| 
*)

subsection{* From bags to sets *}

defs
   (* asSet     :: "('a,'b::bot) Bag \<Rightarrow> ('a,'b::bot) Set" *)
   OclAsSet_def : 
   "OclAsSet    \<equiv> lift1 (strictify (\<lambda>X. Abs_Set_0 \<HolOclLiftLeft>set_of \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>))"

ocl_setup_op [OclAsSet]



section{* Building a canonic representation of bags *}

subsection{* The empty bag *}

constdefs
    OclMtBag   :: "('a, 'b::bot Bag_0) VAL"    
   "OclMtBag  \<equiv> lift0(Abs_Bag_0(\<HolOclLiftLeft>{#}\<HolOclLiftRight>))"

    syntax  (* standard syntax *)
    "_OclMtBag_std"   :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("\<OclMtBag>")
    syntax  (* ascii syntax *)
    "_OclMtBag_ascii" :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("mtBag")
    syntax  (xsymbols)
    "_OclMtBag_math"  :: "('a,'b\<Colon>bot) VAL \<Rightarrow> 'a Boolean"     ("\<MathOclMtBag>")
    parse_translation {* flat(map (emb trans_const) ["OclMtBag"]) *}
    print_translation {* map emb_print ["OclMtBag"] *}




section{* Foundational Properties of MtBag *}

lemma OCL_is_isdef_OclMtBag [simp]:
  "\<MathOclValid> \<MathOclIsDefined> \<MathOclMtBag>"
  by(simp add: OclValid_def OclIsDefined_def OclTrue_def 
               OclMtBag_def ss_lifting')


lemma OCL_is_defopt_OclMtBag [simp]:
  "\<tau> \<MathOclValid> \<MathOclIsDefined> \<MathOclMtBag>"
  by(simp add: valid_elim)


subsection{* Relating the empty bag to its counterpart on Sequences *}

lemma bag2seq_mtSequence_mtBag_UC:
  "(\<MathOclMtSequence> ::('a,'b::bot)Sequence) = ((OclAsSequence (\<MathOclMtBag>::('a,'b::bot)Bag)))"
  by(simp add: OclAsSequence_def OclMtBag_def OclMtSequence_def ss_lifting')

lemmas ss_bag2seq = ss_bag2seq bag2seq_mtSequence_mtBag_UC[symmetric]

text{* and vice versa *}

lemma seq2bag_mtBag_mtSequence_UC:
  "(\<MathOclMtBag>::('a,'b::bot)Bag) = ((OclAsBag (\<MathOclMtSequence> ::('a,'b::bot)Sequence)))"
  by(simp add: OCL_Sequence.OclAsBag_def OclMtBag_def OclMtSequence_def ss_lifting')

lemmas ss_seq2bag = seq2bag_mtBag_mtSequence_UC[symmetric]


subsection{* And the 'cons' operation: including *}

defs 

   (* including     :: "[('a,('b::bot) Bag_0) VAL,('a,'b) VAL] => ('a,'b Bag_0) VAL" *)
   OclIncluding_def  :
   "OclIncluding     \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Bag_0 
                                           (\<HolOclLiftLeft>{#Y#} + \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>))))"

ocl_setup_op [OclIncluding]

lemma OCL_is_def_OclIncluding:
  "\<MathOclIsDefined>(OclIncluding (X::('a,'b::bot)Bag) (Y::('a,'b::bot)VAL)) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  by(rule lift2_strict_is_isdef_fw_Bag_Val[OF OclIncluding_def],
     simp add: neq_commute)
  
lemma OCL_is_defopt_OclIncluding:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(OclIncluding (X::('a,'b::bot)Bag) (Y::('a,'b::bot)VAL))) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  by(rule lift2_strictify_implies_LocalValid_defined_Bag_Val[OF OclIncluding_def],
     simp add: neq_commute)

subsection{* The syntax translation mkBag builds now our bags *}

syntax
   "@OclFinBag"      :: "args \<Rightarrow> ('a,'b Bag_0)VAL"           ("mkBag{(_)}")

translations
  "mkBag{x, xs}"     == "OclIncluding (mkBag{xs}) x"
  "mkBag{x}"         == "OclIncluding OclMtBag x"

lemma "mkBag{1,1,1,0,1} = ?X" oops (*test *)

(* TODO: currently all types have to be given explicitely\<dots> could one
         change that? *)

subsection{* Relating including on Bags to its counterpart on sequences *}

lemma seq2bag_including_UC:
  "((OclAsBag S)::('a,'b::bot)Bag) \<MathOclIncluding> x =
   (OclAsBag ((S ::('a,'b::bot)Sequence) \<MathOclIncluding> (x::('a,'b)VAL)))"
  apply(rule_tac X="S" in Sequence_sem_cases_ext, simp_all)
  apply(case_tac "x \<tau> \<noteq> \<bottom>", frule neq_commute[THEN iffD1])
  apply(simp_all add: OCL_Sequence.OclAsBag_def OCL_Sequence.OclIncluding_def
                      OclIncluding_def ss_lifting' mem_set_multiset_eq union_ac)
  done

(* NOTE: decide which direction is the right one in the simpset *)
lemmas ss_seq2bag = ss_seq2bag seq2bag_including_UC

text{* Singletons can be converted in any direction and don't lose information *}

lemma bag2seq_singleton:
  "((\<MathOclMtSequence>::('a,'b::bot)Sequence) \<MathOclIncluding> x) =
   (\<MathOclAsSequence> ((\<MathOclMtBag>::('a,'b)Bag) \<MathOclIncluding> (x::('a,'b)VAL)))"
  apply(rule ext)
  apply(simp add: OclMtSequence_def OclMtBag_def OclIncluding_def OclAsSequence_def
                  OCL_Sequence.OclIncluding_def ss_lifting')
  done

lemmas ss_bag2seq = ss_bag2seq bag2seq_singleton[symmetric]


subsection{* Further properties of including *}

(* TODO: give this order independence property a real name *)

lemma including_ordIndep:
  "(((B::('a,'b::bot)Bag) \<MathOclIncluding> a) \<MathOclIncluding> b) =
   ((B \<MathOclIncluding> (b::('a,'b)VAL)) \<MathOclIncluding> (a::('a,'b)VAL))"
  apply(rule_tac X="B" in Bag_sem_cases_ext, simp_all)
  apply(case_tac "DEF (a x)", case_tac "DEF (a x)")
  apply(simp add: localValidDefined2sem DEF_def_both, clarify)
  apply(simp_all add: OclIncluding_def union_ac ss_lifting')
  done



section{* A second bag constructor on the OCL level *}
defs
   (* collectionRange :: "[('a,'b::{bot,plus}) VAL,('a,'b) VAL] \<Rightarrow> ('a,'b) Bag" *)
   OclCollectionRange_def :
   "OclCollectionRange \<equiv> lift2 (strictify (\<lambda> x\<Colon>Integer_0. strictify (\<lambda> y. 
                           Abs_Bag_0 \<HolOclLiftLeft>multiset_of (map (\<lambda> z::nat. \<HolOclLiftLeft>(int z) + \<HolOclDropLeft>x\<HolOclDropRight>\<HolOclLiftRight>) 
                                                   [0..<(nat (\<HolOclDropLeft>y\<HolOclDropRight> - \<HolOclDropLeft>x\<HolOclDropRight> + 1))])\<HolOclLiftRight>)))"

ocl_setup_op [OclCollectionRange]

(* TODO: 1. add further rules to simplify use of CollectionRange
            e.g. includes and definedness
         2. is there an induction rule over OCL Integers? *)

lemma OCL_is_defopt_OclCollectionRange [simp]:
  "\<tau> \<MathOclValid> \<MathOclIsDefined> ((OclCollectionRange (a::('a Integer)) b)::('a, Integer_0)Bag) = 
   ((\<tau> \<MathOclValid> \<MathOclIsDefined> a) \<and> (\<tau> \<MathOclValid> \<MathOclIsDefined> b))"
  apply(subgoal_tac "count (multiset_of (map (\<lambda>z. \<HolOclLiftLeft>((int z) + \<HolOclDropLeft>(a \<tau>)\<HolOclDropRight>)\<HolOclLiftRight>) [0..<(nat ((\<HolOclDropLeft>(b \<tau>)\<HolOclDropRight> - \<HolOclDropLeft>(a \<tau>)\<HolOclDropRight>) + 1))])) \<bottom> = 0")
  apply(auto simp: localValidDefined2sem OclCollectionRange_def ss_lifting' count_filter)
  done

subsection{* Relating collectionRange on Bags to its counterpart on Sequences *}

lemma seq2bag_collectionRange_UC:
  "((OclCollectionRange a b)::('a,Integer_0)Bag) =
   OclAsBag ((OclCollectionRange (a::'a Integer) (b::'a Integer))::('a,Integer_0)Sequence)"
  apply(rule ext)
  apply(case_tac "x \<MathOclValid> \<MathOclIsUndefined> a", ocl_hypsubst, simp)
  apply(case_tac "x \<MathOclValid> \<MathOclIsUndefined> b", ocl_hypsubst, simp)
  apply(simp add: localValidDefined2sem DEF_def_both, clarify)
  apply(subgoal_tac "\<bottom> \<notin> set (map (\<lambda>z. \<HolOclLiftLeft>((int z) + \<HolOclDropLeft>(a x)\<HolOclDropRight>)\<HolOclLiftRight>) [0..<(nat ((\<HolOclDropLeft>(b x)\<HolOclDropRight> - \<HolOclDropLeft>(a x)\<HolOclDropRight>) + 1))])")
  apply(subgoal_tac "\<bottom> \<notin> set (map (\<lambda>z. \<HolOclLiftLeft>((int z) + \<HolOclDropLeft>(a x)\<HolOclDropRight>)\<HolOclLiftRight>) [0..<(nat (\<HolOclDropLeft>(b x)\<HolOclDropRight> - \<HolOclDropLeft>(a x)\<HolOclDropRight>))])")
  apply(auto simp: OCL_Sequence.OclAsBag_def OclCollectionRange_def ss_lifting'
                   OCL_Sequence.OclCollectionRange_def)
  done

lemmas ss_seq2bag = ss_seq2bag seq2bag_collectionRange_UC


subsection{* Further properties of collectionRange *}

text{* Having a cp-aware rewriter would be that nice\ldots *}

lemma collectionRange_mtBag_conv:
  "\<lbrakk> \<tau> \<MathOclValid> (b::'a Integer) \<MathOclLess> (a::'a Integer) \<rbrakk> \<Longrightarrow>
   OclCollectionRange a b \<tau> = (\<MathOclMtBag>::('a,Integer_0)Bag) \<tau>"
  apply(simp add: ss_seq2bag)
  apply(rule trans, rule_tac P="OclAsBag" in cp_charn)
  apply(rule OCL_Sequence.collectionRange_mtSequence_conv)
  apply(simp_all add: ss_seq2bag)
  done

lemma collectionRange_singleton_UC[simp]:
  "OclCollectionRange a a = \<MathOclMtBag>::('a,Integer_0)Bag \<MathOclIncluding> (a::('a,Integer_0)VAL)"
  by(simp add: seq2bag_collectionRange_UC seq2bag_including_UC[symmetric]
               seq2bag_mtBag_mtSequence_UC)

lemma collectionRange_expand_including:
  "\<lbrakk> \<tau> \<MathOclValid> (a::'a Integer) \<MathOclLe> b \<rbrakk> \<Longrightarrow>
   ((OclCollectionRange a b)::('a,Integer_0)Bag) \<tau> = 
   (OclCollectionRange a (b - 1)) \<MathOclIncluding> b \<tau>"
  apply(simp add: ss_seq2bag)
  apply(rule trans, rule_tac P="OclAsBag" in cp_charn)
  apply(rule OCL_Sequence.collectionRange_expand_including, simp_all)
  done



section{* Iterate *}

defs

   (* iterate     ::"[ ('a,'d::bot)Bag, 
                       [('a,'b::bot)VAL,('a,'b::bot)VAL] => ('a,'c::bot)VAL,
                       ('a,'c::bot)Bag
                     ] 
                       => ('a,'c::bot) Bag" *)
   OclIterate_def : 
   "OclIterate \<equiv> (lift3' lift_arg0 lift_arg2 lift_arg0) (strictify (\<lambda> S P A.
                           (fold_multiset (\<lambda> x y. P y x) A (\<HolOclDropLeft>Rep_Bag_0 S\<HolOclDropRight>))))"

ocl_setup_op [OclIterate]

text{* Its characterisation on the empty bag *}

lemma iterate_charn_mtBag[simp]:
  "(OclIterate (\<MathOclMtBag>::('a,'b::bot)Bag) 
              (P::(('a \<Rightarrow> 'b) \<Rightarrow> (('a \<Rightarrow> 'c::bot) \<Rightarrow> ('a \<Rightarrow> 'c)))) 
               A) = A"
  by(rule ext, simp add: ss_lifting' OclIterate_def OclMtBag_def)

lemma iterate_charn_including_0:
  assumes ordIndep: "\<forall> a b c \<tau>. (P a (lift0 (P b c \<tau>)) \<tau>) = (P b (lift0 (P a c \<tau>)) \<tau>)"
  and     defB: "\<tau> \<MathOclValid> \<MathOclIsDefined>(B::('a,('b::bot))Bag)"
  and     defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('a \<Rightarrow> 'b::bot))"
  shows "((OclIterate (OclIncluding B x) P A) \<tau>) =
         (P (lift0 (x \<tau>)) (lift0 ((OclIterate (B::(('a,'b::bot)Bag)) P A) \<tau>)) \<tau>)"
  apply(insert defx defB ordIndep)
  apply(simp add: localValidDefined2sem)
  apply(frule DEF_X_Bag[THEN iffD1])
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(clarify)
  apply(simp add: OclIncluding_def OclIterate_def ss_lifting'
                  fold_multiset_union_rev
             del: fold_multiset_union)
  done

lemma iterate_charn_including:
  assumes ordIndep: "\<forall> a b c. (P a (P b c)) = (P b (P a c))"
  and     cpP_1:    "\<And> y. cp (\<lambda> x. P x y)"
  and     cpP_2:    "\<And> x. cp (P x)"
  and     defB: "\<tau> \<MathOclValid> \<MathOclIsDefined>(B::('a,('b::bot))Bag)"
  and     defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('a \<Rightarrow> 'b::bot))"
  shows "((OclIterate (OclIncluding B x) P A) \<tau> ) =
         (P x (OclIterate (B::('a,'b::bot)Bag) P A) \<tau>)"
  apply(rule sym[OF trans]) 
  apply(rule_tac x="x" in cp_subst, simp add: cpP_1)
  apply(rule trans)
  apply(rule_tac x="(B \<MathOclIterate>(u;ua=A | (P u ua)))" in cp_subst)
  apply(insert cpP_2, simp add: cp_by_cpify lift0_def)
  apply(rule sym[OF iterate_charn_including_0])
  apply(simp add: ordIndep cp_by_cpify)
  apply(rule defB, rule defx)
  done

lemma iterate_opt_charn_including:
  assumes ordIndep: "\<forall> a b c. (P a (P b c)) = (P b (P a c))"
  and    P_undef_1: "!!y.  P \<MathOclUndefined> y = \<MathOclUndefined>"
  and    P_undef_2: "!!x.  P x \<MathOclUndefined> = \<MathOclUndefined>"
  and        cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  and        cpP_2: "\<And> x. cp ((P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) x)"
  shows              "(OclIterate (OclIncluding B x) P A) =
                      (P x (OclIterate (B::(('a,'b::bot)Bag)) P A))"
  apply(insert P_undef_1 P_undef_2 cpP_1 cpP_2 ordIndep)
  apply(rule ext)
  apply(case_tac "xa \<MathOclValid> \<MathOclIsDefined> B")
  apply(case_tac "xa \<MathOclValid> \<MathOclIsDefined> x")
  apply(rule iterate_charn_including, simp_all)
  apply(frule isUndefined_charn_local[THEN iffD2])
  apply(rule trans, erule_tac A="x" in cp_charn, simp, simp)
  apply(rule sym[OF trans], erule_tac A="x" in cp_charn, simp, simp)
  apply(frule isUndefined_charn_local[THEN iffD2])
  apply(rule trans, erule_tac A="B" in cp_charn, simp, simp)
  apply(rule sym[OF trans], erule_tac A="B" in cp_charn)
  apply(rule_tac P="P x" in cp_compose2, simp_all)
  done


subsection{* Forward rules for functions defined by iterate *}

lemma f_by_iterate_undef_1:
  "\<lbrakk> \<And> B. f (B::('a,'b::bot)Bag) = 
     OclIterate B (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A \<rbrakk>
   \<Longrightarrow>
   f \<MathOclUndefined> = \<MathOclUndefined>"
  by(simp)
  
lemma f_by_iterate_mtBag:
  "\<lbrakk> \<And> B. f (B::('a,'b::bot)Bag) = 
     OclIterate B (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A;
     \<forall> a b c. (P a (P b c)) = (P b (P a c));
     \<And> x. cp (P x) \<rbrakk>
   \<Longrightarrow>
   f \<MathOclMtBag> = A"
  by(simp)

lemma f_by_iterate_including: 
  assumes f_by_it: "\<And> B. (f (B::('a,'b::bot)Bag)) = 
                         (OclIterate B (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A)"
  and  ordIndep: "\<forall> a b c. (P a (P b c)) = (P b (P a c))"
  and      defB: "\<tau> \<MathOclValid> \<MathOclIsDefined>(B::('a,('b::bot))Bag)"
  and      defx: "\<tau> \<MathOclValid> \<MathOclIsDefined>(x::('a \<Rightarrow> 'b::bot))"
  and     cpP_1: "\<And> y. cp (\<lambda> x. P x y)"
  and     cpP_2: "\<And> x. cp (P x)"
  shows "(f (OclIncluding B x) \<tau>) = (P x (f B) \<tau>)"
  by(simp add: f_by_it iterate_charn_including ordIndep defB defx cpP_1 cpP_2)


lemma f_by_iterate_opt_including:
  assumes f_by_it: "\<And> S. (f (S::('a,'b::bot)Bag)) = 
                         (OclIterate S (P::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) A)"
  and    ordIndep: "\<forall> a b c. (P a (P b c)) = (P b (P a c))"
  and   P_undef_1:  "!!y.  P \<MathOclUndefined> y = \<MathOclUndefined>"
  and   P_undef_2:  "!!x.  P x \<MathOclUndefined> = \<MathOclUndefined>"
  and       cpP_1:  "\<And> y. cp (\<lambda> x. P x y)"
  and       cpP_2:  "\<And> x. cp (P x)"
  shows "f (OclIncluding S x) = P x (f S)"
  by(simp add: f_by_it iterate_opt_charn_including ordIndep
               P_undef_1 P_undef_2 cpP_1 cpP_2)


subsection{* Conversion of iterate from bags to sequences *}

text{* For iterates mapping from bags to a non-sequence *}

lemma iterate_bag2seq:
  assumes ordIndep: "\<forall> a b c. (P a (P b c)) = (P b (P a c))"
  shows             "(OclIterate (B :: (('a,('b::bot))Bag)) 
                                 (P::('a,'b::bot)VAL \<Rightarrow>  ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) 
                                  A) =
                     (OclIterate ((OclAsSequence B)::('a,('b::bot))Sequence) P A)"
  apply(insert ordIndep, rule ext)
  apply(simp add: OclAsSequence_def OclIterate_def OCL_Sequence.OclIterate_def
                  ss_lifting)
  apply(case_tac "B x \<noteq> \<bottom>")
  apply(frule DEF_X_Bag')
  apply(auto simp: fold_multiset_def)
  done

lemma iterate_seq2bag:
  assumes ordIndep: "\<forall> a b c. (P a (P b c)) = (P b (P a c))"
  and        cpP_2: "\<And> x. cp (P x)"
  shows             "(OclIterate (S :: (('a,'b::bot)Sequence)) 
                                 (P::('a,'b::bot)VAL \<Rightarrow>  ('a,'c::bot)VAL \<Rightarrow> ('a,'c::bot)VAL) 
                                  A) =
                     (OclIterate ((OclAsBag S)::('a,'b)Bag) P A)"
  apply(insert ordIndep cpP_2, rule ext)
  apply(simp add: OCL_Sequence.OclAsBag_def OclIterate_def 
                  OCL_Sequence.OclIterate_def ss_lifting cp_by_cpify)
  apply(case_tac "S x \<noteq> \<bottom>")
  apply(frule DEF_X_Sequence', clarify)
  apply(auto simp: mem_set_multiset_eq fold_multiset_def)
  apply(rule list_of_multisetI)
  apply(erule foldl_right_order_independent)
  apply(simp)
  done

text{* and for iterates mapping from bags to bags.
  This lemma characterises the same property on lists and multisets and
  serves as a guide through the proof of the theorem on bags and sequences.
  Because there the view is a little bit obfuscated by the contextpassingness
  and undefinedness cases. *}

lemma foldl_multiset2list2multiset:
  assumes   mtC_conv: "e = multiset_of e'"
  and     stepC_conv: "f = (\<lambda> M a. multiset_of (f' (SOME xs. M = multiset_of xs) a))"
  and   ordInsens_f': "\<And> xs ys a. \<lbrakk> multiset_of xs = multiset_of ys \<rbrakk> \<Longrightarrow>
                          multiset_of (f' xs a) = multiset_of (f' ys a)"
  shows "foldl f e xs = multiset_of (foldl f' e' xs)"
  apply(simp add: mtC_conv stepC_conv)
  apply(rule sym)
  apply(rule_tac xs="xs" in rev_induct)
  apply(simp_all)
  apply(rule ordInsens_f', simp)
  apply(rule someI2_ex)
  apply(rule surj_multiset_of[THEN surjD], simp)
  done

lemma iterate_bag2seq2bag_0:
  assumes ordIndepP : "\<forall> a b c \<tau>. (P a (lift0 (P b c \<tau>)) \<tau>) = (P b (lift0 (P a c \<tau>)) \<tau>)"  
  and     mtC_conv: "A = OclAsBag A'"
  and   stepC_conv: "\<And> B x. P x B = OclAsBag (P' x (OclAsSequence B))"
  and ordInsens_P': "\<And> S S' x \<tau>. \<lbrakk> (OclAsBag S \<tau>) = ((OclAsBag S' \<tau>):: 'c::bot Bag_0) \<rbrakk> \<Longrightarrow> 
                                 OclAsBag (P' x S) \<tau> = ((OclAsBag (P' x S') \<tau>):: 'c::bot Bag_0)"
  shows             "(OclIterate (B :: ('a,'b::bot)Bag) 
                                 (P::('a,'b::bot)VAL \<Rightarrow>  ('a,'c::bot)Bag \<Rightarrow> ('a,'c::bot)Bag) 
                                  A) =
                     OclAsBag (OclIterate ((OclAsSequence B)::('a,'b::bot)Sequence) 
                              (P'::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)Sequence \<Rightarrow> ('a,'c::bot)Sequence)
                               A')" 
  apply(rule ext)
  apply(rule trans[symmetric])
  apply(rule_tac P="OclAsBag" in cp_subst, simp add: OCL_cp_OclAsBag)
  apply(simp add: OclIterate_def OCL_Sequence.OclIterate_def ordIndepP)
(* unfold the one AsSequence that does the conversion *)
  apply(simp add: lift3'_def lift_arg0_def lift_arg2_def strictify_def)
  apply(rule_tac P="\<lambda> t. (?A \<longrightarrow> (?B t) \<and> ?C) \<and> ?Z" and 
                 t="(\<MathOclAsSequence> B x)::'b::bot Sequence_0" in subst)
  apply(simp add: OclAsSequence_def ss_lifting')
  apply(subgoal_tac "P = (\<lambda> x B. OclAsBag (P' x (OclAsSequence B)))")
  apply(simp add: mtC_conv)
  prefer 2
  apply(simp add: stepC_conv[symmetric])
  apply(thin_tac "?Q")
  apply(rule list_of_multisetI)
(*  apply(rule someI2_ex[OF surj_multiset_of[THEN surjD]]) *)
  apply(case_tac "(B x) = \<bottom>")
  apply(simp add: OclAsSequence_def OCL_Sequence.OclAsBag_def
                  ss_lifting')
  apply(frule DEF_X_Bag', clarify, simp)
  oops
(* TODO: stabilize this proof when the changes in cp-ness and so on
         are done.

  apply(simp add: mem_set_multiset_eq DEF_def)
  apply(subgoal_tac "((\<MathOclAsSequence> B x)::'b::bot Sequence_0) \<noteq> \<bottom>")
  apply(simp, rule sym)

  apply(rule_tac xs="xs" in rev_induct)
  apply(simp)
  apply(rule trans)
  apply(rule_tac x=A' in  cp_subst)
  apply(simp)
  apply(simp)
  apply(simp)
  apply(rule trans[symmetric])
  apply(rule_tac x1="P' (lift0 xa) (lift0 (foldl (\<lambda> xa y. (P' (lift0 y) (lift0 xa) x)) (A' x) xsa))" in cp_subst[symmetric]) 
  apply(simp add: OCL_cp_OclAsBag)
  apply(rule ordInsens_P')
  apply(rule trans[symmetric])
  apply(rule_tac x1="\<MathOclAsBag> (lift0 (foldl (\<lambda> xa y. (P' (lift0 y) (lift0 xa) x)) (A' x) xsa))" in cp_subst[symmetric]) back
  apply(simp_all add: OCL_cp_OclAsBag seq2bag_of_bag2seq_UC)
  apply(simp add: OclAsSequence_def ss_lifting)
  done
*)


lemma iterate_bag2seq2bag:
  assumes ordIndepP : "\<forall> a b c. (P a (P b c)) = (P b (P a c))"  
  and         cpP_2 : "\<And> x. cp (P x)"
  and     mtC_conv: "A = OclAsBag A'"
  and   stepC_conv: "\<And> B x. P x B = OclAsBag (P' x (OclAsSequence B))"
  and ordInsens_P': "\<And> S S' x \<tau>. \<lbrakk> (OclAsBag S \<tau>) = ((OclAsBag S' \<tau>):: 'c::bot Bag_0) \<rbrakk> \<Longrightarrow> 
                                 OclAsBag (P' x S) \<tau> = ((OclAsBag (P' x S') \<tau>):: 'c::bot Bag_0)"
  shows             "(OclIterate (B :: ('a,'b::bot)Bag) 
                                 (P::('a,'b::bot)VAL \<Rightarrow>  ('a,'c::bot)Bag \<Rightarrow> ('a,'c::bot)Bag) 
                                  A) =
                     OclAsBag (OclIterate ((OclAsSequence B)::('a,'b::bot)Sequence) 
                              (P'::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)Sequence \<Rightarrow> ('a,'c::bot)Sequence)
                               A')" 
  oops

(* MISSING: dependency iterate_bag2seq2bag
  apply(rule iterate_bag2seq2bag_0, insert cpP_2)
  apply(simp add: cp_by_cpify ordIndepP)
  apply(auto intro!: mtC_conv stepC_conv ordInsens_P')
  done
*)


text{* Note: If one looks at the conversion theorems
  about union, including and so on one sees that asBag can be
  taken inside. The same theorem should hold on iterate. But
  currently I'm not sure about its use, because in all proofs
  about a conversion of iterate the ``reformulation'' theorems
  (the form: bag2seq2bag) are used.

  I don't know either if the ``distributive'' version is equally
  or more powerful. Therefore the following lemma after the lemmas
  illustrating the idea on multisets is currently left unproven.
*}

lemma projection_of_foldl:
  assumes projInsens_f: "\<And> xs ys a. \<lbrakk> P xs = P ys \<rbrakk> \<Longrightarrow>
                                    P (f xs a) = P (f ys a)"
  shows "P (foldl f e xs) = 
         foldl (\<lambda> a b. P (f (SOME ys. a = P ys) b)) (P e) xs"
  apply(rule_tac xs="xs" in rev_induct)
  apply(simp_all)
  apply(rule projInsens_f, simp)
  apply(rule someI2, rule sym, simp_all)
  done

lemma multiset_of_foldl:
  assumes ordInsens_f: "\<And> xs ys a. \<lbrakk> multiset_of xs = multiset_of ys \<rbrakk> \<Longrightarrow>
                          multiset_of (f xs a) = multiset_of (f ys a)"
  and ordIndep_f: "(\<forall>a b c. ((f (f a b) c) = (f (f a c) b)))"
  shows "multiset_of (foldl f e xs) = 
         foldl (\<lambda> M a. multiset_of (f (SOME ys. M = multiset_of ys) a)) 
               (multiset_of e) 
               (SOME ys. multiset_of xs = multiset_of ys)"
  oops
(* TODO: think again about properties of fold_multiset
  apply(rule someI2_ex)
  apply(rule surj_multiset_of[THEN surjD])
  apply(subst foldl_multiset_perm)
  apply(assumption, rule ordIndep_f)
  apply(rule projection_of_foldl, erule ordInsens_f)
  done
*)

lemma iterate_seq2bag':
  assumes ordIndepP : "\<forall> a b c. (P' a (P' b c)) = (P' b (P' a c))"  
  and         P'_def: "P' = (\<lambda> x (B::('a,'c::bot)Bag). OclAsBag (P x ((OclAsSequence B)::('a,'c)Sequence)))"
  and         cpP_2 : "\<And> x. cp (P' x)"
  and    ordInsens_P: "\<And> S S' x \<tau>. \<lbrakk> (OclAsBag S \<tau>) = ((OclAsBag S' \<tau>):: 'c::bot Bag_0) \<rbrakk> \<Longrightarrow> 
                                    OclAsBag (P x S) \<tau> = ((OclAsBag (P x S') \<tau>):: 'c::bot Bag_0)"
  shows             "(OclIterate ((OclAsBag S)::('a,'b::bot)Bag) 
                                  (\<lambda> x (B::('a,'c)Bag). OclAsBag (P x ((OclAsSequence B)::('a,'c)Sequence)))
                                  (OclAsBag A)) =
                     OclAsBag (OclIterate (S::('a,'b::bot)Sequence) 
                              (P::('a,'b)VAL \<Rightarrow> ('a,'c::bot)Sequence \<Rightarrow> ('a,'c::bot)Sequence)
                               A)"
  oops

(*
(* NOTE: using the above lemma it is possible to prove the
  ``reformulation'' version of the conversion theorem:
*)
lemma iterate_bag2seq2bag2_proof_of_concept:
  assumes ordIndepP : "\<forall> a b c. (P a (P b c)) = (P b (P a c))"  
  and         cpP_2 : "\<And> x. cp (P x)"
  and     mtC_conv: "A = OclAsBag A'"
  and   stepC_conv: "\<And> B x. P x B = OclAsBag (P' x (OclAsSequence B))"
  and ordInsens_P': "\<And> S S' x \<tau>. \<lbrakk> (OclAsBag S \<tau>) = ((OclAsBag S' \<tau>):: 'c::bot Bag_0) \<rbrakk> \<Longrightarrow> 
                                 OclAsBag (P' x S) \<tau> = ((OclAsBag (P' x S') \<tau>):: 'c::bot Bag_0)"
  shows             "(OclIterate (B :: ('a,'b::bot)Bag) 
                                 (P::('a,'b::bot)VAL \<Rightarrow>  ('a,'c::bot)Bag \<Rightarrow> ('a,'c::bot)Bag) 
                                  A) =
                     OclAsBag (OclIterate ((OclAsSequence B)::('a,'b::bot)Sequence) 
                              (P'::('a,'b::bot)VAL \<Rightarrow> ('a,'c::bot)Sequence \<Rightarrow> ('a,'c::bot)Sequence)
                               A')"
  apply(subgoal_tac  "P=(\<lambda> x (B::('a,'c)Bag). OclAsBag (P' x ((OclAsSequence B)::('a,'c)Sequence)))")
  prefer 2 apply(rule ext, rule ext, rule stepC_conv)
  apply(subst iterate_seq2bag[symmetric])
  apply(simp_all add: seq2bag_of_bag2seq_UC mtC_conv)
  apply(insert ordIndepP, simp)
  apply(insert cpP_2, simp)
  apply(erule ordInsens_P')
  done
*)

section{* Union *}

defs

(* OclUnion      :: "[('a,('b::bot)Bag_0)VAL, ('a,'b Bag_0)VAL] => ('a,'b Bag_0)VAL" *)
   OclUnion_def  :  "OclUnion      \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Bag_0 
                                           (\<HolOclLiftLeft> \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> + \<HolOclDropLeft>Rep_Bag_0 Y\<HolOclDropRight>\<HolOclLiftRight>))))"

ocl_setup_op [OclUnion]


lemma OCL_is_def_OclUnion:
  "\<MathOclIsDefined>(OclUnion (X::('a,'b::bot)Bag) Y) = (\<MathOclIsDefined> X \<MathOclAnd> \<MathOclIsDefined> Y)"
  apply(rule lift2_strict_is_isdef_fw_Bag_Bag[OF OclUnion_def])
  apply(simp)
  done
  
lemma OCL_is_defopt_OclUnion:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(OclUnion (X::('a,'b::bot)Bag) Y)) = ((\<tau> \<MathOclValid> \<MathOclIsDefined> X) \<and> (\<tau> \<MathOclValid>  \<MathOclIsDefined> Y))"
  apply(rule lift2_strictify_implies_LocalValid_defined_Bag_Bag[OF OclUnion_def])
  apply(simp)
  done

lemma union_seq2bag:
  "\<MathOclAsBag> ((X::('a,'b::bot)Sequence) \<MathOclUnion> Y) = ((\<MathOclAsBag> X)::('a,'b)Bag) \<MathOclUnion> ((\<MathOclAsBag> Y)::('a,'b)Bag)"
  apply(rule Sequence_sem_cases_ext, simp_all)
  apply(rule_tac X="X" in Sequence_sem_cases, simp_all)
  apply(simp_all add: OCL_Sequence.OclAsBag_def OclUnion_def mem_set_multiset_eq
                      OCL_Sequence.OclUnion_def ss_lifting')
  done

lemma union_bag2seq2bag:
  "\<MathOclAsBag> (((\<MathOclAsSequence> (X::('a,'b::bot)Bag))::('a,'b)Sequence) \<MathOclUnion> (\<MathOclAsSequence> Y)) = (X \<MathOclUnion> Y)"
  by(simp add: union_seq2bag seq2bag_of_bag2seq_UC)

(* proofing the correspondence without the direct rewritting shows that one
   just reproves the rewrite theorem union_bag2seq2bag

   BUT: Possibly using the ``distributive'' formulation of the conversion 
        theorem of iterate and asBag could simplify this proof. I just haven't
        seen how at the moment.

lemma union_by_iterate:
  "((X::('a,'b::bot)Bag) \<MathOclUnion> (Y::('a,'b::bot)Bag)) = 
   (Y \<MathOclIterate>(x;y=X | y \<MathOclIncluding> (x::('a,'b::bot)VAL) ))"
  apply(rule_tac t=X in subst[OF seq2bag_of_bag2seq_UC])
  apply(rule_tac t=Y in subst[OF seq2bag_of_bag2seq_UC])
  apply(simp add: union_seq2bag[symmetric] union_by_iterate)
  apply(subst iterate_bag2seq2bag) 
  apply(simp_all add: including_ordIndep seq2bag_of_bag2seq_UC ss_seq2bag[symmetric])
  apply(simp add: OclIncluding_def ss_lift_strict)
  done
*)

lemma union_by_iterate:
  "((X::('a,'b::bot)Bag) \<MathOclUnion> (Y::('a,'b::bot)Bag)) = 
   (Y \<MathOclIterate>(x;y=X | y \<MathOclIncluding> (x::('a,'b::bot)VAL) ))"
  apply(simp add: union_bag2seq2bag[symmetric] union_by_iterate)
oops
(* DEPENDENCY: iterate_bag2seq2bag

  apply(subst iterate_bag2seq2bag) 
  apply(simp_all add: including_ordIndep seq2bag_of_bag2seq_UC ss_seq2bag[symmetric])
  apply(simp add: OclIncluding_def ss_lifting')
  done
  (* NOTE: with an contextpassingness aware rewriter this would be one simp *)
*)

section{* Intersection *}

defs
(* OclIntersection:: "[('a,('b::bot) Bag_0)VAL,('a,'b Bag_0)VAL]\<Rightarrow>('a,'b Bag_0)VAL" *)
   OclIntersection_def: "OclIntersection \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Bag_0 
                                           (\<HolOclLiftLeft> \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> #\<inter> \<HolOclDropLeft>Rep_Bag_0 Y\<HolOclDropRight>\<HolOclLiftRight>))))"

ocl_setup_op [OclIntersection]

(* TODO: adapt ocl_setup_op to the case of Bag x Bag \<longrightarrow> Bag functions
         and to the case that the defintions name is not OclFunc_def.
         Because we need two defs for intersection *)

text{* We'll need a second definition for the mixed case of intersection between
a set and a bag  ... the same case holds for union. *}
 


section{* Includes and excludes *}

defs 
   (* includes     :: "[('a,'b::bot) Bag, ('a,'b) VAL] => 'a Boolean" *)
   OclIncludes_def : "OclIncludes \<equiv> lift2(strictify (\<lambda>X. strictify (\<lambda>x.
                                      \<HolOclLiftLeft>0 < count \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> x\<HolOclLiftRight>)))"

   (* excludes     :: "[('a,'b::bot) Bag, ('a,'b) VAL] => 'a Boolean" *)
   OclExcludes_def : "OclExcludes \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x. 
                                      \<HolOclLiftLeft>count \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> x = 0\<HolOclLiftRight>)))"

ocl_setup_op [OclIncludes, OclExcludes]

text{* We prefer a normal form where excludes is written as not includes *}

lemma exludes_not_includes[simp]: 
  "x \<MathOclExcludes> X = \<MathOclNot> (x::('a, 'b)VAL) \<MathOclIncludes> (X::('a, 'b::bot) Bag)"
  apply(rule Bag_sem_cases_ext, simp_all)
  apply(simp_all add: OclExcludes_def OclNot_def OclIncludes_def
                      ss_lifting')
  done

subsection{* Relating includes and excludes on Bags to their counterparts on Sequences *}

lemma includes_bag2seq:
  "(a::('a,'b)VAL) \<MathOclIncludes> (X::('a,'b::bot)Bag) =
   a \<MathOclIncludes> ((OclAsSequence X)::('a,'b::bot)Sequence)" 
  apply(rule ext)
  apply(rule_tac B="X" in OclAsSequence_charn)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Bag', clarify)
  apply(simp add: OclIncludes_def OCL_Sequence.OclIncludes_def
                  OclAsSequence_def ss_lifting' mem_set_multiset_eq)
  done

lemma excludes_bag2seq:
  "(a::('a,'b)VAL) \<MathOclExcludes> (X::('a,'b::bot)Bag) =
   a \<MathOclExcludes> ((OclAsSequence X)::('a,'b::bot)Sequence)"
  by(simp add: includes_bag2seq)

text{* building a simplifier set to expand all functions on bags to the
  ones on sets, if it is possible. *}

(* TODO: give it the right name *)
lemmas ss_bag2seq = ss_bag2seq includes_bag2seq excludes_bag2seq

text{* and vice versa *}

lemma includes_seq2bag:
  "(a::('a,'b)VAL) \<MathOclIncludes> (X::('a,'b::bot)Sequence) =
   a \<MathOclIncludes> ((OclAsBag X)::('a,'b)Bag)"
  apply(rule Sequence_sem_cases_ext, simp_all) 
  apply(simp_all add: OclIncludes_def OCL_Sequence.OclIncludes_def
                      OCL_Sequence.OclAsBag_def ss_lifting' 
                      mem_set_multiset_eq)
  done

lemma excludes_seq2bag:
  "(a::('a,'b)VAL) \<MathOclExcludes> (X::('a,'b::bot)Sequence) =
   a \<MathOclExcludes> ((OclAsBag X)::('a,'b)Bag)"
  by(simp add: includes_seq2bag)


(* TODO: give it the right name *)
lemmas ss_seq2bag = ss_seq2bag includes_seq2bag excludes_seq2bag


subsection{* Characterisation of includes on the semantic part *}

lemma includes_charn1 :
  "(\<tau> \<MathOclValid> x \<MathOclIncludes> S) \<Longrightarrow> (0 < count \<HolOclDropLeft>Rep_Bag_0 (S \<tau>)\<HolOclDropRight> (x \<tau>))" 
  apply(frule isDefined_if_valid)
  apply(drule OCL_is_defopt_OclIncludes[THEN iffD1])
  apply(simp add: OclLocalValid_def OclIsDefined_def strictify_def
                  OclNot_def OclIncludes_def
                  o_def lift0_def lift1_def lift2_def
                  OclTrue_def OclFalse_def  OclUndefined_def DEF_def) 
  done


lemma includes_charn2 :
  "\<lbrakk> DEF(S \<tau>); 0 < count \<HolOclDropLeft>Rep_Bag_0 (S \<tau>)\<HolOclDropRight> (x \<tau>) \<rbrakk> \<Longrightarrow> \<tau> \<MathOclValid> x \<MathOclIncludes> S"
  apply(frule DEF_X_Bag [THEN iffD1], erule exE)
  apply(simp add: OclLocalValid_def OclIsDefined_def strictify_def
                  OclNot_def OclIncludes_def
                  o_def lift0_def lift1_def lift2_def
                  OclTrue_def OclFalse_def  OclUndefined_def DEF_def)
  apply(auto)
  done


subsection{* Higher Properties of includes and excludes *}


(* We prefer a normal form where excludes is eliminated. *)

lemma includes_of_mtBag [simp]:
  "\<not> (\<tau> \<MathOclValid> (x::('a, 'b)VAL) \<MathOclIncludes> (\<MathOclMtBag>::('a, 'b::bot) Bag))"
  by(simp only: ss_bag2seq, rule includes_of_mtSequence)
  
lemma includes_charn1_including[simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('a,'b)VAL); \<tau> \<MathOclValid> \<MathOclIsDefined> (X::('a,'b::bot)Bag) \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> a \<MathOclIncludes> (X \<MathOclIncluding> a)"
  apply(simp only: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Bag', clarify)
  apply(simp add: localValid2sem OclIncludes_def OclIncluding_def
                  ss_lifting' neq_commute)
  done

lemma includes_charn2_including[simp]:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('a,'b)VAL); \<tau> \<MathOclValid> \<MathOclIsDefined> (b::('a,'b)VAL); \<tau> \<MathOclValid> a \<MathOclIncludes> (X::('a,'b::bot)Bag) \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> a \<MathOclIncludes> (X \<MathOclIncluding> b)"
  apply(case_tac "X \<tau> \<noteq> \<bottom>")
  apply(frule DEF_X_Bag', clarify)
  apply(simp_all only: localValidDefined2sem DEF_def)
  apply(simp_all add: localValid2sem OclIncludes_def OclIncluding_def
                  ss_lifting' neq_commute)
  apply(drule neq_commute[THEN iffD1], simp)
  done

text{* It would be more favourable to have this theorem proven on sets first
  and lift it then to all other collections. Because it doesn't work to lift
  it in the other direction, from sequences to bags, sets and ordered sets. *}

lemma includes_of_collectionRange:
  "(x::('a Integer)) \<MathOclIncludes> ((OclCollectionRange a b)::('a,Integer_0)Bag) = 
   ((a::('a Integer)) \<MathOclLe> x) \<MathOclSand> (x \<MathOclLe> (b::('a Integer)))"
  apply(rule ext)
  apply(case_tac "DEF (x xa)", case_tac "DEF (a xa)", case_tac "DEF (b xa)")
  apply(simp add: DEF_def_both, clarify)
  apply(subgoal_tac "count (multiset_of (map (\<lambda>z. \<HolOclLiftLeft>((int z) + \<HolOclDropLeft>(a xa)\<HolOclDropRight>)\<HolOclLiftRight>) [0..<(nat ((\<HolOclDropLeft>(b xa)\<HolOclDropRight> - \<HolOclDropLeft>(a xa)\<HolOclDropRight>) + 1))])) \<bottom> = 0")
  apply(auto simp: OclIncludes_def OclCollectionRange_def OclLe_def
                   OclSand_def ss_lifting' mem_set_multiset_eq[symmetric])
  apply(arith)
  apply(rule_tac x="nat(\<HolOclDropLeft>(x xa)\<HolOclDropRight> - \<HolOclDropLeft>(a xa)\<HolOclDropRight>)" in image_eqI, auto)
  apply(simp add: count_filter)
  done

(* TODO: move to the right position *)
lemma cp_sor [simp,intro!]: "\<lbrakk> cp P; cp P' \<rbrakk> \<Longrightarrow> cp (\<lambda> X. (P X) \<MathOclSor> (P' X))"
  by(simp add: OclSor_def)

(* TODO: correct lemma in sequence variant *)
lemma includes_of_union [simp]:
  "((a::('a,'b::bot)VAL) \<MathOclIncludes> ((X::('a,'b::bot)Bag) \<MathOclUnion> Y))= 
   ((a \<MathOclIncludes> X) \<MathOclSor> (a \<MathOclIncludes> Y))"
  apply(rule_tac X="X" in Bag_sem_cases_ext, simp_all)
  apply(rule_tac X="Y" in Bag_sem_cases, simp_all)
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> a", ocl_hypsubst, simp add: OclSor_def)
  apply(simp add: localValidDefined2sem DEF_def_both, clarify)
  apply(simp_all add: OclIncludes_def OclUnion_def OclSor_def
                      OclSand_def OclNot_def ss_lifting')
  done

lemma not_includes_of_mtBag:
  "(\<tau> \<MathOclValid> \<MathOclNot> (x::('a, 'b)VAL) \<MathOclIncludes> (\<MathOclMtBag>::('a,'b::bot) Bag)) =  (\<tau> \<MathOclValid> \<MathOclIsDefined> x)"
  by(simp only: ss_bag2seq not_includes_of_mtSequence)




section{* Emptiness *}

defs

 
   (* isEmpty      ::"('a, 'b::bot)Bag => 'a Boolean" *)
   OclIsEmpty_def  : "OclIsEmpty \<equiv> lift1 (strictify (\<lambda>X. 
                                           \<HolOclLiftLeft>{#} =  \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>))"

   (* notEmpty     :: "('a,'b::bot)Bag => 'a Boolean" *)
   OclNotEmpty_def : "OclNotEmpty \<equiv> lift1 (strictify (\<lambda>X. 
                                           \<HolOclLiftLeft>{#}  \<noteq>  \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>))"


ocl_setup_op [OclIsEmpty, OclNotEmpty]



text{* we prefer a canical form were notEmpty is written using isEmpty *}
lemma notEmpty_not_isEmpty_conv[simp]:
  "\<MathOclNotEmpty>(X::('a, 'b::bot) Bag) =  \<MathOclNot> (\<MathOclIsEmpty> X)" 
  apply(rule ext)
  apply(simp add: OclNotEmpty_def OclIsEmpty_def OclNot_def
                  ss_lifting')
  done


subsection{* Relating isEmpty and isNotEmpty on Bags to their counterparts on Sequences *}

lemma isEmpty_bag2seq:
  "\<MathOclIsEmpty>(X::('a, 'b::bot) Bag) = \<MathOclIsEmpty>((OclAsSequence X)::('a,('b::bot))Sequence)"
  apply(rule ext)
  apply(rule_tac B="X" in OclAsSequence_charn, simp_all)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Bag', clarify)
  apply(simp_all add: OclIsEmpty_def OclMtSequence_def mem_set_multiset_eq
                      OclStrictEq_def ss_lifting')
  apply(subst Abs_Sequence_0_inject_charn)
  apply(auto simp: mem_set_multiset_eq)
  done

(* TODO: fix error concerning syntax of \<MathOclIsNotEmpty> *)
lemma isNotEmpty_bag2seq:
  "OclNotEmpty (X::('a,'b::bot)Bag) = OclNotEmpty ((OclAsSequence X)::('a,('b::bot))Sequence)"
  by(simp add: isEmpty_bag2seq)

lemmas ss_bag2seq = ss_bag2seq isEmpty_bag2seq isNotEmpty_bag2seq


text{* from sequences to bags *}


lemma isEmpty_seq2bag:
  "\<MathOclIsEmpty>(X::('a, 'b::bot) Sequence) = \<MathOclIsEmpty>((OclAsBag X)::('a,'b)Bag)"
  apply(rule Sequence_sem_cases_ext, simp_all) 
  apply(simp_all add: OclIsEmpty_def OclStrictEq_def OclMtSequence_def
                      OCL_Sequence.OclAsBag_def ss_lifting' 
                      mem_set_multiset_eq)
  apply(subst Abs_Sequence_0_inject_charn)
  apply(auto simp: mem_set_multiset_eq)
  done

(* TODO: fix error concerning syntax of \<MathOclIsNotEmpty> *)
lemma isNotEmpty_seq2bag:
  "OclNotEmpty (X::('a,'b::bot)Sequence) = OclNotEmpty ((OclAsBag X)::('a,('b::bot))Bag)"
  by(simp only: isEmpty_seq2bag notEmpty_not_isEmpty_conv OCL_Sequence.notEmpty_not_isEmpty_conv)

lemmas ss_seq2bag = ss_seq2bag isEmpty_seq2bag isNotEmpty_seq2bag


subsection{* Further properties of the emptiness test *}


lemma isEmpty_stricteq_mtBag_conv[simp]:
  "(\<MathOclIsEmpty> X) = ((X::('a,'b::bot)Bag) \<MathOclStrictEq> \<MathOclMtBag>)"
  apply(rule ext)
  apply(case_tac "X x \<noteq> \<bottom>")
  apply(frule DEF_X_Bag', clarify)
  apply(simp_all add: OclIsEmpty_def OclMtBag_def OclStrictEq_def
                  Abs_Bag_0_inject_charn ss_lifting')
  apply(blast)
  done

lemma isEmpty_of_mtBag: 
  "\<tau> \<MathOclValid> \<MathOclIsEmpty> \<MathOclMtBag>" 
  by (simp)

lemma isEmpty_of_including:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('a,'b)VAL); \<tau> \<MathOclValid> \<MathOclIsDefined> (X::('a,'b::bot)Bag) \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> \<MathOclNot> (\<MathOclIsEmpty>(X \<MathOclIncluding> a))"
  oops

(* TODO: the relation of OclIncluding and StrictEq is missing *)
  
lemma notEmpty_of_mtBag:
  "\<tau> \<MathOclValid> \<MathOclNot> (\<MathOclNotEmpty> \<MathOclMtBag>)"
  by(simp)

lemma notEmpty_of_including:
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (a::('a,'b)VAL); \<tau> \<MathOclValid> \<MathOclIsDefined> (X::('a,'b::bot)Bag) \<rbrakk> \<Longrightarrow> 
   \<tau> \<MathOclValid> \<MathOclNotEmpty> (X \<MathOclIncluding> a)"
  oops


section{* Size *}

defs 
   (* size         :: "('a,'b::bot Bag_0)VAL=>'a Integer" *)
   OclSize_def     :"(OclSize::('a, 'b::bot Bag_0) VAL => 'a Integer) \<equiv> 
                      lift1(strictify(\<lambda>X. \<HolOclLiftLeft>int (size(\<HolOclDropLeft> Rep_Bag_0 X \<HolOclDropRight>))\<HolOclLiftRight> ))"

ocl_setup_op [OclSize]


subsection{* Relating size on bags to its counterpart on sequences *}

(* TODO: move to right place *)
lemma size_multiset_of: "(size (multiset_of xs)) = (length xs)"
  by(induct xs, auto)

lemma size_bag2seq:
  "(\<MathOclSizeLeft>(X :: (('a,('b::bot))Bag))\<MathOclSizeRight>) =
   (\<MathOclSizeLeft>((OclAsSequence X)::('a,('b::bot))Sequence)\<MathOclSizeRight>)" 
  apply(rule ext)
  apply(rule_tac B="X" in OclAsSequence_charn, simp_all)
  apply(simp_all add: localValidDefined2sem DEF_def) 
  apply(frule DEF_X_Bag', clarify)
  apply(auto simp: OclSize_def OCL_Sequence.OclSize_def
                   OclAsSequence_def mem_set_multiset_eq
                   ss_lifting' size_multiset_of)
  done

lemmas ss_bag2seq = ss_bag2seq size_bag2seq


lemma size_seq2bag:
  "(\<MathOclSizeLeft>(X :: (('a,('b::bot))Sequence))\<MathOclSizeRight>) =
   (\<MathOclSizeLeft>((OclAsBag X)::('a,('b::bot))Bag)\<MathOclSizeRight>)" 
  apply(rule Sequence_sem_cases_ext, simp_all)
  apply(auto simp: OclSize_def OCL_Sequence.OclSize_def
                   OCL_Sequence.OclAsBag_def size_multiset_of
                   ss_lifting' mem_set_multiset_eq)
  done

lemmas ss_seq2bag = ss_seq2bag size_seq2bag


subsection{* Further properties of size *}

text{* The real power of the proofs by conversion *}

lemma size_by_iterate:
  "(\<MathOclSizeLeft>(X::('a,'b::bot)Bag)\<MathOclSizeRight>) = 
    OclIterate X (\<lambda> (x::('a,'b::bot)VAL) (y::'a Integer). y + 1) 0"
  by(simp add: ss_bag2seq OCL_Sequence.size_by_iterate iterate_bag2seq)



section{* Count *}

defs

   (* count        :: "[('a, 'b Bag_0) VAL, ('a, 'b::bot) VAL] => 'a Integer" *)
   OclCount_def    : "OclCount \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>x. 
                                           \<HolOclLiftLeft>int (count \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> x)\<HolOclLiftRight>)))"


ocl_setup_op [OclCount]

subsection{* Relating count on bags to its counterpart on sequences *}

lemma count_bag2seq:
  "(X::('a,'b::bot)Bag) \<MathOclCount> (a::('a,'b)VAL) =
   ((\<MathOclAsSequence> X)::('a,'b::bot)Sequence) \<MathOclCount> a"
  apply(rule ext)
  apply(rule_tac B="X" in OclAsSequence_charn, simp_all)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(frule DEF_X_Bag', clarify)
  apply(auto simp: OclCount_def OCL_Sequence.OclCount_def
                   OclAsSequence_def DEF_def mem_set_multiset_eq
                   ss_lifting')
  apply(induct_tac xs, simp_all)
  done

lemmas ss_bag2seq = ss_bag2seq count_bag2seq 


lemma count_seq2bag:
  "(X::('a,'b::bot)Sequence) \<MathOclCount> (a::('a,'b)VAL) =
   ((\<MathOclAsBag> X)::('a,'b::bot)Bag) \<MathOclCount> a"
  apply(rule Sequence_sem_cases_ext, simp_all)
  apply(auto simp: OclCount_def OCL_Sequence.OclCount_def
                   OCL_Sequence.OclAsBag_def
                   ss_lifting' mem_set_multiset_eq)
  apply(induct_tac c, simp_all)
  done

lemmas ss_seq2bag = ss_seq2bag count_seq2bag 


subsection{* Further properties of count *}


lemma count_by_iterate:
  "((X::('a,'b::bot)Bag) \<MathOclCount> ((a::('a,'b)VAL))) = 
   (OclIterate X
               (\<lambda> (x::('a,'b::bot)VAL) (y::'a Integer).\<MathOclIf>  (x \<MathOclStrictEq> a)\<MathOclThen> (y+1) \<MathOclElse> (y) \<MathOclEndif> )   
               (\<MathOclIf> (\<MathOclIsDefined> a) \<MathOclThen> 0 \<MathOclElse>\<MathOclUndefined> \<MathOclEndif> ) )"
  apply(simp add: ss_bag2seq count_by_iterate)
  apply(subst iterate_bag2seq, simp_all)
  apply((rule allI)+, rule ext)
  apply(simp add: OclIf_def plus_def One_ocl_int_def ss_lifting')
  done

lemma includes_by_count_UC[simp]:
  "(X \<MathOclCount> a) \<MathOclGreater> 0 = (a::('a,'b)VAL) \<MathOclIncludes> (X::('a,'b::bot)Bag)"
  by(simp only: ss_bag2seq OCL_Sequence.includes_by_count_UC)

lemma excludes_by_count_UC[simp]:
  "OclCount S x \<MathOclStrictEq> 0 = \<MathOclNot> ((x::('a,'b)VAL) \<MathOclIncludes> (S::('a,'b::bot)Sequence))"
  by(simp only: ss_bag2seq OCL_Sequence.excludes_by_count_UC)


section{* Forall *}

defs
   (* forall    :: "[('a,'b::bot Bag) VAL, ('a,'b::bot)VAL => 'a Boolean] => 'a Boolean" *)
   OclForAll_def: "OclForAll \<equiv> (lift2' lift_arg0 lift_arg1) (strictify( \<lambda> S. (\<lambda> P.  
                      if  \<forall>x \<in> set_of \<HolOclDropLeft>Rep_Bag_0 S\<HolOclDropRight>.  P x = \<HolOclLiftLeft>True\<HolOclLiftRight> 
                      then \<HolOclLiftLeft>True\<HolOclLiftRight>
                      else if \<exists>x \<in> set_of \<HolOclDropLeft>Rep_Bag_0 S\<HolOclDropRight>. P x = \<HolOclLiftLeft>False\<HolOclLiftRight>
                           then \<HolOclLiftLeft>False\<HolOclLiftRight>
                           else \<bottom>)))"
ocl_setup_op [OclForAll]


subsection{* Relating forall on bags to its counterpart on sequences *}
lemma forall_bag2seq: 
  "(\<MathOclForAll> x \<in> (X :: (('a,('b::bot))Bag)) \<bullet> P (x\<Colon>'a \<Rightarrow> 'b)) = 
   (\<MathOclForAll> x \<in> ((OclAsSequence X)::('a,('b::bot))Sequence)  \<bullet> P x)"
  apply(rule ext)
  apply(rule_tac B="X" in OclAsSequence_charn, simp_all)
  apply(simp_all add: OclForAll_def OCL_Sequence.OclForAll_def OclUndefined_def
                      OclAsSequence_def mem_set_multiset_eq cp_by_cpify
                      ss_lifting' localValidDefined2sem)
  done

lemmas ss_bag2seq = ss_bag2seq forall_bag2seq


subsection{* Further properties of forall *}
lemma forall_by_iterate:
  "(\<MathOclForAll> x \<in> (B::('a,'b::bot)Bag) \<bullet> (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x) = 
   (B \<MathOclIterate>(x;y= \<MathOclTrue> | (P x) \<MathOclAnd> y))"
  apply(simp add: forall_by_iterate ss_bag2seq)
  apply(subst iterate_bag2seq, simp_all add: and_assoc)
  apply(simp add: and_commute)
  done



section{* Exists *}

subsection{* Relating exists on bags to its counterpart on sequences *}

lemma exists_bag2seq:
  "(\<MathOclExists> x \<in> (B :: (('a,('b::bot))Bag)) \<bullet> P (x\<Colon>'a \<Rightarrow> 'b)) = 
   (\<MathOclExists> x \<in> ((OclAsSequence B)::('a,('b::bot))Sequence)  \<bullet> P x)"
  by(simp add: OclExists_def forall_bag2seq)

lemmas ss_bag2seq = ss_bag2seq exists_bag2seq


subsection{* Further properties of exists *}


lemma exists_by_iterate:
  "(\<MathOclExists> x \<in> (B::('a,'b::bot)Bag) \<bullet> (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x) = 
   (B \<MathOclIterate>(x;y= \<MathOclFalse> | (P x) \<MathOclOr> y))" 
  apply(simp add: exists_by_iterate ss_bag2seq)
  apply(subst iterate_bag2seq, simp_all)
  apply(subst or_commute)
  apply(simp add: or_assoc)
  apply(simp add: or_commute)
  done


section{* IncludesAll *}

defs
  
   (* includesAll  :: "[('a, 'b::bot Bag_0) VAL, ('a, 'b Bag_0) VAL] => 'a Boolean" *)
   OclIncludesAll_def: "OclIncludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                                           \<HolOclLiftLeft>set_of \<HolOclDropLeft>Rep_Bag_0 Y\<HolOclDropRight> \<subseteq> set_of \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight>\<HolOclLiftRight>)))" 

ocl_setup_op [OclIncludesAll]

subsection{* Relating includesAll on bags to its counterpart ond sequences *}


lemma includesAll_bag2seq:
  "OclIncludesAll (X::('a,'b::bot)Bag) (Y::('a,'b)Bag) =
   OclIncludesAll ((OclAsSequence X)::('a,'b)Sequence) ((OclAsSequence Y)::('a,'b)Sequence)"
  apply(rule ext)
  apply(rule_tac B="X" in OclAsSequence_charn, simp_all)
  apply(rule_tac B="Y" in OclAsSequence_charn, simp_all)
  apply(simp add: OclIncludesAll_def OCL_Sequence.OclIncludesAll_def
                  OclAsSequence_def mem_set_multiset_eq
                  localValidDefined2sem ss_lifting')
  done

lemmas ss_bag2seq = ss_bag2seq includesAll_bag2seq


subsection{* Further properties of includesAll *}

lemma includesAll_by_iterate:
  "OclIncludesAll X (Y::('a,'b::bot)Bag) = 
   (Y \<MathOclIterate>(x;y= \<MathOclIf> (\<MathOclIsDefined> X) \<MathOclThen> \<MathOclTrue> \<MathOclElse> \<MathOclUndefined> \<MathOclEndif> | ((x::('a,'b)VAL) \<MathOclIncludes> X) \<MathOclSand> y))"
  apply(simp add: includesAll_by_iterate ss_bag2seq)
  apply(subst iterate_bag2seq)
  apply(simp_all add: OCL_is_def_OclAsSequence)
  apply(auto, rule ext, auto simp: OclSand_def ss_lifting')
  done



section{* ExcludesAll *}

defs

   (* excludesAll  :: "[('a, 'b::bot Bag_0) VAL, ('a, 'b Bag_0) VAL] => 'a Boolean" *)
   OclExcludesAll_def: "OclExcludesAll \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. 
                                           \<HolOclLiftLeft>set_of \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> \<inter> set_of \<HolOclDropLeft>Rep_Bag_0 Y\<HolOclDropRight> = {}\<HolOclLiftRight>)))"

ocl_setup_op [OclExcludesAll]


subsection{* Relating excludesAll on bags to its counterpart ond sequences *}

lemma excludesAll_bag2seq:
  "OclExcludesAll (X::('a,'b::bot)Bag) (Y::('a,'b)Bag) =
   OclExcludesAll ((OclAsSequence X)::('a,'b)Sequence) ((OclAsSequence Y)::('a,'b)Sequence)"
  apply(rule ext)
  apply(rule_tac B="X" in OclAsSequence_charn, simp_all)
  apply(rule_tac B="Y" in OclAsSequence_charn, simp_all)
  apply(simp add: OclExcludesAll_def OCL_Sequence.OclExcludesAll_def
                  OclAsSequence_def mem_set_multiset_eq
                  localValidDefined2sem ss_lifting')
  done

lemmas ss_bag2seq = ss_bag2seq excludesAll_bag2seq


subsection{* Further properties of excludesAll *}

lemma excludesAll_by_iterate:
  "OclExcludesAll X (Y::('a,'b::bot)Bag) = 
   (Y \<MathOclIterate>(x;y= \<MathOclIf> (\<MathOclIsDefined> X) \<MathOclThen> \<MathOclTrue> \<MathOclElse> \<MathOclUndefined> \<MathOclEndif> | ((x::('a,'b)VAL) \<MathOclExcludes> X) \<MathOclSand> y))"
  apply(simp add: excludesAll_by_iterate ss_bag2seq)
  apply(subst iterate_bag2seq)
  apply(simp_all add: OCL_is_def_OclAsSequence)
  apply(auto, rule ext, auto simp: OclSand_def ss_lifting')
  done


section{* Select *}

defs
   (* select    :: "[('a,'c::bot Bag) VAL, ('a,'b::bot)VAL => 'a Boolean] *)
   OclSelect_def:  "OclSelect \<equiv>  (lift2' lift_arg0 lift_arg1) (strictify (\<lambda> S P.
                       if    \<forall> x \<in> set_of \<HolOclDropLeft>Rep_Bag_0 S\<HolOclDropRight>. DEF (P x)
                       then  Abs_Bag_0(\<HolOclLiftLeft>MCollect \<HolOclDropLeft>Rep_Bag_0 S\<HolOclDropRight> (\<lambda> x. P x = \<HolOclLiftLeft>True\<HolOclLiftRight>)\<HolOclLiftRight>)
                       else  \<bottom>))"

ocl_setup_op [OclSelect]


subsection{* Relating select on bags to its counterpart on sequences *}

lemma select_bag2seq2bag:
  "(OclSelect (X::('a,'b::bot)Bag) (P::('a,'b)VAL \<Rightarrow> 'a Boolean)) =
   (OclAsBag (OclSelect ((OclAsSequence X)::('a,'b)Sequence) P))" 
  apply(rule ext)
  apply(case_tac "X x \<noteq> \<bottom>")
  apply(frule DEF_X_Bag', clarify)
  apply(auto simp: OclAsSequence_def OCL_Sequence.OclAsBag_def
                   OclSelect_def OCL_Sequence.OclSelect_def
                   ss_lifting')
  done


lemmas ss_bag2seq2bag = select_bag2seq2bag


subsection{* Further properties of select *}


lemma select_by_iterate:
  "(\<MathOclSelectLeft>x : (B::('a,'b::bot)Bag) | ((P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x)\<MathOclSelectRight>) = 
   (B \<MathOclIterate>(x; y = \<MathOclMtBag> | 
                             \<MathOclIf>  (P x)\<MathOclThen> (y \<MathOclIncluding> x) \<MathOclElse> y \<MathOclEndif> ))"
oops
(* DEPENDENCY: iterate_bag2seq2bag

  apply(simp add: select_by_iterate ss_bag2seq2bag)
  apply(subst iterate_bag2seq2bag, simp_all)
(* TODO: cases of OclIf would be very nice *)
  apply((rule allI)+, rule ext)
  apply(simp add: OclIf_def OclIncluding_def ss_lifting')
  apply(case_tac "\<not> DEF (P a x)", simp add: DEF_def)
  apply(case_tac "\<not> DEF (P b x)", simp add: DEF_def)
  apply(case_tac "\<not> DEF (a x)", simp add: DEF_def)
  apply(case_tac "\<not> DEF (b x)", simp add: DEF_def)
  apply(case_tac "\<not> DEF (c x)", simp add: DEF_def)
  apply(simp only: DEF_def_both, clarify)
  apply(frule DEF_X_Bag', clarify)
  apply(auto simp: union_ac)
  apply(rule ext)
(* NOTE: ocl_hypsubst does not work here\<dots> and would be very useful*)
(* NOTE (as well ;-): the required assumptions are satisfyable\<dots> but its quite
                      hard work to do it without real support for OclIf and
                      would be much easier with a cp aware reweriter *)
(* TODO: finish *)
  oops
*)

section{* Reject *}


defs
   (* reject    :: "[('a,'c::bot Bag) VAL, ('a,'b::bot)VAL => 'a Boolean] *)
   OclReject_def:  "OclReject \<equiv>  (lift2' lift_arg0 lift_arg1) (strictify (\<lambda> S P.
                       if    \<forall> x \<in> set_of \<HolOclDropLeft>Rep_Bag_0 S\<HolOclDropRight>. DEF (P x)
                       then  Abs_Bag_0(\<HolOclLiftLeft>MCollect \<HolOclDropLeft>Rep_Bag_0 S\<HolOclDropRight> (\<lambda> x. P x = \<HolOclLiftLeft>False\<HolOclLiftRight>)\<HolOclLiftRight>)
                       else  \<bottom>))"

ocl_setup_op [OclReject]



section{* Excluding *}

defs
   (* excluding     :: "[('a,('b::bot) Bag_0) VAL,('a,'b) VAL] => ('a,'b Bag_0) VAL" *)
   OclExcluding_def  :
   "OclExcluding     \<equiv> lift2 (strictify (\<lambda>X. strictify (\<lambda>Y. Abs_Bag_0 
                                           (\<HolOclLiftLeft> MCollect \<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> (\<lambda> a. a \<noteq> Y) \<HolOclLiftRight>))))"

ocl_setup_op [OclExcluding]

(*
section{* Flatten *}

text{* More to come . . .  *}

*)
section{* Collect *}

text{* Note currently this collect corresponds to a collectNested. This will be fixed later. *}

defs

   (* collect    ::"[('a,'d::bot Bag) VAL, ('a,'b::bot)VAL => ('a,'c::bot)VAL] *)
   OclCollect_def: "OclCollect S P \<equiv>  \<lambda> \<tau>. 
                    if DEF (S \<tau>)
                    then if \<forall>x \<in> set_of \<HolOclDropLeft>Rep_Bag_0 (S \<tau>)\<HolOclDropRight>. DEF (P (\<lambda>\<tau>. x) \<tau>)
                         then Abs_Bag_0 \<HolOclLiftLeft>Abs_multiset ((count \<HolOclDropLeft>Rep_Bag_0 (S \<tau>)\<HolOclDropRight>) o (inv (\<lambda> x. P (lift0 x) \<tau>)))\<HolOclLiftRight> 
                         else \<bottom>
                    else \<bottom>"

ocl_setup_op [OclCollect]


section{* One *}

(* TODO: make a semantic definition and prove this fact *)
defs
   (* one         ::"[('a,'c::bot Bag) VAL, ('a,'b::bot)VAL => 'a Boolean] => 'a Boolean" *)
   OclOne_def     : "OclOne (S::('a,('b::bot) Bag_0)VAL) P  \<equiv>
                     \<MathOclSizeLeft>(\<MathOclSelectLeft>x : S | P (x\<Colon>'a\<Colon>type \<Rightarrow> 'b\<Colon>bot)\<MathOclSelectRight>)\<MathOclSizeRight> \<MathOclStrictEq> 1"

ocl_setup_op [OclOne]

(*
section{* IsUnique *}

text{* More to come . .  *}


section {* Any *}

text{* More to come . .  *}
*)

section{* Sum *}

text{* Like collect sum is specified as mapping to other ocl functions \ldots  this is because
       for every type supporting ${plus,zero}$ the sum function must be available and it
       make no sense to add a new definition for every one of these *}

defs
   (* sum          :: "('a,'b::{bot,plus,zero})Bag) => ('a,'b::{bot,plus,zero})VAL" *)
   OclSum_def      :
   "OclSum S       \<equiv> OclIterate (S::('a,'b::{bot,plus,zero})Bag) (\<lambda> x y. x + y) 0"



section{* Complement *}

(* TODO: is this really needed in bags? 
         it's not prescribed by the standard\<dots> is it? *)
         
defs

(* OclComplement :: "('a, 'b::bot Bag_0) VAL => ('a, 'b Bag_0) VAL"  *)
   OclComplement_def:"OclComplement   \<equiv> lift1 (strictify(\<lambda>X. Abs_Bag_0 
                                           (\<HolOclLiftLeft> -(\<HolOclDropLeft>Rep_Bag_0 X\<HolOclDropRight> - {#\<bottom>#})\<HolOclLiftRight>)))"

ocl_setup_op[OclComplement]


subsection{* Stuff that should be moved into the right theory  *}

(* MOVE this theory to OCL_Set.thy when the whole reordering of the theories is done

lemma OCL_is_defopt_OclAsSet: 
  "(\<tau> \<MathOclValid> \<MathOclIsDefined> (((self::('a,'b::bot)Bag) ->asSet())::(('a, 'b::bot)Set))) = (\<tau> \<MathOclValid> \<MathOclIsDefined> self)"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule iffI) prefer 2
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(simp add: OclAsSet_def strictify_def lift1_def DEF_def)
  apply(fold DEF_def)
  apply(erule DEF_Set_if_DEF_Bag)
  apply(auto simp: OclAsSet_def strictify_def lift1_def DEF_def)
done

lemma OCL_is_def_OclAsSet:
  "(\<MathOclIsDefined> (((self::('a,'b::bot)Bag) ->asSet())::(('a, 'b::bot)Set))) = (\<MathOclIsDefined> self)"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule ext)
  apply(rule arg_cong) back
  apply(subgoal_tac "((St \<MathOclValid> (\<MathOclIsDefined> \<MathOclAsSet> self)) = (St \<MathOclValid> (\<MathOclIsDefined> self)))")
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule OCL_is_defopt_OclAsSet)
  done

lemma OclAsSet_finite:
  "(\<tau> \<MathOclValid> \<MathOclIsDefined>(S::('a,'b::bot)Bag)) =
   (\<tau> \<MathOclValid> \<MathOclIsDefined>(((OclAsSet S)::('a,'b::bot)Set)->size()))"
  apply(simp add: OclIsDefined_def lift0_def lift1_def lift2_def
                  strictify_def OclLocalValid_def OclTrue_def OclFalse_def)
  apply(rule iffI)
  apply(frule DEF_X_Bag[THEN iffD1])  
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(simp_all add: OclAsSet_def OCL_Set.OclSize_def
                      strictify_def lift1_def DEF_def)
  apply(auto)
  done

lemma OclAsSet_mtBag_mtSet_conv[simp]:
  "((OclAsSet (\<MathOclMtBag>::('a,'b::bot)Bag))) = (\<MathOclMtSet>::('a,'b::bot)Set)"
  apply(insert Abs_Bag_Mt_DEF)
  apply(simp add: OclAsSet_def OclMtBag_def OclMtSet_def
                  lift0_def lift1_def strictify_def DEF_def)
  done

subsection{* Fundamental relations between OclAsSequence and OclAsSet *}

lemma bag2seq2set_eq_bag2set[simp]:
  "((OclAsSet ((OclAsSequence (B::('a,'b::bot)Bag))::('a,'b::bot)Sequence))::('a,'b::bot)Set) = 
   ((OclAsSet B)::('a,'b::bot)Set)"
  apply(rule ext)
  apply(rule_tac B="B" in OclAsSequence_charn)
  apply(simp_all add: cp2_by_cp)
  apply(drule localValidDefined2sem[THEN iffD1])
  apply(simp add: OclAsSet_def OCL_Sequence.OclAsSet_def
                  lift0_def lift1_def strictify_def)
  apply(frule DEF_X_Bag_0[THEN iffD1])
  apply(auto simp: mem_set_multiset_eq)
  done

 
subsection{* Relating includes and excludes on Bags to their counterparts on Sets *}

lemma includes_bag2set:
  "(x::('a,'b)VAL) \<MathOclIncludes> (S::('a,'b::bot)Bag) =
   x \<MathOclIncludes> ((OclAsSet S)::('a,'b::bot)Set)" 
  by(simp add: includes_seq2set includes_bag2seq)

lemma excludes_bag2set:
  "(x::('a,'b)VAL) \<MathOclExcludes> (S::('a,'b::bot)Bag) =
   x \<MathOclExcludes> ((OclAsSet S)::('a,'b::bot)Set)"
  by(simp add: includes_bag2set)
*)


(* MOVE to theory OCL_Set.thy after reordering is finished

section{* Relating the operations for which it is possible to the corresponding set operations *}

lemma isEmpty_bag2set:
  "OclIsEmpty(S::('a, 'b::bot) Bag) = OclIsEmpty ((OclAsSet S)::('a,('b::bot))Set)"
  by(simp add: isEmpty_bag2seq isEmpty_seq2set)

lemma isNotEmpty_bag2set:
  "OclNotEmpty(S::('a, 'b::bot) Bag) = OclNotEmpty ((OclAsSet S)::('a,('b::bot))Set)"
  by(simp only: notEmpty_eq_not_isEmpty OCL_Set.not_empty isEmpty_bag2set)

lemma forall_bag2set: 
  "(\<MathOclForAll> x \<in> (S :: (('a,('b::bot))Bag)) \<bullet> P (x\<Colon>'a \<Rightarrow> 'b)) = 
   (\<MathOclForAll> x \<in> ((OclAsSet S)::('a,('b::bot))Set)  \<bullet> P x)"
  by(simp add: forall_bag2seq forall_seq2set)

lemma exists_bag2set:
  "(\<MathOclExists> x \<in> (S :: (('a,('b::bot))Bag)) \<bullet> P (x\<Colon>'a \<Rightarrow> 'b)) = 
   (\<MathOclExists> x \<in> ((OclAsSet S)::('a,('b::bot))Set)  \<bullet> P x)"
  by(simp add: OclExists_def forall_bag2set)

lemma includesAll_bag2set:
  "OclIncludesAll (S1::('a, 'b::bot)Bag) (S2::('a, 'b::bot)Bag) =
   OclIncludesAll ((OclAsSet S1)::('a, 'b::bot)Set) ((OclAsSet S2)::('a, 'b::bot)Set)"
  by(simp add: includesAll_bag2seq includesAll_seq2set)

lemma excludesAll_bag2set:
  "OclExcludesAll (S1::('a, 'b::bot)Bag) (S2::('a, 'b::bot)Bag) =
   OclExcludesAll ((OclAsSet S1)::('a, 'b::bot)Set) ((OclAsSet S2)::('a, 'b::bot)Set)"
  by(simp add: excludesAll_bag2seq excludesAll_seq2set)

*)



end


