(* ****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_CharactersticSet.thy -
 * Copyright (C) 2001,2002,2003,2004 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                   Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNES 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ***************************************************************************** *)
header {* CharactersticSets *}

theory OCL_CharactersticSet
imports 
   "$HOLOCL_HOME/src/library/basic/OCL_Real"
   "$HOLOCL_HOME/src/library/basic/OCL_String"   
   "$HOLOCL_HOME/src/library/OclAny/OCL_OclAny"
begin
 text_raw{* \label{thy:OCL_CharactersticSet} *}

section{* Embedding Characteristic Sets into \OCL *}

lemma Integer_0_Defined [simp]:
"\<tau> \<MathOclValid> \<MathOclIsDefined> Integer_0"
by(simp add: localValidDefined2sem Abs_Set_Mt_DEF Integer_0_def lift0_def)

lemma   in_Integer_0: 
"\<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow>
 \<tau> \<MathOclValid> (x::'a Integer) \<MathOclIncludes> (Integer_0::('a,Integer_0)Set)"
apply (simp add: OCL_Set.OclIncludes_def Integer_0_def localValidDefined2sem
                 OclLocalValid_def OclIsDefined_def OclTrue_def
                 lift0_def lift1_def lift2_def strictify_def 
                 Abs_Set_rangelift_DEF[simplified DEF_def])
by    (auto simp:DEF_def not_down_exists_lift)

lemma   in_Integer_0R[simp]: 
"(\<tau> \<MathOclValid> (x::'a Integer) \<MathOclIncludes> (Integer_0::('a,Integer_0)Set)) = 
 (\<tau> \<MathOclValid> \<MathOclIsDefined> x)"
by (auto simp: in_Integer_0
               OCL_Logic_core.isDefined_if_valid
                  [THEN OCL_Set.OCL_is_defopt_OclIncludes
                           [THEN iffD1]])

text{* The relevance of the latter type of rules become apparent if we 
       look at the following example: *}
lemma  test:
"(\<MathOclForAll> x \<in> Integer_0 \<bullet> (x + (OclNegative x) \<MathOclStrictEq> 0)) = \<MathOclTrue>"
oops

text{* First of all, such an expression is syntactically illegal
       in \OCL. In Standard syntax, this is presented as
       \begin{ocl}
           Integer_0 ->forAll( x | x + (-x) = 0)
       \end{ocl}
       \inlineocl{Integer_0} is considered a type name which is not part of 
       the expression language; therefore, no quantification
       over them is possible in standard OCL. Since we interpret \OCL types
       as sets in the sense of the \theory{OCL\_Set} theory, we can
       quantify and reason over it in \holocl.

       In particular, rule @{text "forAllI"} eliminates the quantifier
       and reduces this to 
       \begin{equation}
          \isasymtau  \isasymTurnstile  x \isasymin  Integer_0 
          \isasymLongrightarrow  (x + (OclNegative x) \isasymMathOclStrictEq  0)) = \isasymMathOclTrue 
       \end{equation}

       which can be further simplified by @{text "in_Integer_0R"} 
       to
       \begin{equation}
          \isasymtau  \isasymTurnstile   \isasymMathOclIsDefined  x 
           \isasymLongrightarrow  (x + (- x) \isasymMathOclStrictEq  0)) = \isasymMathOclTrue 
       \end{equation}
       This is a valid statement in \OCL (although we ommit
       the proof here). In contrast, 
       \begin{equation}
          (x + (-x) \isasymMathOclStrictEq 0)) = \isasymMathOclTrue
       \end{equation}
       in itself is not valid since $x$ may be undefined.
       Thus, quantifiers range over defined values only
       (as one would expect). These theorems
       also implicitly state that the characteristic sets
       of the basic types are non-empty.
*}

lemma Real_Defined [simp]:
"\<tau> \<MathOclValid> \<MathOclIsDefined> Real_0"
by(simp add: localValidDefined2sem Real_0_def lift0_def)

lemma   in_Real_0: 
"\<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow>
 \<tau> \<MathOclValid> (x::'a Real) \<MathOclIncludes> (Real_0::('a,Real_0)Set)"
apply (simp add: OCL_Set.OclIncludes_def Real_0_def localValidDefined2sem
                 OclLocalValid_def OclIsDefined_def OclTrue_def
                 lift0_def lift1_def lift2_def strictify_def 
                 Abs_Set_rangelift_DEF[simplified DEF_def])
by    (auto simp:DEF_def not_down_exists_lift)

lemma   in_RealR_0[simp]: 
"(\<tau> \<MathOclValid> (x::'a Real) \<MathOclIncludes> (Real_0::('a,Real_0)Set)) = 
 (\<tau> \<MathOclValid> \<MathOclIsDefined> x)"
by (auto simp: in_Real_0
               isDefined_if_valid
                  [THEN OCL_Set.OCL_is_defopt_OclIncludes
                           [THEN iffD1]])



lemma Boolean_Defined [simp]:
"\<tau> \<MathOclValid> \<MathOclIsDefined> Boolean_0"
by(simp add: localValidDefined2sem Boolean_0_def lift0_def)

lemma   in_Boolean: 
"\<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow>
 \<tau> \<MathOclValid> (x::'a Boolean) \<MathOclIncludes> (Boolean_0::('a,Boolean_0)Set)"
apply (simp add: OCL_Set.OclIncludes_def Boolean_0_def localValidDefined2sem
                 OclLocalValid_def OclIsDefined_def OclTrue_def
                 lift0_def lift1_def lift2_def strictify_def 
                 Abs_Set_rangelift_DEF[simplified DEF_def])
by    (auto simp:DEF_def not_down_exists_lift)

lemma   in_BooleanR[simp]: 
"(\<tau> \<MathOclValid> (x::'a Boolean) \<MathOclIncludes> (Boolean_0::('a,Boolean_0)Set)) = 
 (\<tau> \<MathOclValid> \<MathOclIsDefined> x)"
by (auto simp: in_Boolean
               isDefined_if_valid
                  [THEN OCL_Set.OCL_is_defopt_OclIncludes
                           [THEN iffD1]])


lemma String_0_Defined [simp]:
"\<tau> \<MathOclValid> \<MathOclIsDefined> String_0"
by(simp add: localValidDefined2sem String_0_def lift0_def)

lemma   in_String_0: 
"\<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow>
 \<tau> \<MathOclValid> (x::'a String) \<MathOclIncludes> (String_0::('a,String_0)Set)"
apply (simp add: OCL_Set.OclIncludes_def String_0_def localValidDefined2sem
                 OclLocalValid_def OclIsDefined_def OclTrue_def
                 lift0_def lift1_def lift2_def strictify_def 
                 Abs_Set_rangelift_DEF[simplified DEF_def])
by    (auto simp:DEF_def not_down_exists_lift)

lemma   in_StringR[simp]: 
"(\<tau> \<MathOclValid> (x::'a String) \<MathOclIncludes> (String_0::('a,String_0)Set)) = 
 (\<tau> \<MathOclValid> \<MathOclIsDefined> x)"
by (auto simp: in_String_0
               isDefined_if_valid
                  [THEN OCL_Set.OCL_is_defopt_OclIncludes
                           [THEN iffD1]])



lemma OclAny_Defined [simp]:
"\<tau> \<MathOclValid> \<MathOclIsDefined> OclAny"
by(simp add: localValidDefined2sem OclAny_def level1.OclAny_def lift0_def)

lemma   in_OclAny: 
"\<tau> \<MathOclValid> \<MathOclIsDefined> x \<Longrightarrow>
 \<tau> \<MathOclValid> (x::('a,'b) OclAny) \<MathOclIncludes> (OclAny::('a,'b OclAny_0)Set)"
apply (simp add: OCL_Set.OclIncludes_def OclAny_def level1.OclAny_def
                 localValidDefined2sem
                 OclLocalValid_def OclIsDefined_def OclTrue_def
                 lift0_def lift1_def lift2_def strictify_def 
                 Abs_Set_rangelift_DEF[simplified DEF_def])
by    (auto simp:DEF_def not_down_exists_lift)


lemma   in_OclAnyR[simp]: 
" \<tau> \<MathOclValid> (x::('a,'b) OclAny) \<MathOclIncludes> (OclAny::('a,'b OclAny_0)Set) = 
 (\<tau> \<MathOclValid> \<MathOclIsDefined> x)"
by (auto simp: in_OclAny
               isDefined_if_valid
                  [THEN OCL_Set.OCL_is_defopt_OclIncludes
                           [THEN iffD1]])
end
 
