(* ****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Collection_requirements.thy -
 * Copyright (C) 2001-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNES 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ***************************************************************************** *)

header {* \OCL Collection requirements *}
theory OCL_Collection_requirements
imports 
  "$HOLOCL_HOME/src/OCL_Library"
begin
 text_raw{* \label{thy:OCL_Collection_requirements} *}

text {*
  This theory contains the requirements for the collection types defined in 
  the \OCL standard library (Set, OrderedSet, Bag, and 
  Sequence)~\cite[Chapter 11.7]{omg:ocl:2003}. 
*}
(**************************************************************************)
section {* Collection *}
(**************************************************************************)
text_raw {* 
\begin{oclrequirement}{11.7.1-a}{size() : Integer} 
The number of elements in the collection \emph{self}.
\begin{ocl}
post: result = self->iterate(elem; acc : Integer = 0 | acc + 1)
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_7_1-a": (* REQ_bag_11_7_1_1: *)
"\<MathOclSizeLeft>(self::('a,'b::bot)Bag)\<MathOclSizeRight> = OclIterate self (\<lambda> (elem::('a,'b::bot)VAL) (acc::'a Integer). acc + 1) 0"
    by(simp add: OCL_Bag.size_by_iterate)
(* TODO: correct includes_by_count_UC such that it is in normal form (OclLess instead of OclGreater) *)
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_7_1-a":   (* REQ_seq_11_7_1_1: *)
"\<MathOclSizeLeft>(self::('a,'b::bot)Sequence)\<MathOclSizeRight> = OclIterate self (\<lambda> (elem::('a,'b::bot)VAL) (acc::'a Integer). acc + 1) 0"
  by(simp add: OCL_Sequence.size_by_iterate)
(* TODO: correct includes_by_count_UC such that it is in normal form (OclLess instead of OclGreater) *)
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.7.1-b}{includes(object : T) : Boolean}
True if object is an element of self, false otherwise.
\begin{ocl}
post: result = (self->count(object) > 0)
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_7_1-b":     
  "((obj::('a,'b)VAL) \<MathOclIncludes> (self::('a,'b::bot)Bag)) =  ((self \<MathOclCount> obj) \<MathOclGreater> 0)"
  by(simp only: OCL_Bag.includes_by_count_UC)

text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_7_1-b":     
  "((obj::('a,'b)VAL) \<MathOclIncludes> (self::('a,'b::bot)Sequence)) =  ((self \<MathOclCount> obj) \<MathOclGreater> 0)"
  by(simp only: OCL_Sequence.includes_by_count_UC)
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.7.1-c}{excludes(object : T) : Boolean}
True if object is not an element of self, false otherwise.
\begin{ocl}
post: result = (self->count(object) = 0)
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_7_1-c":
  "((obj::('a,'b)VAL) \<MathOclExcludes> (self::('a,'b::bot)Bag) ) = ((self \<MathOclCount> obj) \<MathOclStrictEq> 0)"
  oops
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_7_1-c":
  "((obj::('a,'b)VAL) \<MathOclExcludes> (self::('a,'b::bot)Sequence) ) = ((self \<MathOclCount> obj) \<MathOclStrictEq> 0)"
  by(simp)
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.7.1-d}{count(object : T) : Integer}
The number of times that object occurs in the collection self.
\begin{ocl}
post: result = self->iterate( elem; acc : Integer = 0 |
         if elem = object then acc + 1 else acc endif)
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_7_1-d":
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(obj::('a,'b)VAL) \<rbrakk> \<Longrightarrow>
   ((self::('a,'b::bot)Bag) \<MathOclCount> obj ) \<tau> = 
   (self \<MathOclIterate>(x;y=0 | (\<MathOclIf> (x \<MathOclStrictEq> obj) \<MathOclThen> (y + 1)\<MathOclElse> y\<MathOclEndif>))) \<tau>"
  by(simp add: OCL_Bag.count_by_iterate OCL_Bag.OclIterate_def 
               OclIf_def localValid2sem ss_lifting')
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_7_1-d":
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(obj::('a,'b)VAL) \<rbrakk> \<Longrightarrow>
   ((self::('a,'b::bot)Sequence) \<MathOclCount> obj ) \<tau> = 
   (self \<MathOclIterate>(x;y=0 | (\<MathOclIf> (x \<MathOclStrictEq> obj) \<MathOclThen> (y + 1)\<MathOclElse> y\<MathOclEndif>))) \<tau>"
  by(simp add: OCL_Sequence.count_by_iterate OCL_Sequence.OclIterate_def 
               OclIf_def localValid2sem ss_lifting')
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.7.1-e}{includesAll(c2 : Collection(T)) : Boolean}
Does self contain all the elements of c2 ?
\begin{ocl}
post: result = c2->forAll(elem | self->includes(elem))
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
text{* 
  \OCLglitch{Wrong definition of the correspondence of includesAll and iterate}

  If @{text "self = \<MathOclUndefined>"} and @{text "(obj = \<MathOclMtSet>)"} then the left-hand side of 
  the equation evaluates to @{text "\<MathOclUndefined>"} whereas the right-hand side evaluates 
  to @{text "\<MathOclTrue>"}.
*}
(* TODO: how can one prove such a statement by a counterexample? *)
lemma "req_set_11_7_1-e": 
  "(((self::('\<tau>,'\<alpha>::bot)Set) ->includesAll(obj)) 
   = (obj ->forAll(elem | (self ->includes((elem::('\<tau>,'\<alpha>::bot)VAL)))))) = False"
  apply(simp add: OclIncludes_def OclIncludesAll_def OclForAll_def
                  lift0_def lift1_def lift2_def strictify_def DEF_def)
  oops
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_7_1-e": 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(self::('a,'b::bot)Bag) \<rbrakk> \<Longrightarrow>
   (OclIncludesAll self obj) \<tau> = (\<MathOclForAll> elem \<in> obj \<bullet> ((elem::('a,'b)VAL) \<MathOclIncludes> self)) \<tau>"
  oops
(* missing: cp of forall 
  apply(rule_tac X="self" in Bag_sem_cases)
  apply(rule_tac X="obj" in Bag_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(auto simp: OclForAll_def OclIncludesAll_def OclStrongEq_def 
                   OclIncludes_def ss_lifting' localValid2sem)
  done
*)
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_7_1-e": 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(self::('a,'b::bot)Sequence) \<rbrakk> \<Longrightarrow>
   (OclIncludesAll self obj) \<tau> = (\<MathOclForAll> elem \<in> obj \<bullet> ((elem::('a,'b)VAL) \<MathOclIncludes> self)) \<tau>"
  apply(rule_tac X="self" in Sequence_sem_cases)
  apply(rule_tac X="obj" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(auto simp: OCL_Sequence.OclForAll_def OCL_Sequence.OclIncludesAll_def OclStrongEq_def 
                   OCL_Sequence.OclIncludes_def ss_lifting' localValid2sem)
  done
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.7.1-f}{excludesAll(c2 : Collection(T)) : Boolean}
Does self contain none of the elements of c2 ?
\begin{ocl}
post: result = c2->forAll(elem | self->excludes(elem))
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
text{* 
  \OCLglitch{Wrong definition of the correspondence of excludesAll and iterate}
  If @{text "self = \<MathOclUndefined>"} and @{text "obj = \<MathOclMtSet>"} then the 
  left-hand side of the equation evaluates to
  @{text "\<MathOclUndefined>"} whereas the right-hand side evaluates to 
  @{text "\<MathOclTrue>"}.
*}
lemma "req_set_11_7_1-f":
  "((self::('\<tau>,'\<alpha>::bot)Set) ->excludesAll(obj)) 
   = (obj ->forAll(elem | (self ->excludes((elem::('\<tau>,'\<alpha>::bot)VAL)))))"
  apply(rule ext)
  apply(simp add: OclExcludes_def OclExcludesAll_def OclForAll_def
                  lift0_def lift1_def lift2_def strictify_def DEF_def)
  apply(auto)
  oops
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_7_1-f": 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(self::('a,'b::bot)Bag) \<rbrakk> \<Longrightarrow>
   \<tau> \<MathOclValid> (OclExcludesAll self obj) \<MathOclStrongEq> (\<MathOclForAll> elem \<in> obj \<bullet> ((elem::('a,'b)VAL) \<MathOclExcludes> self))"
  oops
(* missing: cp of forall
  apply(rule_tac X="self" in Bag_sem_cases)
  apply(rule_tac X="obj" in Bag_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(auto simp: OclForAll_def OclExcludesAll_def OclStrongEq_def OclIncludes_def
                   OclNot_def ss_lifting' localValid2sem)
  done
*)
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_7_1-f": 
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(self::('a,'b::bot)Sequence) \<rbrakk> \<Longrightarrow>
   \<tau> \<MathOclValid> (OclExcludesAll self obj) \<MathOclStrongEq> (\<MathOclForAll> elem \<in> obj \<bullet> ((elem::('a,'b)VAL) \<MathOclExcludes> self))"
  apply(rule_tac X="self" in Sequence_sem_cases)
  apply(rule_tac X="obj" in Sequence_sem_cases)
  apply(simp_all add: localValidDefined2sem DEF_def)
  apply(auto simp: OCL_Sequence.OclForAll_def OCL_Sequence.OclExcludesAll_def OclStrongEq_def 
                   OCL_Sequence.OclIncludes_def ss_lifting' localValid2sem OclNot_def)
  done
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.7.1-g}{isEmpty() : Boolean}
Is self the empty collection?
\begin{ocl}
post: result = ( self->size() = 0 )
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
text{* Due to our extension to infinite sets, the usual connection between
       isEmpty and size holds only under a precondition *}
lemma REQ11_7_1_7: 
  "\<MathOclValid> \<MathOclIsDefined>(self ->size()) 
     \<Longrightarrow> self ->isEmpty() = (self::('\<tau>,'\<alpha>::bot)Set) ->size() \<MathOclStrictEq> 0"
  apply (rule ext)
  apply (drule_tac \<tau> = x in valid_elim)
  apply (subgoal_tac "finite (Lifting.drop (Rep_Set_0 (self x)))")
  prefer 2
  apply (erule finite_Sets)
  apply (simp add:  OclIsEmpty_def OclSize_def OclStrictEq_def 
    OCL_Integer.Zero_ocl_int_def)
  apply (simp only: lift0_def lift1_def lift2_def 
    strictify_def o_def not_def)
  apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def 
    split add: split_if up.split)
  apply (simp only: lift0_def lift1_def lift2_def 
    strictify_def o_def not_def OclIsDefined_def DEF_def OclTrue_def)
  apply auto
  done
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_7_1-g": 
  "(\<MathOclIsEmpty> self) = \<MathOclSizeLeft>(self::('a,'b::bot)Bag)\<MathOclSizeRight> \<MathOclStrictEq> 0"
  apply(rule Bag_sem_cases_ext, simp_all)
  apply(simp_all add: OCL_Bag.OclSize_def OclMtBag_def OclStrictEq_def 
                      Zero_ocl_int_def ss_lifting')
  done
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_7_1-g": 
  "(\<MathOclIsEmpty> self) = \<MathOclSizeLeft>(self::('a,'b::bot)Sequence)\<MathOclSizeRight> \<MathOclStrictEq> 0"
  by simp
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.7.1-h}{notEmpty() : Boolean}
Is self not the empty collection?
\begin{ocl}
post: result = ( self->size() <> 0 )
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
lemma "req_set_11_7_1-h": 
"OclNotEmpty(X::('\<tau>, '\<alpha>::bot) Set) =  \<MathOclNot> (OclIsEmpty(X))" 
apply (rule ext)
apply (simp add:  OclNotEmpty_def OclIsEmpty_def OclNot_def)
apply (simp only: lift1_def lift2_def 
                  strictify_def o_def not_def)
apply (simp_all (no_asm_use) add: UU_fun_def DEF_def strictify_def 
                             split add: split_if up.split)
done
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_7_1-h": 
  "(\<MathOclNotEmpty> self) =  (\<MathOclSizeLeft>(self::('a,'b::bot)Bag)\<MathOclSizeRight>) `<>` 0"
  by(simp add: "req_bag_11_7_1-g"[symmetric])
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_7_1-h": 
  "(\<MathOclNotEmpty> self) = (\<MathOclSizeLeft>(self::('a,'b::bot)Sequence)\<MathOclSizeRight>) `<>` 0"
  by(simp add: "req_seq_11_7_1-g"[symmetric])
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.7.1-i}{sum() : T}
The addition of all elements in self. Elements must be of a type supporting the + operation. The + operation must take one
parameter of type T and be both associative: (a+b)+c = a+(b+c), and commutative: a+b = b+a. Integer and Real fulfill this
condition.
\begin{ocl}
post: result = self->iterate( elem; acc : T = 0 | acc + elem )
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_7_1-i":
  "(\<MathOclSum> S) = ((S::('a,'b::{bot,plus,zero})Bag) \<MathOclIterate>(x;y=0 | (x + y)))"
  by(simp add: OCL_Bag.OclSum_def)
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_7_1-i":
  "(\<MathOclSum> S) = ((S::('a,Integer_0)Sequence) \<MathOclIterate>(x;y=0 | (x + y)))"
  by(simp add: OCL_Sequence.OclSum_def)
text{* Note: That this requirement is not fulfilled completely as it currently holds
  only for integers. But this will change in the future.
*}
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.7.1-j}{product(c2: Collection(T2)) : Set( Tuple( first: T, second: T2) )}
The cartesian product operation of self and c2.
\begin{ocl}
post: result = 
self->iterate(e1; acc: Set(Tuple(first: T, second: T2)) = Set{} |
 c2->iterate(e2; acc2: Set(Tuple(first: T, second: T2)) = acc |
             acc2->including (Tuple{first = e1, second = e2}) ) )
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_7_1-j":
  "(OclProduct (self::('a,'b::bot)Bag) (c2::('a,'c::bot)Bag)) =
   (self \<MathOclIterate>(e1; acc=OclMtSet | 
           (c2 \<MathOclIterate>(e2; acc2=acc | 
                 (acc2 \<MathOclIncluding> (OclTuple e1 e2))))))"
  oops
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_7_1-j":
  "(OclProduct (self::('a,'b::bot)Sequence) (c2::('a,'c::bot)Sequence)) =
   (self \<MathOclIterate>(e1; acc=OclMtSet | 
           (c2 \<MathOclIterate>(e2; acc2=acc | 
                 (acc2 \<MathOclIncluding> (OclTuple e1 e2))))))"
  oops
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

(**************************************************************************)
section {* Set *}
(**************************************************************************)

(**************************************************************************)
section {* OrderedSet *}
(**************************************************************************)

(**************************************************************************)
section {* Bag *}
(**************************************************************************)
text_raw {* 
\begin{oclrequirement}{11.7.7-b}{= (bag : Bag(T)) : Boolean}
True if self and bag contain the same elements, the same number of times.
\begin{ocl}
post: 
result=(self->forAll(elem | self->count(elem) = bag->count(elem)) 
       and
       bag->forAll(elem | bag->count(elem) = self->count(elem)))
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{union(bag : Bag(T)) : Bag(T)}
The union of self and bag.
\begin{ocl}
post: result->forAll( elem | 
      result->count(elem) = self->count(elem) + bag->count(elem))
post: self ->forAll( elem | 
      result->count(elem) = self->count(elem) + bag->count(elem))
post: bag ->forAll( elem | 
      result->count(elem) = self->count(elem) + bag->count(elem))
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{union(set : Set(T)) : Bag(T)}
The union of self and set.
\begin{ocl}
post: result->forAll(elem | 
      result->count(elem) = self->count(elem) + set->count(elem))
post: self ->forAll(elem | 
      result->count(elem) = self->count(elem) + set->count(elem))
post: set ->forAll(elem | 
      result->count(elem) = self->count(elem) + set->count(elem))
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{intersection(bag : Bag(T)) : Bag(T)}
The intersection of self and bag.
\begin{ocl}
post: result->forAll(elem |
  result->count(elem)=self->count(elem).min(bag->count(elem)))
post: self->forAll(elem |
  result->count(elem)=self->count(elem).min(bag->count(elem)))
post: bag->forAll(elem |
  result->count(elem)=self->count(elem).min(bag->count(elem)))
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{intersection(set : Set(T)) : Set(T)}
The intersection of self and set.
\begin{ocl}
post: result->forAll(elem|
  result->count(elem) = self->count(elem).min(set->count(elem)))
post: self ->forAll(elem|
  result->count(elem) = self->count(elem).min(set->count(elem)))
post: set ->forAll(elem|
  result->count(elem) = self->count(elem).min(set->count(elem)))
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{including(object : T) : Bag(T)}
The bag containing all elements of self plus object.
\begin{ocl}
post: result->forAll(elem |
     if elem = object then
        result->count(elem) = self->count(elem) + 1
     else
        result->count(elem) = self->count(elem)
     endif)
post: self->forAll(elem |
     if elem = object then
        result->count(elem) = self->count(elem) + 1
     else
        result->count(elem) = self->count(elem)
     endif)
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{excluding(object : T) : Bag(T)}
The bag containing all elements of self apart from all occurrences of object.
\begin{ocl}
post: result->forAll(elem |
     if elem = object then
        result->count(elem) = 0
     else
        result->count(elem) = self->count(elem)
     endif)
post: self->forAll(elem |
     if elem = object then
        result->count(elem) = 0
     else
        result->count(elem) = self->count(elem)
     endif)
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{count(object : T) : Integer}
The number of occurrences of object in self.
\end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{flatten() : Bag(T2)}
If the element type is not a collection type this result in the same bag. If the element type is a collection type, the r
is the bag containing all the elements of all the elements of self.
\begin{ocl}
post: 
  result = if self.type.elementType.oclIsKindOf(CollectionType) 
      then
        self->iterate(c; acc : Bag() = Bag{} |
            acc->union(c->asBag() ) )
      else
        self
      endif
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{asBag() : Bag(T)}
A Bag identical to self. This operation exists for convenience reasons.
\begin{ocl}
post: result = self
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{asSequence() : Sequence(T)}
A Sequence that contains all the elements from self, in undefined order.
\begin{ocl}
post: 
  result->forAll(elem | self->count(elem)=result->count(elem))
post: 
  self ->forAll(elem | self->count(elem)=result->count(elem))
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{asSet() : Set(T)}
The Set containing all the elements from self, with duplicates removed.
\begin{ocl}
post: result->forAll(elem | self ->includes(elem))
post: self ->forAll(elem | result->includes(elem))
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.7-b}{asOrderedSet() : OrderedSet(T)}
An OrderedSet that contains all the elements from self, in undefined order, with duplicates removed.
\begin{ocl}
post: result->forAll(elem | self ->includes(elem))
post: self ->forAll(elem | result->includes(elem))
post: self ->forAll(elem | result->count(elem) = 1)
\end{ocl}*}
text_raw {* \end{oclrequirement} *}

(**************************************************************************)
section {* Sequence *}
(**************************************************************************)

text_raw {* 
\begin{oclrequirement}{11.7.5-a}{count(object : T) : Integer}
The number of occurrences of object in self.
\end{oclrequirement}
*}

text_raw {* 
\begin{oclrequirement}{11.7.5-b}{= (s : Sequence(T)) : Boolean}
True if self contains the same elements as s in the same order.
\begin{ocl}
post:
  result=Sequence{1..self->size()}->forAll(index : Integer |
                self->at(index) = s->at(index))
                and
                self->size() = s->size()
\end{ocl}*}
lemma "req_11_7_5-b":
  "((S1::('a,'b::bot)Sequence) \<MathOclStrictEq> S2) =
   (\<MathOclForAll> i \<in> ((OclCollectionRange 1 \<MathOclSizeLeft>S1\<MathOclSizeRight>)::('a,Integer_0)Sequence) \<bullet> ((OclAt S1 i) \<MathOclStrictEq> (OclAt S2 i)) ) \<MathOclAnd> 
   (\<MathOclSizeLeft>S1\<MathOclSizeRight> \<MathOclStrictEq> \<MathOclSizeLeft>S2\<MathOclSizeRight>)"
  oops
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-c}{union (s : Sequence(T)) : Sequence(T)}
The sequence consisting of all elements in self, followed by all elements in s.
\begin{ocl}
post: result->size() = self->size() + s->size()
post: Sequence{1..self->size()}->forAll(index : Integer |
                          self->at(index) = result->at(index))
post: Sequence{1..s->size()}->forAll(index : Integer |
\end{ocl}*}
lemma "req_11_7_5-c-1":
  "\<MathOclSizeLeft>((self::('a,'b::bot)Sequence) \<MathOclUnion> s )\<MathOclSizeRight> = (\<MathOclSizeLeft>self\<MathOclSizeRight> + \<MathOclSizeLeft>s\<MathOclSizeRight>)"
  by(simp)
lemma "req_11_7_5-c-2":
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(self::('a,'b::bot)Sequence) \<rbrakk> \<Longrightarrow>
     \<tau> \<MathOclValid> \<MathOclForAll> index  \<in> ((OclCollectionRange 1 \<MathOclSizeLeft>self\<MathOclSizeRight>)::('a,Integer_0)Sequence) \<bullet>  
      ((OclAt (self \<MathOclUnion> s ) index) \<MathOclStrictEq> (OclAt self index))"
  apply(rule OCL_Sequence.ForAllI)
  apply(simp add: OCL_Sequence.OCL_is_defopt_OclCollectionRange)
  apply(simp add: OCL_Sequence.includes_of_collectionRange)
  oops
lemma "req_11_7_5-c-3":
  "\<MathOclValid> \<MathOclForAll> index  \<in> ((OclCollectionRange 1 \<MathOclSizeLeft>s\<MathOclSizeRight>)::('a,Integer_0)Sequence) \<bullet>  
      ((OclAt s index) \<MathOclStrictEq> (OclAt (self \<MathOclUnion> s ) (index + \<MathOclSizeLeft>self\<MathOclSizeRight>)))"
  oops
text_raw {* \end{oclrequirement} *}



text_raw {* 
\begin{oclrequirement}{11.7.5-d}{flatten() : Sequence(T2)}
If the element type is not a collection type this result in the same self. If the element type is a collection type, the result
is the seuqnce containing all the elements of all the elements of self. The order of the elements is partial.
\begin{ocl}
post:
  result = if self.type.elementType.oclIsKindOf(CollectionType) 
       then
         self->iterate(c; acc : Sequence() = Sequence{} |
            acc->union(c->asSequence() ) )
       else
         self
       endif
\end{ocl}*}
text{* The postcondition in the standard is not typeable because it must have these two
types at the same time which is not possible:
\begin{verbatim}
(('a,'b::bot Sequence_0)Sequence) \<Rightarrow> ('a,'b)Sequence
('a,'b::bot)Sequence \<Rightarrow> ('a,'b)Sequence
\end{verbatim}
therefore we split the requirement into two parts:
\begin{enumerate}
\item characterises the case of a flatten of a sequence of sequences
\item characterises the case of a flatten of a sequence of a not collection type. This 
  follows by definition of flatten on sequences of non collection types
\end{enumerate}
*}
lemma "req_11_7_5-d-1":
  "(\<MathOclFlattenLeft>(self::('a,'b::bot Sequence_0)Sequence)\<MathOclFlattenRight>::('a,'b)Sequence) = 
   (self \<MathOclIterate>(x;y=\<MathOclMtSequence> | (y \<MathOclUnion> x )))"
  by(simp add: OCL_Sequence.flatten_by_iterate)
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-e}{append (object: T) : Sequence(T)}
The sequence of elements, consisting of all elements of self, followed by object.
\begin{ocl}
post: result->size() = self->size() + 1
post: result->at(result->size() ) = object
post: Sequence{1..self->size() }->forAll(index : Integer |
       result->at(index) = self ->at(index))
\end{ocl}*}
lemma "req_11_7_5-e-1":
  "\<tau> \<MathOclValid> (\<MathOclIsDefined> (obj::('a,'b::bot)VAL)) \<Longrightarrow> 
   (\<MathOclSizeLeft>OclAppend (self::('a,'b)Sequence) obj\<MathOclSizeRight> \<tau>) = ((\<MathOclSizeLeft>self\<MathOclSizeRight> + 1) \<tau>)"
  by(simp)
lemma "req_11_7_5-e-2":
  "\<tau> \<MathOclValid> (\<MathOclIsDefined> (self::('a,'b::bot)Sequence)) \<Longrightarrow> 
   (OclAt (OclAppend self (obj::('a,'b)VAL)) \<MathOclSizeLeft>OclAppend self obj\<MathOclSizeRight> \<tau>) = (obj \<tau>)"
  apply(case_tac "\<tau> \<MathOclValid> \<MathOclIsUndefined> obj")
  apply(ocl_hypsubst, simp_all)
  apply(rule OCL_Sequence.at_last_of_including)
  apply(rule subst, rule sym)
  apply(rule_tac A="\<MathOclSizeLeft>self \<MathOclIncluding> obj \<MathOclSizeRight>" in cp_charn)
  apply(rule OCL_Sequence.size_of_including)
  apply(simp_all)
  oops
lemma "req_11_7_5-e-3":
  "\<MathOclValid> \<MathOclForAll> index  \<in> ((OclCollectionRange 1 \<MathOclSizeLeft>(self::('a,'b::bot)Sequence)\<MathOclSizeRight>)::('a,Integer_0)Sequence) \<bullet>  
      ((OclAt self index) \<MathOclStrictEq> (OclAt (OclAppend self (obj::('a,'b)VAL) ) index))"
  oops
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-f}{prepend(object : T) : Sequence(T)}
The sequence consisting of object, followed by all elements in self.
\begin{ocl}
post: result->size = self->size() + 1
post: result->at(1) = object
post: Sequence{1..self->size()}->forAll(index : Integer |
      self->at(index) = result->at(index + 1))
\end{ocl}*}
lemma "req_11_7_5-f-1":
  "\<tau> \<MathOclValid> (\<MathOclIsDefined> (obj::('a,'b::bot)VAL)) \<Longrightarrow> 
   (\<MathOclSizeLeft>OclPrepend (self::('a,'b)Sequence)  obj\<MathOclSizeRight> \<tau>) = ((\<MathOclSizeLeft>self\<MathOclSizeRight> + 1) \<tau>)"
  apply(simp)
  apply(rule trans)
  apply(rule_tac A="\<MathOclSizeLeft>?S \<MathOclIncluding> obj \<MathOclSizeRight>" in cp_charn)
  apply(rule OCL_Sequence.size_of_including)
  apply(simp_all add: localValidDefined2sem)
  apply(simp add: localValid2sem OCL_Sequence.OclSize_def OclStrictEq_def
                  One_ocl_int_def Zero_ocl_int_def plus_def ss_lifting')
  oops
lemma "req_11_7_5-f-2":
  "\<tau> \<MathOclValid> (\<MathOclIsDefined> (self::('a,'b::bot)Sequence)) \<Longrightarrow> 
   (OclAt (OclPrepend self (obj::('a,'b)VAL)) 1 \<tau>) = (obj \<tau>)"
  by(simp only: at_first_of_prepend)
lemma "req_11_7_5-f-3":
  "\<MathOclValid> \<MathOclForAll> index  \<in> ((OclCollectionRange 1 \<MathOclSizeLeft>(self::('a,'b::bot)Sequence)\<MathOclSizeRight>)::('a,Integer_0)Sequence) \<bullet>  
      ((OclAt self index) \<MathOclStrictEq> (OclAt (OclPrepend self (obj::('a,'b)VAL) ) (index+1)))"
  oops
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-g}{insertAt(index : Integer, object : T) : Sequence(T)}
The sequence consisting of self with object inserted at position index.
\begin{ocl}
post: result->size = self->size() + 1
post: result->at(index) = object
post: Sequence{1..(index - 1)}->forAll(i : Integer |
      self->at(i) = result->at(i))
post: Sequence{(index + 1)..self->size()}->forAll(i : Integer |
      self->at(i) = result->at(i + 1))
\end{ocl}*}
text{* \OCLglitch{no preconditions about range of index. But they are
  surely needed because of the correspondence to \emph{at()}} 
*}
lemma "req_11_7_5-g-1":
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> (obj::('a,'b::bot)VAL); \<tau> \<MathOclValid> 1 \<MathOclLe> index; \<tau> \<MathOclValid> index \<MathOclLe> (\<MathOclSizeLeft>self\<MathOclSizeRight> + 1) \<rbrakk> \<Longrightarrow>
   \<MathOclSizeLeft>OclInsertAt self index obj\<MathOclSizeRight> \<tau> = (\<MathOclSizeLeft>(self::('a,'b)Sequence)\<MathOclSizeRight> + 1) \<tau>"
  by(simp add: size_of_insertAt)
lemma "req_11_7_5-g-2":
  "\<lbrakk> \<tau> \<MathOclValid> 1 \<MathOclLe> index; \<tau> \<MathOclValid> index \<MathOclLe> (\<MathOclSizeLeft>self\<MathOclSizeRight> + 1) \<rbrakk> \<Longrightarrow>
   OclAt (OclInsertAt (self::('a,'b::bot)Sequence) index obj) index \<tau>  = (obj::('a,'b)VAL) \<tau>"
  by(simp add: at_of_insertAt)
lemma "req_11_7_5-g-3":
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> obj;  \<tau> \<MathOclValid> 1 \<MathOclLe> i; \<tau> \<MathOclValid> index \<MathOclLe> (\<MathOclSizeLeft>(self::('a,'b::bot)Sequence)\<MathOclSizeRight> + 1) \<rbrakk> \<Longrightarrow>
   \<tau> \<MathOclValid> \<MathOclForAll> i  \<in> ((OclCollectionRange 1 (index - 1))::('a,Integer_0)Sequence) \<bullet>  
      ((OclAt self i) \<MathOclStrictEq> (OclAt (OclInsertAt self index (obj::('a,'b)VAL)) i))"
  oops
text{* \OCLglitch{the valid range must be from \inlineocl{index} to 
  \inlineocl{self->size()}} 
*}
lemma "req_11_7_5-g-4":
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined> obj;  \<tau> \<MathOclValid> 1 \<MathOclLe> i; \<tau> \<MathOclValid> index \<MathOclLe> (\<MathOclSizeLeft>(self::('a,'b::bot)Sequence)\<MathOclSizeRight> + 1) \<rbrakk> \<Longrightarrow>
   \<tau> \<MathOclValid> \<MathOclForAll> i  \<in> ((OclCollectionRange index \<MathOclSizeLeft>(self::('a,'b::bot)Sequence)\<MathOclSizeRight>)::('a,Integer_0)Sequence) \<bullet>  
      ((OclAt self i) \<MathOclStrictEq> (OclAt (OclInsertAt self index (obj::('a,'b)VAL)) (i+1)))"
  oops
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-h}{subSequence(lower : Integer, upper : Integer) : Sequence(T)}
The sub-sequence of self starting at number lower, up to and including element number upper.
\begin{ocl}
pre : 1 <= lower
pre : lower <= upper
pre : upper <= self->size()
post: result->size() = upper -lower + 1
post: Sequence{lower..upper}->forAll( index |
       result->at(index - lower + 1) =
                  self->at(index))
\end{ocl}*}
lemma "req_11_7_5-h-2":
  "\<lbrakk> \<tau> \<MathOclValid> 1 \<MathOclLe> lower; \<tau> \<MathOclValid> lower \<MathOclLe> upper - 1; \<tau> \<MathOclValid> upper \<MathOclLe> \<MathOclSizeLeft>self\<MathOclSizeRight> \<rbrakk> \<Longrightarrow> 
   \<MathOclSizeLeft>(OclSubSequence (self::('a,'b::bot)Sequence) lower upper)\<MathOclSizeRight>  \<tau> = (upper - lower + 1) \<tau>"
  by(simp add: size_of_subsequence)
text {* \todo{correct core lemma assumption a < b to a <= b} *}
lemma "req_11_7_5-h-2":
  "\<lbrakk> \<tau> \<MathOclValid> 1 \<MathOclLe> lower; \<tau> \<MathOclValid> lower \<MathOclLe> upper - 1; \<tau> \<MathOclValid> upper \<MathOclLe> \<MathOclSizeLeft>self\<MathOclSizeRight> \<rbrakk> \<Longrightarrow>
   \<tau> \<MathOclValid> \<MathOclForAll> index  \<in> ((OclCollectionRange lower upper)::('a,Integer_0)Sequence) \<bullet>  
          (OclAt (OclSubSequence self lower upper) (index - lower + 1) \<MathOclStrictEq>
           OclAt (self::('a,'b::bot)Sequence) index)"
  oops
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-i}{at(i : Integer) : T}
The i-th element of sequence.
\begin{ocl}
pre : i >= 1 and i <= self->size()
\end{ocl}*}
text{* \todo{How shall this requirement be formulated? Its only a precondition.} *}
lemma "req_11_7_5-i":
  "\<lbrakk> \<tau> \<MathOclValid> 1 \<MathOclLe> i; \<tau> \<MathOclValid> i \<MathOclLe> \<MathOclSizeLeft>S\<MathOclSizeRight> \<rbrakk> \<Longrightarrow> 
   (\<tau> \<MathOclValid> \<MathOclIsDefined>((OclAt (S::('a,'b::bot)Sequence) i)::('a,'b)VAL))"
  apply(rotate_tac 1, frule isDefined_if_valid)
  apply(simp only: OCL_is_defopt_OclLe OCL_Sequence.OCL_is_defopt_OclSize OCL_is_defopt_OclAt)
  oops (* works in interactive mode \<dots> done*)
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-j}{indexOf(obj : T) : Integer}
The index of object obj in the sequence.
\begin{ocl}
pre : self->includes(obj)
post : self->at(i) = obj
\end{ocl}*}
lemma "req_11_7_5-j":
  "\<lbrakk> \<tau> \<MathOclValid> (obj::('a,'b)VAL) \<MathOclIncludes> (self::('a,'b::bot)Sequence) \<rbrakk> \<Longrightarrow>
   OclAt self (OclIndexOf self obj) \<tau> = obj \<tau>"
  by(simp add: at_indexOf)
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-k}{first() : T}
The first element in self.
\begin{ocl}
post: result = self->at(1)
\end{ocl}*}
lemma "req_11_7_5-k":
  "OclFirst (self::('a,'b::bot)Sequence) = ((OclAt self 1)::('a,'b)VAL)"
  by(rule first_at_UC)
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-l}{last() : T}
The last element in self.
\begin{ocl}
post: result = self->at(self->size() )
\end{ocl}*}
lemma "req_11_7_5-l":
  "OclLast (self::('a,'b::bot)Sequence) = ((OclAt self \<MathOclSizeLeft>self\<MathOclSizeRight>)::('a,'b)VAL)"
  by(rule last_at_UC)
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-m}{including(object : T) : Sequence(T)}
The sequence containing all elements of self plus object added as the last element.
\begin{ocl}
post: result = self.append(object)
\end{ocl}*}
lemma "req_11_7_5-m":
  "self \<MathOclIncluding> (obj::('a,'b)VAL) = OclAppend (self::('a,'b::bot)Sequence) obj"
  by(rule append_including_UC[symmetric])
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-n}{excluding(object : T) : Sequence(T)}
The sequence containing all elements of self apart from all occurrences of object.
The order of the remaining elements is not changed.
\begin{ocl}
post:result->includes(object) = false
post: result->size() = self->size() - self->count(object)
post: result = self->iterate(elem; acc : Sequence(T)
   = Sequence{}|
      if elem = object then acc else acc->append(elem) endif )
\end{ocl}*}
lemma "req_11_7_5-n-1":
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(obj::('a,'b)VAL);  \<tau> \<MathOclValid> \<MathOclIsDefined>(self::('a,'b::bot)Sequence) \<rbrakk> \<Longrightarrow>
   (obj \<MathOclIncludes> (self \<MathOclExcluding> obj)) \<tau> = \<MathOclFalse> \<tau>"
  by(simp add: OCL_Sequence.includes_of_excluding)
lemma "req_11_7_5-n-2":
  "\<MathOclSizeLeft>(self::('a,'b::bot)Sequence) \<MathOclExcluding> (obj::('a,'b)VAL)\<MathOclSizeRight> =
   \<MathOclSizeLeft>self\<MathOclSizeRight> - OclCount self obj"
  by(simp add: OCL_Sequence.size_of_excluding)
lemma "req_11_7_5-n-3":
  "\<lbrakk> \<tau> \<MathOclValid> \<MathOclIsDefined>(obj::('a,'b)VAL) \<rbrakk> \<Longrightarrow>
   ((self::('a,'b::bot)Sequence) \<MathOclExcluding> (obj::('a,'b)VAL)) \<tau> = 
   (self \<MathOclIterate>(elem; acc=\<MathOclMtSequence> | (\<MathOclIf> (elem \<MathOclStrictEq> obj) \<MathOclThen> acc \<MathOclElse> (OclAppend acc elem )\<MathOclEndif>))) \<tau>"
  by(simp add: OCL_Sequence.excluding_by_iterate localValid2sem 
               OCL_Sequence.OclIterate_def OclIf_def ss_lifting')
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-o}{asBag() : Bag(T)}
The Bag containing all the elements from self, including duplicates.
\begin{ocl}
post: 
  result->forAll(elem | self->count(elem)=result->count(elem))
post: 
  self->forAll(elem | self->count(elem)=result->count(elem))
\end{ocl}*}
lemma "req_11_7_5-o-1":
  "\<MathOclValid> \<MathOclForAll> elem  \<in> ((\<MathOclAsBag> (self::('a,'b::bot)Sequence))::('a,'b::bot)Bag) \<bullet>  
   ((self \<MathOclCount> elem) \<MathOclStrictEq> (((\<MathOclAsBag> self)::('a,'b::bot)Bag) \<MathOclCount> elem))"
  oops
lemma "req_11_7_5-o-2":
  "\<MathOclValid> \<MathOclForAll> elem  \<in> (self::('a,'b::bot)Sequence) \<bullet>  
   ((self \<MathOclCount> elem) \<MathOclStrictEq> (((\<MathOclAsBag> self)::('a,'b::bot)Bag) \<MathOclCount> elem))"
  oops
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-p}{asSequence() : Sequence(T)}
The Sequence identical to the object itself. This operation exists for convenience reasons.
\begin{ocl}
post: result = self
\end{ocl}*}
lemma "req_11_7_5-p":
  "((\<MathOclAsSequence> (self::('a,'b::bot)Sequence))::('a,'b::bot)Sequence) = self"
  by(simp)
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-q}{asSet() : Set(T)}
The Set containing all the elements from self, with duplicated removed.
\begin{ocl}
post: result->forAll(elem | self ->includes(elem))
post: self ->forAll(elem | result->includes(elem))
\end{ocl}*}
lemma "req_11_7_5-q-1":
  "\<MathOclValid> \<MathOclForAll> elem  \<in> ((\<MathOclAsSet> (self::('a,'b::bot)Sequence))::('a,'b Set_0)VAL) \<bullet> ((elem::('a,'b)VAL) \<MathOclIncludes> self)"
  oops
lemma "req_11_7_5-q-2":
  "\<MathOclValid> \<MathOclForAll> elem  \<in> (self::('a,'b::bot)Sequence) \<bullet> ((elem::('a,'b)VAL) \<MathOclIncludes>((\<MathOclAsSet> self)::('a,'b Set_0)VAL))"
  oops
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.7.5-r}{asOrderedSet() : OrderedSet(T)}
An OrderedSet that contains all the elements from self, in the same order, with duplicates removed.
\begin{ocl}
post: result->forAll(elem | self ->includes(elem))
post: self ->forAll(elem | result->includes(elem))
post: self ->forAll(elem | result->count(elem) = 1)
post: self ->forAll(elem1, elem2 |
             self->indexOf(elem1) < self->indexOf(elem2)
      implies result->indexOf(elem1) < result->indexOf(elem2) )
\end{ocl}*}
lemma "req_11_7_5-q-1":
  "\<MathOclValid> \<MathOclForAll> elem  \<in> ((\<MathOclAsOrderedSet> (self::('a,'b::bot)Sequence))::('a,'b OrderedSet_0)VAL) \<bullet> ((elem::('a,'b)VAL) \<MathOclIncludes> self)"
  oops
lemma "req_11_7_5-q-2":
  "\<MathOclValid> \<MathOclForAll> elem  \<in> (self::('a,'b::bot)Sequence) \<bullet> ((elem::('a,'b)VAL) \<MathOclIncludes>((\<MathOclAsOrderedSet> self)::('a,'b::bot OrderedSet_0)VAL))"
  oops
text_raw {* \end{oclrequirement} *}

end
