(* ****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * OCL_Iterator_requirements.thy -
 * Copyright (C) 2003-2006 Achim D. Brucker <brucker@inf.ethz.ch>   
 *                         Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNES 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ***************************************************************************** *)

header {* \OCL Iterator requirements *}
theory OCL_Iterator_requirements
imports 
  "$HOLOCL_HOME/src/OCL_Library"
begin
 text_raw{* \label{thy:OCL_Iterator_requirements} *}

text {*
  This theory contains the requirements for iterators defined in 
  the \OCL standard library~\cite[Chapter 11.9]{omg:ocl:2003}. 
*}

(**************************************************************************)
section {* Collection *}
(**************************************************************************)
text_raw {* 
\begin{oclrequirement}{11.9.1-a}{exists}
Results in true if body evaluates to true for at least one element in the source collection.
\begin{ocl}
source->exists(iterators | body) =
         source->iterate(iterators; result : Boolean = false 
                         | result or body)
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_9_1-a":
  "(\<MathOclExists> x \<in> (source::('a,'b::bot)Bag) \<bullet> (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x) = 
   (source \<MathOclIterate>(x;y= \<MathOclFalse> | (P x) \<MathOclOr> y))" 
  by(rule OCL_Bag.exists_by_iterate)
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_9_1-a":
  "(\<MathOclExists> x \<in> (source::('a,'b::bot)Sequence) \<bullet> (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x) = 
   (source \<MathOclIterate>(x;y= \<MathOclFalse> | (P x) \<MathOclOr> y))" 
  by(rule OCL_Sequence.exists_by_iterate)
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.9.1-b}{forAll} 
Results in true if the body expression evaluates to true for each element in the 
source collection; otherwise, result is false
\begin{ocl}
source->forAll(iterators | body ) =
          source->iterate(iterators; result : Boolean = true 
                          | result and body)
\end{ocl}
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma "req_bag_11_9_1-b":
  "(\<MathOclForAll> x \<in> (source::('a,'b::bot)Bag) \<bullet> (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x) = 
   (source \<MathOclIterate>(x;y= \<MathOclTrue> | (P x) \<MathOclAnd> y))"
  by(rule OCL_Bag.forall_by_iterate)
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_9_1-b":
  "(\<MathOclForAll> x \<in> (source::('a,'b::bot)Sequence) \<bullet> (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x) = 
   (source \<MathOclIterate>(x;y= \<MathOclTrue> | (P x) \<MathOclAnd> y))"
  by(rule OCL_Sequence.forall_by_iterate)
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.9.1-c}{isUnique} 
Results in true if body evaluates to a different value for each element in 
the source collection; otherwise, result is false.
\begin{ocl}
source->isUnique (iterators | body) = 
source->collect(iterators 
             | Tuple{iter = Tuple{iterators}, value = body})
   ->forAll (x, y | (x.iter <> y.iter) 
                    implies (x.value <> y.value))
\end{ocl}
isUnique may have at most one iterator variable.
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.9.1-d}{any} 
Returns any element in the source collection for which body 
evaluates to true. If there is more than one element for which body is true, one 
of them is returned. There must be at least one element fulfilling body, otherwise the result of this
IteratorExp is OclUndefined.
\begin{ocl}
source->any(iterator | body) =
    source->select(iterator | body)->asSequence()->first()
\end{ocl}
any may have at most one iterator variable.
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma"req_bag_11_9_1-d":
  "OCL_Collection.OclAny (Source::('a,'b::bot)Bag) (P::(('a,'b::bot)VAL \<Rightarrow> 'a Boolean)) =
   OclFirst ((\<MathOclAsSequence> (\<MathOclSelectLeft>x : source | (P x)\<MathOclSelectRight>)::('a,'b::bot)Sequence))"
  oops
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_9_1-d":
  "OCL_Collection.OclAny (source::('a,'b::bot)Sequence) (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) =
   OclFirst ((\<MathOclAsSequence> (\<MathOclSelectLeft>x : source | (P x)\<MathOclSelectRight>)::('a,'b::bot)Sequence))"
  by(rule OCL_Sequence.any_by_first_of_select)
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.9.1-e}{one} 
Results in true if there is exactly one element in the source 
collection for which body is true.
\begin{ocl}
source->one(iterator | body) =
   source->select(iterator | body)->size() = 1
\end{ocl}
one may have at most one iterator variable.
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma"req_bag_11_9_1-e":
  "(OclOne (source::('a,'b::bot)Bag) (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean)) = 
   (\<MathOclSizeLeft>(\<MathOclSelectLeft>x : source | (P x)\<MathOclSelectRight>)\<MathOclSizeRight> \<MathOclStrictEq> 1)"
  by(simp add: OCL_Bag.OclOne_def)
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_9_1-e":
  "(OclOne (source::('a,'b::bot)Sequence) (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean)) = 
   (\<MathOclSizeLeft>(\<MathOclSelectLeft>x : source | (P x)\<MathOclSelectRight>)\<MathOclSizeRight> \<MathOclStrictEq> 1)"
  by(rule OCL_Sequence.one_by_size_of_select)
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}

text_raw {* 
\begin{oclrequirement}{11.9.1-f}{collect} 
The Collection of elements which results from applying body to every member of the source set. The result is flattened.
Notice that this is based on collectNested, which can be of different type depending on the type of source. collectNested
is defined individually for each subclass of CollectionType.
\begin{ocl}
source->collect (iterators | body) 
     = source->collectNested (iterators | body)->flatten()
\end{ocl}
collect may have at most one iterator variable.
\begin{description} 
\item[Set]\mbox{}\\
*}
text_raw{* 
  \item[OrderedSet]\mbox{}\\
*}
text_raw{* 
  \item[Bag]\mbox{}\\
*}
lemma"req_bag_11_7_4-f":
   "OclCollect S P \<equiv> \<MathOclFlattenLeft>(OclCollectNested S P)\<MathOclFlattenRight>"
  oops
text_raw{* 
  \item[Sequence]\mbox{}\\
*}
lemma "req_seq_11_9_1-f":
   "OclCollect S P \<equiv> \<MathOclFlattenLeft>(OclCollectNested S P)\<MathOclFlattenRight>"
  oops
text_raw {* 
  \end{description} 
  \end{oclrequirement} 
*}


(**************************************************************************)
section {* Set *}
(**************************************************************************)

(**************************************************************************)
section {* Bag *}
(**************************************************************************)

(**************************************************************************)
section {* Sequence *}
(**************************************************************************)

text_raw {* 
\begin{oclrequirement}{11.9.4-a}{select(expression : OclExpression) : Sequence(T)}
The subsequence of the source sequence for which body is true.
\begin{ocl}
source->select(iterator | body) =
  source->iterate(iterator; result : Sequence(T)=Sequence{} |
            if body then result->including(iterator)
                 else result
            endif)
\end{ocl}
select may have at most one iterator variable.
*}
lemma "req_11_9_4-a":
  "(\<MathOclSelectLeft>x : (source::('a,'b::bot)Sequence) | ((P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x)\<MathOclSelectRight>) = 
   (source \<MathOclIterate>(iterator; result = \<MathOclMtSequence> | 
                             \<MathOclIf> (P iterator)\<MathOclThen> (OclIncluding result iterator) \<MathOclElse> result \<MathOclEndif>))"
  by(rule OCL_Sequence.select_by_iterate)
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.9.4-b}{reject}
The subsequence of the source sequence for which body is false.
\begin{ocl}
source->reject(iterator | body) =
     source->select(iterator | not body)
\end{ocl}
reject may have at most one iterator variable.
*}
lemma "req_11_9_4-b":
  "OclReject source P = (\<MathOclSelectLeft>x : (source::('a,'b::bot)Sequence) | (\<MathOclNot> (P::('a,'b::bot)VAL \<Rightarrow> 'a Boolean) x)\<MathOclSelectRight>)"
  by(rule OCL_Sequence.reject_by_select)
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.9.4-c}{collectNested}
The Sequence of elements which results from applying body to every member of the source sequence.
\begin{ocl}
source->collect(iterators | body) =
     source->iterate(iterators; result : Sequence(body.type) 
               = Sequence{} |
               result->append(body ) )
\end{ocl}
collectNested may have at most one iterator variable.
*}
lemma "req_11_9_4-c":
  "((OclCollectNested (source::('a,'b::bot)Sequence) (P::('a,'b)VAL \<Rightarrow> ('a,'c)VAL))::('a,'c::bot)Sequence) =
   (source \<MathOclIterate>(iterators; result=\<MathOclMtSequence> |  result \<MathOclIncluding> (P iterators)))"
  by(rule OCL_Sequence.collectNested_by_iterate)
text_raw {* \end{oclrequirement} *}

text_raw {* 
\begin{oclrequirement}{11.9.4-d}{sortedBy}
Results in the Sequence containing all elements of the source collection. The element for which body has the lowest value
comes first, and so on. The type of the body expression must have the < operation defined. The < operation must return a
Boolean value and must be transitive i.e. if a < b and b < c then a < c.
\begin{ocl}
source->sortedBy(iterator | body) =
  iterate( iterator ; result : Sequence(T) : Sequence {} |
      if result->isEmpty() then
        result.append(iterator)
      else
        let position : Integer = result->indexOf (
             result->select (item 
               | body (item) > body (iterator)) ->first() )
        in
         result.insertAt(position, iterator)
      endif
\end{ocl}
sortedBy may have at most one iterator variable.
*}
text_raw {* \end{oclrequirement} *}
end
