(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * mdr_encoder.sml - generic mdr import interface for HOL-OCL
 * Copyright (C) 2001-2005  Achim D. Brucker <brucker@inf.ethz.ch>   
 *                          Burkhart Wolff   <bwolff@inf.ethz.ch>    
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)

(**
 *  The UML Repository Encoder.
 * 
 *)
signature REP_ENCODER =
sig
    type mdr   = {  theory        : theory,
                    universe      : typ,
	      	    classifiers   : Rep_Core.Classifier list
	         }
(**
 * the empty UML repository
 *)
val empty           : mdr
(**
 * add classifiers to the given theory. 
 * @return updated repository structure
 *)	   
val add_classifiers : mdr -> Rep_Core.Classifier list -> theory -> mdr
end
   
(** 
 *
 *)
structure rep_encoder : REP_ENCODER  = 
struct

open rep_encoder_common
open Rep_Core 

(**************************************************************************************)
type mdr   = {  theory: Theory.theory,
                universe     : typ,
		classifiers   : Rep_Core.Classifier list
	     }
(**
 * the default (empty) repository, only containing OclAny.
 *)		       	    
val empty = {
    theory = theory "OCL",
    universe    = (UT(TFree("'alpha_OclAny", typeS))), 
    classifiers = []
}:mdr
    
fun add_classifiers (cd:mdr) cl thy =
    let
	val _ = trace 1 ""
	val _ = trace 1 "Starting Rep Encoder"
	val _ = trace 1 "===================="
        val base_universe   = #universe cd
        val base_cl         = #classifiers cd
	val _               = map (fn c => trace 9 ((string_of_path (name_of c))
						    ^" extends "
						    ^(string_of_path (parent_name_of c)))) cl
	val thy_name        = Context.theory_name thy
	val cl              = map (fn c => update_thyname thy_name c) cl 
	val cl              = map (fn c => Rep_Core.normalize_init c) cl 
	val cl              = topsort_cl cl
        val classifiers     = topsort_cl (base_cl@cl)
					  


	val _ = isa_packages_quiet()
	
	val _ = if !quick_and_dirty then trace 1 "  Mode: quick and dirty (interactive)" 
					else ()

	val timer =  Timer.startRealTimer ()
	val startTime = Timer.checkRealTimer timer 

	(*************************************)	  
       	(* zero level: build universe        *)
	val _ = trace 1 ""
	val _ = trace 1 "  Encoding Level 0:"
	val _ = trace 1 "  ================="
	val (thy,universe) = rep_encoder_level0.build_universe base_universe cl thy 
       	(*             add definitions       *)
	val thy = rep_encoder_level0.add_level0 universe cl thy 
	val level0Time = Timer.checkRealTimer timer
	(*************************************)	  
       	(* first level:                      *)
	val _ = trace 1 ""
	val _ = trace 1 "  Encoding Level 1:"
	val _ = trace 1 "  ================="
	val thy = rep_encoder_level1.add_level1 universe cl thy 
	val level1Time = Timer.checkRealTimer timer
	(*************************************)	  
	(* second level:                     *)
	val _ = trace 1 ""
	val _ = trace 1 "  Encoding Level 2:"
	val _ = trace 1 "  ================="
	val thy = rep_encoder_level2.add_level2 universe cl thy 
	val level2Time = Timer.checkRealTimer timer

	val l0Duration = Time.-(level0Time,startTime)
	val l1Duration = Time.-(level1Time,level0Time)
	val l2Duration = Time.-(level2Time,level1Time)
		 
	val _ = trace 1 ""
	val _ = trace 1 "  Used Times:"
	val _ = trace 1 "  ==========="
	val _ = trace 1 ("    Encoding Level 0: "^(Time.toString(l0Duration)))
	val _ = trace 1 ("    Encoding Level 1: "^(Time.toString(l1Duration)))
	val _ = trace 1 ("    Encoding Level 2: "^(Time.toString(l2Duration)))

	val _ = context thy
		
    in
	{ theory     = thy,
	 universe    = universe,  
         classifiers = classifiers
	}:mdr
    end 
end;    
	  
 
