(*****************************************************************************
 *          HOL-OCL - a shallow embedding of OCL in Isabelle/HOL              
 *                                                                            
 * mdr_encoder_level1.sml - generic mdr import interface for HOL-OCL
 * Copyright (C) 2001-2005  Achim D. Brucker <brucker@inf.ethz.ch>   
 *                                                                            
 * This file is part of HOL-OCL.                                              
 *                                                                            
 * HOL-OCL is free software; you can redistribute it and/or modify it under   
 * the terms of the GNU General Public License as published by the Free       
 * Software Foundation; either version 2 of the License, or (at your option)  
 * any later version.                                                         
 *                                                                            
 * HOL-OCL is distributed in the hope that it will be useful, but WITHOUT ANY 
 * WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS 
 * FOR A PARTICULAR PURPOSE. See the GNU General Public License for more 
 * details.                                                              
 *                                                                            
 * You should have received a copy of the GNU General Public License along    
 * with this program; if not, write to the Free Software Foundation, Inc.,    
 * 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.                  
 ******************************************************************************)


signature REP_ENCODER_LEVEL1 =
sig
    val  add_level1    : typ -> Rep_Core.Classifier list -> theory -> theory
end


structure rep_encoder_level1 : REP_ENCODER_LEVEL1 =
struct
open Rep_OclType
open Rep_Core
open rep_encoder_common

fun mkFstLevel u thy lhs rhs  = #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
				 Const("==", dummyT) 
				      $ (lhs)
				      $ (Const ("Lifting.lift1",dummyT)$rhs)
				 ],propT))




(********************************************************************)
fun add_constructors_1 C u cl  thy =  
    let

	val thy_name = Context.theory_name thy
	val name = name_of C
	val ctype_name = type_charset_name_of name
	val _ = trace 3  ("        "^(string_of_path name))
	val parent = parent_name_of C
	val mk = hc_path l1 ((mk_class_of o name_of) C)
	val mk0 = hc_path l0 ((mk_class_of o name_of) C)
	val mk_parent = hc_path l1 ((mk_class_of o parent_name_of) C)
	val get = hc_path l1 ((get_class_of o name_of) C)
	val get0 = hc_path l0 ((get_class_of o name_of) C)
	val parent_of =  hc_path l1 ( (get_class_of o parent_name_of)  C)
	val is =  hc_path l1 ((is_class_of o name_of) C)
	val isType =  hc_path l1 ((is_classType_of o name_of) C)
        val is0 =  hc_path l0 ((is_class_of o name_of) C)
        val isType0 =  hc_path l0 ((is_classType_of o name_of) C)
        val is_parent = hc_path l1 ((is_class_of o parent_name_of) C)
	val is_univ =  hc_path l1 ((is_univ_class_of o name_of) C)
	val is_univ0 =  hc_path l0 ((is_univ_class_of o name_of) C)
	val is_univ_parent =  hc_path l1 ((is_univ_class_of o parent_name_of) C)
	val class = name
	val holes = map (fn x => TFree(x,typeS)) (holes_of_type  u C)
        val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes) 

	val holes_par = map (fn x => TFree(x,typeS)) 
			    (holes_of_typestr  u (parent_name_of C))
        val class_u_par = if (parent_name_of C) = (name_of OclAnyC) then 
			      the(get_type_universe u (key_name_of (name_of OclAnyC))) (* TODO *)
			  else Type((string_of_path (hc_path univ (parent_name_of C))),holes_par)

	val class2parent = hc_path l1 (class2get_parent  name (local_name_of parent))
	val parent2class = hc_path l1 (parent2class_of  name (local_name_of parent))
	val class2parent0 = hc_path l0 (class2get_parent  name (local_name_of parent)) 
	val parent2class0 = hc_path l0 (parent2class_of  name (local_name_of parent))
			   
	fun add_consts_1  thy = 
	    let
		val mk_type      = mk_funT(V(u,class_u),V(u,u))
		val get_type     = mk_funT(V(u,u),V(u,class_u)) 
		val is_type      = mk_funT(V(u,class_u),BOOLT(St(u)))
		val is_univ_type = mk_funT(V(u,u),BOOLT(St(u)))
	    	val class2parent_type   = mk_funT(V(u,class_u),V(u,class_u_par))
		val parent2class_type   = mk_funT(V(u,class_u_par),V(u,class_u))
		val characteristic_type   = (V(u,SetT(class_u)))
	    in
		add_consts_i  (SOME (hc_path l1 (package_of C)) ) 
			      [(local_name_of mk,mk_type,NoSyn), 
			       (local_name_of get,get_type,NoSyn),
			       (local_name_of is,is_type,NoSyn), 
			       (local_name_of isType,is_type,NoSyn), 
			       (local_name_of is_univ,is_univ_type,NoSyn),
			       (local_name_of class2parent,class2parent_type, NoSyn),
			       (local_name_of parent2class,parent2class_type, NoSyn),	       
			       (local_name_of ctype_name,characteristic_type, NoSyn),
			       (local_name_of name,characteristic_type, NoSyn)
			      ] thy
	    end
	    
	fun add_defs C thy = 
	    let
		val mk_def_name    = (local_name_of o mk_def_of) mk
	        val get_def_name   = (local_name_of o mk_def_of) get
	        val is_def_name    = local_name_of (mk_def_of is)
	        val isType_def_name    = local_name_of (mk_def_of isType)
	        val is_univ_def_name = local_name_of (mk_def_of is_univ)
	        val class_def_name = local_name_of (mk_def_of name)
       		val parent_name =  (short_parent_name_of C)
		val class2parent_def_name = local_name_of (mk_def_of class2parent)
		val parent2class_def_name = local_name_of (mk_def_of parent2class)
				  

	
		val get_def_i =  mkFstLevel u thy (Const(local_name_of get,dummyT))  (Const(string_of_path get0,dummyT))

	
		val mk_def_i =  mkFstLevel u thy (Const(local_name_of mk,dummyT))  (Const(string_of_path mk0,dummyT))

		val is_univ_def_i = mkFstLevel u thy (Const(local_name_of is_univ,dummyT)) 
					      (Const("Fun.comp",dummyT)$(Const ("Lifting.up.lift",dummyT))$(Const(string_of_path is_univ0,dummyT)))

	
		val is_def_i = mkFstLevel u thy (Const(local_name_of is,dummyT)) 
					      (Const("Lifting.strictify",dummyT)$(Const("Fun.comp",dummyT)$(Const ("Lifting.up.lift",dummyT))$(Const(string_of_path is0,dummyT))))

	
		val isType_def_i = mkFstLevel u thy (Const(local_name_of isType,dummyT)) 
					      (Const("Lifting.strictify",dummyT)$(Const("Fun.comp",dummyT)$(Const ("Lifting.up.lift",dummyT))$(Const(string_of_path isType0,dummyT))))


		val class2parent_def_i =  mkFstLevel u thy (Const(local_name_of class2parent,dummyT))  (Const(string_of_path class2parent0,dummyT))

		val parent2class_def_i =  mkFstLevel u thy (Const(local_name_of parent2class,dummyT))  (Const(string_of_path parent2class0,dummyT))


		val characterstic_def_i = #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
							       Const("==", dummyT) 
						  $ Const(string_of_path name,dummyT)
						  $ (Term.absfree("St",dummyT, 
								  (Const("OCL_Set_type.Abs_Set_0",dummyT) 
									$ (Const ("Lifting.up.lift",dummyT) 
										 $  (Const ("op -",dummyT)  
											   $ (Const ("image",dummyT) 
												    $ ( Term.absfree("x",dummyT, 
														     Free("x",dummyT)
															 $Free("St",dummyT))) 
												    $ (Const ("image",dummyT)
													     $Const(string_of_path get,
														    dummyT)
													     $Const ("UNIV",dummyT) )) 
											   $Const ("{}",dummyT)
					     )))))],propT))

		val cset_type_def_i = #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
						 Const("==", dummyT) 
						      $ (Const(string_of_path ctype_name,dummyT))
						      $ (Const("OCL_Collection.OclSelect",dummyT)$Const(string_of_path name,dummyT)
						      $ Term.absfree("obj",dummyT, Const(string_of_path isType,dummyT)$Free("obj",dummyT)))
						 ],propT))

	

	    in
		add_definitions_i  false (SOME (hc_path l1 (package_of C)) )
				   [  (local_name_of(mk_def_of name),characterstic_def_i),
				      (get_def_name,get_def_i),
				      (mk_def_name,mk_def_i),
				      (is_univ_def_name,is_univ_def_i),
				      (local_name_of(mk_def_of  ctype_name),cset_type_def_i),
				      (class2parent_def_name,class2parent_def_i) ,   
				      (parent2class_def_name,parent2class_def_i),
				      (is_def_name,is_def_i),
				      (isType_def_name,isType_def_i)
				   ] thy 	
	    end 
	val thy = add_consts_1  thy
	val thy = add_defs  C thy   
    in 
	thy
    end


  
fun add_attributes_1 C u  cl thy =
    let
	val name = name_of C
	val _        = trace 3 ("      "^(string_of_path name))
	val thy_name = Context.theory_name thy
	val holes = map (fn x => TFree(x,typeS)) (holes_of_type  u C)
        val class_u =  Type((string_of_path (hc_path univ (name_of C))),holes) 
				
	fun add_consts  thy_path thy (a: attribute) = 
	    let
		val attr_name = #name a
		val attr   = #attr_type a

		val attr_type = mk_funT(V(u,class_u),V(u, (tilde u attr)))
		val attrn     = (hc_path l1 (attr_of name attr_name))
		val attr_pre  = (hc_path l1 (attr_pre_of name attr_name))
		val _ = trace 9  ("Adding: "^(string_of_path attrn))
	    in
		add_consts_i (SOME thy_path) [(local_name_of attrn,    attr_type,NoSyn),
					      (local_name_of attr_pre,attr_type,NoSyn)] thy
	    end
	    
	fun add_defs thy_path  thy (a:attribute) = 
	    let
		val attr_n = #name a
		val attr   = #attr_type a

		val attr_name    = (hc_path l1 (attr_of name attr_n))
		val attr_pre_name     = (hc_path l1 (attr_pre_of name attr_n))
		val attr_name0   = (hc_path l0 (attr_of name attr_n))
 
		val attr_def_name = local_name_of (mk_def_of attr_name)
		val attr_pre_def_name = local_name_of (mk_def_of attr_pre_name)

		fun mk_attr_def_i attr = 
		    case attr of 
			(Set(Classifier s))           => 
			Term.absfree("X",dummyT,
				     Term.absfree("St",dummyT,
						  (Const("OCL_Set_type.Abs_Set_0",dummyT)
							$ (Const("Lifting.up.lift",dummyT)
								$(Const("image",dummyT) 
								       $Term.absfree("r",dummyT,
										     Const("Datatype.option.option_case",dummyT)
											  $Const("Lifting.UU",dummyT)
											  $Term.absfree("x",dummyT,(Const(string_of_path (hc_path l0 (get_class_of s)),dummyT)$Free("x",dummyT)))
											  $(Const("fst",dummyT)$Free("St",dummyT)$Free("r",dummyT)))
								       $(Const ("Lifting.drop",dummyT)
									       $(Const("OCL_Set_type.Rep_Set_0",dummyT)
										      $(Const(string_of_path attr_name0,dummyT)$(Free("X",dummyT)$Free("St",dummyT))))))))))
		      | (Sequence(Classifier s))           => 	
			Term.absfree("X",dummyT,
				     Term.absfree("St",dummyT,
						  (Const("OCL_Sequence_type.Abs_Sequence_0",dummyT)
							$ (Const("Lifting.up.lift",dummyT)
								$(Const("map",dummyT) 
								       $Term.absfree("r",
										     dummyT,
										     Const("Datatype.option.option_case",dummyT)
											  $Const("Lifting.UU",dummyT)
											  $Term.absfree("x",dummyT,(Const(string_of_path (hc_path l0 (get_class_of s)),dummyT)$Free("x",dummyT)))
											  $(Const("fst",dummyT)$Free("St",dummyT)$Free("r",dummyT)))
								       $(Const ("Lifting.drop",dummyT)
									       $(Const("OCL_Sequence_type.Rep_Sequence_0",dummyT)
										      $(Const(string_of_path attr_name0,dummyT)$(Free("X",dummyT)$Free("St",dummyT))))))))))
			
		      | (Classifier s) =>   Term.absfree("X",dummyT,
							Term.absfree("St",dummyT,
								     Const("Datatype.option.option_case",dummyT)
									  $Const("Lifting.UU",dummyT)
									  $(Term.absfree("x",dummyT,
											 Const(string_of_path (hc_path l0 (get_class_of s)),dummyT)$Free("x",dummyT)))
									  $ ((Const("fst",dummyT)$Free("St",dummyT))
										 $(Const((string_of_path attr_name0),dummyT)
											$(Free("X",dummyT)$Free("St",dummyT))))))
		      | _ =>  Const("Lifting.lift1",dummyT)$(Const("Lifting.strictify",dummyT)
				  $Const(string_of_path attr_name0,dummyT))

		fun mk_attr_pre_def_i attr = 
		    case attr of 
			(Set(Classifier s))           => 
			Term.absfree("X",dummyT,
				     Term.absfree("St",dummyT,
						  (Const("OCL_Set_type.Abs_Set_0",dummyT)
							$ (Const("Lifting.up.lift",dummyT)
								$(Const("image",dummyT) 
								       $Term.absfree("r",dummyT,Const("Datatype.option.option_case",dummyT)
												     $Const("Lifting.UU",dummyT)
												     $Term.absfree("x",dummyT,(Const(string_of_path (hc_path l0 (get_class_of s)),dummyT)$Free("x",dummyT)))
												     $(Const("snd",dummyT)
													    $Free("St",dummyT)$Free("r",dummyT)))
								       $(Const ("Lifting.drop",dummyT)$(Const("OCL_Set_type.Rep_Set_0",dummyT)
													     $(Const(string_of_path attr_name0,dummyT)
														    $(Free("X",dummyT)$Free("St",dummyT))))))))))
		      | (Sequence(Classifier s))           => 	
			Term.absfree("X",dummyT,
				     Term.absfree("St",dummyT,
						  (Const("OCL_Sequence_type.Abs_Sequence_0",dummyT)
							$ (Const("Lifting.up.lift",dummyT)$(Const("map",dummyT) 
												 $Term.absfree("r",dummyT,Const("Datatype.option.option_case",dummyT)
															       $Const("Lifting.UU",dummyT)
															       $Term.absfree("x",dummyT,(Const(string_of_path (hc_path l0 (get_class_of s)),dummyT)$Free("x",dummyT)))
															       $(Const("snd",dummyT)$Free("St",dummyT)$Free("r",dummyT)))
												 $(Const ("Lifting.drop",dummyT)$(Const("OCL_Sequence_type.Rep_Sequence_0",dummyT)
																       $(Const(string_of_path attr_name0,dummyT)$(Free("X",dummyT)$Free("St",dummyT))))))))))
			
		      | (Classifier s) =>   Term.absfree("X",dummyT,Term.absfree("St",dummyT,Const("Datatype.option.option_case",dummyT)
												 $Const("Lifting.UU",dummyT)
												 $(Term.absfree("x",dummyT,Const(string_of_path (hc_path l0 (get_class_of s)),dummyT)$Free("x",dummyT)))
												 $ ((Const("snd",dummyT)$Free("St",dummyT))
													$(Const((string_of_path attr_name0),dummyT)$(Free("X",dummyT)$Free("St",dummyT))))))
		      | _ =>  Const("Lifting.lift1",dummyT)$(Const("Lifting.strictify",dummyT)
				  $Const(string_of_path attr_name0,dummyT))
			    

		val attr_def_i = #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
				 Const("==", dummyT) 
				      $ Const(string_of_path(hc_path l1 ((name_of C)@[local_name_of attr_name])),dummyT)
				      $ (mk_attr_def_i attr)
				 ],propT))
	        val attr_pre_def_i = #1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([
				 Const("==", dummyT) 
				      $ Const(string_of_path(hc_path l1 ((name_of C)@[local_name_of attr_pre_name])),dummyT)
				      $ (mk_attr_pre_def_i attr)
				 ],propT))
	
		val thy = add_definitions_i false (SOME thy_path) [(attr_def_name,attr_def_i),
						 (attr_pre_def_name,attr_pre_def_i)] thy 	
	    
		
		val thy = prove_basic_properties (SOME thy_path)
				   [local_name_of attr_name  
			         ,local_name_of attr_pre_name ]  thy
	    in
		thy
	    end 
	    

	val thy = foldl (fn (a,thy) => add_consts ((hc_path l1 (name_of C))) thy a) 
			thy (attributes_of C)
    
	val thy = foldl (fn (a,thy) => add_defs ((hc_path l1 (name_of C))) thy a) 
			thy (attributes_of C)  
    in 
	thy
    end


 
fun prove_1 C u  cl thy =
    let
	val thy_name = Context.theory_name thy
	val class  = name_of C
		     val _ = trace 3 ("      "^(string_of_path class))
	val parent = parent_name_of C

	val (oldpath,thy) = set_thy_path (SOME (hc_path l1 (name_of C))) thy

	val is_parent = string_of_path (hc_path l1 ((is_class_of o parent_name_of) C))
	val parent_of = string_of_path (hc_path l1 ((get_class_of o parent_name_of) C))
	val mk_parent = string_of_path (hc_path l1 ((mk_class_of o parent_name_of) C))
	val is_parent_univ = string_of_path (hc_path l1 ((is_univ_class_of o parent_name_of) C))

	fun prover_level_1_i thy prems goal l0_thms =
	    let
		val isDefined_if_valid   = get_thm thy (Name "isDefined_if_valid")
		val localValidDefined2sem = get_thm thy (Name "localValidDefined2sem")
		val iffD1                 = get_thm thy (Name "iffD1")
	(*	fun term_case_tac (a_i:term) = res_terminst_tac [] [((("P",0),boolT),a_i)]  case_split_thm; *)

		val goal_ci = (cterm_of (sign_of thy) 
					(#1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([goal],propT))))
			      
		val _ = trace 9  (string_of_cterm goal_ci)


	    in 
		quick_and_dirty_prove_goalw_cterm thy  prems goal_ci
			    (fn p => [cut_facts_tac p 1, 
				      TRY(forward_tac [isDefined_if_valid] 1),
				      TRY(forward_tac [localValidDefined2sem RS iffD1] 1),  
				      unfold_level_1_tac thy,
				      asm_full_simp_tac (HOL_DEF_ss) 1,
				     case_tac "((op = ) ((obj) (St)) (Lifting.up.down))" 1,  
				      auto_tac (OCL_cla,OCL_DEF_ss addsimps (l0_thms))
			    ])
	    end


	    fun mkGoal_i prems goal = Const ("==>",propT --> propT --> propT) $
				     (Const ("Trueprop",boolT --> propT) $ prems) $
				     (Const ("Trueprop",boolT --> propT) $ goal)

	    fun prover_level_1b thy prems goal l0_thms =
		let
		    val isDefined_if_valid   = get_thm thy (Name "isDefined_if_valid")
		    val localValidDefined2sem = get_thm thy (Name "localValidDefined2sem")
		    val iffD1                 = get_thm thy (Name "iffD1")
		    val simp = (get_def thy "OclLocalValid")   
			       ::(get_def thy "lift1") 
			       ::(get_def thy "lift0") 
			       ::(get_def thy "OclTrue") 
			       ::(get_def thy "strictify") 
			       ::(get_def thy "DEF")
			       ::[] 
		    val goal_ci = (cterm_of (sign_of thy) 
					    (#1(Sign.infer_types (Sign.pp thy) thy (K NONE) (K NONE) [] true ([goal],propT))))
				  
		    val _ = trace 9  (string_of_cterm goal_ci)
			    
		in 
		    quick_and_dirty_prove_goalw_cterm thy  prems goal_ci
				       (fn p => [cut_facts_tac p 1, 
						 unfold_level_1_tac thy,
				      asm_full_simp_tac (HOL_DEF_ss) 1,
						 auto_tac (OCL_cla,OCL_DEF_ss addsimps l0_thms)
				       ])
		end
		

	fun is_mk_class class parent thy = 
	    let
		
		val name="is_mk_"^(local_name_of class)
		val  goal_i = mkGoal_i (Const ("OCL_Logic_core.OclLocalValid",dummyT)
					      $ Free ("St",dummyT) $
					      (Const((string_of_path (hc_path l1 (is_class_of class))),dummyT) $ Free ("obj",dummyT)))
				       (Const ("OCL_Logic_core.OclLocalValid",dummyT)
					      $ Free ("St",dummyT) 
					      $ (Const(string_of_path (hc_path l1 (is_univ_class_of class)),dummyT) 
						      $ (Const(string_of_path (hc_path l1 (mk_class_of class)),dummyT) 
							      $ Free ("obj",dummyT))))
	        val _ = trace 9  ("proving "^name)
		val simps = [get_thm  thy (Name ((string_of_path (hc_path l0 (rev(tl(rev(class))))))
			     ^"."^name ))] 
		val thm = prover_level_1_i thy [] goal_i simps 
	    in
		 (fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end    



	fun get_mk_class_id class parent thy =
	    let
		val name = "get_mk_"^(local_name_of class)^"_id"
	

		val  goal_i = mkGoal_i (Const ("OCL_Logic_core.OclLocalValid",dummyT)
					      $ Free ("St",dummyT) $
					      (Const((string_of_path (hc_path l1 (is_class_of class))),dummyT) $ Free ("obj",dummyT)))
				       (Const ("op =",dummyT)
					      $ (Const(string_of_path (hc_path l1 (get_class_of class)),dummyT) 
						      $ (Const(string_of_path (hc_path l1 (mk_class_of class)),dummyT) 
							      $ Free ("obj",dummyT))$Free("St",dummyT))
					      $ (Free("obj",dummyT)$Free("St",dummyT)))
			      
		val _ = trace 9  ("proving "^name)

		val simps = [get_thm  thy (Name ((string_of_path (hc_path l0 (rev(tl(rev(class))))))
			     ^"."^name))]
		val thm = prover_level_1_i thy [] goal_i simps 

	    in
		 (fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end

	fun mk_get_class_id class parent thy =
	    (* mk_get_A_id_0: "is_A_uni_0 univ ==> mk_A_0(get_A_0 univ) = univ"  *)
	    let
		val name = "mk_get_"^(local_name_of class)^"_id"
		val  goal_i = mkGoal_i (Const ("OCL_Logic_core.OclLocalValid",dummyT)
					      $ Free ("St",dummyT) $
					      (Const((string_of_path (hc_path l1 (is_univ_class_of class))),dummyT) $ Free ("univ",dummyT)))
				       (Const ("op =",dummyT)
					      $ (Const(string_of_path (hc_path l1 (mk_class_of class)),dummyT) 
						      $ (Const(string_of_path (hc_path l1 (get_class_of class)),dummyT) 
							      $ Free ("univ",dummyT))$Free("St",dummyT))
					      $ (Free("univ",dummyT)$Free("St",dummyT)))

		val _ = trace 9  ("proving "^name)
		val simps = [get_thm  thy (Name ((string_of_path (hc_path l0 (rev(tl(rev(class))))))
			     ^"."^name))]
		val thm = prover_level_1_i thy [] goal_i simps 
	    in
		(fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end

	fun is_class_univ_implies_is_get class parent thy =
	    (* is_D_univ_0 univ ==> is_D_0(get_D_0 univ) *)
	    let
		val name="is_"^(local_name_of class)^"_univ_implies_is_get"
		val goal_i = Const ("==>",propT --> propT --> propT) $
				     (Const ("Trueprop",boolT --> propT) $ (Const("Lifting.DEF",dummyT)$(Const((string_of_path (hc_path l1 (get_class_of class))),dummyT)$Free("univ",dummyT)$Free("St",dummyT))

)) $
				 (Const ("==>",propT --> propT --> propT) $

				 ((Const ("Trueprop",boolT --> propT) $ 
				 (Const ("OCL_Logic_core.OclLocalValid",dummyT)
					      $ Free ("St",dummyT) $
					      (Const((string_of_path (hc_path l1 (is_univ_class_of class))),dummyT) $ Free ("univ",dummyT)))
				)) $ ( Const ("Trueprop",boolT --> propT) $
				 (Const ("OCL_Logic_core.OclLocalValid",dummyT)
					$ Free ("St",dummyT) 
					$ (Const((string_of_path (hc_path l1 (is_class_of class))),dummyT) 
						$ (Const(string_of_path (hc_path l1 (get_class_of class)),dummyT) 
							$ Free ("univ",dummyT))))))
			     
			     


		val _ = trace 9  ("proving "^name)
		val simps = [get_thm  thy (Name ((string_of_path (hc_path l0 (rev(tl(rev(class))))))
					   ^"."^name))]
		val thm = prover_level_1_i  thy [] goal_i simps 	
	    in
		 (fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end

	      
	fun is_parent_mk_class class parent thy = 
	    let
		val name="is_"^(local_name_of parent)^"_mk_"^(local_name_of class)

		val  goal_i = mkGoal_i (Const ("OCL_Logic_core.OclLocalValid",dummyT)
					      $ Free ("St",dummyT) $
					      (Const((string_of_path (hc_path l1 (is_class_of class))),dummyT) $ Free ("obj",dummyT)))
				       (Const ("OCL_Logic_core.OclLocalValid",dummyT)
					      $ Free ("St",dummyT) 
					      $ (Const(is_parent_univ,dummyT) 
						      $ (Const(string_of_path (hc_path l1 (mk_class_of class)),dummyT) 
							      $ Free ("obj",dummyT))))
	        val _ = trace 9  ("proving "^name)
		val simps = [get_thm  thy (Name ((string_of_path (hc_path l0 (rev(tl(rev(class))))))
					   ^"."^name)),
			     get_def thy (is_parent_univ)]
		val thm = prover_level_1_i thy [] goal_i simps

	    in
		 (fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end    
	    
	    
	    
	fun is_class_univ_implies_parent_univ class parent thy =
	    (*@ $is\_\template{C}\_univ_0 obj \Implies is\_A\_univ_0 obj$ @*)
	    let
		val name="is_"^(local_name_of class)^"_univ_implies_"^(local_name_of parent)^"_univ"
		val goal_i  = mkGoal_i (Const ("OCL_Logic_core.OclLocalValid",dummyT)
					     $ Free ("St",dummyT) $
					     (Const((string_of_path (hc_path l1 (is_univ_class_of class))),dummyT) $ Free ("obj",dummyT)))
				      (Const ("OCL_Logic_core.OclLocalValid",dummyT)
					     $ Free ("St",dummyT) $
					     (Const (is_parent_univ,dummyT) $ Free ("obj",dummyT)))

	        val _ = trace 9  ("proving "^name)

		val simps = [get_thm  thy (Name ((string_of_path (hc_path l0 (rev(tl(rev(class))))))
					   ^"."^name)),
			    get_def thy is_parent_univ ]
		val thm = prover_level_1b thy [] goal_i simps 	
 	    in
		 (fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end 



	fun cast_class_id class parent thy =
	    (*   cast_C_id_0: "is_C_0 obj ==> (A_2_C_0(C_2_A_0(obj))) = obj" *)
	    let
		val name = "cast_"^(local_name_of class)^"_id"
		val  goal_i = mkGoal_i (Const ("OCL_Logic_core.OclLocalValid",dummyT)
					      $ Free ("St",dummyT) $
					      (Const((string_of_path (hc_path l1 (is_class_of class))),dummyT) $ Free ("obj",dummyT)))
				       (Const ("op =",dummyT)
					      $ (Const(string_of_path (hc_path l1 ((parent2class_of class (local_name_of parent)))),dummyT) 
						      $ (Const(string_of_path (hc_path l1 ((class2get_parent class (local_name_of parent)))),dummyT) 
							      $ Free ("obj",dummyT))$Free("St",dummyT))
					      $ (Free("obj",dummyT)$Free("St",dummyT)))

		val _ = trace 9  ("proving "^name)
		val simps = [get_thm  thy (Name ((string_of_path (hc_path l0 (rev(tl(rev(class))))))
					   ^"."^name))]
		val thm = prover_level_1_i thy [] goal_i simps 	
	    in
		(fst(PureThy.add_thms [((name,thm),[])] (thy))) 
	    end  

	fun cp_operators class parent thy =
	    let	    

		fun cp_goal thy t = (cterm_of (sign_of thy) 
					      (#1(Sign.infer_types (Sign.pp thy)thy (K NONE) (K NONE) [] true (
						  [

						     (Const ("Trueprop",boolT --> propT) $ 
						         (Const("Lifting.cp",dummyT)$t))
						 ],propT))))


		val name_get = "cp_get_"^(local_name_of class)
		val goal_get_i = cp_goal thy (Const((string_of_path (hc_path l1 (get_class_of class))),dummyT))
		val simps_get = [get_def thy (string_of_path (hc_path l1 (get_class_of class)))(*,
				 get_thm thy "cp_lift1" *)]

		val thm_get = quick_and_dirty_prove_goalw_cterm thy [] goal_get_i
					  (fn p => [cut_facts_tac p 1, 
						    simp_tac (OCL_ss addsimps simps_get) 1
					  ])
		val name_mk = "cp_mk_"^(local_name_of class)
		val goal_mk_i = cp_goal thy (Const((string_of_path (hc_path l1 (mk_class_of class))),dummyT))

		val simps_mk = [get_def thy (string_of_path (hc_path l1 (mk_class_of class))) (*,
				 get_thm thy "cp_lift1" *)]

		val thm_mk = quick_and_dirty_prove_goalw_cterm thy  [] goal_mk_i
					  (fn p => [cut_facts_tac p 1, 
						    simp_tac (OCL_ss addsimps simps_mk) 1
					  ])

		val name_is = "cp_is_"^(local_name_of class)
		val goal_is_i = cp_goal thy (Const((string_of_path (hc_path l1 (is_class_of class))),dummyT))
		val simps_is = [get_def thy (string_of_path (hc_path l1 (is_class_of class))) (*,
				 get_thm thy "cp_lift1" *)]

		val thm_is = quick_and_dirty_prove_goalw_cterm thy [] goal_is_i
					  (fn p => [cut_facts_tac p 1, 
						    simp_tac (OCL_ss addsimps simps_is) 1
					  ])

		val name_is_univ = "cp_is_univ_"^(local_name_of class)
		val goal_is_univ_i = cp_goal thy (Const((string_of_path (hc_path l1 (is_univ_class_of class))),dummyT))
		val simps_is_univ = [get_def thy (string_of_path (hc_path l1 (is_univ_class_of class))) (*,
				 get_thm thy "cp_lift1" *)]

		val thm_is_univ = quick_and_dirty_prove_goalw_cterm thy [] goal_is_univ_i
					  (fn p => [cut_facts_tac p 1, 
						    simp_tac (OCL_ss addsimps simps_is_univ) 1
					  ])

		(* extend simpset *)
		val rsim = simpset_ref_of thy
		val rcla = claset_ref_of thy
		val cp_compose2 = get_thm (theory "OCL") (Name "cp_compose2")
	    in 
		rsim := !rsim addsimps [thm_get RS cp_compose2, 
					thm_mk RS cp_compose2, 
					thm_is RS cp_compose2, 
					thm_is_univ RS cp_compose2];

		rcla := !rcla addSIs [thm_get RS cp_compose2, 
					thm_mk RS cp_compose2, 
					thm_is RS cp_compose2, 
					thm_is_univ RS cp_compose2];

		(fst(PureThy.add_thms [((name_get,thm_get),[]),
				       ((name_mk, thm_mk),[]) ,
				       ((name_is, thm_is),[]), 
				       ((name_is_univ, thm_is_univ),[]) 
		 		      ] (thy))) 
	    end

	fun cp_attributes class  thy =
	    let	    

		fun cp_goal thy t = (cterm_of (sign_of thy) 
					      (#1(Sign.infer_types (Sign.pp thy)thy (K NONE) (K NONE) [] true (
						  [

						     (Const ("Trueprop",boolT --> propT) $ 
						         (Const("Lifting.cp",dummyT)$t))
						 ],propT))))

		fun cp (attr:attribute) = 
		    let
			
			val  cp_ss = (OCL_ss addsimps [get_thm thy (Name "cp_def")]);
			val simps_get = [] (* [get_def thy attrn] *)
			val attr_name  = (#name attr)
			val attrn     = string_of_path (hc_path l1 (attr_of (name_of class)  attr_name))
			val attrn0     = string_of_path (hc_path l0 (attr_of (name_of class)  attr_name))
	
			fun prove_value goal = quick_and_dirty_prove_goalw_cterm thy [] goal

								 (fn p => [cut_facts_tac p 1, 
									 rewrite_goal_tac [get_thm thy (Name (attrn^"_def"))] 1,
									 simp_tac (OCL_ss) 1
								 ])
			fun prove_ref goal = quick_and_dirty_prove_goalw_cterm thy [] goal
								 (fn p => [cut_facts_tac p 1, 
									 rewrite_goal_tac [get_thm thy (Name (attrn^"_def"))] 1,
									 simp_tac (OCL_ss) 1,
									 TRY(rtac (get_thm thy (Name "cp_option_case")) 1),
									 TRY(ALLGOALS (asm_full_simp_tac OCL_ss)),
									 TRY(simp_tac (OCL_ss addsimps [get_thm thy (Name "cp_def")]) 1),
									 TRY(res_inst_tac [("x","\\<lambda> x x'. (((fst x') ("^attrn0^"(x)))) ")] 
										      (get_thm thy (Name "exI")) 1),
									  TRY(simp_tac (OCL_ss) 1)
								 ])

			fun prove_setref get_class goal = quick_and_dirty_prove_goalw_cterm thy [] goal
								  (fn p => [cut_facts_tac p 1, 
									 rewrite_goal_tac [get_thm thy (Name (attrn^"_def"))] 1,
									  simp_tac (OCL_ss) 1,
									  TRY(simp_tac cp_ss  1),
									  TRY(res_inst_tac [("x",
											     "\\<lambda> XST ST'. Abs_Set_0  \\<HolOclLiftLeft>(\\<lambda>r. "
											     ^"option_case \\<bottom> "
											     ^get_class ^"((fst ST') r)) ` "
											     ^"\\<HolOclDropLeft>Rep_Set_0 ("
												  ^attrn0
												  ^" (XST))\\<HolOclDropRight>\\<HolOclLiftRight> " 
											   )] 
											   (get_thm thy (Name "exI")) 1),
									  TRY(simp_tac (OCL_ss) 1)
								  ])

			fun prove_seqref get_class goal = quick_and_dirty_prove_goalw_cterm thy [] goal
								  (fn p => [cut_facts_tac p 1, 
									  simp_tac (OCL_ss addsimps [get_thm thy (Name (attrn^"_def"))]) 1,
									  TRY(simp_tac cp_ss  1),
									  TRY(res_inst_tac [("x",
											     "\\<lambda> XST ST. Abs_Sequence_0 "
											     ^"\\<HolOclLiftLeft>map (\\<lambda>r. option_case \\<bottom> "
											     ^get_class
											     ^"((fst ST) r))"
											     ^"\\<HolOclDropLeft>Rep_Sequence_0 ("
											     ^attrn0
											     ^" (XST))\\<HolOclDropRight>\\<HolOclLiftRight>"
											   )] 
											   (get_thm thy (Name "exI")) 1),
									  TRY(simp_tac (OCL_ss) 1)
								  ])




		   


			val goal  = cp_goal thy (Const(attrn,dummyT))

			val thm =  (case (#attr_type attr) of 
				       (Set(Classifier s))      => prove_setref (string_of_path (hc_path l0 (get_class_of s))) goal
				     | (Sequence(Classifier s)) => prove_seqref (string_of_path (hc_path l0 (get_class_of s))) goal
				     | (Classifier s)           => prove_ref goal
				     | _                        => prove_value goal
				   )
		    in
			(("cp_"^attr_name,thm),[])
		    end
		     
		val cp_thms = (map cp (attributes_of C))
		val rsim = simpset_ref_of thy
		val rcla = claset_ref_of thy
				      
		val cp_compose2 = get_thm (theory "OCL") (Name "cp_compose2")
	    in 
		rsim := !rsim addsimps (map ((fn t => t RS cp_compose2) o snd o fst) cp_thms);
		rcla := !rcla addSIs (map ((fn t => t RS cp_compose2) o snd o fst) cp_thms);
		(fst(PureThy.add_thms cp_thms thy))
	    end
	    


	val thy = is_class_univ_implies_is_get class parent thy 
	val thy = get_mk_class_id  class parent thy   
	val thy = is_mk_class class  parent thy
	val thy = mk_get_class_id class parent thy   
	val thy = is_parent_mk_class class  parent thy  
 	val thy = is_class_univ_implies_parent_univ class parent thy  
	val thy = cast_class_id class parent thy 		  
	val thy = cp_operators class parent thy 		  
	val thy = cp_attributes C thy 		  

	val (_,thy) = set_thy_path oldpath thy
    in
	thy
    end



fun add_level1 universe cl thy = 
    let
	(*************************************)	  
       	(* add first level definitions       *)
	val _ = trace 2 "    Adding constructors" 
	val thy = foldl (fn (c,thy) => add_constructors_1 
					   c universe  cl thy) 
			thy cl   	      
		  
	val _ = trace 2 "    Adding attributes" 
	val thy = foldl (fn (c,thy) => add_attributes_1
					   c universe cl  thy) 
			thy cl 	      

	val _ = trace 2 "    Proving ..." 
	val thy = foldl (fn (c,thy) => prove_1 
					   c universe cl thy) 
			thy cl   	      
    in 
	thy
    end


end
